/*
 * last modified---
 * 	07-17-15/jet/ elide unnecessary unit conversions in getRelativePrice()
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	06-05-13/jet/ SilentVault client version
 * 	10-12-11/kbw/ remove obsolete URL data source stuff; switch to TimerTask
 * 	06-08-11/kbw/ add support for silver (XAG)
 * 	03-11-11/kbw/ add fetchPrices() and parsePrices(); obtain data from OFS
 * 	02-22-11/kbw/ add getRelativePrice()
 * 	11-17-10/kbw/ do status bar changes in a separate thread from actual fetch
 * 	09-20-10/kbw/ add notification on progress bar when fetching prices
 * 	10-05-09/kbw/ return 1.0 as price for gAu
 * 	05-27-09/kbw/ relocate from ./ui, add plugin, tell preference UI on dnld
 * 	05-12-09/kbw/ new
 *
 * purpose---
 * 	provide a class to fetch, store, and cache backing asset prices
 */

package org.silentvault.client;

import javax.swing.JPanel;
import javax.swing.JProgressBar;
import java.awt.Color;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.silentvault.client.ui.wallet.WTabManager;
import org.silentvault.client.ui.wallet.HomePane;
import org.silentvault.client.ui.VSPreferences;
import org.silentvault.client.ui.PreferencesUI;

import java.util.HashMap;
import java.util.Date;
import java.util.Iterator;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;


/**
 * This class provides an object to retrieve, store, cache, and fetch on demand
 * the fiat currency prices of a backing asset (gold or silver, e.g.).  Once
 * the initial fetch is performed, it will automatically update the data when
 * another price is asked for after a set expiration time (by default, 5 min.).
 * This object is also intended to convert between fiat currency types, as
 * defined.
 *
 * @author Kevin Wilkerson
 */
public class AssetPriceCache extends TimerTask {
	// BEGIN data members
	/**
	 * the maximum time allowed before data must be requested anew (in msec)
	 */
	private final long		M_DataStale = 180000L;		// 3 minutes

	/**
	 * the plugin which owns us
	 */
	private WalletClient		m_Plugin;

	/**
	 * the XML contents last fetched from the data source
	 */
	private String				m_FetchedXML;

	/**
	 * the date/time at which the last fetch was performed
	 */
	private Date				m_LastFetch;

	/**
	 * return status of last background fetch
	 */
	private boolean			m_FetchStat;

	/**
	 * the timer which handles market price downloads off the OFS
	 */
	private Timer				m_PriceDownloadTimer;

	/**
	 * the hash map of prices for 1 gram worth of gold
	 */
	private HashMap<String, Double>	m_PriceMap;

	/**
	 * the hash map of currency names
	 */
	private HashMap<String, String>	m_NameMap;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the plugin which created us
	 */
	public AssetPriceCache(WalletClient plugin) {
		m_Plugin = plugin;
		m_PriceMap = new HashMap<String, Double>();
		m_NameMap = new HashMap<String, String>();
		m_FetchedXML = "";
		m_PriceDownloadTimer = new Timer("pricefetch", true);
	}

	/**
	 * request a price in a given fiat currency
	 * @param fiatCode the type of currency for which price is desired
	 * @return the price, or null if none is recorded for this currency type
	 */
	public Double getCurrPrice(String fiatCode) {
		if (fiatCode == null || fiatCode.isEmpty()) {
			return null;
		}
		if (fiatCode.equalsIgnoreCase("GAU")) {
			// ratio is 1 by definition
			return new Double(1.0);
		}
		return m_PriceMap.get(fiatCode);
	}

	/**
	 * method to obtain the ratio between two currencies by interpolation
	 * @param curr1 the currency whose price in curr2 will be returned
	 * NB: if this is "GAU" then this method reduces to a call to getCurrPrice()
	 * @param curr2 the fiat currency whose price we want
	 * NB: if this is "GAU" then this method reduces to a call to getCurrPrice()
	 * @return the amount of curr2 which equals 1.0 in curr1, or 0.0 if not found
	 */
	public Double getRelativePrice(String curr1, String curr2) {
		if (curr1 == null || curr1.isEmpty() || curr2 == null || curr2.isEmpty())
		{
			return null;
		}

		if (curr1.equals(curr2)) {
			// ratio is 1 by definition
			return new Double(1.0);
		}

		// if we want the price of gold, simply do a market price lookup
		if (curr1.equalsIgnoreCase("GAU")) {
			return getCurrPrice(curr2);
		}
		// if we want other priced in gold, return reciprocal of market price
		if (curr2.equalsIgnoreCase("GAU")) {
			Double cpr = getCurrPrice(curr1);
			return cpr == null ? null : new Double(1.0 / cpr.doubleValue());
		}

		// otherwise get the value of 1 GAU in each, and return the ratio
		Double value1 = getCurrPrice(curr1);
		if (value1 == null) {
			Log.error("Could not obtain " + curr1 + " price data");
			return new Double(0.0);
		}
		Double value2 = getCurrPrice(curr2);
		if (value2 == null) {
			Log.error("Could not obtain " + curr2 + " price data");
			return new Double(0.0);
		}
		return new Double(value2.doubleValue() / value1.doubleValue());
	}

	/**
	 * utility method to set the status bar to indicate we are fetching prices
	 * @param starting true if we're starting, false if we're finishing
	 */
	private void setStatBar(boolean starting) {
		WTabManager tm = m_Plugin.getWalletTabManager();
		HomePane hp = tm.getHomePane();
		JPanel statBar = m_Plugin.getStatusBar();

		// if we're logged in and have a status bar, say that we're fetching
		if (starting) {
			if (hp != null) {
				JProgressBar pb = new JProgressBar();
				pb.setIndeterminate(true);
				pb.setString("Downloading market prices, please wait");
				pb.setStringPainted(true);
				pb.setOpaque(true);
				pb.setBackground(Color.LIGHT_GRAY);
				pb.setFont(hp.M_DataFont);
				statBar.removeAll();
				statBar.add(pb);
				statBar.revalidate();
				statBar.repaint();
			}
		}
		else {
			// take down stat bar if we set one
			if (hp != null) {
				statBar.removeAll();
				statBar.invalidate();
				statBar.revalidate();
				statBar.repaint();
			}
		}
	}

	/**
	 * utility method to obtain prices from the OFS
	 * @return true on success, false on any failure/timeout
	 */
	private synchronized boolean fetchPrices() {
		// set the stat bar to indicate we are fetching
		setStatBar(true);

		VscState state = m_Plugin.getVscState();
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot download asset prices without OFS connection");
			return false;
		}

		// build message to send
		OFSMessage fetch = new OFSMessage();
		fetch.setType(IQ.Type.GET);
		fetch.setFrom(conn.getUser());
		fetch.setTo(conn.getServiceName());
		fetch.setPacketID("prices_" + state.getNextOFSid());
		fetch.setOpcode("REQ_lookup_prices");

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(fetch.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(fetch);

		// wait for reply or timeout
		OFSMessage rep = (OFSMessage) repColl.nextResult(m_Plugin.getTimeout());
		if (rep == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_price_data");
			rep.setFrom(fetch.getTo());
			rep.setTo(fetch.getFrom());
			rep.setPacketID(fetch.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("market price data timeout from OFS");
		}

		// clear any stat bar changes
		setStatBar(false);
		m_Plugin.setDefaultStatusBar();

		// process results
		if (rep.getType() == IQ.Type.ERROR) {
			Log.error("Unable to obtain market asset price data, code "
						+ rep.getErrcode() + ", error: " + rep.getErrmsg());
			return false;
		}
		String priceData = rep.getPriceData();
		m_FetchedXML = new String(StringUtils.decodeBase64(priceData));

		// parse the retrieved XML -- this code depends on the data source format
		Element pricesElt = EncodingUtils.buildElement(m_FetchedXML);
		if (pricesElt == null
			 || !pricesElt.getName().equalsIgnoreCase("GoldPrice"))
		{
			Log.error("fetched prices would not parse as XML");
			return false;
		}

		// parse the price data and load it into our maps
		return parsePrices(pricesElt);
	}

	/**
	 * utility method to parse prices
	 * @param pricesElt the Element containing the price map
	 * @return true on success
	 */
	private boolean parsePrices(Element pricesElt) {
		// clear the maps to avoid inadvertently retaining a stale price or name
		if (!m_PriceMap.isEmpty()) {
			m_PriceMap.clear();
		}
		if (!m_NameMap.isEmpty()) {
			m_NameMap.clear();
		}

		// iterate over elements
		Double price = null;
		String curr = null;
		String name = null;
		Iterator priceIt = pricesElt.elementIterator();
		while (priceIt.hasNext()) {
			Element prElt = (Element) priceIt.next();
			try {
				if (prElt.getName().equalsIgnoreCase("Price")) {
					curr = prElt.attributeValue("currencycode");
					if (curr == null || curr.isEmpty()) {
						throw new DocumentException("missing currencycode attribute");
					}
					name = prElt.attributeValue("currencyname");
					if (name == null || name.equals("currencyname")) {
						throw new DocumentException("missing currencyname attribute");
					}
					price = Double.parseDouble(prElt.getText());
				}
				else {
					throw new DocumentException("non-price element seen");
				}
				m_PriceMap.put(curr, price);
				m_NameMap.put(curr, name);
			}
			catch (DocumentException de) {
				Log.error("error parsing fetched price", de);
				return false;
			}
			catch (NumberFormatException nfe) {
				Log.error("error parsing price value", nfe);
				return false;
			}
		}
		return true;
	}

	/**
	 * retrieve the descriptive name of a currency
	 * @param fiatCode the code for a fiat currency
	 * @return its long name, e.g. "Pound Sterling", or null if not found
	 */
	public String getCurrencyName(String fiatCode) {
		if (fiatCode == null || fiatCode.isEmpty()) {
			return null;
		}
		return m_NameMap.get(fiatCode);
	}

	/**
	 * retrieve the available set of fiat currencies for which we have prices
	 * @return the set of currency codes
	 */
	public Set<String> getSupportedCurrencies() { return m_PriceMap.keySet(); }

	// method to extend TimerTask
	/**
	 * utility method to allow price fetching in a background thread
	 */
	public void run() {
		m_FetchStat = fetchPrices();
		if (m_FetchStat) {
			m_LastFetch = new Date();

			// tell the preferences UI about the latest list of currencies
			VSPreferences vsp = m_Plugin.getUserPreferences();
			Object prefUI = vsp.getGUI();
			if (prefUI != null && prefUI instanceof PreferencesUI) {
				PreferencesUI ui = (PreferencesUI) prefUI;
				ui.buildUI();
			}
			vsp.load();
		}
	}

	/**
	 * initialize the price-fetching timer
	 */
	public void initPriceFetching() {
		// start timer recurring
		if (m_PriceDownloadTimer != null) {
			// NB: run() is invoked manually by startup code, the first time, to
			// prevent collisions in Swing with WalletClient.init() run after we
			// return from this call.
			m_PriceDownloadTimer.schedule(this, M_DataStale, M_DataStale);
		}
	}

	/**
	 * cancel the price-fetching timer
	 */
	public void cancelPriceFetching() {
		if (m_PriceDownloadTimer != null) {
			m_PriceDownloadTimer.cancel();
			m_PriceDownloadTimer.purge();
			m_PriceDownloadTimer = null;
		}
	}

	// END methods
}
