/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	11-12-12/kbw/ use setLastActivity() rather than setLastRecvDHT(), so that
 * 					  we count this as network activity against inactivity timeout
 * 	04-18-11/kbw/ use ConcurrentHashMap for m_DHTrecords instead of Hashtable
 * 	04-04-11/kbw/ add deque to keep track of previous mark requests
 * 	03-16-10/kbw/ support silent DHT queries after successful payments
 * 	02-24-10/kbw/ getState() -> getVsState()
 * 	04-21-09/kbw/ add revolving filter map, to support N outstanding queries
 * 	12-09-08/kbw/ new
 *
 * purpose---
 * 	listen for messages from the DHT which are replies to IQ packets we sent
 *
 * notes---
 * 	see underlying messages in vsc_dht.x
 */

package org.silentvault.client;

import javax.swing.*;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.filter.OrFilter;

import org.silentvault.client.ui.wallet.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.ArrayDeque;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;


/**
 * class to handle processing of DHT reply packets, within the context of the
 * jabber:iq:voucher-safe#vsc-dht namespace.
 *
 * @author Kevin Wilkerson
 */
public final class DHTListener implements PacketListener, Runnable {
	// BEGIN data members

	/**
	 * back pointer reference to the Plugin object
	 */
	private WalletClient						m_PluginPtr;

	/**
	 * the live connection to the DHT on which we are listening
	 */
	private XMPPConnection					m_DHTConn;

	/**
	 * the list of records we most recently retrieved from the DHT
	 */
	private ConcurrentHashMap<Integer, String>	m_DHTrecords;

	/**
	 * a map of currently outstanding listen filters, representing all requests
	 * for which we are awaiting replies, keyed by their packet IDs
	 */
	private Hashtable<String, PacketIDFilter>	m_FilterMap;

	/**
	 * a deque of the record mark requests we've sent, awaiting replies
	 */
	private ArrayDeque<DHTMessage>		m_RecordMarks;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the reference to the plugin of which we are a part
	 */
	public DHTListener(WalletClient plugin) {
		m_PluginPtr = plugin;
		m_DHTrecords = new ConcurrentHashMap<Integer, String>();
		m_FilterMap = new Hashtable<String, PacketIDFilter>();
		m_RecordMarks = new ArrayDeque<DHTMessage>();
	}

	// method required to implement the PacketListener interface
	/**
	 * deal with a reply packet from the DHT
	 * @param packet the packet (will really be a DHTMessage)
	 */
	public void processPacket(Packet packet) {
		// verify we got what's expected
		if (packet instanceof DHTMessage) {
			DHTMessage thisMsg = (DHTMessage) packet;
			String opcode = thisMsg.getOpcode();
			//Log.debug("Listener got DHT message opcode " + opcode);

			// record message
			VscState stateObj = m_PluginPtr.getVscState();
			stateObj.setLastActivity(thisMsg);
			WTabManager tm = m_PluginPtr.getWalletTabManager();

			// see if it's an error
			HomePane hp = tm.getHomePane();
			if (thisMsg.getType() == IQ.Type.ERROR) {
				Log.error("Got an error: code " + thisMsg.getErrcode() + ", "
							+ thisMsg.getErrmsg());
				if (opcode.equalsIgnoreCase("REP_receipt_stored")) {
					// tell the IncomingPane about this error
					tm.getIncomingPane().recordReceiptUpload(thisMsg);
				}
				else {
					// display something generic
					tm.showError("DHT error code " + thisMsg.getErrcode() + ", "
									+ thisMsg.getErrmsg());
				}
				removeFilter(packet.getPacketID());
				hp.setCursor(null);
				return;
			}

			// success -- take further actions based on the opcode
			if (opcode.equalsIgnoreCase("REP_records")) {
				// record records returned
				m_DHTrecords.putAll(thisMsg.getRecordList());
				hp.setCursor(null);

				// use the housekeeping pane to parse the list
				tm.queueTransition(new Integer(WTabManager.M_WORK_DHT_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_receipt_stored")) {
				// tell the incoming pane about this
				tm.getIncomingPane().recordReceiptUpload(thisMsg);
				hp.setCursor(null);
			}
			else if (opcode.equalsIgnoreCase("REP_records_marked")) {
				// examine stored queue of mark requests
				if (m_RecordMarks.isEmpty()) {
					Log.warning("REP_records_marked received from DHT with no "
									+ "extant mark requests");
					hp.setCursor(null);
					return;
				}
				Iterator<DHTMessage> recIt = m_RecordMarks.iterator();
				while (recIt.hasNext()) {
					DHTMessage lastMark = recIt.next();
					if (thisMsg.getPacketID().equals(lastMark.getPacketID())) {
						// remove the records referenced in this reply from curr list
						ArrayList<Integer> marklist = lastMark.getMarkList();
						for (Integer recId : marklist) {
							m_DHTrecords.remove(recId);
						}
						recIt.remove();

						// have the WTabManager tell the HousekeepingPane about it
						tm.queueTransition(new Integer(WTabManager.M_WORK_DHT_REGEN));
						new Thread(this).start();
						break;
					}
					else {
						Log.warning("Reply ID " + lastMark.getPacketID() + " received"
										+ " out of order, expected "
										+ thisMsg.getPacketID());
					}
				}
				hp.setCursor(null);
			}
			else {
				Log.error("Unrecognized DHT reply opcode " + opcode + ", XML: "
							+ packet.toXML());
				tm.showError("Bad DHT reply opcode: " + opcode);
			}
		}
		else {
			Log.error("Non-DHTMessage packet received, with XML: "
						+ packet.toXML());
		}

		// in any event, remove ourselves as listener
		removeFilter(packet.getPacketID());
	}

	/**
	 * obtain the list of records most recently retrieved from the DHT
	 */
	public ConcurrentHashMap<Integer, String> getDHTrecs() {
		return m_DHTrecords;
	}

	/**
	 * record a mark request, pending reply
	 * @param request the record mark request sent to the DHT
	 */
	public void recordMarkReq(DHTMessage request) {
		if (request != null
			 && request.getOpcode().equalsIgnoreCase("REQ_mark_records"))
		{
			m_RecordMarks.push(request);
		}
	}

	/**
	 * get the existing DHT connection, if any
	 * @return the connection, or null if there isn't a live one
	 */
	public XMPPConnection getDHTConnection() {
		if (m_DHTConn != null) {
			if (!m_DHTConn.isConnected()) {
				m_DHTConn.disconnect();
				m_DHTConn = null;
			}
		}
		return m_DHTConn;
	}

	/**
	 * record the connection to the DHT, so we can close it later
	 * @param conn the connection (must be currently live)
	 */
	public void setDHTConnection(XMPPConnection conn) {
		if (conn.isConnected()) {
			if (m_DHTConn != null) {
				// blow away old one first
				m_DHTConn.disconnect();
			}
			m_DHTConn = conn;
		}
	}

	/**
	 * close the connection to the DHT
	 */
	public void closeDHTConnection() {
		if (m_DHTConn != null) {
			m_DHTConn.disconnect();
			m_DHTConn = null;
		}
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this message listener thread by Swing.  Also, a single thread is used for
	 * handling all listener events, so we don't want it to block for very long.
	 */
	public void run() {
		// obtain UI tab manager from plugin and do the queued state transition
		m_PluginPtr.getWalletTabManager().makeTransition();
	}

	/**
	 * method to add a filter to the list of packets we're accepting replies for
	 * @param ID the packet ID we're expecting in a reply
	 * @param filter the filter to add, which watches for the ID
	 */
	public void addFilter(String ID, PacketIDFilter filter) {
		if (ID == null || ID.isEmpty() || filter == null
			|| m_FilterMap.containsKey(ID))
		{
			// redundant
			return;
		}
		m_FilterMap.put(ID, filter);
		resetConnFilter();
	}

	/**
	 * method to remove a filter from the list of replies we're listening for
	 * @param ID the ID of the packet filter to remove
	 */
	public void removeFilter(String ID) {
		if (ID == null || ID.isEmpty() || !m_FilterMap.containsKey(ID)) {
			// nothing to do
			return;
		}
		m_FilterMap.remove(ID);
		resetConnFilter();
	}

	/**
	 * private utility method to rebuild and reset the filter on this connection,
	 * ORing together all filters extant in the map
	 */
	private synchronized void resetConnFilter() {
		// rebuild filter of all outstanding requests
		OrFilter of = new OrFilter();
		for (String packetId : m_FilterMap.keySet()) {
			PacketIDFilter filt = m_FilterMap.get(packetId);
			of.addFilter(filt);
		}
		m_DHTConn.removePacketListener(this);

		// add new listener if map is non-empty
		if (!m_FilterMap.isEmpty()) {
			m_DHTConn.addPacketListener(this, of);
		}
	}

	// END methods
}
