/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	12-07-08/kbw/ new
 *
 * purpose---
 * 	class to manipulate messages sent from/to a DHT server
 *
 * notes---
 * 	see underlying message definitions in vsc_dht.x
 */

package org.silentvault.client;

import org.jivesoftware.smack.packet.IQ;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;


/**
 * Holds all the pieces of a message which is part of the
 * jabber:iq:voucher-safe#vsc-dht protocol between a client and a DHT
 * using the element vsc_dht.
 * The class also knows how to emit itself as XML.
 * It is used as a utility class to assist in recording
 * incoming reply messages and forming correct get/set stanzas.
 *
 * @author Kevin Wilkerson
 */
public final class DHTMessage extends IQ {
	// BEGIN data members

	/**
	 * The opcode of this request (begins with REQ_) or reply (begins with REP_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String				m_opcode;

	/**
	 * errcode is used when the IQ.Type is error.
	 * It represents the nature of the error.  For ease of use and readability,
	 * standard HTTP return values are used.  Thus 4xx represents a fatal error
	 * while 5xx represents a temporary error.
	 */
	private int					m_errcode;

	/**
	 * errmsg is used when the IQ.Type is error.
	 * It represents an optional textual elaboration of the problem, to be
	 * used for logging or exception generation purposes.
	 */
	private String				m_errmsg;

	/**
	 * recipient_hash is used with all REQ_ opcodes.
	 * It represents the hash of the VS# assigned by the VP, which is the key
	 * under which data will be recorded in the DHT's database.
	 */
	private String				m_recipient_hash;

	/**
	 * secret_hash is used with opcodes REQ_get_records and REQ_mark_records.
	 * It represents a hash value known only to the VSC, used to distinguish
	 * which records to retrieve or mark.
	 */
	private String				m_secret_hash;

	/**
	 * record_list is used with opcode REP_records.
	 * It represents the list of records returned by the DHT.  Each <record/>
	 * consists of an ID attribute and a base64-encoded encrypted string.
	 * The ID attribute is made into an Integer and used as the key of the
	 * Hashtable; the encoded String becomes the associated value.
	 */
	private Hashtable<Integer, String>	m_Records;

	/**
	 * receipt_blob is used with opcode REQ_store_receipt.
	 * It represents the encrypted and base64-encoded data blob (a receipt)
	 * which will be stored on the DHT.
	 */
	private String				m_receipt_blob;

	/**
	 * token is used with opcode REQ_store_receipt.
	 * It represents the XML representation of a XMLToken object, which is
	 * offered as payment to the DHT for storing the indicated data.
	 * The value will be as returned by XMLToken.toXML(), including the <token>
	 * and </token> delimiters.
	 */
	private String				m_token_xml;

	/**
	 * mark_list is used with opcode REQ_mark_records.
	 * It represents the ID numbers of records previously returned which the
	 * client now wishes to mark as read.
	 */
	private ArrayList<Integer>	m_MarkList;

	// END data members


	// BEGIN methods

	/**
	 * base constructor
	 */
	public DHTMessage() {
		super();
		m_opcode = m_errmsg = m_recipient_hash = m_secret_hash = "";
		m_receipt_blob = m_token_xml = "";
		m_Records = new Hashtable<Integer, String>();
		m_MarkList = new ArrayList<Integer>();
	}


	// get accessor methods
	/**
	 * obtain the current opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_opcode; }

	/**
	 * obtain current error code, valid only when the IQ.Type = error
	 * @return the error
	 */
	public int getErrcode() { return m_errcode; }

	/**
	 * obtain the textual representation of the error, valid only when the
	 * IQ.Type = error
	 * @return the error message
	 */
	public String getErrmsg() { return m_errmsg; }

	/**
	 * get the recipient hash (valid for all REQ_ opcodes)
	 * @return the hash, a base64-encoded SHA1 160-bit hash of the VS#
	 */
	public String getRecipient_hash() { return m_recipient_hash; }

	/**
	 * get the secret hash (valid for opcodes REQ_get_records and
	 * REQ_mark_records)
	 * @return the hash, a base64-encoded SHA1 160-bit hash of a VSC secret
	 */
	public String getSecret_hash() { return m_secret_hash; }

	/**
	 * get the record list hash table (valid when opcode = REP_records)
	 * @return the list of records
	 */
	public Hashtable<Integer, String> getRecordList() { return m_Records; }

	/**
	 * get the receipt blob (valid when opcode = REQ_store_receipt)
	 * @return the receipt, a base64-encoded and encrypted string
	 */
	public String getReceipt_blob() { return m_receipt_blob; }

	/**
	 * get the XML representation of the payment token (valid when opcode =
	 * REQ_store_receipt)
	 * @return the token, in XML form
	 */
	public String getToken() { return m_token_xml; }

	/**
	 * get the IDs of records to be marked (valid when opcode = REQ_mark_records)
	 * @return the array of record IDs to mark as read
	 */
	public ArrayList<Integer> getMarkList() { return m_MarkList; }


	// set access methods
	/**
	 * set the opcode value
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_opcode = new String(opcode);
		}
	}

	/**
	 * set the error code value (needed when IQ.Type = error)
	 * @param ecode the error code
	 */
	public void setErrcode(int ecode) { m_errcode = ecode; }

	/**
	 * set the error message value (needed when IQ.Type = error)
	 * @param emsg the error message
	 */
	public void setErrmsg(String emsg) {
		if (emsg != null) {
			m_errmsg = new String(emsg);
		}
	}

	/**
	 * set the recipient hash (always needed for all REQ_ opcodes)
	 * @param hash the hash value, base64-encoded
	 */
	public void setRecipient_hash(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_recipient_hash = new String(hash);
		}
	}

	/**
	 * set the secret hash (needed for opcodes REQ_get_records and
	 * REQ_mark_records)
	 * @param hash the hash value, base64-encoded
	 */
	public void setSecret_hash(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_secret_hash = new String(hash);
		}
	}

	/**
	 * add a record to the list of records which have been retrieved
	 * (needed when opcode = REP_records)
	 * @param id the record ID number to use as the key
	 * @param rec the encoded/encrypted record text
	 */
	public void addRecord(int id, String rec) {
		if (rec != null && !rec.isEmpty() && id > 0) {
			Integer key = new Integer(id);
			m_Records.put(key, rec);
		}
	}

	/**
	 * set the receipt value (needed for opcode REQ_store_receipt)
	 * @param receipt the base64-encoded, encrypted text of the receipt record
	 */
	public void setReceipt_blob(String receipt) {
		if (receipt != null && !receipt.isEmpty()) {
			m_receipt_blob = new String(receipt);
		}
	}

	/**
	 * set the token XML value (needed when opcode = REQ_store_receipt)
	 * @param token the token to be offered as payment
	 */
	public void setToken(XMLToken token) {
		if (token != null) {
			m_token_xml = token.toXML();
		}
	}

	/**
	 * add a record to the list of records to be marked (needed when opcode =
	 * REQ_mark_records)
	 * @param id the ID of the record to put on the mark list
	 */
	public void markRecord(int id) {
		if (id > 0) {
			Integer rec = new Integer(id);
			m_MarkList.add(rec);
		}
	}


	// method required to extend the IQ class
	/**
	 * provide the child element as XML
	 * @return the extended child element as a XML string
	 */
	public String getChildElementXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<vsc_dht xmlns=\"jabber:iq:voucher-safe#vsc-dht\">");
		xml.append("<opcode>" + m_opcode + "</opcode>");

		// if this is an error stanza, supply code and message
		if (getType() == IQ.Type.ERROR) {
			xml.append("<errcode>" + m_errcode + "</errcode>");
			xml.append("<errmsg>" + m_errmsg + "</errmsg>");
		}
		else {
			// based on the value of the opcode, emit additional fields

			if (m_opcode.equalsIgnoreCase("REQ_get_records")) {
				xml.append("<recipient_hash>" + m_recipient_hash);
				xml.append("</recipient_hash>");
				xml.append("<secret_hash>" + m_secret_hash + "</secret_hash>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_records")) {
				xml.append("<record_list>");

				// emit all records in the list, with their ids as attributes
				if (m_Records.size() > 0) {
					Enumeration<Integer> recIds = m_Records.keys();
					while (recIds.hasMoreElements()) {
						Integer id = recIds.nextElement();
						String rec = m_Records.get(id);
						xml.append("<record id=\"" + id + "\">");
						xml.append(rec + "</record>");
					}
				}
				xml.append("</record_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_store_receipt")) {
				xml.append("<recipient_hash>" + m_recipient_hash);
				xml.append("</recipient_hash>");
				xml.append("<receipt_blob>" + m_receipt_blob + "</receipt_blob>");
				xml.append(m_token_xml);
			}
			else if (m_opcode.equalsIgnoreCase("REP_receipt_stored")) {
				// no additional fields
			}
			else if (m_opcode.equalsIgnoreCase("REQ_mark_records")) {
				xml.append("<recipient_hash>" + m_recipient_hash);
				xml.append("</recipient_hash>");
				xml.append("<secret_hash>" + m_secret_hash + "</secret_hash>");
				xml.append("<mark_list>");

				// emit all record ids on the list
				if (m_MarkList.size() > 0) {
					Object[] ids = m_MarkList.toArray();
					for (int idx = 0; idx < ids.length; idx++) {
						Integer id = (Integer) ids[idx];
						xml.append("<id>" + id + "</id>");
					}
				}
				xml.append("</mark_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_records_marked")) {
				// no additional fields
			}
			else {
				Log.error("Improper opcode " + m_opcode);
			}
		}

		xml.append("</vsc_dht>");
		return xml.toString();
	}

	// END methods
}
