/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	12-07-08/kbw/ new
 *
 * purpose---
 * 	Parse {@link DHTMessage} packets.  Parses the <tt>vsc_dht</tt>
 * 	sub-document and creates an instance of {@link DHTMessage}.
 *
 * notes---
 * 	see underlying message definitions in vsc_dht.x
 */

package org.silentvault.client;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.IQProvider;


/**
 * Implements the IQProvider interface to receive and parse incoming messages
 * from a DHT (Distributed Hash Table) according to opcode in the vsc_dht
 * element of the jabber:iq:voucher-safe#vsc-dht namespace.  Creates a
 * {@link DHTMessage} to hold the parsed message.
 *
 * @author Kevin Wilkerson
 */
public class DHTMessageProvider implements IQProvider {
	// members
	private static final String	M_EndTag = "vsc_dht";
	
	// methods
	/**
	 * constructor
	 */
	public DHTMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		DHTMessage dhtMsg = new DHTMessage();
		// NB: these values should be vsc_dht and jabber:iq:voucher-safe#vsc-dht
		dhtMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		dhtMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		String opcode = "";
		int eventType = parser.next();
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		dhtMsg.setOpcode(opcode);

		// Now based on the opcode, pull out the other data.  (replies only)

		boolean done = false;
		if (opcode.equalsIgnoreCase("REP_records")) {
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						dhtMsg.setErrcode(Integer.parseInt(parser.nextText()));
						dhtMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						dhtMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("record_list")) {
						// look for nested <record id="N">...</record> elements
						parseRecordList(parser, dhtMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_records_marked")
					|| opcode.equalsIgnoreCase("REP_receipt_stored"))
		{
			// no special fields for these, beyond the opcode
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						dhtMsg.setErrcode(Integer.parseInt(parser.nextText()));
						dhtMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						dhtMsg.setErrmsg(parser.nextText());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return dhtMsg;
	}

	/**
	 * private sub-method to parse the supplied list of DHT records
	 * @param parser the parser, positioned at the start of the list
	 * @param dhtMsg the message being built
	 */
	private void parseRecordList(XmlPullParser parser, DHTMessage dhtMsg)
		throws Exception
	{
		if (parser == null || dhtMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("record")) {
					// obtain the record Id N, from attribute id="N"
					if (parser.getAttributeCount() < 1) {
						throw new Exception("Record id missing on DHT <record>");
					}
					String idTxt = parser.getAttributeValue(0);
					if (idTxt.isEmpty()) {
						throw new Exception("Record id missing on DHT <record>");
					}
					Integer id = null;
					try {
						id = Integer.parseInt(idTxt);
					}
					catch (NumberFormatException nfe) {
						throw new Exception("Non-numeric record id on DHT <record>",
													nfe);
					}
					String rec = parser.nextText();
					if (rec.isEmpty()) {
						throw new Exception("Empty DHT <record/>");
					}

					// add to list
					dhtMsg.addRecord(id, rec);
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("record_list")) {
					done = true;
				}
			}
		}
	}
}
