/*
 *	last modified---
 *		11-09-13/jet/ disable error CCs unless on test network
 * 	06-05-13/jet/ SilentVault client version
 *		03-14-10/kbw/ report error messages via chat to listener
 *		02-25-10/kbw/ new
 *
 *	purpose---
 *		class for writing messages to VSC error log
 */

package org.silentvault.client;

import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.XMPPConnection;

import java.util.*;
import java.io.*;


/**
 * This class mimics the behavior of org.jivesoftware.spark.util.Log, for use
 * with the stand-alone client.
 *
 * @author Kevin Wilkerson
 */
public class Log {
	// global variables -- for all instances
	private static File					m_Logfile = null;
	private static FileOutputStream	m_LogStrm = null;
	private static PrintStream			m_LogOut = null;
	private static boolean				m_Init = false;

	/**
	 * the owning client object
	 */
	private static WalletClient		m_Client;

	/**
	 * the recipient of error report messages
	 */
	private static String				m_ErrorDest;

	/**
	 * constructor
	 * @param lfilename the log file filename to write to
	 * @param client the wallet client which owns us
	 */
	public Log(String lfilename, WalletClient client) {
		m_Logfile = new File(lfilename);
		m_Client = client;
		m_ErrorDest = "";
	}

	/**
	 * set the error message recipient
	 * @param recip the JID where we'll CC our logged errors
	 */
	public void setErrRecip(String recip) {
		if (recip == null || recip.isEmpty()) {
			return;
		}
		m_ErrorDest = new String(recip);
	}

	// get accessor methods
	public File getFile() { return m_Logfile; }

	/**
	 * initializer -- this reassigns syserr and sysout for the caller
	 * @return true on success, false otherwise
	 */
	public boolean Init() {
		if (m_Logfile == null)
			return false;
		try {
			// create empty file if it doesn't exist
			if (!m_Logfile.exists()) {
				m_Logfile.createNewFile();
			}
			m_LogStrm = new FileOutputStream(m_Logfile.getPath(), true);
			m_LogOut = new PrintStream(m_LogStrm);
			System.setOut(m_LogOut);
			System.setErr(m_LogOut);
		}
		catch (FileNotFoundException fnfe)
		{
			System.err.println("Unable to create logfile for writing: "
									 + m_Logfile);
			return false;
		}
		catch (IOException ioe)
		{
			System.err.println("Unable to initialize logfile: " + m_Logfile);
			return false;
		}
		m_Init = true;
		return true;
	}

	/**
	 * de-initializer
	 */
	public void deInit() {
		m_LogOut.close();
		m_Init = false;
	}

	/**
	 * debug method
	 * @param message the debug message
	 */
	public static void debug(String message) {
		if (message == null || message.isEmpty())
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer dOut = new StringBuffer(1024);
		dOut.append("DEBUG " + dnow.toString() + ":\n\t");
		dOut.append(message);
		System.out.println(dOut.toString());
	}

	/**
	 * error method
	 * @param message the error message
	 */
	public static void error(String message) {
		if (message == null || message.isEmpty())
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer eOut = new StringBuffer(1024);
		eOut.append("ERROR " + dnow.toString() + ":\n\t");
		eOut.append(message);
		System.err.println(eOut.toString());
		if (m_Client.isTestNetwork()) {
			sendErrorCC(eOut.toString());
		}
	}

	/**
	 * error method
	 * @param message the error message
	 * @param ex exception related to the message
	 */
	public static void error(String message, Throwable ex) {
		if (message == null || message.isEmpty() || ex == null)
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer eOut = new StringBuffer(1024);
		eOut.append("ERROR " + dnow.toString() + ":\n\t");
		eOut.append(message);
		eOut.append("\nCaused by: ");
		eOut.append(ex.getMessage() + "\n");
		System.err.println(eOut.toString());
		ex.printStackTrace(m_LogOut);
		if (m_Client.isTestNetwork()) {
			sendErrorCC(eOut.toString());
		}
	}

	/**
	 * error method
	 * @param ex exception describing the error
	 */
	public static void error(Throwable ex) {
		if (ex == null)
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer eOut = new StringBuffer(256);
		eOut.append("ERROR " + dnow.toString() + ":\n\t");
		eOut.append(ex.getMessage() + "\n");
		System.err.println(eOut.toString());
		ex.printStackTrace(m_LogOut);
		if (m_Client.isTestNetwork()) {
			sendErrorCC(eOut.toString());
		}
	}

	/**
	 * helper method to CC logged errors to a network JID on the server, to
	 * facilitate QA
	 * @param errTxt the text of the error message to send
	 */
	private static void sendErrorCC(String errTxt) {
		// copy the error message to the designated recipient
		if (!m_ErrorDest.isEmpty()) {
			XMPPConnection conn = m_Client.getOFSConnection();
			if (conn != null) {
				Message errMsg = new Message(m_ErrorDest);
				// if we're logged in, put the wallet Id in the subject and body
				String text = null;
				String vsNum = m_Client.getLoginSecrets().getVSnumber();
				if (vsNum.isEmpty()) {
					errMsg.setSubject("Error report outside of wallet context");
					text = errTxt;
				}
				else {
					errMsg.setSubject("Error report from wallet Id " + vsNum
											+ " operation");
					text = "Wallet Id " + vsNum + ": " + errTxt;
				}
				errMsg.setBody(text);
				conn.sendPacket(errMsg);
			}
		}
	}

	/**
	 * warning method
	 * @param message the warning message
	 */
	public static void warning(String message) {
		if (message == null || message.isEmpty())
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer wOut = new StringBuffer(1024);
		wOut.append("WARNING " + dnow.toString() + ":\n\t");
		wOut.append(message);
		System.out.println(wOut.toString());
	}

	/**
	 * warning method
	 * @param message the error message
	 * @param ex exception related to the message
	 */
	public static void warning(String message, Throwable ex) {
		if (message == null || message.isEmpty())
			return;
		if (!m_Init)
			return;

		Date dnow = new Date();
		StringBuffer wOut = new StringBuffer(1024);
		wOut.append("WARNING " + dnow.toString() + ":\n\t");
		wOut.append(message);
		wOut.append("\nCaused by: ");
		wOut.append(ex.getMessage() + "\n");
		System.out.println(wOut.toString());
		ex.printStackTrace(m_LogOut);
	}
}
