/*
 * last modified---
 * 	10-11-13/jet/ add m_key_server_domain and methods
 * 	06-05-13/jet/ SilentVault client version
 * 	01-16-13/kbw/ add REQ_redeem_voucher / REP_voucher_redeemed
 * 	03-14-11/kbw/ add REQ_lookup_prices / REP_price_data; prioritize opcodes
 * 	04-28-10/kbw/ add REQ_delete_voucher / REP_voucher_deleted
 * 	03-22-10/kbw/ add <voucher_publisher/> to REQ_initiate_login and
 * 					  REQ_lookup_recovery
 * 	09-25-09/kbw/ add <voucher_publisher/> to REQ_discover_net
 * 	07-12-09/kbw/ add <vs_prefix/> to REQ_init_key
 * 	07-05-09/kbw/ actually remove <confirmDetails/> from REP_payment_sent
 * 	05-22-09/kbw/ add REQ_recovery_ques / REP_questions, and
 * 					  REQ_lookup_recovery / REP_recovery_blob
 * 	04-08-09/kbw/ add m_longphrase_len
 * 	02-24-09/kbw/ support <syncPending/> element in replies where needed
 * 	02-16-09/kbw/ add payment-related messages
 * 	11-24-08/kbw/ add support for REQ_close_safe / REP_safe_closed
 * 	10-29-08/kbw/ switch to <certificate/> in REP_safe_data
 * 	09-19-08/kbw/ add <hashkey/> to a REQ_perform_login message, so it can
 * 					  be associated with the previous REQ_initiate_login request
 * 	09-11-08/kbw/ new
 *
 * purpose---
 * 	class to manipulate VSC messages sent to/from the OFS
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs.x
 * 	If you modify this file, make the equivalent change in VSCMessage in
 * 	the Openfire plugin and parallel clients.
 */

package org.silentvault.client;

import org.jivesoftware.smack.packet.IQ;

import java.util.*;
import java.net.InetSocketAddress;


/**
 * Holds all the pieces of a message which is part of the
 * jabber:iq:voucher-safe#vsc-ofs protocol between a VSC and an OFS
 * using the element vsc_ofs.  The class also knows how to emit itself as XML.
 * It is used as a utility class to assist in fabricating outgoing messages to
 * the OFS and processing result/error stanza replies.
 *
 * @author Kevin Wilkerson
 */
public final class OFSMessage extends IQ {
	// BEGIN data members
	
	/**
	 * The opcode of this request (begins with REQ_) or reply (begins with REP_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String				m_opcode;

	/**
	 * errcode is used when the IQ.Type is error.
	 * It represents the nature of the error.  For ease of use and readability,
	 * standard HTTP return values are used.  Thus 4xx represents a fatal error
	 * while 5xx represents a temporary error.
	 */
	private int					m_errcode;

	/**
	 * errmsg is used when the IQ.Type is error.
	 * It represents an optional textual elaboration of the problem, to be
	 * used for logging or exception generation purposes.
	 */
	private String				m_errmsg;

	/**
	 * hashkey is used with opcodes REQ_initiate_login and REQ_perform_login.
	 * It represents the hash of the voucher safe number and associated PIN.
	 */
	private String				m_hashkey;

	/**
	 * long_phrase_len is used with opcode REP_long-phrase_challenge.
	 * It represents the total character length of the long-phrase, to aid the
	 * VSC in UI display.
	 */
	private int					m_longphrase_len;

	/**
	 * long_phrase_index1 is used with opcode REP_long-phrase_challenge.
	 * It represents a 1-based index into the safe's long-phrase string.
	 */
	private int					m_longphrase_idx1;

	/**
	 * long_phrase_index2 is used with opcode REP_long-phrase_challenge.
	 * It represents a 1-based index into the safe's long-phrase string.
	 */
	private int					m_longphrase_idx2;

	/**
	 * long_phrase_index3 is used with opcode REP_long-phrase_challenge.
	 * It represents a 1-based index into the safe's long-phrase string.
	 */
	private int					m_longphrase_idx3;

	/**
	 * long_phrase_char1 is used with opcode REQ_perform_login.
	 * It represents a character extracted from the safe's long-phrase string,
	 * at index m_longphrase_idx1.
	 */
	private byte				m_longphrase_char1;

	/**
	 * long_phrase_char2 is used with opcode REQ_perform_login.
	 * It represents a character extracted from the safe's long-phrase string,
	 * at index m_longphrase_idx2.
	 */
	private byte				m_longphrase_char2;

	/**
	 * long_phrase_char3 is used with opcode REQ_perform_login.
	 * It represents a character extracted from the safe's long-phrase string,
	 * at index m_longphrase_idx3.
	 */
	private byte				m_longphrase_char3;

	/**
	 * loginblob is used with opcode REP_login-blob.
	 * It represents the base64-encoded login blob for the safe, encrypted to
	 * the safe's passphrase.
	 */
	private String				m_loginblob;

	/**
	 * voucher-publisher-list is a list of ids of voucher publishers, in the
	 * form <voucher_publisher>vpid@voucher-safe.org</voucher_publisher>.
	 */
	private final List<String>		m_VoucherPublishers;

	/**
	 * voucher_publisher is the selected VP, used with the opcodes
	 * REQ_select_publisher, REQ_init_key, REQ_init_safe, REQ_close_safe,
	 * REQ_send_payment, REQ_pickup_payment, REQ_revalidate, REQ_reclaim_payment,
	 * REQ_delete_spent_tokens, REQ_purchase_tokens, REQ_recovery_ques,
	 * REQ_discover_net, REQ_lookup_recovery, REQ_initiate_login,
	 * REQ_delete_voucher, and REQ_redeem_voucher
	 */
	private String				m_SelectedVP;

	/**
	 * algorithm is used with opcode REP_key_params, and represents the
	 * algorithm to be used in generating a keypair for the safe (DSA, RSA, etc.)
	 */
	private String				m_algorithm;

	/**
	 * min_length is used with opcode REP_key_params, and represents the
	 * minimum key length in bits to be used generating a keypair for the safe
	 */
	private int					m_min_length;

	/**
	 * public_key is used with opcode REQ_init_key, and represents the
	 * public key generated by the client for use with the safe
	 */
	private String				m_public_key;

	/**
	 * vs_prefix is used with opcode REQ_init_key, and represents the user's
	 * selected identifying prefix for the VS number being created.  To this will
	 * be appended the system base suffix (if any), plus a sequence number t
	 * guarantee uniqueness.
	 */
	private String				m_vs_prefix;

	/**
	 * certificate is used with opcode REP_safe_data, and represents
	 * a cert made from the safe's generated public key, now signed by the VP
	 */
	private String				m_certificate;

	/**
	 * key_server_domain is used with opcode REP_safe_data, and represents the
	 * XMPP domain of the PKS associated with the coordinates in m_key_server
	 */
	private String				m_key_server_domain;

	/**
	 * key_server_addr is used with opcode REP_safe_data, and represents
	 * the socket address of the key server where the safe's signed certificate
	 * has been stored, and where the VP's public key can be looked up
	 */
	private ArrayList<InetSocketAddress>	m_key_servers;

	/**
	 * cap_blob is used with opcode REP_safe_data, and represents the
	 * capabilities blob for the safe, encrypted with the public key
	 */
	private String				m_cap_blob;

	/**
	 * pin_length is used with opcode REP_safe_data, and represents the
	 * required length in digits of the PIN number to be generated
	 */
	private short				m_pin_length;

	/**
	 * recovery_ques_1 is used with opcode REP_safe_data, and represents the
	 * first question which the user must answer to build the login details
	 * recovery block
	 */
	private String				m_recovery_ques1;

	/**
	 * recovery_ques_2 is used with opcode REP_safe_data, and represents the
	 * second question which the user must answer to build the login details
	 * recovery block
	 */
	private String				m_recovery_ques2;

	/**
	 * recovery_ques_3 is used with opcode REP_safe_data, and represents the
	 * third question which the user must answer to build the login details
	 * recovery block
	 */
	private String				m_recovery_ques3;

	/**
	 * recovery_ques_4 is used with opcode REP_safe_data, and represents the
	 * fourth question which the user must answer to build the login details
	 * recovery block
	 */
	private String				m_recovery_ques4;

	/**
	 * recovery_ques_5 is used with opcode REP_safe_data, and represents the
	 * fifth question which the user must answer to build the login details
	 * recovery block
	 */
	private String				m_recovery_ques5;

	/**
	 * details_blob is used with opcode REQ_init_safe, and represents the
	 * login details blob for the safe, encrypted using the passphrase
	 */
	private String				m_details_blob;

	/**
	 * status_msg is used with opcodes REP_safe_init and REP_safe_closed,
	 * and represents a result message forwarded from the VP
	 */
	private String				m_status_msg;

	/**
	 * recovery_index is used with opcode REQ_lookup_recovery, and represents
	 * the index hash at which the recovery details sought should be stored
	 * (ie hash(hash(VS#+5 recovery answers))
	 */
	private String				m_recovery_index;

	/**
	 * recovery_blob is used with opcode REP_recovery_blob, and represents
	 * the recovery data for the VS, which is encrypted with 
	 * hash(VS#+5 recovery answers)
	 */
	private String				m_recovery_blob;

	/**
	 * network_conf_blob is used with opcode REP_network_details, and
	 * represents a list of current network endpoints, signed by the Issuer's key
	 */
	private String				m_network_conf_blob;

	/**
	 * closure_block is used with opcode REQ_close_safe, and represents a
	 * detail specification encrypted with the VP's public key
	 */
	private String				m_closure_block;

	/**
	 * signature is used with opcode REQ_close_safe, and represents the
	 * private key signature of the VS# being closed, on the unencrypted
	 * <closure_block/>
	 */
	private String				m_signature;

	/**
	 * oldVS_rwcap is used with opcode REQ_close_safe, and represents the
	 * voucher folder SDS read-write cap for the safe being closed,
	 * encrypted with the VP's public key (hence opaque to the OFS).
	 */
	private String				m_old_VS_rwcap;

	/**
	 * newVS_rwcap is used with opcode REQ_close_safe, and represents the
	 * voucher folder SDS read-write cap for the safe receiving the value,
	 * encrypted with the VP's public key (hence opaque to the OFS).
	 */
	private String				m_new_VS_rwcap;

	/**
	 * payDetails is used with opcodes REQ_send_payment and
	 * REQ_reclaim_payment, and represents the details of the payment.
	 * It is encrypted with the public key of the VP, and base64-encoded.
	 */
	private String				m_PayDetails;

	/**
	 * A voucherRequest is used with these opcodes: REQ_send_payment,
	 * REQ_pickup_payment, REQ_revalidate, REQ_reclaim_payment, and
	 * REQ_purchase_tokens.  It describes the input vouchers and output values,
	 * along with folder hashes and caps, etc.  (See {@link VoucherRequest}.)
	 */
	private VoucherRequest	m_VouchReq;

	/**
	 * A confirmation element is used with opcodes REP_payment_confirmed,
	 * REP_payment_reverted, and REP_voucher_redeemed.  It describes a new
	 * voucher created for this VS, is signed by the VP and encrypted with the
	 * our public key.
	 */
	private String				m_Confirmation;

	/**
	 * validation is used with opcode REP_validation.  It contains a list
	 * of one or more vouchers issued to the VS, signed by the VP, and encrypted
	 * with our public key.
	 */
	private String				m_Validation;

	/**
	 * purgeDetails is used with opcode REQ_delete_spent_tokens, and contains
	 * a signed and encrypted list of tokens to be reported to the VP as having
	 * been spent to a DHT or SDS node (basically, a token delete request).
	 */
	private String				m_PurgeDetails;

	/**
	 * purged is used with opcode REP_tokens_purged, and contains a
	 * confirmation list signed by the VP of the tokens which were actually
	 * purged.  It is encrypted with the VS's pubkey, thus opaque to us.
	 */
	private String				m_Purged;

	/**
	 * purchased is used with opcode REP_tokens_purchased.  It contains a
	 * list of new tokens issued to the VS, along with a voucher for any change,
	 * signed by the VP and encrypted with our public key.
	 */
	private String				m_Purchased;

	/**
	 * redeemDetails is used with opcode REQ_redeem_voucher.  It contains a
	 * signed and encrypted specification of an expired voucher which the VP
	 * is requested to return to the Issuer for disposition.  It is encrypted
	 * with the VP's pubkey.
	 */
	private String				m_RedeemDetails;

	/**
	 * deleteDetails is used with opcode REQ_delete_voucher.  It contains a
	 * signed and encrypted specification of an expired voucher which the VP
	 * is requested to delete off the SDS.
	 */
	private String				m_DeleteDetails;

	/**
	 * priceData is used with opcode REP_price_data.  It contains an XML body
	 * which can be parsed into a list of gold or other asset prices in various
	 * national fiat currencies.  (E.g. from dgcsc.org/goldprices.xml).  To
	 * prevent transport problems, this XML data is base64-encoded.
	 */
	private String				m_price_data;

	/**
	 * syncPending is used with many reply (REP_*) opcodes.  It is normally
	 * set to false, but when true indicates that the VP had to pend one or more
	 * operations with the SDS or DHT.  In such cases, we must not expect the
	 * immediate consistency of records retrieved from the SDS.  This field does
	 * not indicate a permanent failure, merely a soft (temporary delay) failure.
	 */
	private boolean			m_SyncPending;

	// END data members


	// BEGIN methods
	
	/**
	 * constructor
	 */
	public OFSMessage() {
		super();
		m_VoucherPublishers = new ArrayList<String>();
		m_opcode = m_errmsg = m_hashkey = m_loginblob = m_status_msg = "";
		m_SelectedVP = m_algorithm = m_public_key = m_certificate = "";
		m_cap_blob = m_recovery_ques1 = m_recovery_ques2 = m_recovery_ques3 = "";
		m_recovery_ques4 = m_recovery_ques5 = m_details_blob = "";
		m_network_conf_blob = m_closure_block = m_signature = m_Purchased = "";
		m_old_VS_rwcap = m_new_VS_rwcap = m_PayDetails = m_vs_prefix = "";
		m_Confirmation = m_Validation = m_PurgeDetails = m_Purged = "";
		m_recovery_index = m_recovery_blob = m_DeleteDetails = m_price_data = "";
		m_RedeemDetails = m_key_server_domain = "";
		m_key_servers = new ArrayList<InetSocketAddress>();
	}

	// get accesssor methods
	/**
	 * obtain the current opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_opcode; }

	/**
	 * obtain current error code, valid only when the IQ.Type = error
	 * @return the error
	 */
	public int getErrcode() { return m_errcode; }

	/**
	 * obtain the textual representation of the error, valid only when the
	 * IQ.Type = error
	 * @return the error message
	 */
	public String getErrmsg() { return m_errmsg; }

	/**
	 * get the hashkey (valid when opcode = REQ_initiate_login
	 * or REQ_perform_login)
	 * @return the hashkey
	 */
	public String getHashkey() { return m_hashkey; }

	/**
	 * get the long-phrase length
	 * (valid when opcode = REP_long-phrase_challenge)
	 * @return the length
	 */
	public int getLong_phrase_len() { return m_longphrase_len; }

	/**
	 * get the first long-phrase index
	 * (valid when opcode = REP_long-phrase_challenge)
	 * @return the 1st 1-based index
	 */
	public int getLong_phrase_index1() { return m_longphrase_idx1; }

	/**
	 * get the second long-phrase index
	 * (valid when opcode = REP_long-phrase_challenge)
	 * @return the 2nd 1-based index
	 */
	public int getLong_phrase_index2() { return m_longphrase_idx2; }

	/**
	 * get the third long-phrase index
	 * (valid when opcode = REP_long-phrase_challenge)
	 * @return the 3rd 1-based index
	 */
	public int getLong_phrase_index3() { return m_longphrase_idx3; }

	/**
	 * get the first long-phrase character
	 * (valid when opcode = REQ_perform_login)
	 * @return the long-phrase character at index1
	 */
	public byte getLong_phrase_char1() { return m_longphrase_char1; }

	/**
	 * get the second long-phrase character
	 * (valid when opcode = REQ_perform_login)
	 * @return the long-phrase character at index2
	 */
	public byte getLong_phrase_char2() { return m_longphrase_char2; }

	/**
	 * get the third long-phrase character
	 * (valid when opcode = REQ_perform_login)
	 * @return the long-phrase character at index3
	 */
	public byte getLong_phrase_char3() { return m_longphrase_char3; }

	/**
	 * get the encrypted login blob (valid when opcode = REP_login-blob)
	 * @return the login blob (base64 encoded)
	 */
	public String getLoginblob() { return m_loginblob; }

	/**
	 * get the list of possible voucher publishers
	 * (valid when opcode = REP_publisher_list)
	 * @return the list of VP ids
	 */
	public List<String> getVoucherPublishers() { return m_VoucherPublishers; }

	/**
	 * get the selected VP (valid when opcode = REQ_select_publisher,
	 * REQ_init_key, REQ_init_safe, REQ_close_safe, REQ_send_payment,
	 * REQ_pickup_payment, REQ_revalidate, REQ_reclaim_payment,
	 * REQ_delete_spent_tokens, REQ_purchase_tokens, REQ_recovery_ques,
	 * REQ_discover_net, REQ_lookup_recovery, REQ_initiate_login,
	 * REQ_redeem_voucher, or REQ_delete_voucher)
	 * @return the id of the voucher publisher
	 */
	public String getVoucherPublisher() { return m_SelectedVP; }

	/**
	 * get the algorithm to be used for key generation
	 * (valid when opcode = REP_key_params)
	 * @return the name of the algorithm
	 */
	public String getAlgorithm() { return m_algorithm; }

	/**
	 * get the minimum key length to be used for keypair generation
	 * @return the minimum length, in bits
	 */
	public int getMin_length() { return m_min_length; }

	/**
	 * get the public key (valid when opcode = REQ_init_key)
	 * @return the public key, base64 encoded
	 */
	public String getPublic_key() { return m_public_key; }

	/**
	 * get the user's VS# prefix (valid when opcode = REQ_init_key)
	 * @return the chosen prefix
	 */
	public String getVSprefix() { return m_vs_prefix; }

	/**
	 * get the signed certificate (valid when opcode = REP_safe_data)
	 * @return the signed certificate, base64 encoded
	 */
	public String getCertificate() { return m_certificate; }

	/**
	 * get the XMPP domain name of the key server where VS and VP's keys are
	 * found (valid when opcode = REP_safe_data)
	 */
	public String getKey_server_domain() { return m_key_server_domain; }

	/**
	 * get the socket addresses of the key server where VS and VP's keys are
	 * found (valid when opcode = REP_safe_data)
	 * @return the network address list
	 */
	public ArrayList<InetSocketAddress> getKey_server_addr() {
		return m_key_servers;
	}

	/**
	 * get the safe capabilities blob
	 * (valid when opcode = REP_safe_data)
	 * @return the base64-encoded blob, encrypted to the public key
	 */
	public String getCap_blob() { return m_cap_blob; }

	/**
	 * get the required PIN length (valid when opcode = REP_safe_data)
	 * @return the length, in digits
	 */
	public short getPin_length() { return m_pin_length; }

	/**
	 * get the text of the first login details recovery question
	 * (valid when opcode = REP_safe_data)
	 * @return the question text
	 */
	public String getRecovery_ques_1() { return m_recovery_ques1; }

	/**
	 * get the text of the second login details recovery question
	 * (valid when opcode = REP_safe_data)
	 * @return the question text
	 */
	public String getRecovery_ques_2() { return m_recovery_ques2; }

	/**
	 * get the text of the third login details recovery question
	 * (valid when opcode = REP_safe_data)
	 * @return the question text
	 */
	public String getRecovery_ques_3() { return m_recovery_ques3; }

	/**
	 * get the text of the fourth login details recovery question
	 * (valid when opcode = REP_safe_data)
	 * @return the question text
	 */
	public String getRecovery_ques_4() { return m_recovery_ques4; }

	/**
	 * get the text of the fifth login details recovery question
	 * (valid when opcode = REP_safe_data)
	 * @return the question text
	 */
	public String getRecovery_ques_5() { return m_recovery_ques5; }

	/**
	 * get the value of the login details blob
	 * (valid when opcode = REQ_init_safe)
	 * @return the blob, base64-encoded encrypted to the safe's passphrase
	 */
	public String getDetails_blob() { return m_details_blob; }

	/**
	 * get the status message
	 * (valid when opcode = REP_safe_init or REP_safe_closed)
	 * @return the message text, forwarded from the VP (or SDS)
	 */
	public String getStatus_msg() { return m_status_msg; }

	/**
	 * get the recovery index hash (valid when opcode = REQ_lookup_recovery)
	 * @return the hash, forwarded from the SDS
	 */
	public String getRecovery_index() { return m_recovery_index; }

	/**
	 * get the recovery blob (valid when opcode = REP_recovery_blob)
	 * @return the blob, base64-encoded and encrypted with the hash of the VS#
	 * and the answers to the 5 recovery questions
	 */
	public String getRecovery_blob() { return m_recovery_blob; }

	/**
	 * get the network configuration blob
	 * (valid when opcode = REP_network_details)
	 * @return the blob, base64-encoded and encrypted to the Issuer's pubkey
	 */
	public String getNetwork_conf_blob() { return m_network_conf_blob; }

	/**
	 * get the closure details block (valid when opcode = REQ_close_safe)
	 * @return the block, base64-encoded and encrypted with the VP's pubkey
	 */
	public String getClosure_block() { return m_closure_block; }

	/**
	 * get the private key signature on the clear-text data in the closure_block
	 * (valid when opcode = REQ_close_safe)
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_signature; }

	/**
	 * get the rw cap of the safe being closed
	 * (valid when opcode = REQ_close_safe)
	 * @return the capability, base64-encoded and encrypted with VP's pubkey
	 */
	public String getOldVS_rwcap() { return m_old_VS_rwcap; }

	/**
	 * get the rw cap of the safe receiving the residual value
	 * (valid when opcode = REQ_close_safe)
	 * @return the capability, base64-encoded and encrypted with VP's pubkey
	 */
	public String getNewVS_rwcap() { return m_new_VS_rwcap; }

	/**
	 * get the payment details blob
	 * (valid when opcode is REQ_send_payment or REQ_reclaim_payment)
	 * @return the details, base64-encoded and encrypted with VP's pubkey
	 */
	public String getPayDetails() { return m_PayDetails; }

	/**
	 * get the voucher request input/output data
	 * (valid for opcodes REQ_send_payment, REQ_pickup_payment, REQ_revalidate,
	 * REQ_reclaim_payment, and REQ_purchase_tokens)
	 * @return the {@link VoucherRequest} object
	 */
	public VoucherRequest getVouchReq() { return m_VouchReq; }

	/**
	 * get voucher confirmation details (valid for opcodes REP_payment_confirmed,
	 * REP_voucher_redeemed, and REP_payment_reverted)
	 * @return the confirmation, base64-encoded and encrypted with VS's pubkey
	 */
	public String getConfirmation() { return m_Confirmation; }

	/**
	 * get voucher validation blob (valid for opcode REP_validation)
	 * @return the validation, base64-encoded and encrypted with VS's pubkey
	 */
	public String getValidation() { return m_Validation; }

	/**
	 * get token purging instructions (valid for opcode REQ_delete_spent_tokens)
	 * @return the instructions, base64-encoded and encrypted with VP's pubkey
	 */
	public String getPurgeDetails() { return m_PurgeDetails; }

	/**
	 * get list of tokens actually purged (valid for opcode REP_tokens_purged)
	 * @return the list, base64-encoded and encrypted with our pubkey
	 */
	public String getPurged() { return m_Purged; }

	/**
	 * get list of purchased tokens (valid for opcode REP_tokens_purchased)
	 * @return the list, base64-encoded and encrypted with our pubkey
	 */
	public String getPurchased() { return m_Purchased; }

	/**
	 * get voucher redemption specification (valid for opcode REQ_redeem_voucher)
	 * @return the spec, base64-encoded and encrypted with VP's pubkey
	 */
	public String getRedeemDetails() { return m_RedeemDetails; }

	/**
	 * get voucher deletion specification (valid for opcode REQ_delete_voucher)
	 * @return the spec, base64-encoded and encrypted with VP's pubkey
	 */
	public String getDeleteDetails() { return m_DeleteDetails; }

	/**
	 * get price data body (valid for opcode REP_price_data)
	 * @return the downloaded price data, in base64-encoded XML format
	 */
	public String getPriceData() { return m_price_data; }

	/**
	 * determine whether a pending synchronization condition exists
	 * (valid for most reply opcodes)
	 * @return true if sync is delayed/pending, false otherwise
	 */
	public boolean syncPending() { return m_SyncPending; }


	// set access methods
	/**
	 * set the opcode value
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_opcode = new String(opcode);
		}
	}

	/**
	 * set the error code value (needed when IQ.Type = error)
	 * @param ecode the error code
	 */
	public void setErrcode(int ecode) { m_errcode = ecode; }

	/**
	 * set the error message value (needed when IQ.Type = error)
	 * @param emsg the error message
	 */
	public void setErrmsg(String emsg) {
		if (emsg != null) {
			m_errmsg = new String(emsg);
		}
	}

	/**
	 * set the hashkey (needed when opcode =  REQ_initiate_login
	 * or REQ_perform_login)
	 * @param hash the hash value
	 */
	public void setHashkey(String hash) {
		if (hash != null) {
			m_hashkey = new String(hash);
		}
	}

	/**
	 * set the long-phrase length
	 * (needed when opcode = REP_long-phrase_challenge)
	 * @param len the length
	 */
	public void setLong_phrase_len(int len) { m_longphrase_len = len; }

	/**
	 * set the first long-phrase index
	 * (needed when opcode = REP_long-phrase_challenge)
	 * @param idx the index (1-based)
	 */
	public void setLong_phrase_index1(int idx) { m_longphrase_idx1 = idx; }

	/**
	 * set the second long-phrase index
	 * (needed when opcode = REP_long-phrase_challenge)
	 * @param idx the index (1-based)
	 */
	public void setLong_phrase_index2(int idx) { m_longphrase_idx2 = idx; }

	/**
	 * set the third long-phrase index
	 * (needed when opcode = REP_long-phrase_challenge)
	 * @param idx the index (1-based)
	 */
	public void setLong_phrase_index3(int idx) { m_longphrase_idx3 = idx; }

	/**
	 * set the first long-phrase character
	 * (needed when opcode = REQ_perform_login)
	 * @param ch the character at idx1
	 */
	public void setLong_phrase_char1(byte ch) { m_longphrase_char1 = ch; }

	/**
	 * set the second long-phrase character
	 * (needed when opcode = REQ_perform_login)
	 * @param ch the character at idx2
	 */
	public void setLong_phrase_char2(byte ch) { m_longphrase_char2 = ch; }

	/**
	 * set the third long-phrase character
	 * (needed when opcode = REQ_perform_login)
	 * @param ch the character at idx3
	 */
	public void setLong_phrase_char3(byte ch) { m_longphrase_char3 = ch; }

	/**
	 * set the encrypted login blob (needed when opcode = REP_login-blob)
	 * @param blob the login blob
	 */
	public void setLoginblob(String blob) {
		if (blob != null) {
			m_loginblob = new String(blob);
		}
	}

	/**
	 * set the list of voucher publishers
	 * (needed when opcode = REP_publisher_list)
	 * @param pubs a List or other Collection of voucher publisher ids
	 */
	public void setVoucherPublishers(Collection<String> pubs) {
		m_VoucherPublishers.addAll(pubs);
	}

	/**
	 * append a single voucher publisher to the list, or set as the selected VP
	 * (needed when opcode = REQ_select_publisher, REQ_init_key, REQ_init_safe,
	 * REQ_close_safe, REQ_send_payment, REQ_pickup_payment, REQ_revalidate,
	 * REQ_reclaim_payment, REQ_delete_spent_tokens, REQ_purchase_tokens,
	 * REQ_recovery_ques, REQ_discover_net, REQ_lookup_recovery,
	 * REQ_initiate_login, REQ_redeem_voucher, or REQ_delete_voucher)
	 * @param vpid the id of a single, or the selected, voucher publisher
	 */
	public void setVoucher_publisher(String vpid) {
		if (vpid != null && !vpid.isEmpty()) {
			if (m_opcode.equalsIgnoreCase("REP_publisher_list")) {
				m_VoucherPublishers.add(vpid);
			}
			else {
				m_SelectedVP = new String(vpid);
			}
		}
	}

	/**
	 * set the algorithm for keypair generation
	 * (needed when opcode = REP_key_params)
	 * @param algo the algorithm (DSA, RSA, etc.)
	 */
	public void setAlgorithm(String algo) {
		if (algo != null && !algo.isEmpty()) {
			m_algorithm = new String(algo);
		}
	}

	/**
	 * set the minimum bits for keypair generation
	 * (needed when opcode = REP_key_params)
	 * @param len the minimum length, in bits
	 */
	public void setMin_length(int len) {
		if (len > 0) {
			m_min_length = len;
		}
	}

	/**
	 * set the public key (at this point unsigned by the VP)
	 * (needed when opcode = REQ_init_key)
	 * @param pubkey the base64-encoded key
	 */
	public void setPublic_key(String pubkey) {
		if (pubkey != null && !pubkey.isEmpty()) {
			m_public_key = new String(pubkey);
		}
	}

	/**
	 * set the user's VS number prefix (needed when opcode = REQ_init_key)
	 * @param prefix the prefix selected by the user
	 */
	public void setVSprefix(String prefix) {
		if (prefix != null && !prefix.isEmpty()) {
			m_vs_prefix = new String(prefix);
		}
	}

	/**
	 * set the signed public key cert (needed when opcode = REP_safe_data)
	 * @param cert the base64-encoded cert
	 */
	public void setCertificate(String cert) {
		if (cert != null && !cert.isEmpty()) {
			m_certificate = new String(cert);
		}
	}

	/**
	 * set the XMPP domain of the key server (needed when opcode = REP_safe_data)
	 * @param domain the domain
	 */
	private void setKey_server_domain(String domain) {
		if (domain != null && !domain.isEmpty()) {
			m_key_server_domain = new String(domain);
		}
	}

	/**
	 * set the socket address of the key server
	 * (needed when opcode = REP_safe_data)
	 * @param addr the address, in the form domain:::host:port::host2:port2
	 */
	public void setKey_server_addr(String addr) {
		if (addr != null && !addr.isEmpty()) {
			String[] config = addr.split(":::");
			if (config.length == 2) {
				setKey_server_domain(config[0]);
				// remainder should be a sequence of host:port addrs delimited by ::
				String[] paths = config[1].split("::");
				for (int iii = 0; iii < paths.length; iii++) {
					String[] address = paths[iii].split(":");
					if (address.length == 2) {
						InetSocketAddress keyServ = new InetSocketAddress(address[0],
																Integer.parseInt(address[1]));
						m_key_servers.add(keyServ);
					}
				}
			}
		}
	}

	/**
	 * set the encrypted capabilities blob (needed when opcode = REP_safe_data)
	 * @param caps the base64-encoded blob, encrypted with the public key
	 */
	public void setCap_blob(String caps) {
		if (caps != null && !caps.isEmpty()) {
			m_cap_blob = new String(caps);
		}
	}

	/**
	 * set the required PIN length (needed when opcode = REP_safe_data)
	 * @param len the length, in digits
	 */
	public void setPin_length(short len) {
		if (len > 0) {
			m_pin_length = len;
		}
	}

	/**
	 * set the text of the first recovery question for the login recovery blob
	 * (needed when opcode = REP_safe_data)
	 * @param text the question text
	 */
	public void setRecovery_ques_1(String text) {
		if (text != null && !text.isEmpty()) {
			m_recovery_ques1 = new String(text);
		}
	}

	/**
	 * set the text of the second recovery question for the login recovery blob
	 * (needed when opcode = REP_safe_data)
	 * @param text the question text
	 */
	public void setRecovery_ques_2(String text) {
		if (text != null && !text.isEmpty()) {
			m_recovery_ques2 = new String(text);
		}
	}

	/**
	 * set the text of the third recovery question for the login recovery blob
	 * (needed when opcode = REP_safe_data)
	 * @param text the question text
	 */
	public void setRecovery_ques_3(String text) {
		if (text != null && !text.isEmpty()) {
			m_recovery_ques3 = new String(text);
		}
	}

	/**
	 * set the text of the fourth recovery question for the login recovery blob
	 * (needed when opcode = REP_safe_data)
	 * @param text the question text
	 */
	public void setRecovery_ques_4(String text) {
		if (text != null && !text.isEmpty()) {
			m_recovery_ques4 = new String(text);
		}
	}

	/**
	 * set the text of the fifth recovery question for the login recovery blob
	 * (needed when opcode = REP_safe_data)
	 * @param text the question text
	 */
	public void setRecovery_ques_5(String text) {
		if (text != null && !text.isEmpty()) {
			m_recovery_ques5 = new String(text);
		}
	}

	/**
	 * set the value of the login details blob
	 * (needed when opcode = REQ_init_safe)
	 * @param blob the details, base64-encoded and encrypted with the passphrase
	 */
	public void setDetails_blob(String blob) {
		if (blob != null && !blob.isEmpty()) {
			m_details_blob = new String(blob);
		}
	}

	/**
	 * set the value of the returned status message
	 * (needed when opcode = REP_safe_init or REP_safe_closed)
	 * @param msg the message text (can be "")
	 */
	public void setStatus_msg(String msg) {
		if (msg != null) {
			m_status_msg = new String(msg);
		}
	}

	/**
	 * set the value of the recovery index hash
	 * (needed when opcode = REQ_lookup_recovery)
	 * @param hash the index hash, base64-encoded
	 */
	public void setRecovery_index(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_recovery_index = new String(hash);
		}
	}

	/**
	 * set the value of the recovery blob
	 * (needed when opcode = REP_recovery_blob)
	 * @param blob the recovery data, base64-encoded and 3DES encrypted
	 */
	public void setRecovery_blob(String blob) {
		if (blob != null && !blob.isEmpty()) {
			m_recovery_blob = new String(blob);
		}
	}

	/**
	 * set the value of the network details blob
	 * (needed when opcode = REP_network_details)
	 * @param blob the details blob, base64-encoded encrypted with Issuer pubkey
	 */
	public void setNetwork_conf_blob(String blob) {
		if (blob != null && !blob.isEmpty()) {
			m_network_conf_blob = new String(blob);
		}
	}

	/**
	 * set the value of the closure_block blob
	 * (needed when opcode = REQ_close_safe)
	 * @param block the closure block, base64-encoded encrypted with VP pubkey
	 */
	public void setClosure_block(String block) {
		if (block != null && !block.isEmpty()) {
			m_closure_block = new String(block);
		}
	}

	/**
	 * set the value of the signature on the (decrypted) closure_block
	 * (needed when opcode = REQ_close_safe)
	 * @param sig the signature, base64-encoded
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_signature = new String(sig);
		}
	}

	/**
	 * set the value of the voucher rw cap for the safe being closed
	 * (needed when opcode = REQ_close_safe)
	 * @param cap the capability, base64-encoded and encrypted with VP pubkey
	 */
	public void setOldVS_rwcap(String cap) {
		if (cap != null && !cap.isEmpty()) {
			 m_old_VS_rwcap = new String(cap);
		}
	}

	/**
	 * set the value of the voucher rw cap for the safe receiving value
	 * (needed when opcode = REQ_close_safe)
	 * @param cap the capability, base64-encoded and encrypted with VP pubkey
	 */
	public void setNewVS_rwcap(String cap) {
		if (cap != null && !cap.isEmpty()) {
			m_new_VS_rwcap = new String(cap);
		}
	}

	/**
	 * set the payment details blob
	 * (needed when opcode is REQ_send_payment or REQ_reclaim_payment)
	 * @param payDet the details, base64-encoded and encrypted with VP's pubkey
	 */
	public void setPayDetails(String payDet) {
		if (payDet != null && !payDet.isEmpty()) {
			m_PayDetails = new String(payDet);
		}
	}

	/**
	 * set the voucher request input/output data
	 * (needed for opcodes REQ_send_payment, REQ_pickup_payment, REQ_revalidate,
	 * REQ_reclaim_payment, and REQ_purchase_tokens)
	 * @param vr the parsed {@link VoucherRequest} object
	 */
	public void setVouchReq(VoucherRequest vr) {
		m_VouchReq = vr;
	}

	/**
	 * set voucher confirmation details (needed for opcodes REP_payment_confirmed
	 * REP_voucher_redeemed, and REP_payment_reverted)
	 * @param conf the confirmation, base64-encoded and encrypted with VS pubkey
	 */
	public void setConfirmation(String conf) {
		if (conf != null && !conf.isEmpty()) {
			m_Confirmation = new String(conf);
		}
	}

	/**
	 * set voucher validation blob (needed for opcode REP_validation)
	 * @param val the validation, base64-encoded and encrypted with VS's pubkey
	 */
	public void setValidation(String val) {
		if (val != null && !val.isEmpty()) {
			m_Validation = new String(val);
		}
	}

	/**
	 * set token purging instructions (needed for opcode REQ_delete_spent_tokens)
	 * @param inst the instructions, base64-encoded and encrypted with VP pubkey
	 */
	public void setPurgeDetails(String inst) {
		if (inst != null && !inst.isEmpty()) {
			m_PurgeDetails = new String(inst);
		}
	}

	/**
	 * set list of tokens actually purged (needed for opcode REP_tokens_purged)
	 * @param pur the list, base64-encoded and encrypted with VS's pubkey
	 */
	public void setPurged(String pur) {
		if (pur != null && !pur.isEmpty()) {
			m_Purged = new String(pur);
		}
	}

	/**
	 * set voucher redemption details (needed for opcode REQ_redeem_voucher)
	 * @param red the instructions, base64-encoded and encrypted with VP pubkey
	 */
	public void setRedeemDetails(String red) {
		if (red != null && !red.isEmpty()) {
			m_RedeemDetails = new String(red);
		}
	}

	/**
	 * set voucher deletion details (needed for opcode REQ_delete_voucher)
	 * @param del the instructions, base64-encoded and encrypted with VP pubkey
	 */
	public void setDeleteDetails(String del) {
		if (del != null && !del.isEmpty()) {
			m_DeleteDetails = new String(del);
		}
	}

	/**
	 * set list of purchased tokens (needed for opcode REP_tokens_purchased)
	 * @param purch the list, base64-encoded and encrypted with VS's pubkey
	 */
	public void setPurchased(String purch) {
		if (purch != null && !purch.isEmpty()) {
			m_Purchased = new String(purch);
		}
	}

	/**
	 * set price data to be returned (needed for opcode REP_price_data)
	 * @param prices the price data, in base64-encoded XML format
	 */
	public void setPriceData(String prices) {
		if (prices != null && !prices.isEmpty()) {
			m_price_data = new String(prices);
		}
	}

	/**
	 * set whether or not a pending record synchronization condition exists
	 * (needed for almost all REP_* opcodes)
	 * @param sync true if so, false if not
	 */
	public void setSyncPending(boolean sync) { m_SyncPending = sync; }


	// method required to extend the IQ class
	/**
	 * provide the child element as XML
	 * @return the extended child element as a XML string
	 */
	public String getChildElementXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<vsc_ofs xmlns=\"jabber:iq:voucher-safe#vsc-ofs\">");
		xml.append("<opcode>" + m_opcode + "</opcode>");

		// if this is an error stanza, supply code and message
		if (getType() == IQ.Type.ERROR) {
			xml.append("<errcode>" + m_errcode + "</errcode>");
			xml.append("<errmsg>" + m_errmsg + "</errmsg>");
		}
		else {
			/* Based on the value of the opcode, emit additional fields.
			 * NB: order in which opcodes are listed is intended to reflect
			 * the relative frequency of each opcode, for efficiency.
			 */
			if (m_opcode.equalsIgnoreCase("REP_publisher_list")) {
				if (m_VoucherPublishers.isEmpty()) {
					xml.append("<voucher-publisher-list/>");
				}
				else {
					xml.append("<voucher-publisher-list>");
					for (String vpid : m_VoucherPublishers) {
						xml.append("<voucher_publisher>");
						xml.append(vpid + "</voucher_publisher>");
					}
					xml.append("</voucher-publisher-list>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_price_data")) {
				xml.append("<priceData>" + m_price_data + "</priceData>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_initiate_login")) {
				xml.append("<hashkey>" + m_hashkey + "</hashkey>");
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_long-phrase_challenge")) {
				xml.append("<long_phrase_len>" + m_longphrase_len);
				xml.append("</long_phrase_len>");
				xml.append("<long_phrase_index1>" + m_longphrase_idx1
								+ "</long_phrase_index1>");
				xml.append("<long_phrase_index2>" + m_longphrase_idx2
								+ "</long_phrase_index2>");
				xml.append("<long_phrase_index3>" + m_longphrase_idx3
								+ "</long_phrase_index3>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_perform_login")) {
				xml.append("<hashkey>" + m_hashkey + "</hashkey>");
				byte[] ary = { m_longphrase_char1,
									m_longphrase_char2,
									m_longphrase_char3 };
				String lpChars = new String(ary);
				xml.append("<long_phrase_char1>" + lpChars.charAt(0)
								+ "</long_phrase_char1>");
				xml.append("<long_phrase_char2>" + lpChars.charAt(1)
								+ "</long_phrase_char2>");
				xml.append("<long_phrase_char3>" + lpChars.charAt(2)
								+ "</long_phrase_char3>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_login-blob")) {
				xml.append("<loginblob>" + m_loginblob + "</loginblob>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_discover_net")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_network_details")) {
				xml.append("<network_conf_blob>" + m_network_conf_blob);
				xml.append("</network_conf_blob>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_delete_spent_tokens")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<purgeDetails>" + m_PurgeDetails + "</purgeDetails>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_tokens_purged")) {
				xml.append("<purged>" + m_Purged + "</purged>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_purchase_tokens")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				if (m_VouchReq != null) {
					xml.append(m_VouchReq.toXML());
				}
				else {
					xml.append("<voucherRequest/>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_tokens_purchased")) {
				xml.append("<purchased>" + m_Purchased + "</purchased>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_send_payment")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<payDetails>" + m_PayDetails + "</payDetails>");
				if (m_VouchReq != null) {
					xml.append(m_VouchReq.toXML());
				}
				else {
					xml.append("<voucherRequest/>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_payment_sent")) {
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_pickup_payment")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				if (m_VouchReq != null) {
					xml.append(m_VouchReq.toXML());
				}
				else {
					xml.append("<voucherRequest/>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_payment_confirmed")) {
				xml.append("<confirmation>" + m_Confirmation + "</confirmation>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_revalidate")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				if (m_VouchReq != null) {
					xml.append(m_VouchReq.toXML());
				}
				else {
					xml.append("<voucherRequest/>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_validation")) {
				xml.append("<validation>" + m_Validation + "</validation>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_reclaim_payment")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<payDetails>" + m_PayDetails + "</payDetails>");
				if (m_VouchReq != null) {
					xml.append(m_VouchReq.toXML());
				}
				else {
					xml.append("<voucherRequest/>");
				}
			}
			else if (m_opcode.equalsIgnoreCase("REP_payment_reverted")) {
				xml.append("<confirmation>" + m_Confirmation + "</confirmation>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_select_publisher")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_key_params")) {
				xml.append("<algorithm>" + m_algorithm + "</algorithm>");
				xml.append("<min_length>" + m_min_length + "</min_length>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_init_key")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<vs_prefix>" + m_vs_prefix + "</vs_prefix>");
				xml.append("<public_key>" + m_public_key + "</public_key>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_safe_data")) {
				xml.append("<certificate>" + m_certificate);
				xml.append("</certificate>");
				if (!m_key_servers.isEmpty()) {
					xml.append("<key_server_addr>" + m_key_server_domain + ":::");
					boolean firstIsa = true;
					for (InetSocketAddress isa : m_key_servers) {
						if (!firstIsa) {
							xml.append("::");
						}
						else {
							firstIsa = false;
						}
						xml.append(isa.getHostName() + ":" + isa.getPort());
					}
					xml.append("</key_server_addr>");
				}
				else {
					xml.append("<key_server_addr/>");
				}
				xml.append("<cap_blob>" + m_cap_blob + "</cap_blob>");
				xml.append("<pin_length>" + m_pin_length + "</pin_length>");
				xml.append("<recovery_ques_1>" + m_recovery_ques1);
				xml.append("</recovery_ques_1>");
				xml.append("<recovery_ques_2>" + m_recovery_ques2);
				xml.append("</recovery_ques_2>");
				xml.append("<recovery_ques_3>" + m_recovery_ques3);
				xml.append("</recovery_ques_3>");
				xml.append("<recovery_ques_4>" + m_recovery_ques4);
				xml.append("</recovery_ques_4>");
				xml.append("<recovery_ques_5>" + m_recovery_ques5);
				xml.append("</recovery_ques_5>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_init_safe")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<details_blob>" + m_details_blob + "</details_blob>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_safe_init")) {
				xml.append("<status_msg>" + m_status_msg + "</status_msg>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_recovery_ques")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_questions")) {
				xml.append("<recovery_ques_1>" + m_recovery_ques1);
				xml.append("</recovery_ques_1>");
				xml.append("<recovery_ques_2>" + m_recovery_ques2);
				xml.append("</recovery_ques_2>");
				xml.append("<recovery_ques_3>" + m_recovery_ques3);
				xml.append("</recovery_ques_3>");
				xml.append("<recovery_ques_4>" + m_recovery_ques4);
				xml.append("</recovery_ques_4>");
				xml.append("<recovery_ques_5>" + m_recovery_ques5);
				xml.append("</recovery_ques_5>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_lookup_recovery")) {
				xml.append("<recovery_index>" + m_recovery_index);
				xml.append("</recovery_index>");
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_recovery_blob")) {
				xml.append("<recovery_blob>" + m_recovery_blob);
				xml.append("</recovery_blob>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_close_safe")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<closure_block>" + m_closure_block);
				xml.append("</closure_block>");
				xml.append("<signature>" + m_signature + "</signature>");
				xml.append("<oldVS_rwcap>" + m_old_VS_rwcap + "</oldVS_rwcap>");
				xml.append("<newVS_rwcap>" + m_new_VS_rwcap + "</newVS_rwcap>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_safe_closed")) {
				xml.append("<status_msg>" + m_status_msg + "</status_msg>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_redeem_voucher")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<redeemDetails>" + m_RedeemDetails);
				xml.append("</redeemDetails>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_voucher_redeemed")) {
				xml.append("<confirmation>" + m_Confirmation + "</confirmation>");
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_delete_voucher")) {
				xml.append("<voucher_publisher>" + m_SelectedVP);
				xml.append("</voucher_publisher>");
				xml.append("<deleteDetails>" + m_DeleteDetails);
				xml.append("</deleteDetails>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_voucher_deleted")) {
				xml.append("<syncPending>" + m_SyncPending + "</syncPending>");
			}
			// else: there are no opcode-specific values
		}

		xml.append("</vsc_ofs>");
		return xml.toString();
	}

	// END methods
}
