/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	10-28-08/kbw/ add REQ_get_cert / REP_cert support
 * 	09-12-08/kbw/ new
 *
 * purpose---
 * 	class to manipulate VSC message sent to/from the PKS
 *
 * notes---
 * 	see underlying message definitions in sys_pks.x
 */

package org.silentvault.client;

import org.jivesoftware.smack.packet.IQ;


/**
 * Holds all the pieces of a message which is part of the
 * jabber:iq:voucher-safe#sys-pks namespace between a VSC and a PKS
 * (Public Key Server) using the element sys_pks.
 * Used to facilitate building messages sent to a PKS or received
 * from it in reply.  Knows how to emit itself as an XML child element.
 *
 * @author Kevin Wilkerson
 */
public class PKSMessage extends IQ {
	// BEGIN data members

	/**
	 * The opcode of this request (begins with REQ_) or reply (begins with REP_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String					m_opcode;

	/**
	 * errcode is used when the IQ.Type is error.
	 * It represents the nature of the error.  For ease of use and readability,
	 * standard HTTP return values are used.  Thus 4xx represents a fatal error
	 * while 5xx represents a temporary error.
	 */
	private int						m_errcode;

	/**
	 * errmsg is used when the IQ.Type is error.
	 * It represents an optional textual elaboration of the problem, to be
	 * used for logging or exception generation purposes.
	 */
	private String					m_errmsg;

	/**
	 * user_id is used with opcodes REQ_get_pubkey and REQ_get_cert.
	 * It represents the alias to be looked up by the key server.
	 */
	private String					m_user_id;

	/**
	 * public_key is used with opcode REP_pubkey.
	 * It represents the base64-encoded public key retrieved.
	 */
	private String					m_public_key;

	/**
	 * certificate is used with opcode REP_cert.
	 * It represents the base64-encoded public key certificate retrieved.
	 */
	private String					m_certificate;

	// END data members


	// BEGIN methods
	
	/**
	 * constructor
	 */
	public PKSMessage() {
		super();
		m_opcode = m_errmsg = m_user_id = m_public_key = m_certificate = "";
	}

	// get accessor methods
	/**
	 * obtain the current opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_opcode; }

	/**
	 * obtain current error code, valid only when the IQ.Type = error
	 * @return the error
	 */
	public int getErrcode() { return m_errcode; }

	/**
	 * obtain the textual representation of the error, valid only when the
	 * IQ.Type = error
	 * @return the error message
	 */
	public String getErrmsg() { return m_errmsg; }

	/**
	 * get the user_id (valid when opcode = REQ_get_pubkey)
	 * @return the user id
	 */
	public String getUser_id() { return m_user_id; }

	/**
	 * get the public key (valid when opcode = REP_pubkey)
	 * @return the key, base64-encoded
	 */
	public String getPublic_key() { return m_public_key; }

	/**
	 * get the certificate (valid when opcode = REP_cert)
	 * @return the certificate, base64-encoded
	 */
	public String getCertificate() { return m_certificate; }


	// set access methods
	/**
	 * set the opcode value
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_opcode = new String(opcode);
		}
	}

	/**
	 * set the error code value
	 * @param ecode the error code
	 */
	public void setErrcode(int ecode) { m_errcode = ecode; }

	/**
	 * set the error message value
	 * @param emsg the error message
	 */
	public void setErrmsg(String emsg) {
		if (emsg != null) {
			m_errmsg = new String(emsg);
		}
	}

	/**
	 * set the user_id (needed when opcode = REQ_get_pubkey)
	 * @param id the id
	 */
	public void setUser_id(String id) {
		if (id != null && !id.isEmpty()) {
			m_user_id = new String(id);
		}
	}

	/**
	 * set the public_key (needed when opcode = REP_pubkey)
	 * @param key the encoded key
	 */
	public void setPublic_key(String key) {
		if (key != null && !key.isEmpty()) {
			m_public_key = new String(key);
		}
	}

	/**
	 * set the certificate (needed when opcode = REP_cert)
	 * @param cert the base64-encoded certificate value
	 */
	public void setCertificate(String cert) {
		if (cert != null && !cert.isEmpty()) {
			m_certificate = new String(cert);
		}
	}

	// method required to extend the IQ class
	/**
	 * provide the child element as XML
	 * @return the extended child element as a XML string
	 */
	public String getChildElementXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<sys_pks xmlns=\"jabber:iq:voucher-safe#sys-pks\">");
		xml.append("<opcode>" + m_opcode + "</opcode>");

		// if this is an error stanza, supply code and message
		if (getType() == IQ.Type.ERROR) {
			xml.append("<errcode>" + m_errcode + "</errcode>");
			xml.append("<errmsg>" + m_errmsg + "</errmsg>");
		}
		else {
			// based on the value of the opcode, emit additional fields

			if (m_opcode.equalsIgnoreCase("REQ_get_pubkey")) {
				xml.append("<user_id>" + m_user_id + "</user_id>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_get_cert")) {
				xml.append("<user_id>" + m_user_id + "</user_id>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_pubkey")) {
				xml.append("<public_key>" + m_public_key + "</public_key>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_cert")) {
				xml.append("<certificate>" + m_certificate + "</certificate>");
			}
			else {
				Log.error("invalid opcode, " + m_opcode);
			}
		}

		xml.append("</sys_pks>");
		return xml.toString();
	}
}
