/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	11-07-08/kbw/ add support for REP_cert
 * 	09-12-08/kbw/ new
 *
 * purpose---
 * 	Parse {@link PKSMessage} packets.  Parses the <tt>sys_pks</tt>
 * 	sub-document and creates an instance of {@link PKSMessage}.
 *
 * notes---
 * 	see underlying message definitions in sys_pks.x
 */

package org.silentvault.client;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.IQProvider;


/**
 * Implements the IQProvider interface to receive and parse incoming messages
 * from a PKS (Public Key Server) according to opcode in the sys_pks element
 * of the jabber:iq:voucher-safe#sys-pks namespace.  Creates a
 * {@link PKSMessage} to hold the parsed message.
 *
 * @author Kevin Wilkerson
 */
public class PKSMessageProvider implements IQProvider {
	// members
	private static final String	M_EndTag = "sys_pks";
	
	/**
	 * constructor
	 */
	public PKSMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		PKSMessage pksMsg = new PKSMessage();
		// NB: these values should be sys_pks and jabber:iq:voucher-safe#sys-pks
		pksMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		pksMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		String opcode = "";
		int eventType = parser.next();
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		pksMsg.setOpcode(opcode);

		// Now based on the opcode, pull out the other data.  (Only replies)

		if (opcode.equalsIgnoreCase("REP_pubkey")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						pksMsg.setErrcode(Integer.parseInt(parser.nextText()));
						pksMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						pksMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("public_key")) {
						pksMsg.setPublic_key(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_cert")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						pksMsg.setErrcode(Integer.parseInt(parser.nextText()));
						pksMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						pksMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("certificate")) {
						pksMsg.setCertificate(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return pksMsg;
	}
}
