/*
 * last modified---
 * 	08-09-13/jet/ force logout if we get a fail on listing vouchers or tokens
 * 	06-05-13/jet/ SilentVault client version
 * 	02-24-10/kbw/ getState() -> getVsState()
 * 	04-21-09/kbw/ add revolving filter map, to support N outstanding queries
 * 	04-16-09/kbw/ begin adding usage of real UI elements
 * 	02-10-09/kbw/ handle filenames in voucher list (use Hashtable)
 * 	01-27-08/kbw/ new
 *
 * purpose---
 * 	listen for messages from the SDS which are replies to IQ packets we sent
 *
 * notes---
 * 	see underlying messages in vsc_sds.x
 */

package org.silentvault.client;

import javax.swing.*;

import org.dom4j.DocumentException;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.filter.OrFilter;

import org.silentvault.client.ui.wallet.*;

import java.util.Hashtable;
import java.util.ArrayList;


/**
 * class to handle processing of SDS reply packets, within the context of the
 * jabber:iq:voucher-safe#vsc-sds namespace.
 *
 * @author Kevin Wilkerson
 */
public final class SDSListener implements PacketListener, Runnable {
	// BEGIN data members

	/**
	 * back pointer reference to the Plugin object
	 */
	private WalletClient		m_PluginPtr;

	/**
	 * the live connection to the SDS on which we are listening
	 */
	private XMPPConnection	m_SDSConn;

	/**
	 * a map of currently outstanding listen filters, representing all requests
	 * for which we are awaiting replies, keyed by their packet IDs
	 */
	private Hashtable<String, PacketIDFilter>	m_FilterMap;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the reference to the plugin of which we are a part
	 */
	public SDSListener(WalletClient plugin) {
		m_PluginPtr = plugin;
		m_FilterMap = new Hashtable<String, PacketIDFilter>();
	}

	// method required to implement the PacketListener interface
	/**
	 * deal with a reply packet from the SDS
	 * @param packet the packet (will really be a SDSMessage)
	 */
	public void processPacket(Packet packet) {
		// verify we got what's expected
		if (packet instanceof SDSMessage) {
			SDSMessage thisMsg = (SDSMessage) packet;
			String opcode = thisMsg.getOpcode();
			//Log.debug("Listener got SDS message opcode " + opcode);
			WTabManager tm = m_PluginPtr.getWalletTabManager();

			// record message
			VscState stateObj = m_PluginPtr.getVscState();
			stateObj.setLastActivity(thisMsg);

			// see if it's an error
			if (thisMsg.getType() == IQ.Type.ERROR) {
				Log.error("Got an error: code " + thisMsg.getErrcode() + ", "
							+ thisMsg.getErrmsg());

				// some opcodes should have errors passed through below
				if (opcode.equalsIgnoreCase("REP_payment_stored")) {
					// pass error to confirm pane
					ConfirmPane cp = tm.getConfirmPane();
					cp.processPmtStoreReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_receipt_stored")) {
					// pass error to NewReceiptPane
					NewReceiptPane nrp = tm.getNewReceiptPane();
					nrp.processRctStoreReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_payment_deleted")) {
					// tell the HousekeepingPane
					HousekeepingPane hkp = tm.getHousekeepingPane();
					hkp.confirmPaymentDeletion(false);

					// update display, in separate thread
					tm.queueTransition(new Integer(
															WTabManager.M_WORK_PAYMENT_REGEN));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_receipts_deleted")) {
					// tell ReceiptsPane about the failure
					ReceiptsPane rp = tm.getReceiptsPane();
					rp.confirmReceiptDeletions(thisMsg.getReceipt_del_list());
				}
				else if (opcode.equalsIgnoreCase("REP_voucher_list")) {
					tm.showError("cannot list vouchers in wallet",
									"SDS error code " + thisMsg.getErrcode()
									+ ", " + thisMsg.getErrmsg(),
									"we will now log out, please try again later");
					tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_token_list")) {
					tm.showError("cannot list tokens in wallet",
									"SDS error code " + thisMsg.getErrcode()
									+ ", " + thisMsg.getErrmsg(),
									"we will now log out, please try again later");
					tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
					new Thread(this).start();
				}
				else {
					// payment or receipt list, show something generic
					tm.showError("error accessing SDS", "error code "
									+ thisMsg.getErrcode() + ", " + thisMsg.getErrmsg());
				}

				// retire this ID from the list of active packet filters
				removeFilter(packet.getPacketID());
				return;
			}

			// take further actions based on the opcode
			HomePane hp = tm.getHomePane();
			if (opcode.equalsIgnoreCase("REP_voucher_list")) {
				Hashtable<String, String> vouchers = thisMsg.getVoucher_list();

				// tell the Home Pane about the list of vouchers
				hp.recordVouchers(vouchers);

				// and queue up the state transition for execution in a separate
				// thread to update display -- this will also decode the vouchers

				tm.queueTransition(new Integer(WTabManager.M_HOME_VOUCH_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_token_list")) {
				ArrayList<String> tokens = thisMsg.getToken_list();
				//Log.debug("Got REP_token_list " + tokens.size() + " tokens");

				// For each token on the list, parse it into a real XMLToken object.
				int tCnt = 0;
				ArrayList<XMLToken> parsedTokens
					= new ArrayList<XMLToken>(tokens.size());
				for (String tokStr : tokens) {
					XMLToken tok = null;
					try {
						tok = XMLToken.getTokenRep(tokStr, false);
						tCnt++;
						parsedTokens.add(tok);
					}
					catch (DocumentException de) {
						Log.error("Token index " + tCnt + " would not parse!");
					}
				}

				// tell Home Pane about the list of tokens
				hp.recordTokens(parsedTokens, tCnt);

				// and queue up the state transition for execution in a separate
				// thread to update display

				tm.queueTransition(new Integer(WTabManager.M_HOME_TOKEN_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_payment_list")) {
				// pass to HousekeepingPane's table and clear wait cursor
				HousekeepingPane hkp = tm.getHousekeepingPane();
				hkp.recordPayments(thisMsg.getPayment_list());
				hp.setCursor(null);

				// decode and update display, in separate thread
				tm.queueTransition(new Integer(WTabManager.M_WORK_PAYMENT_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_receipt_list")) {
				// pass to ReceiptsPane's table and clear wait cursor
				ReceiptsPane rp = tm.getReceiptsPane();
				rp.recordReceipts(thisMsg.getReceipt_list());
				hp.setCursor(null);

				// decode and update display, in separate thread
				tm.queueTransition(new Integer(WTabManager.M_WORK_RECEIPT_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_payment_stored")) {
				// tell the ConfirmPane
				ConfirmPane cp = tm.getConfirmPane();
				cp.processPmtStoreReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_payment_deleted")) {
				// tell the HousekeepingPane
				HousekeepingPane hkp = tm.getHousekeepingPane();
				hkp.confirmPaymentDeletion(true);

				// update display, in separate thread
				tm.queueTransition(new Integer(WTabManager.M_WORK_PAYMENT_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_receipt_stored")) {
				// pass reply to NewReceiptPane
				NewReceiptPane nrp = tm.getNewReceiptPane();
				nrp.processRctStoreReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_receipts_deleted")) {
				// tell ReceiptsPane about the removals
				ReceiptsPane rp = tm.getReceiptsPane();
				rp.confirmReceiptDeletions(thisMsg.getReceipt_del_list());

				// update display
				tm.queueTransition(new Integer(WTabManager.M_RECEIPT_LIST));
				new Thread(this).start();
			}
			else {
				Log.error("Unrecognized SDS reply opcode " + opcode + ", XML: "
							+ packet.toXML());
				tm.showError("Bad SDS reply opcode: " + opcode);
			}
		}
		else {
			Log.error("Non-SDSMessage packet received, with XML: "
						+ packet.toXML());
		}

		// retire this ID from the list of active packet filters
		removeFilter(packet.getPacketID());
	}

	/**
	 * get the existing SDS connection, if any
	 * @return the connection, or null if there isn't a live one
	 */
	public XMPPConnection getSDSConnection() {
		if (m_SDSConn != null) {
			if (!m_SDSConn.isConnected()) {
				m_SDSConn.disconnect();
				m_SDSConn = null;
			}
		}
		return m_SDSConn;
	}

	/**
	 * record the connection to the SDS, so we can close it later (overwrites)
	 * @param conn the connection (must be currently live)
	 */
	public void setSDSConnection(XMPPConnection conn) {
		if (conn.isConnected()) {
			if (m_SDSConn != null) {
				// blow away old one first
				m_SDSConn.disconnect();
			}
			m_SDSConn = conn;
		}
	}

	/**
	 * close the connection to the SDS
	 */
	public void closeSDSConnection() {
		if (m_SDSConn != null) {
			m_SDSConn.disconnect();
			m_SDSConn = null;
		}
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this message listener thread by Swing.  Also, a single thread is used for
	 * handling all listener events, so we don't want it to block for very long.
	 */
	public void run() {
		// obtain UI tab manager from plugin and do the queued state transition
		m_PluginPtr.getWalletTabManager().makeTransition();
	}

	/**
	 * method to add a filter to the list of packets we're accepting replies for
	 * @param ID the packet ID we're expecting in a reply
	 * @param filter the filter to add, which watches for the ID
	 */
	public void addFilter(String ID, PacketIDFilter filter) {
		if (ID == null || ID.isEmpty() || filter == null
			|| m_FilterMap.containsKey(ID))
		{
			// redundant
			return;
		}
		m_FilterMap.put(ID, filter);
		resetConnFilter();
	}

	/**
	 * method to remove a filter from the list of replies we're listening for
	 * @param ID the ID of the packet filter to remove
	 */
	public void removeFilter(String ID) {
		if (ID == null || ID.isEmpty() || !m_FilterMap.containsKey(ID)) {
			// nothing to do
			return;
		}
		m_FilterMap.remove(ID);
		resetConnFilter();
	}

	/**
	 * private utility method to rebuild and reset the filter on this connection,
	 * ORing together all filters extant in the map
	 */
	private synchronized void resetConnFilter() {
		// rebuild filter of all outstanding requests
		OrFilter of = new OrFilter();
		for (String packetId : m_FilterMap.keySet()) {
			PacketIDFilter filt = m_FilterMap.get(packetId);
			of.addFilter(filt);
		}
		m_SDSConn.removePacketListener(this);

		// add new listener if map is non-empty
		if (!m_FilterMap.isEmpty()) {
			m_SDSConn.addPacketListener(this, of);
		}
	}

	// END methods
}
