/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	02-10-09/kbw/ index vouchers by their filename, as well
 * 	01-15-09/kbw/ make caps char[] and add finalize()
 * 	12-30-08/kbw/ new
 *
 * purpose---
 * 	class to manipulate messages sent to/from a SDS
 *
 * notes---
 * 	see underlying message definitions in vsc_sds.x
 */

package org.silentvault.client;

import org.jivesoftware.smack.packet.IQ;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Arrays;


/**
 * Holds all the pieces of a message which is part of the
 * jabber:iq:voucher-safe#vsc-sds protocol between a VSC and a SDS
 * using the element vsc_sds.
 * The class also knows how to emit itself as XML.
 * It is used as a utility class to assist in recording
 * incoming reply messages and forming correct get/set stanzas.
 *
 * @author Kevin Wilkerson
 */
public final class SDSMessage extends IQ {
	// BEGIN data members

	/**
	 * The opcode of this request (begins with REQ_) or reply (begins with REP_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String				m_opcode;

	/**
	 * errcode is used when the IQ.Type is error.
	 * It represents the nature of the error.  For ease of use and readability,
	 * standard HTTP return values are used.  Thus 4xx represents a fatal error
	 * while 5xx represents a temporary error.
	 */
	private int					m_errcode;

	/**
	 * errmsg is used when the IQ.Type is error.
	 * It represents an optional textual elaboration of the problem, to be
	 * used for logging or exception generation purposes.
	 */
	private String				m_errmsg;

	/**
	 * folder_index is used with all REQ_ opcodes.  It specifies the
	 * index of the voucher/token or payment/receipt folder for the given VS.
	 * (This folder must always exist.)  This value is base64-encoded.
	 */
	private String				m_folder_index;

	/**
	 * readcap is used with opcodes REQ_list_vouchers, REQ_list_tokens,
	 * REQ_list_pending_pmts, and REQ_list_receipts.  It represents the read
	 * access capability for the given folder index, and is base64-encoded.
	 */
	private char[]				m_readcap;

	/**
	 * readwritecap is used with opcodes REQ_store_pending_pmt,
	 * REQ_delete_pending_pmt, REQ_store_receipt, and REQ_delete_receipts.
	 * It represents the read-write access capability for the given folder index,
	 * and is base64-encoded.
	 */
	private char[]				m_readwritecap;

	/**
	 * This voucher list is used with opcode REP_voucher_list.  It represents
	 * the voucher_blob elements (each a base64-encoded, encrypted blob)
	 * found within the voucher_list.  (Each string on this list can be
	 * decoded, decrypted, and parsed into a XMLVoucher object.)  The index is
	 * the filename at which the voucher is stored in the folder.
	 */
	private Hashtable<String, String>	m_voucher_list;

	/**
	 * This token list is used with opcode REP_token_list.  It represents
	 * the token elements found within the token_list.  (Each string on
	 * this list can be parsed into a XMLToken object.)
	 */
	private ArrayList<String>	m_token_list;

	/**
	 * This payment list is used with opcode REP_payment_list
	 * It represents the payment_blob elements, indexed by their filename
	 * attributes.  (Each string at a given filename key can be decoded and
	 * decrypted into a XMLPayment object.)
	 */
	private Hashtable<String, String>	m_payment_list;

	/**
	 * This receipt list is used with opcode REP_receipt_list
	 * It represents the receipt_blob elements, indexed by their filename
	 * attributes.  (Each string at a given filename key can be decoded and
	 * decrypted into a XMLReceipt object.)
	 */
	private Hashtable<String, String>	m_receipt_list;

	/**
	 * payment_blob is used with opcode REQ_store_pending_pmt.
	 * It represents the encrypted and base64-encoded data blob (a pending
	 * payment specification) received from the OFS/VP following a
	 * REQ_send_payment / REP_payment_sent exchange.
	 */
	private String				m_payment_blob;

	/**
	 * payment is used with opcodes REQ_store_receipt and
	 * REQ_store_pending_pmt.  It represents the list of tokens which are
	 * offered as payment to the SDS for storing the relevant data.
	 * The value of all strings in this list will be as returned by
	 * XMLToken.toXML(), including the token and /token delimiters.
	 * Note: this list normally has length one.
	 */
	private ArrayList<String>	m_payment_tokens;

	/**
	 * filename is used with opcodes REQ_delete_pending_pmt,
	 * REP_receipt_stored, and REP_payment_stored.
	 * It is used to indicate the specific payment or
	 * receipt record stored or to be removed.
	 */
	private String				m_filename;

	/**
	 * receipt_blob is used with opcode REQ_store_receipt.
	 * It represents the encrypted and base64-encoded data blob (a receipt)
	 * which was retrieved from the DHT and is now being stored in the folder.
	 */
	private String				m_receipt_blob;

	/**
	 * This receipt list is used with opcodes REQ_delete_receipts and
	 * REP_receipts_deleted.
	 * It represents the filenames of receipts to be deleted.
	 */
	private ArrayList<String>	m_receipt_del_list;

	// END data members


	// BEGIN methods

	/**
	 * base constructor
	 */
	public SDSMessage() {
		super();
		m_opcode = m_errmsg = m_folder_index = "";
		m_payment_blob = m_filename = m_receipt_blob = "";
		m_voucher_list = new Hashtable<String, String>();
		m_token_list = new ArrayList<String>();
		m_payment_tokens = new ArrayList<String>();
		m_receipt_del_list = new ArrayList<String>();
		m_payment_list = new Hashtable<String, String>();
		m_receipt_list = new Hashtable<String, String>();
	}


	// get accessor methods
	/**
	 * obtain the current opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_opcode; }

	/**
	 * obtain current error code, valid only when the IQ.Type = error
	 * @return the error
	 */
	public int getErrcode() { return m_errcode; }

	/**
	 * obtain the textual representation of the error, valid only when the
	 * IQ.Type = error
	 * @return the error message
	 */
	public String getErrmsg() { return m_errmsg; }

	/**
	 * obtain the folder index hash (valid for all REQ_ opcodes)
	 * @return the index, base64-encoded
	 */
	public String getFolder_index() { return m_folder_index; }

	/**
	 * obtain the folder read capability (valid for REQ_list_* opcodes)
	 * @return a clone of the cap, base64-encoded
	 */
	public char[] getReadcap() { return m_readcap.clone(); }

	/**
	 * obtain the folder read-write capability (valid for REQ_store_* and
	 * REQ_delete_* opcodes)
	 * @return a clone of the cap, base64-encoded
	 */
	public char[] getReadwritecap() { return m_readwritecap.clone(); }

	/**
	 * obtain the list of vouchers (valid for opcode REP_voucher_list)
	 * @return the list, which may be empty
	 */
	public Hashtable<String, String> getVoucher_list() { return m_voucher_list; }

	/**
	 * obtain the list of tokens (valid for opcode REP_token_list)
	 * @return the list, which may be empty
	 */
	public ArrayList<String> getToken_list() { return m_token_list; }

	/**
	 * obtain the list of payments (valid for opcode REP_payment_list)
	 * @return the list, which may be empty
	 */
	public Hashtable<String, String> getPayment_list() { return m_payment_list; }

	/**
	 * obtain the list of receipts (valid for opcode REP_receipt_list)
	 * @return the list, which may be empty
	 */
	public Hashtable<String, String> getReceipt_list() { return m_receipt_list; }

	/**
	 * get the payment blob (valid when opcode = REQ_store_pending_pmt)
	 * @return the payment, a base64-encoded and encrypted string
	 */
	public String getPayment_blob() { return m_payment_blob; }

	/**
	 * get the receipt blob (valid when opcode = REQ_store_receipt)
	 * @return the receipt, a base64-encoded and encrypted string
	 */
	public String getReceipt_blob() { return m_receipt_blob; }

	/**
	 * obtain the list of tokens being offered to the SDS as payment (valid for
	 * opcodes REQ_store_receipt and REQ_store_pending_pmt)
	 * @return the list, which may be empty
	 */
	public ArrayList<String> getPaymentTokens() { return m_payment_tokens; }

	/**
	 * get the filename (valid when opcode is one of REQ_delete_pending_pmt,
	 * REP_payment_stored, or REP_receipt_stored)
	 * @return the filename to be deleted, or that was just stored
	 */
	public String getFilename() { return m_filename; }

	/**
	 * obtain the list of receipt filenames to be deleted (valid when opcode
	 * is REQ_delete_receipts or REP_receipts_deleted)
	 * @return the list, which may be empty
	 */
	public ArrayList<String> getReceipt_del_list() { return m_receipt_del_list; }


	// set access methods
	/**
	 * set the opcode value
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_opcode = new String(opcode);
		}
	}

	/**
	 * set the error code value (needed when IQ.Type = error)
	 * @param ecode the error code
	 */
	public void setErrcode(int ecode) { m_errcode = ecode; }

	/**
	 * set the error message value (needed when IQ.Type = error)
	 * @param emsg the error message
	 */
	public void setErrmsg(String emsg) {
		if (emsg != null) {
			m_errmsg = new String(emsg);
		}
	}
	
	/**
	 * set the folder index value (needed for all REQ_ opcodes)
	 * @param findex the base64-encoded index of the relevant folder
	 */
	public void setFolder_index(String findex) {
		if (findex != null && !findex.isEmpty()) {
			m_folder_index = new String(findex);
		}
	}

	/**
	 * set the folder read capability (needed for REQ_list_* opcodes)
	 * @param rcap the base64-encoded read cap for the relevant folder
	 */
	public void setReadcap(String rcap) {
		if (rcap != null && !rcap.isEmpty()) {
			m_readcap = rcap.toCharArray();
		}
	}

	/**
	 * set the folder read-write capability (needed for REQ_store_* and
	 * REQ_delete_* opcodes)
	 * @param rwcap the base64-encoded read-write cap for the relevant folder
	 */
	public void setReadwritecap(String rwcap) {
		if (rwcap != null && !rwcap.isEmpty()) {
			m_readwritecap = rwcap.toCharArray();
		}
	}

	/**
	 * add a voucher blob to the list of vouchers (needed for opcode
	 * REP_voucher_list)
	 * @param vfile the voucher's filename
	 * @param vblob the voucher blob, base64-encoded
	 */
	public void addVoucherToList(String vfile, String vblob) {
		if (vblob != null && !vblob.isEmpty()
			 && vfile != null && !vfile.isEmpty())
		{
			m_voucher_list.put(vfile, vblob);
		}
	}

	/**
	 * add a token to the list of tokens (needed for opcode REP_token_list)
	 * @param token the token as a XML string
	 */
	public void addTokenToList(String token) {
		if (token != null && !token.isEmpty()) {
			m_token_list.add(token);
		}
	}

	/**
	 * add a payment to the list of pending payments (needed for opcode
	 * REP_payment_list)
	 * @param fname the filename in which the payment is stored
	 * @param pblob the base64-encoded, encrypted blob representing the payment
	 */
	public void addPaymentToList(String fname, String pblob) {
		if (fname == null || fname.isEmpty()
			 || pblob == null || pblob.isEmpty()) {
			return;
		}
		m_payment_list.put(fname, pblob);
	}

	/**
	 * add a receipt to the list of receipts (needed for opcode REP_receipt_list)
	 * @param fname the filename in which the receipt is stored
	 * @param rblob the base64-encoded, encrypted blob representing the receipt
	 */
	public void addReceiptToList(String fname, String rblob) {
		if (fname == null || fname.isEmpty()
			 || rblob == null || rblob.isEmpty()) {
			return;
		}
		m_receipt_list.put(fname, rblob);
	}

	/**
	 * set the payment value (needed for opcode REQ_store_pending_pmt)
	 * @param payment the base64-encoded, encrypted text of the payment record
	 */
	public void setPayment_blob(String payment) {
		if (payment != null && !payment.isEmpty()) {
			m_payment_blob = new String(payment);
		}
	}

	/**
	 * set the receipt value (needed for opcode REQ_store_receipt)
	 * @param receipt the base64-encoded, encrypted text of the receipt record
	 */
	public void setReceipt_blob(String receipt) {
		if (receipt != null && !receipt.isEmpty()) {
			m_receipt_blob = new String(receipt);
		}
	}

	/**
	 * add a token to the list of tokens offered to the SDS for storing records
	 * (needed for opcodes REQ_store_receipt and REQ_store_pending_pmt)
	 * @param xmlTok the token in XML representation
	 */
	public void addPaymentToken(String xmlTok) {
		if (xmlTok != null && !xmlTok.isEmpty()) {
			m_payment_tokens.add(xmlTok);
		}
	}

	/**
	 * set the filename value (needed for opcode REQ_delete_pending_pmt)
	 * @param fname the filename containing the pending payment to be deleted
	 */
	public void setFilename(String fname) {
		if (fname != null && !fname.isEmpty()) {
			m_filename = new String(fname);
		}
	}

	/**
	 * add a receipt to the list of receipts to be removed from the folder
	 * (needed for opcodes REQ_delete_receipts and REP_receipts_deleted)
	 * @param fname the filename in which the receipt is stored
	 */
	public void addReceiptToDelList(String fname) {
		if (fname != null && !fname.isEmpty()) {
			m_receipt_del_list.add(fname);
		}
	}


	// method required to extend the IQ class
	/**
	 * provide the child element as XML
	 * @return the extended child element as a XML string
	 */
	public String getChildElementXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<vsc_sds xmlns=\"jabber:iq:voucher-safe#vsc-sds\">");
		xml.append("<opcode>" + m_opcode + "</opcode>");

		// if this is an error stanza, supply code and message
		if (getType() == IQ.Type.ERROR) {
			xml.append("<errcode>" + m_errcode + "</errcode>");
			xml.append("<errmsg>" + m_errmsg + "</errmsg>");
		}
		else {
			// based on the value of the opcode, emit additional fields
			if (m_opcode.equalsIgnoreCase("REQ_list_vouchers")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readcap>" + new String(m_readcap) + "</readcap>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_voucher_list")) {
				xml.append("<voucher_list>");
				for (String vouchFile : m_voucher_list.keySet()) {
					String vouchBlob = m_voucher_list.get(vouchFile);
					xml.append("<voucher_blob filename=\"" + vouchFile + "\">");
					xml.append(vouchBlob + "</voucher_blob>");
				}
				xml.append("</voucher_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_list_tokens")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readcap>" + new String(m_readcap) + "</readcap>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_token_list")) {
				xml.append("<token_list>");
				for (String token : m_token_list) {
					xml.append(token);
				}
				xml.append("</token_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_list_pending_pmts")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readcap>" + new String(m_readcap) + "</readcap>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_payment_list")) {
				xml.append("<payment_list>");
				Enumeration<String> fnames = m_payment_list.keys();
				while (fnames.hasMoreElements()) {
					String file = fnames.nextElement();
					xml.append("<payment_blob filename=\"" + file + "\">");
					xml.append(m_payment_list.get(file) + "</payment_blob>");
				}
				xml.append("</payment_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_list_receipts")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readcap>" + new String(m_readcap) + "</readcap>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_receipt_list")) {
				xml.append("<receipt_list>");
				Enumeration<String> fnames = m_receipt_list.keys();
				while (fnames.hasMoreElements()) {
					String file = fnames.nextElement();
					xml.append("<receipt_blob filename=\"" + file + "\">");
					xml.append(m_receipt_list.get(file) + "</receipt_blob>");
				}
				xml.append("</receipt_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_store_pending_pmt")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readwritecap>" + new String(m_readwritecap));
				xml.append("</readwritecap>");
				xml.append("<payment_blob>" + m_payment_blob + "</payment_blob>");
				xml.append("<payment>");
				for (String pmt : m_payment_tokens) {
					xml.append(pmt);
				}
				xml.append("</payment>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_payment_stored")) {
				xml.append("<filename>" + m_filename + "</filename>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_delete_pending_pmt")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readwritecap>" + new String(m_readwritecap));
				xml.append("</readwritecap>");
				xml.append("<filename>" + m_filename + "</filename>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_store_receipt")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readwritecap>" + new String(m_readwritecap));
				xml.append("</readwritecap>");
				xml.append("<receipt_blob>" + m_receipt_blob + "</receipt_blob>");
				xml.append("<payment>");
				for (String pmt : m_payment_tokens) {
					xml.append(pmt);
				}
				xml.append("</payment>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_receipt_stored")) {
				xml.append("<filename>" + m_filename + "</filename>");
			}
			else if (m_opcode.equalsIgnoreCase("REQ_delete_receipts")) {
				xml.append("<folder_index>" + m_folder_index + "</folder_index>");
				xml.append("<readwritecap>" + new String(m_readwritecap));
				xml.append("</readwritecap>");
				xml.append("<receipt_del_list>");
				for (String rct : m_receipt_del_list) {
					xml.append("<filename>" + rct + "</filename>");
				}
				xml.append("</receipt_del_list>");
			}
			else if (m_opcode.equalsIgnoreCase("REP_receipts_deleted")) {
				xml.append("<receipt_del_list>");
				for (String rct : m_receipt_del_list) {
					xml.append("<filename>" + rct + "</filename>");
				}
				xml.append("</receipt_del_list>");
			}
			// else: there are no special fields
		}

		xml.append("</vsc_sds>");
		return xml.toString();
	}

	/**
	 * finalize object when garbage-collected
	 */
	@Override
	protected void finalize() throws Throwable {
		// zero out sensitive data arrays if they are present
		try {
			if (m_readcap != null) {
				Arrays.fill(m_readcap, ' ');
			}
			if (m_readwritecap != null) {
				Arrays.fill(m_readwritecap, ' ');
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
