/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	06-17-09/kbw/ handle <token>s base64-encoded
 * 	02-10-09/kbw/ add filename attributes to voucher list
 * 	01-15-09/kbw/ add <receipt_del_list/> to REP_receipts_deleted
 * 	12-30-08/kbw/ new
 *
 * purpose---
 * 	Parse {@link SDSMessage} packets.  Parses the <tt>vsc_sds</tt>
 * 	sub-document and creates an instance of {@link SDSMessage}.
 *
 * notes---
 * 	see underlying message definitions in vsc_sds.x
 */

package org.silentvault.client;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.util.StringUtils;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.IQProvider;


/**
 * Implements the IQProvider interface to receive and parse incoming messages
 * from a SDS (Secure Data Store) according to opcode in the vsc_sds
 * element of the jabber:iq:voucher-safe#vsc-sds namespace.  Creates a
 * {@link SDSMessage} to hold the parsed message.
 *
 * @author Kevin Wilkerson
 */
public final class SDSMessageProvider implements IQProvider {
	// members
	private static final String	M_EndTag = "vsc_sds";
	
	// methods
	/**
	 * constructor
	 */
	public SDSMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		SDSMessage sdsMsg = new SDSMessage();
		// NB: these values should be vsc_sds and jabber:iq:voucher-safe#vsc-sds
		sdsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		sdsMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		String opcode = "";
		int eventType = parser.next();
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		sdsMsg.setOpcode(opcode);

		// Now based on the opcode, pull out the other data.  (replies only)

		boolean done = false;
		if (opcode.equalsIgnoreCase("REP_voucher_list")) {
			// a <voucher_list/> should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("voucher_list")) {
						// look for nested <voucher_blob/> elements
						parseVoucherList(parser, sdsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_token_list")) {
			// a <token_list/> should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("token_list")) {
						// look for nested <token/> elements
						parseTokenList(parser, sdsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_payment_list")) {
			// a <payment_list/> should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("payment_list")) {
						// look for nested <payment_blob/> elements
						parsePaymentList(parser, sdsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_receipt_list")) {
			// a <receipt_list/> should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("receipt_list")) {
						// look for nested <receipt_blob/> elements
						parseReceiptList(parser, sdsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_payment_stored")
					|| opcode.equalsIgnoreCase("REP_receipt_stored"))
		{
			// a <filename/> should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("filename")) {
						sdsMsg.setFilename(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if(opcode.equalsIgnoreCase("REP_receipts_deleted")) {
			// a confirmation list of files actually deleted should be present
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("receipt_del_list")) {
						// look for nested <filename/> elements
						parseReceiptDelList(parser, sdsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_payment_deleted")) {
			// no special fields for this, beyond the opcode
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						sdsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						sdsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						sdsMsg.setErrmsg(parser.nextText());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return sdsMsg;
	}

	/**
	 * utility sub-method to parse the supplied list of voucher blobs
	 * @param parser the parser, positioned at the start of the list
	 * @param sdsMsg the message being built
	 */
	private void parseVoucherList(XmlPullParser parser, SDSMessage sdsMsg)
		throws Exception
	{
		if (parser == null || sdsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("voucher_blob")) {
					// obtain the file name, from attribute filename="..."
					if (parser.getAttributeCount() < 1) {
						throw new Exception("filename missing on <voucher_blob>");
					}
					String fname = parser.getAttributeValue(0).trim();
					if (fname.isEmpty()) {
						throw new Exception("filename missing on <voucher_blob>");
					}

					// obtain voucher blob itself
					String voucher = parser.nextText().trim();
					if (voucher.isEmpty()) {
						throw new Exception("Empty <voucher_blob/>");
					}

					// add this voucher blob to list
					sdsMsg.addVoucherToList(fname, voucher);
				}
				else {
					throw new Exception("Invalid element in voucher_list, "
												+ parser.getName());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("voucher_list")) {
					done = true;
				}
			}
		}
	}

	/**
	 * utility sub-method to parse the supplied list of payment tokens
	 * @param parser the parser, positioned at the start of the list
	 * @param sdsMsg the message being built
	 */
	private void parseTokenList(XmlPullParser parser, SDSMessage sdsMsg)
		throws Exception
	{
		if (parser == null || sdsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				String tagNm = parser.getName();
				if (tagNm.equals("token")) {
					String tokenContents = parser.nextText().trim();
					if (tokenContents.isEmpty()) {
						throw new Exception("Empty <token/>");
					}

					// Add this token to the list.  The SDS ships it to us base64
					// encoded so that we don't have to parse it in detail here.
					String token
						= new String(StringUtils.decodeBase64(tokenContents));
					sdsMsg.addTokenToList(token);
				}
				else {
					throw new Exception("Invalid element in token_list, " + tagNm);
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("token_list")) {
					done = true;
				}
			}
		}
	}

	/**
	 * utility sub-method to parse the supplied list of pending payments
	 * @param parser the parser, positioned at the start of the list
	 * @param sdsMsg the message being built
	 */
	private void parsePaymentList(XmlPullParser parser, SDSMessage sdsMsg)
		throws Exception
	{
		if (parser == null || sdsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("payment_blob")) {
					// obtain the file name, from attribute filename="..."
					if (parser.getAttributeCount() < 1) {
						throw new Exception("filename missing on <payment_blob>");
					}
					String fname = parser.getAttributeValue(0).trim();
					if (fname.isEmpty()) {
						throw new Exception("filename missing on <payment_blob>");
					}

					// obtain payment blob itself
					String payment = parser.nextText().trim();
					if (payment.isEmpty()) {
						throw new Exception("Empty <payment_blob/>");
					}

					// add this payment to the list
					sdsMsg.addPaymentToList(fname, payment);
				}
				else {
					throw new Exception("Invalid element in payment_list, "
												+ parser.getName());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("payment_list")) {
					done = true;
				}
			}
		}
	}

	/**
	 * utility sub-method to parse the supplied list of receipts
	 * @param parser the parser, positioned at the start of the list
	 * @param sdsMsg the message being built
	 */
	private void parseReceiptList(XmlPullParser parser, SDSMessage sdsMsg)
		throws Exception
	{
		if (parser == null || sdsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("receipt_blob")) {
					// obtain the file name, from attribute filename="..."
					if (parser.getAttributeCount() < 1) {
						throw new Exception("filename missing on <receipt_blob>");
					}
					String fname = parser.getAttributeValue(0).trim();
					if (fname.isEmpty()) {
						throw new Exception("filename missing on <receipt_blob>");
					}

					// obtain receipt blob itself
					String receipt = parser.nextText().trim();
					if (receipt.isEmpty()) {
						throw new Exception("Empty <receipt_blob/>");
					}

					// add this receipt to the list
					sdsMsg.addReceiptToList(fname, receipt);
				}
				else {
					throw new Exception("Invalid element in receipt_list, "
												+ parser.getName());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("receipt_list")) {
					done = true;
				}
			}
		}
	}

	/**
	 * utility sub-method to parse the supplied list of deleted receipt filenames
	 * @param parser the parser, positioned at the start of the list
	 * @param sdsMsg the message being built
	 */
	private void parseReceiptDelList(XmlPullParser parser, SDSMessage sdsMsg)
		throws Exception
	{
		if (parser == null || sdsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("filename")) {
					String fname = parser.nextText().trim();
					if (fname.isEmpty()) {
						throw new Exception("filename missing in <receipt_del_list>");
					}

					// add this receipt filename to the list
					sdsMsg.addReceiptToDelList(fname);
				}
				else {
					throw new Exception("Invalid element in receipt_del_list, "
												+ parser.getName());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("receipt_del_list")) {
					done = true;
				}
			}
		}
	}
}
