/*
 * last modified---
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	Parse {@link SVGSVXMessage} packets.  Parses the <tt>vsc_ofs</tt>
 * 	sub-document and creates an instance of {@link SVGSVXMessage}.
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs-svx.xml
 */

package org.silentvault.client;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.packet.PacketExtension;
import org.jivesoftware.smack.provider.IQProvider;
import org.jivesoftware.smack.provider.PacketExtensionProvider;


/**
 * Implements the IQProvider interface to receive and parse replies from an OFS
 * using the vsc_ofs element within the jabber:iq:silent-vault#vsc-ofs
 * namespace.  Implements the PacketExtensionProvider interface to receive and
 * parse incoming messages from a OFS-SVX, according to the vsc_ofs element of
 * the jabber:message:silent-vault#vsc-ofs namespace.
 * Uses the XmlPullParser to interpret each message according to its opcode,
 * and creates a {@link SVGSVXMessage} object to hold it, or to represent the
 * packet extention.
 *
 * @author Justin Turrell
 */
public final class SVGSVXMessageProvider
	implements IQProvider, PacketExtensionProvider
{
	// members
	private static final String	M_EndTag = "vsc_ofs";
	
	// methods
	/**
	 * constructor
	 */
	public SVGSVXMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		SVGSVXIQ ofsMsg = new SVGSVXIQ();
		// NB: these values should be vsc_ofs and jabber:iq:silent-vault#vsc-ofs
		ofsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		ofsMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		int eventType = parser.next();
		String opcode = "";
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				Log.error("Missing opcode");
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			Log.error("Missing opcode");
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		ofsMsg.setOpcode(opcode);

		/* Now based on the opcode, pull out the other data.
		 * NB: any reply can have errcode/errmsg, and we cannot tell (from here)
		 * what the overall type of the IQ packet is.
		 */
		if (opcode.equalsIgnoreCase("REP_svx_message")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrCode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrMsg(parser.nextText());
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("rep_blob")) {
						ofsMsg.setRepBlob(parser.nextText());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			Log.error("Unsupported opcode " + opcode);
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return ofsMsg;
	}

	/**
	 * parse the packet extension (used with a Message packet)
	 * @param parser the parser, positioned at the start of the extension
	 * @return the extension element
	 */
	public PacketExtension parseExtension(XmlPullParser parser) throws Exception
	{
		SVGSVXMessage ofsMsg = new SVGSVXMessage();
		// NB: these values should be vsc_ofs and
		// 	 jabber:message:silent-vault#vsc-ofs
		ofsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));

		// first, determine the opcode
		int eventType = parser.next();
		String opcode = "";
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				Log.error("Missing opcode");
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			Log.error("Missing opcode");
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		ofsMsg.setOpcode(opcode);

		/* Now based on the opcode, pull out the other data.
		 * NB: no errors are passed, these are actually Message types not IQs.
		 */
		if (opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("level")) {
						ofsMsg.setLevel(parser.nextText());
					}
					else if (parser.getName().equals("text")) {
						ofsMsg.setText(parser.nextText());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			Log.error("Unsupported opcode " + opcode);
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return ofsMsg;
	}
}
