/*
 * last modified---
 * 	08-24-16/jet/ add pluginVersion field to REP_login_phase1 and
 * 					  REP_create_acct1
 * 	11-12-14/jet/ put back REP_xfer_confirmed
 * 	10-03-14/jet/ split off generic items from SSMBlock; add <pluginClass/>
 * 	10-02-14/jet/ cosmetic
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	class to instantiate a reply message from the SVM to the VSC
 *
 * notes---
 * 	maintain in parallel with org.silentsilver.market.SSMBlock etc.
 *
 * 	these messages are delivered embedded inside a <rep_blob/> element
 * 	within a REP_svm_message message routed from the OFS
 *
 * 	see underlying message definitions in vsc_svm.xml and vsc_ofs-svm.xml
 */

package org.silentvault.client;

import org.dom4j.Element;

import org.jivesoftware.smack.util.Base64;

import java.util.Date;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.InvalidKeyException;


/**
 * This class instantiates server replies generated by the SVM to a connected
 * client VSC via the OFS gateway.  It knows how to emit itself as a
 * correctly formatted XML body based on the embedded opcode.  It also
 * deals with validating signatures, and representing error returns.
 *
 * This class should be extended for a fully functional SVM, and in particular
 * the methods getXML() and parseBlock() overridden.
 *
 * @author Justin Turrell
 */
public class SVMBlock {
	// BEGIN data members
	/**
	 * whether this object is representing an error (if true) or a normal reply
	 */
	protected boolean			m_IsError;

	/**
	 * opcode to be sent to the client
	 */
	protected String			m_Opcode;

	/**
	 * if this is an error message, the numeric HTTP-style error code
	 */
	protected int				m_ErrCode;

	/**
	 * if this is an error message, the textual error message
	 */
	protected String			m_ErrMsg;

	/**
	 * signature of SVM on data block, base64-encoded
	 */
	protected String			m_Signature;

	/**
	 * challenge ramdom data, base64-encoded (used with opcodes REP_create_acct1
	 * and REP_login_phase1)
	 */
	protected String			m_Challenge;

	/**
	 * the class we must have available to support the SVM (used with opcodes
	 * REP_create_acct1 and REP_login_phase1)
	 */
	protected String			m_PluginClass;

	/**
	 * the version we must have available to support the SVM (used with opcodes
	 * REP_create_acct1 and REP_login_phase1)
	 */
	protected String			m_PluginVersion;

	/**
	 * session identifier assigned by SVM at successful login (used with opcodes
	 * REP_acct_created and REP_logged_in)
	 */
	protected String			m_SessionId;

	/**
	 * series Id (such as SS-S[0-6]) (used with opcode REQ_wallet_xfer)
	 */
	protected String			m_SeriesId;

	/**
	 * wallet transfer Id assigned (used with opcode REP_wallet_xfer)
	 */
	protected int				m_TransId;

	/**
	 * wallet Id involved on SVM's behalf in a wallet transfer in or out
	 * (used with opcode REP_wallet_xfer)
	 */
	protected String			m_WalletId;

	/**
	 * new balance after an inbound wallet transfer (used with opcode
	 * REP_xfer_confirmed)
	 */
	protected double			m_NewBal;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param isErr true if this block represents a returned error message
	 */
	public SVMBlock(boolean isErr) {
		m_IsError = isErr;
		m_Opcode = m_ErrMsg = m_Signature = m_Challenge = m_SessionId = "";
		m_SeriesId = m_WalletId = m_PluginClass = m_PluginVersion = "";
	}

	/**
	 * nullary constructor (equivalent to invoking SVMBlock(false))
	 */
	public SVMBlock() {
		this(false);
	}


	// get accessor methods
	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the error code, if any
	 * @return the code, or 0 if we're not an error message
	 */
	public int getErrCode() {
		if (m_IsError) {
			return m_ErrCode;
		}
		return 0;
	}

	/**
	 * obtain the error message, if any
	 * @return the error text, or the empty string if we're not an error message
	 */
	public String getErrMsg() {
		if (m_IsError) {
			return m_ErrMsg;
		}
		return "";
	}

	/**
	 * obtain whether this block is representing an error
	 * @return true if error
	 */
	public boolean isError() { return m_IsError; }

	/**
	 * obtain the server's signature
	 * @return the digital privkey signature provided on other fields
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the challenge data
	 * @return the generated random data, base64-encoded
	 */
	public String getChallenge() { return m_Challenge; }

	/**
	 * obtain the plugin class
	 * @return the name of the class required to talk to this SVM instance
	 */
	public String getPluginClass() { return m_PluginClass; }

	/**
	 * obtain the plugin version
	 * @return the version of the class required to talk to this SVM instance
	 */
	public String getPluginVersion() { return m_PluginVersion; }

	/**
	 * obtain the session Id
	 * @return the Id assigned to the client's session
	 */
	public String getSessionId() { return m_SessionId; }

	/**
	 * obtain the series Id
	 * @return the series name, i.e. SS-S[0-6]
	 */
	public String getSeriesId() { return m_SeriesId; }

	/**
	 * obtain the wallet transfer Id
	 * @return tId the id assigned in the database
	 */
	public int getTransId() { return m_TransId; }

	/**
	 * obtain the wallet Id used in a transfer by the SVM
	 * @return the id (VS#) of the wallet
	 */
	public String getWalletId() { return m_WalletId; }

	/**
	 * obtain the new balance after a wallet transfer
	 * @return the balance in the affected asset/series
	 */
	public double getNewBal() { return m_NewBal; }


	// set accessor methods
	/**
	 * set the opcode returned in the client reply
	 * @param op the opcode to set (assumed to be valid)
	 */
	public void setOpcode(String op) {
		if (op != null && op.startsWith("REP_")) {
			m_Opcode = new String(op);
		}
		else {
			Log.error("Improper outgoing opcode, " + op);
		}
	}

	/**
	 * set the error code (also sets m_IsError to true)
	 * @param code the error code value to set
	 */
	public void setErrCode(int code) {
		if (code > 0) {
			m_ErrCode = code;
			m_IsError = true;
		}
	}

	/**
	 * set the error text (also sets m_IsError to true)
	 * @param text the error message text
	 */
	public void setErrMsg(String text) {
		if (text != null && !text.isEmpty()) {
			m_ErrMsg = new String(text);
			m_IsError = true;
		}
	}

	/**
	 * set the server's signature
	 * @param sig the privkey signature of the SVM on other fields
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_Signature = new String(sig);
		}
	}

	/**
	 * set the ramdom challenge data
	 * @param chal the data, base64-encoded
	 */
	public void setChallenge(String chal) {
		if (chal != null && !chal.isEmpty()) {
			m_Challenge = new String(chal);
		}
	}

	/**
	 * set the plugin class name
	 * @param plugin the name of the class which we must have included
	 */
	public void setPluginClass(String plugin) {
		if (plugin != null && !plugin.isEmpty()) {
			m_PluginClass = new String(plugin);
		}
	}

	/**
	 * set the plugin class version
	 * @param version the version of the class which we must have included
	 * (can be empty)
	 */
	public void setPluginVersion(String version) {
		if (version != null) {
			m_PluginVersion = new String(version);
		}
	}

	/**
	 * set the sessionId
	 * @param sId the id assigned to the user's session
	 */
	public void setSessionId(String sId) {
		if (sId != null && !sId.isEmpty()) {
			m_SessionId = new String(sId);
		}
	}

	/**
	 * set the series
	 * @param sId the id (name) of the series
	 */
	public void setSeriesId(String sId) {
		if (sId != null && !sId.isEmpty()) {
			m_SeriesId = new String(sId);
		}
	}

	/**
	 * set the wallet transfer Id
	 * @param tId the id from the database
	 */
	public void setTransId(int tId) {
		if (tId > 0) {
			m_TransId = tId;
		}
	}

	/**
	 * set the wallet Id used in a transfer
	 * @param wId the id (VS#) of the wallet used by the SVM
	 */
	public void setWalletId(String wId) {
		if (wId != null && wId.indexOf('-') != -1) {
			m_WalletId = new String(wId);
		}
	}

	/**
	 * set the new balance in the affected asset after a wallet transfer
	 * @param bal the amount of the user's new balance in asset
	 */
	public void setNewBal(double bal) {
		if (bal >= 0.0) {
			m_NewBal = bal;
		}
	}


	/**
	 * method to parse a decrypted block from the SVM, and set field values.
	 * Derived classes should @Override this method.
	 * @param svmBlock the decrypted text, including delimiters
	 * @return true on successful parse, false otherwise
	 */
	public boolean parseBlock(String svmBlock) {
		if (svmBlock == null || !svmBlock.startsWith("<svmBlock>")) {
			Log.error("Missing or improper SVM block, cannot parse");
			return false;
		}
		Element svmElt = EncodingUtils.buildElement(svmBlock);
		if (svmElt == null) {
			Log.error("<svmBlock/> parse failure");
			return false;
		}

		// <opcode/> must always be present
		Element opElt = svmElt.element("opcode");
		if (opElt == null) {
			Log.error("Missing svmBlock opcode");
			return false;
		}
		String opcode = opElt.getTextTrim();
		setOpcode(opcode);

		// try to determine if we're an error (will have errcode and/or errmsg)
		Element eCodeElt = svmElt.element("errcode");
		if (eCodeElt != null) {
			setErrCode(Integer.parseInt(eCodeElt.getText()));
		}
		else if (m_IsError) {
			Log.error("Missing errcode for error " + opcode);
			return false;
		}

		Element eMsgElt = svmElt.element("errmsg");
		if (eMsgElt != null) {
			setErrMsg(eMsgElt.getTextTrim());
		}
		else if (m_IsError) {
			Log.error("Missing errmsg for error " + opcode);
			return false;
		}
		if (m_IsError) {
			// done after these fields
			return true;
		}

		// otherwise examine the opcode for further fields
		if (opcode.equalsIgnoreCase("REP_create_acct1")
			 || opcode.equalsIgnoreCase("REP_login_phase1"))
		{
			// need the challenge data
			Element chalElt = svmElt.element("challenge");
			if (chalElt == null) {
				Log.error("Missing challenge for " + opcode);
				return false;
			}
			setChallenge(chalElt.getTextTrim());

			// need a plugin class name
			Element plugElt = svmElt.element("pluginClass");
			if (plugElt == null) {
				Log.error("Missing pluginClass for " + opcode);
				return false;
			}
			setPluginClass(plugElt.getTextTrim());

			// optionally need a plugin class version (which can be empty)
			Element versElt = svmElt.element("pluginVersion");
			if (versElt == null) {
				Log.warning("Missing pluginVersion for " + opcode);
			}
			else {
				setPluginVersion(versElt.getTextTrim());
			}
		}
		else if (opcode.equalsIgnoreCase("REP_acct_created")
					|| opcode.equalsIgnoreCase("REP_logged_in"))
		{
			// need session Id
			Element sessElt = svmElt.element("sessionId");
			if (sessElt == null) {
				Log.error("Missing sessionId for " + opcode);
				return false;
			}
			setSessionId(sessElt.getTextTrim());

			// need SVM's signature
			Element sigElt = svmElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_logged_out")) {
			// no fields here
		}
		else if (opcode.equalsIgnoreCase("REP_wallet_xfer")) {
			// need entered transId
			Element tIdElt = svmElt.element("transId");
			if (tIdElt == null) {
				Log.error("Missing transId for " + opcode);
				return false;
			}
			setTransId(Integer.parseInt(tIdElt.getText()));


			// need SVM's signature
			Element sigElt = svmElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_xfer_confirmed")) {
			// need echoed transId
			Element tIdElt = svmElt.element("transId");
			if (tIdElt == null) {
				Log.error("Missing transId for " + opcode);
				return false;
			}
			setTransId(Integer.parseInt(tIdElt.getText()));

			// need new balance after transfer
			Element balElt = svmElt.element("newBal");
			if (balElt == null) {
				Log.error("Missing newBal for " + opcode);
				return false;
			}
			setNewBal(Double.parseDouble(balElt.getText()));

			// need SVM's signature
			Element sigElt = svmElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else {
			Log.error("Unrecognized reply opcode, " + opcode);
			return false;
		}
		return true;
	}

	/**
	 * method to construct the block XML, based on opcode and member values.
	 * Any signature should be checked using the string returned by this method.
	 * Missing required output fields generate log errors, but do not stop the
	 * generation of output.  Derived classes should @Override this method.
	 * @return the complete block for the indicated reply, absent the delimiters
	 * and the SVM's signature (if required)
	 */
	public String getXML() {
		StringBuilder xml = new StringBuilder(2048);

		// if we're an error message, this is just the code and error text
		xml.append("<opcode>" + m_Opcode + "</opcode>");
		if (m_IsError) {
			xml.append("<errcode>" + m_ErrCode + "</errcode>");
			xml.append("<errmsg>" + m_ErrMsg + "</errmsg>");
			return xml.toString();
		}

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// base our output on the opcode
		if (m_Opcode.equals("REP_create_acct1")
			 || m_Opcode.equals("REP_login_phase1"))
		{
			if (m_Challenge.isEmpty()) {
				Log.error("Missing challenge for " + m_Opcode);
			}
			xml.append("<challenge>" + m_Challenge + "</challenge>");
			if (m_PluginClass.isEmpty()) {
				Log.error("Missing pluginClass for " + m_Opcode);
			}
			xml.append("<pluginClass>" + m_PluginClass + "</pluginClass>");
			xml.append("<pluginVersion>" + m_PluginVersion + "</pluginVersion>");
		}
		else if (m_Opcode.equals("REP_acct_created")
					|| m_Opcode.equals("REP_logged_in"))
		{
			if (m_SessionId.isEmpty()) {
				Log.error("Missing sessionId for " + m_Opcode);
			}
			xml.append("<sessionId>" + m_SessionId + "</sessionId>");
		}
		else if (m_Opcode.equals("REP_logged_out")) {
			// no special fields for this
		}
		else if (m_Opcode.equals("REP_wallet_xfer")) {
			xml.append("<transId>" + m_TransId + "</transId>");
			xml.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equals("REP_xfer_confirmed")) {
			xml.append("<transId>" + m_TransId + "</transId>");
			xml.append("<newBal>" + df.format(m_NewBal) + "</newBal>");
		}
		else {
			Log.error("Unrecognized reply opcode, " + m_Opcode);
		}
		return xml.toString();
	}

	/**
	 * validate the signature on the block, given the SVM's public key
	 * @param pubKey the public key required to validate the signature
	 * @return true if the signature validates, false otherwise
	 */
	public boolean sigVerify(PublicKey pubKey) {
		String blockXML = getXML();
		if (blockXML.isEmpty() || m_Signature.isEmpty()) {
			Log.error("Missing block data or signature for sig check");
			return false;
		}
		if (pubKey == null) {
			Log.error("No key passed for sig check");
			return false;
		}

		Signature vSig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initVerify(pubKey);
			vSig.update(blockXML.getBytes());
			return vSig.verify(Base64.decode(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate client signature", se);
		}
		return false;
	}

	// END methods
}
