/*
 * last modified---
 * 	03-05-15/jet/ make PIN a string interpreted as an int, not an int
 * 	10-03-14/jet/ new, generic items abstracted from SSMClientBlock
 *
 * purpose---
 * 	class to instantiate a message from the VSC to the SVM
 *
 * notes---
 * 	maintain in parallel with org.silentsilver.market.ClientBlock etc.
 *
 * 	these messages are delivered embedded inside a <msg_blob/> element
 * 	within a REQ_svx_message message routed through the OFS-SVG
 * 	
 * 	see underlying message definitions in vsc_svm.xml
 */

package org.silentvault.client;

import org.jivesoftware.smack.util.Base64;

import java.util.Date;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * This class instantiates client requests sent to the SVM from a connected
 * VSC via the OFS gateway.  It knows how to emit the various message fields
 * based on the contained opcode, and affix client signatures.  Only outbound
 * message fields need to be supported here.
 *
 * This base class should be extended for any full function SVM, and in
 * particular the methods signXML() and buildSigData() should be overridden.
 *
 * @author Justin Turrell
 */
public class SVMClientBlock {
	// BEGIN data members
	/**
	 * native XML as sent in the msgBlob element by the client
	 */
	protected String			m_ClientBlockXML;

	/**
	 * opcode sent by the client
	 */
	protected String			m_Opcode;

	/**
	 * session identifier assigned by SVM at login.  Used with all opcodes
	 * employed after login, given by b64(sha512(wallet ID + PIN + login time)).
	 */
	protected String			m_SessionId;

	/**
	 * privkey signature of client, used with opcodes that make changes on the
	 * server to the user's account
	 */
	protected String			m_Signature;

	/**
	 * walletId of the VSC (used with opcodes REQ_create_acct1 and
	 * REP_wallet_xfer)
	 */
	protected String			m_WalletId;

	/**
	 * PIN entered by the VSC (should be int, used with opcode REQ_create_acct1)
	 */
	protected String			m_PIN;

	/**
	 * account hash, constructed as b64(sha512(wallet ID + PIN)).  Used with
	 * opcodes REQ_create_acct2, REQ_login_phase1, and REQ_login_phase2.
	 */
	protected String			m_AcctHash;

	/**
	 * JID at which VSC is connected to the OFS (used with opcodes
	 * REQ_create_acct2 and REQ_login_phase2)
	 */
	protected String			m_JID;

	/**
	 * reply to a login challenge (used with opcodes REQ_create_acct2 and
	 * REQ_login_phase2)
	 */
	protected String			m_ChallengeReply;

	/**
	 * the series ID involved in an operation (used with opcode REQ_wallet_xfer)
	 */
	protected String			m_SeriesId;

	/**
	 * the order quantity (used with opcode REQ_wallet_xfer)
	 */
	protected Double			m_Quantity;

	/**
	 * wallet transfer direction, either in or out (used with opcode
	 * REQ_wallet_xfer)
	 */
	protected String			m_Direction;

	/**
	 * wallet escrow transaction Id (assigned by server)
	 */
	protected int				m_TransId;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public SVMClientBlock() {
		m_Opcode = m_SessionId = m_WalletId = m_AcctHash = m_ChallengeReply = "";
		m_Signature = m_SeriesId = m_ClientBlockXML = m_Direction = m_JID = "";
		m_PIN = "";
		m_Quantity = new Double(0.0);
	}


	// get access methods
	/**
	 * obtain original XML as sent from client
	 * @return the unparsed XML
	 */
	public String getClientBlockXML() { return m_ClientBlockXML; }

	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the session identifier
	 * @return the ID assigned to the user's session
	 */
	public String getSessionId() { return m_SessionId; }

	/**
	 * obtain the client's signature, for opcodes which require one
	 * @return the digital privkey signature provided on m_ClientBlockXML
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the wallet ID
	 * @return the ID of the wallet for which the account is to be created
	 */
	public String getWalletId() { return m_WalletId; }

	/**
	 * obtain the PIN
	 * @return the PIN value to be used to build the hash for the new account
	 */
	public String getPIN() { return m_PIN; }

	/**
	 * obtain the account hash value
	 * @return the hash used to look up the account in the database
	 */
	public String getAcctHash() { return m_AcctHash; }

	/**
	 * obtain the connected JID
	 * @return the user@host/resource at which the user is connected at the OFS
	 */
	public String getJID() { return m_JID; }

	/**
	 * obtain the challenge reply sent
	 * @return the data, base64-encoded
	 */
	public String getChallengeReply() { return m_ChallengeReply; }

	/**
	 * obtain the series ID
	 * @return the series, or voucher asset name
	 */
	public String getSeriesId() { return m_SeriesId; }

	/**
	 * obtain the order quantity requested
	 * @return the quantity
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * obtain the wallet transfer direction
	 * @return the direction of transfer, either in or out
	 */
	public String getDirection() { return m_Direction; }

	/**
	 * obtain transfer Id of wallet transfer
	 * @return the Id assigned by the server
	 */
	public int getTransId() { return m_TransId; }


	// set access methods
	/**
	 * re-record (override) the XML received in the client message
	 * @param xml the original unparsed XML, exclusive of delimiters
	 */
	public void setClientBlockXML(String xml) {
		if (xml != null && !xml.isEmpty()) {
			m_ClientBlockXML = new String(xml);
		}
	}

	/**
	 * set the opcode received in the client message
	 * @param op the opcode to set (assumed to be valid)
	 */
	public void setOpcode(String op) {
		if (op != null && op.startsWith("REQ_")) {
			m_Opcode = new String(op);
		}
		else {
			Log.error("Improper incoming opcode, " + op);
		}
	}

	/**
	 * set the session identifier
	 * @param sessId the ID to record
	 */
	public void setSessionId(String sessId) {
		if (sessId != null && !sessId.isEmpty()) {
			m_SessionId = new String(sessId);
		}
	}

	/**
	 * set the client's signature
	 * @param sig the privkey signature of the client on other fields
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_Signature = new String(sig);
		}
	}

	/**
	 * set the wallet ID for a new account
	 * @param wId the wallet ID (VS#) for which we're to create an account
	 */
	public void setWalletId(String wId) {
		if (wId != null && wId.indexOf('-') != -1) {
			m_WalletId = new String(wId);
		}
		else {
			Log.error("Illegal wallet Id, " + wId);
		}
	}

	/**
	 * set the PIN for construction of a new account hash
	 * @param pin the PIN to use
	 */
	public void setPIN(String pin) {
		if (pin != null && !pin.isEmpty()) {
			m_PIN = new String(pin);
		}
	}

	/**
	 * set the account hash value
	 * @param hash the hash value used to find the account in the database
	 */
	public void setAcctHash(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_AcctHash = new String(hash);
		}
	}

	/**
	 * set the JID at which client is connected to the OFS
	 * @param jid the JID of the VSC (assumed to be valid)
	 */
	public void setJID(String jid) {
		if (jid != null && !jid.isEmpty()) {
			m_JID = new String(jid);
		}
	}

	/**
	 * set the challenge reply
	 * @param reply the reply, base64-encoded decrypted hash sent by SVM
	 */
	public void setChallengeReply(String reply) {
		if (reply != null && !reply.isEmpty()) {
			m_ChallengeReply = new String(reply);
		}
	}

	/**
	 * set the series ID
	 * @param sId the series involved in the operation, one of SS-S[1-6]
	 */
	public void setSeriesId(String sId) {
		if (sId != null && !sId.isEmpty()) {
			m_SeriesId = new String(sId);
		}
		else {
			Log.error("Unrecognized series, " + sId);
		}
	}

	/**
	 * set the order quantity
	 * @param amt the amount of the particular series involved, in DSO
	 */
	public void setQuantity(double amt) {
		if (amt > 0.0) {
			m_Quantity = new Double(amt);
		}
	}

	/**
	 * set the wallet transfer direction
	 * @param dir the direction, either in or out
	 */
	public void setDirection(String dir) {
		if (dir != null && (dir.equals("in") || dir.equals("out"))) {
			m_Direction = new String(dir);
		}
		else {
			Log.error("Illegal wallet transfer direction, " + dir);
		}
	}

	/**
	 * set the wallet transfer Id
	 * @param tId the Id to be set
	 */
	public void setTransId(int tId) {
		m_TransId = tId;
	}


	/**
	 * generate the data which would be encrypted (and possibly signed)
	 */
	protected void buildSigData() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// build data based on the opcode
		StringBuilder sigData = new StringBuilder(1024);
		sigData.append("<opcode>" + m_Opcode + "</opcode>");
		if (m_Opcode.equalsIgnoreCase("REQ_create_acct1")) {
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
			sigData.append("<PIN>" + m_PIN + "</PIN>");
			// NB: this opcode is unsigned
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_create_acct2")
					|| m_Opcode.equalsIgnoreCase("REQ_login_phase2"))
		{
			sigData.append("<acctHash>" + m_AcctHash + "</acctHash>");
			sigData.append("<JID>" + m_JID + "</JID>");
			sigData.append("<challengeReply>" + m_ChallengeReply);
			sigData.append("</challengeReply>");
			// NB: these opcodes are unsigned
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_login_phase1")) {
			sigData.append("<acctHash>" + m_AcctHash + "</acctHash>");
			// NB: this opcode is unsigned
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_logout")) {
			sigData.append("<sessionId>" + m_SessionId + "</sessionId>");
			// NB: these opcodes are unsigned
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_wallet_xfer")) {
			sigData.append("<sessionId>" + m_SessionId + "</sessionId>");
			sigData.append("<direction>" + m_Direction + "</direction>");
			sigData.append("<seriesId>" + m_SeriesId + "</seriesId>");
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_confirm_xfer")) {
			sigData.append("<sessionId>" + m_SessionId + "</sessionId>");
			sigData.append("<transId>" + m_TransId + "</transId>");
		}
		else {
			// opcode is not signed by client
			Log.error(m_Opcode + " not recognized");
		}
		m_ClientBlockXML = sigData.toString();
	}

	/**
	 * apply the signature to the block, given a private key
	 * @param privKey the private key required to create the signature
	 * @return true if the signature was applied, false otherwise
	 */
	public boolean signXML(PrivateKey privKey) {
		if (m_ClientBlockXML.isEmpty()) {
			Log.error("No client block XML to sign");
			return false;
		}
		// these opcodes are unsigned
		if (m_Opcode.equalsIgnoreCase("REQ_create_acct1")
			 || m_Opcode.equalsIgnoreCase("REQ_create_acct2")
			 || m_Opcode.equalsIgnoreCase("REQ_login_phase1")
			 || m_Opcode.equalsIgnoreCase("REQ_login_phase2")
			 || m_Opcode.equalsIgnoreCase("REQ_logout"))
		{
			return true;
		}

		Signature vSig = null;
		byte[] sig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initSign(privKey);
			vSig.update(m_ClientBlockXML.getBytes());
			sig = vSig.sign();
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate client signature", se);
		}
		if (sig == null) {
			return false;
		}
		setSignature(Base64.encodeBytes(sig, Base64.DONT_BREAK_LINES));
		return true;
	}

	// END methods
}
