/*
 * last modified---
 * 	04-02-15/jet/ restore SVX message filter in resetConnFilter()
 * 	02-18-15/jet/ add sendRequest
 * 	10-02-14/jet/ SSM -> SVM
 * 	06-11-13/jet/ new
 *
 * purpose---
 * 	handle interface to the OFS-SVG plugin which proxies the SVM
 *
 * notes---
 * 	see underlying messages in vsc_ofs-svm.xml and vsc_svm.xml
 */

package org.silentvault.client;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.util.Base64;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.*;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.filter.OrFilter;
import org.jivesoftware.smack.filter.PacketFilter;

import org.silentvault.client.ui.svm.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.security.PublicKey;
import java.net.HttpURLConnection;


/**
 * Handles messaging to/from the SVM, which is proxied by the SVG plugin at
 * the OFS.  The actual namespace is jabber:iq:silent-market#vsc-ofs, but the
 * 'virtual namespace' wrapped by REQ_svm_message / REP_svm_message exchanges
 * is jabber:iq:silent-market#vsc-svm.  This class acts as a listener for
 * requests sent from the SVM via the OFS, for replies to its own requests
 * made to the SVM, and for broadcast messages sent from the SVM.
 *
 * This class should be extended by any fully implemented SVM client-side, in
 * order to support unique opcodes and message fields.  In particular, the
 * method processPacket() should be overridden.
 *
 * @author Justin Turrell
 */
public class SVMListener implements PacketListener, Runnable {
	// data members
	/**
	 * reference to client application
	 */
	protected WalletClient				m_Plugin;

	/**
	 * a map of currently outstanding listen filters, representing all requests
	 * for which we are awaiting replies, keyed by the packet IDs
	 */
	protected Hashtable<String, PacketIDFilter>	m_FilterMap;

	// methods
	/**
	 * constructor
	 * @param plugin the reference to the app of which we are a part
	 */
	public SVMListener(WalletClient plugin) {
		m_Plugin = plugin;
		m_FilterMap = new Hashtable<String, PacketIDFilter>();
	}

	// implement the PacketListener interface
	public void processPacket(Packet packet) {
		MTabManager mtm = m_Plugin.getMarketplaceTabManager();

		// check for a Message notify
		if (packet instanceof Message) {
			/* Look for the extension we recognize.  Due to the filtering, *only*
			 * messages with this extension should be sent here.
			 */
			PacketExtension pExt
				= packet.getExtension("vsc_ofs",
											"jabber:message:silent-market#vsc-ofs");
			if (pExt == null || !(pExt instanceof SVGSVMMessage)) {
				Log.error("Message from SVM without valid extension");
				return;
			}
			// the extension will be embodied in a SVGSVMMessage
			SVGSVMMessage broadCst = (SVGSVMMessage) pExt;
			String opcode = broadCst.getOpcode();
			if (opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
				// based on the severity level, display an appropriate popup
				String level = broadCst.getLevel();
				if (level.equals("info")) {
					mtm.showInfo(broadCst.getText());
				}
				else if (level.equals("warning")) {
					mtm.showWarning(broadCst.getText());
				}
				else {
					mtm.showError(broadCst.getText());

					// if this is a shutdown, disconnect forcibly from the SVM
					// (handle in derived class in SVM-specific plugin)
				}
			}
			else if (opcode.equalsIgnoreCase("NFY_ticker_event")) {
				// TBD
			}
			else {
				Log.error("Weird opcode on Message packet extension, " + opcode);
			}
		}
		// check for IQ reply
		else if (packet instanceof IQ) {
			if (!(packet instanceof SVGSVMIQ)) {
				Log.error("Unexpected IQ packet, " + packet.toXML());
				return;
			}
			SVGSVMIQ svmMsg = (SVGSVMIQ) packet;

			// get the SVM's reply blob
			String repBlob = svmMsg.getRepBlob();
			if (repBlob.isEmpty()) {
				// the only time this is okay is if it's an error packet
				if (svmMsg.getType() == IQ.Type.ERROR) {
					// show error to user
					mtm.showError(svmMsg.getErrMsg(),
										"code " + svmMsg.getErrCode());
				}
				else {
					Log.error("Empty SVM reply blob, cannot process");
				}
				return;
			}
			SVMBlock svmReply = null;
			String reply = null;
			String opcode = null;

			// attempt decrypt
			VsSecrets secrets = m_Plugin.getLoginSecrets();
			String decBlob = EncodingUtils.getStrFromBase64PubkeyEnc(repBlob,
																			secrets.getPrivKey());
			if (decBlob == null) {
				/* Only way we will not have any encryption is if an error
				 * occurred before the SVM had our key available.
				 */
				if (svmMsg.getType() == IQ.Type.ERROR) {
					// try to parse as error in clear
					svmReply = new SVMBlock(true);
					reply = "<svmBlock>" + repBlob + "</svmBlock>";
					if (!svmReply.parseBlock(reply)) {
						Log.error("Unable to parse SVM error reply block");
						// show error to user
						mtm.showError(svmMsg.getErrMsg(),
											"code " + svmMsg.getErrCode());
						return;
					}
					opcode = svmReply.getOpcode();

					// if this is a REQ_ value, then we got a timeout
					if (opcode.startsWith("REQ_")) {
						// show error to user
						mtm.showError(svmReply.getErrMsg(),
										"code " + svmReply.getErrCode());
						return;
					}
				}
				else {
					Log.error("Unable to decrypt SVM reply");
					return;
				}
			}
			else {
				// parse what the SVM sent in its block -- might be an error
				svmReply = new SVMBlock(false);
				reply = "<svmBlock>" + decBlob + "</svmBlock>";
				if (!svmReply.parseBlock(reply)) {
					Log.error("Unable to parse SVM reply block");
					mtm.showError("SVM encrypted reply message did not parse");
					return;
				}
				opcode = svmReply.getOpcode();
			}

			// errors and certain reply opcodes are not signed
			if (!svmReply.isError()
				 && !(opcode.equalsIgnoreCase("REP_create_acct1")
				 		|| opcode.equalsIgnoreCase("REP_login_phase1")
				 		|| opcode.equalsIgnoreCase("REP_logged_out")))
			{
				// validate SVM's signature
				PublicKey svmKey = secrets.getSVMKey();
				if (svmKey == null) {
					Log.error("No SVM pubkey available for sig check");
					return;
				}
				if (!svmReply.sigVerify(svmKey)) {
					Log.error("SVM signature check failed on " + opcode);
					return;
				}
			}
			//Log.debug("Valid SVM reply seen, opcode " + opcode);

			// evaluate opcode and pass to appropriate ui handler
			if (opcode.equalsIgnoreCase("REP_create_acct1")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = mtm.getLoginPane();
				lp.recordCreate1(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_acct_created")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = mtm.getLoginPane();
				lp.recordCreate2(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_login_phase1")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = mtm.getLoginPane();
				lp.recordLogin1(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_logged_in")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = mtm.getLoginPane();
				lp.recordLogin2(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_logged_out")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = mtm.getLoginPane();
				lp.recordLogout(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_wallet_xfer")) {
				// there is no default handler for this
				mtm.showError("Unhandled opcode, " + opcode);
			}
			else if (opcode.equalsIgnoreCase("REP_xfer_confirmed")) {
				// there is no default handler for this
				mtm.showError("Unhandled opcode, " + opcode);
			}
			else {
				Log.error("Unrecognized SVM reply opcode, " + opcode);
			}
		}
		else {
			// must be Presence, not supported in this namespace
			Log.error("Unexpected (Presence?) packet: " + packet.toXML());
		}

		// retire this ID from the list of active packet filters
		removeFilter(packet.getPacketID());
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this thread by Swing.  Also, we need this callback to return quickly so
	 * that further messages may be processed.
	 */
	public void run() {
		// obtain UI tab manager from plugin and do the queued state transition
		m_Plugin.getMarketplaceTabManager().makeTransition();
	}

	/**
	 * add a filter to the list of packets for which we're accepting replies
	 * @param ID the packet ID we're expecting in a reply
	 * @param filter the filter to add, which watches for the ID
	 */
	public void addFilter(String ID, PacketIDFilter filter) {
		if (ID == null || ID.isEmpty() || filter == null
			 || m_FilterMap.containsKey(ID))
		{
			return;
		}
		m_FilterMap.put(ID, filter);
		resetConnFilter();
	}

	/**
	 * remove a filter from the list of replies for which we're listening
	 * @param ID the ID of the packet filter to remove
	 */
	public void removeFilter(String ID) {
		if (ID == null || ID.isEmpty() || !m_FilterMap.containsKey(ID)) {
			// nothing to do
			return;
		}
		m_FilterMap.remove(ID);
		resetConnFilter();
	}

	/**
	 * utility method to rebuild and reset the filter on this connection, by
	 * ORing together all filters extant in the map
	 */
	private synchronized void resetConnFilter() {
		// rebuild filter of all outstanding requests
		OrFilter of = new OrFilter();
		for (String packetId : m_FilterMap.keySet()) {
			PacketIDFilter filt = m_FilterMap.get(packetId);
			of.addFilter(filt);
		}
		XMPPConnection conn = m_Plugin.getOFSConnection();
		conn.removePacketListener(this);
		// add back the generic Message filter if we have one
		PacketFilter msgFilter = m_Plugin.getSVMMsgFilter();
		if (msgFilter != null) {
			conn.addPacketListener(this, msgFilter);
		}

		// add new asynch reply listener if map is non-empty
		if (!m_FilterMap.isEmpty()) {
			conn.addPacketListener(this, of);
		}
	}

	/**
	 * prepare and send a request message to the SVM (via OFS), given a client
	 * block.  NB: this method blocks on the reply from the OFS/SVM.
	 * @param cBlock the client block to send
	 * @return true if message sent successfully
	 */
	public boolean sendRequest(SVMClientBlock cBlock) {
		if (cBlock == null || !cBlock.getOpcode().startsWith("REQ_")) {
			Log.error("Missing or improper SVM client block");
			return false;
		}

		// build output XML and sign it (if required)
		cBlock.buildSigData();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!cBlock.signXML(secrets.getPrivKey())) {
			Log.error("Unable to sign output XML");
			return false;
		}

		// encrypt block, including any signature, with SVM's pubkey
		StringBuilder clearBlock = new StringBuilder(5120);
		clearBlock.append("<clientBlock>");
		clearBlock.append(cBlock.getClientBlockXML());
		if (!cBlock.getSignature().isEmpty()) {
			clearBlock.append("<signature>" + cBlock.getSignature());
			clearBlock.append("</signature>");
		}
		clearBlock.append("</clientBlock>");
		String encBlock
			= EncodingUtils.makeBase64PubkeyEncStr(clearBlock.toString(),
																secrets.getSVMKey());
		if (encBlock == null || encBlock.isEmpty()) {
			Log.error("Error encrypting client block for SVM");
			return false;
		}

		// prepare wrapper message to OFS
		XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		String opcode = cBlock.getOpcode();
		SVGSVMIQ svmReq = new SVGSVMIQ();
		if (opcode.equals("REQ_login_phase1")
			 || opcode.equals("REQ_login_phase2")
			 || opcode.equals("REQ_logout"))
		{
			svmReq.setType(IQ.Type.GET);
		}
		else {
			svmReq.setType(IQ.Type.SET);
		}
		svmReq.setFrom(conn.getUser());
		svmReq.setTo(conn.getServiceName());
		String pId = "svm_" + stData.getNextOFSid();
		svmReq.setPacketID(pId);
		svmReq.setOpcode("REQ_svm_message");
		svmReq.setSVMId(secrets.getSVMId());
		svmReq.setMsgBlob(encBlock);

		// construct a packet collector to listen for reply
		PacketIDFilter pFilter = new PacketIDFilter(pId);
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		//Log.debug("Sending " + opcode + " to SVG, " + clearBlock.toString());
		conn.sendPacket(svmReq);
		stData.setLastActivity(svmReq);

		// wait for reply
		SVGSVMIQ rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new SVGSVMIQ();
			rep.setOpcode("REP_svm_message");
			rep.setFrom(svmReq.getTo());
			rep.setTo(svmReq.getFrom());
			rep.setPacketID(pId);
			rep.setType(IQ.Type.ERROR);
			SVMBlock fakeRep = new SVMBlock(true);
			fakeRep.setErrCode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			fakeRep.setErrMsg("request timed out");
			fakeRep.setOpcode(opcode);
			rep.setRepBlob(fakeRep.getXML());
		}
		else {
			rep = (SVGSVMIQ) reply;
		}
		this.processPacket(rep);
		return true;
	}
}
