/*
 * last modified---
 * 	01-16-15/jet/ add m_SVMHomePane
 * 	10-03-14/jet/ new
 *
 * purpose---
 * 	bean class describing plugin support for SVM client-side classes
 */

package org.silentvault.client;

import org.silentvault.client.ui.svm.MTabManager;

import javax.swing.JPanel;


/**
 * This class describes the classes which must be supplied client-side in
 * order to extend the functionality of an SVM beyond that provided by the
 * base classes.  A plugin will provide an extension of this class, and an
 * extension for each of the data member classes.  The SVM side will indicate
 * the SVMPlugin sub-class which is required, and the client will verify that
 * such a class is available before continuing with a login or account creation.
 *
 * @author Justin Turrell
 */
public class SVMPlugin {
	// data members
	/**
	 * owning master plugin or application
	 */
	protected WalletClient		m_Plugin;

	/**
	 * class describing the listener to be used to handle messages from the SVM
	 */
	protected SVMListener		m_SVMListener;

	/**
	 * class describing the encrypted client block sent to the SVM
	 */
	protected Class				m_ClientBlockClass;

	/**
	 * class describing the encrypted reply block sent by the SVM
	 */
	protected Class				m_SVMBlockClass;

	/**
	 * class describing the home pane to be used with the UI for the SVM
	 */
	protected JPanel				m_SVMHomePane;

	/**
	 * class describing the tab manager to be used to manage the UI for the SVM
	 */
	protected MTabManager		m_SVMTabManager;

	// methods
	/**
	 * constructor
	 * @param plugin the owning master plugin or appl
	 */
	public SVMPlugin(WalletClient plugin) {
		m_Plugin = plugin;
	}

	// get methods
	/**
	 * obtain the message listener
	 * @return the callback for replies from the SVM
	 */
	public SVMListener getSVMListener() { return m_SVMListener; }

	/**
	 * obtain the client block representation
	 * @return instance of the representation of the client-side encrypted block
	 */
	public SVMClientBlock getClientBlock() {
		if (m_ClientBlockClass == null) {
			return null;
		}
		Object newBlock = null;
		try {
			newBlock = m_ClientBlockClass.newInstance();
			return (SVMClientBlock) newBlock;
		}
		catch (Exception eee) {
			Log.error("Unable to create " + m_ClientBlockClass.getName()
						+ " instance", eee);
		}
		return null;
	}

	/**
	 * obtain the server block representation
	 * @return the representation of the server-side encrypted block
	 */
	public SVMBlock getReplyBlock() {
		if (m_SVMBlockClass == null) {
			return null;
		}
		Object newBlock = null;
		try {
			newBlock = m_SVMBlockClass.newInstance();
			return (SVMBlock) newBlock;
		}
		catch (Exception eee) {
			Log.error("Unable to create " + m_SVMBlockClass.getName()
						+ " instance", eee);
		}
		return null;
	}

	/**
	 * obtain the home pane
	 * @return the home pane required for the UI for this SVM
	 */
	public JPanel getHomePane() { return m_SVMHomePane; }

	/**
	 * obtain the tab manager
	 * @return the tab manager required for the UI for this SVM
	 */
	public MTabManager getTabManager() { return m_SVMTabManager; }


	// set methods
	/**
	 * set the owning top-level plugin
	 * @param plugin the plugin which created us
	 */
	public void setPlugin(WalletClient plugin) {
		m_Plugin = plugin;
	}

	/**
	 * set the SVM message listener
	 * @param listener the listener subclass to be used
	 */
	public void setSVMListener(SVMListener listener) {
		if (listener != null) {
			m_SVMListener = listener;
		}
	}

	/**
	 * set the client block class
	 * @param clientBlock the client block representation class name
	 */
	public void setClientBlock(String clientBlock) {
		if (clientBlock != null && !clientBlock.isEmpty()) {
			try {
				m_ClientBlockClass
					= m_Plugin.getClass().getClassLoader().loadClass(clientBlock);
			}
			catch (ClassNotFoundException cnfe) {
				Log.error("Could not load SVM client block class, " + clientBlock,
							cnfe);
			}
		}
	}

	/**
	 * set the reply block class
	 * @param replyBlock the SVM block representation class name
	 */
	public void setReplyBlock(String replyBlock) {
		if (replyBlock != null && !replyBlock.isEmpty()) {
			try {
				m_SVMBlockClass
					= m_Plugin.getClass().getClassLoader().loadClass(replyBlock);
			}
			catch (ClassNotFoundException cnfe) {
				Log.error("Could not load SVM block class, " + replyBlock, cnfe);
			}
		}
	}

	/**
	 * set the home pane
	 * @param hPane the pane to use
	 */
	public void setHomePane(JPanel hPane) {
		if (hPane != null) {
			m_SVMHomePane = hPane;
		}
	}

	/**
	 * set the tab manager
	 * @param mTabMgr the marketplace tab manager
	 */
	public void setTabManager(MTabManager mTabMgr) {
		if (mTabMgr != null) {
			m_SVMTabManager = mTabMgr;
		}
	}
	// end methods
}
