/*
 * last modified---
 * 	08-06-14/jet/ cosmetic
 * 	07-30-13/jet/ handle expired status like filled
 * 	07-18-13/jet/ add m_EscrowWallet to OfferStatus
 * 	06-07-13/jet/ new
 *
 * purpose---
 * 	class to instantiate a reply message from the SVX to the VSC
 *
 * notes---
 * 	maintain in parallel with org.silentvault.exchange.SVXBlock
 *
 * 	these messages are delivered embedded inside a <rep_blob/> element
 * 	within a REP_svx_message message routed through the OFS
 *
 * 	see underlying message definitions in vsc_svx.xml and vsc_ofs-svx.xml
 */

package org.silentvault.client;

import org.dom4j.Element;

import org.jivesoftware.smack.util.Base64;

import java.util.Date;
import java.util.Locale;
import java.util.ArrayList;
import java.util.Iterator;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * This class instantiates server replies generated by the SVX to a connected
 * client VSC via the OFS gateway.  It knows how to emit itself as a
 * correctly formatted XML body based on the embedded opcode.  It also
 * deals with validating signatures, and representing error returns.
 *
 * @author Justin Turrell
 */
public final class SVXBlock {
	// BEGIN data members
	/**
	 * whether this object is representing an error (if true) or a normal reply
	 */
	private boolean			m_IsError;

	/**
	 * opcode to be sent to the client
	 */
	private String				m_Opcode;

	/**
	 * if this is an error message, the numeric HTTP-style error code
	 */
	private int					m_ErrCode;

	/**
	 * if this is an error message, the textual error message
	 */
	private String				m_ErrMsg;

	/**
	 * echo of the ID (of key on PKS) of the SVX (used with opcode REP_config)
	 */
	private String				m_SVXId;

	/**
	 * signature of SVX on data block, base64-encoded (used with virtually every
	 * reply opcode) 
	 */
	private String				m_Signature;

	/**
	 * inner class describing an asset in which the SVX permits escrowed trades
	 */
	public final class TradeAsset {
		/**
		 * name of the asset (also the voucher asset name)
		 */
		public String		m_AssetName;

		/**
		 * smallest trade amount allowed
		 */
		public double		m_MinTrade;

		/**
		 * commission rate (in percent) charged by SVX for such exchanges
		 */
		public float		m_Commission;

		/**
		 * constructor
		 * @param asset the (voucher) asset name of the supported asset
		 */
		public TradeAsset(String asset) {
			m_AssetName = new String(asset);
		}
	}

	/**
	 * list of exchangeable assets supported by this SVX (used with opcode
	 * REP_config)
	 */
	private ArrayList<TradeAsset>	m_TradeAssets;

	/**
	 * inner class describing an asset for which the SVX sells vouchers for
	 * external payments
	 */
	public final class SaleAsset {
		/**
		 * name of voucher asset offered for sale
		 */
		public String		m_AssetName;

		/**
		 * colon-separated list of external currencies accepted as payment
		 */
		public String		m_Currencies;

		/**
		 * minimum purchase amount
		 */
		public double		m_MinTrade;

		/**
		 * sale commission (fee %) charged on sales of this asset type
		 */
		public float		m_Commission;

		/**
		 * constructor
		 * @param asset the (voucher) asset name of the supported asset
		 * @param currs the list of currencies for which this asset is sold
		 */
		public SaleAsset(String asset, String currs) {
			m_AssetName = new String(asset);
			m_Currencies = new String(currs);
		}
	}

	/**
	 * list of saleable assets supported by this SVX (used with opcode
	 * REP_config)
	 */
	private ArrayList<SaleAsset>	m_SaleAssets;

	/**
	 * inner class describing an asset for which the SVX redeems vouchers for
	 * external payments
	 */
	public final class PurchAsset {
		/**
		 * name of voucher asset purchased by SVX
		 */
		public String		m_AssetName;

		/**
		 * colon-separated list of external currencies available as payment
		 */
		public String		m_Currencies;

		/**
		 * minimum sale amount
		 */
		public double		m_MinTrade;

		/**
		 * purchase commission fee (in percent) charged on buy-backs of this asset
		 */
		public float		m_Commission;

		/**
		 * constructor
		 * @param asset the (voucher) asset name of the supported asset
		 * @param currs the list of currencies with which this asset is bought
		 */
		public PurchAsset(String asset, String currs) {
			m_AssetName = new String(asset);
			m_Currencies = new String(currs);
		}
	}

	/**
	 * list of buyable assets supported by this SVX (used with opcode
	 * REP_config)
	 */
	private ArrayList<PurchAsset>	m_PurchAssets;

	/**
	 * order Id assigned to purchase or sale order (used with opcodes
	 * REP_purchase_terms and REP_sale_terms)
	 */
	private Integer			m_OrderId;

	/**
	 * external wallet belong to the SVX (used with opcode REP_purchase_terms).
	 * example: a bitcoin address.
	 */
	private String				m_ExtWallet;

	/**
	 * quantity of an order or exchange offer (used with opcodes
	 * REP_purchase_terms, REP_sale_terms, REP_voucher_sold, REP_offer_funded,
	 * REP_offer_pickup, and REP_offer_canceled)
	 */
	private Double				m_Quantity;

	/**
	 * amount of commission to be charged by SVX for the transaction
	 * (used with opcodes REP_purchase_terms and REP_sale_terms)
	 */
	private Double				m_Commission;

	/**
	 * total amount including commission, which must be paid by or to VSC
	 * (used with opcodes REP_sale_terms and REP_purchase_terms)
	 */
	private Double				m_Total;

	/**
	 * internal wallet ID (VS#) belonging to SVX used for escrow or as source
	 * of outgoing voucher payment (used with opcodes REP_sale_terms,
	 * REP_offer_entered, and REP_accept_entered)
	 */
	private String				m_WalletId;

	/**
	 * spend ID from external network, if any (used with opcode REP_voucher_sold)
	 */
	private String				m_SpendId;

	/**
	 * offer Id assigned to an exchange offer at this SVX (used with opcodes
	 * REP_offer_entered, REP_offer_funded, REP_accept_entered, 
	 * REP_offer_pickup, and REP_offer_canceled)
	 */
	private Integer			m_OfferId;

	/**
	 * echo confirmation of voucher asset type (used with opcode
	 * REP_offer_funded)
	 */
	private String				m_Asset;

	/**
	 * offer activation timestamp (used with opcode REP_offer_funded)
	 */
	private Date				m_Timestamp;

	/**
	 * inner class to describe an open offer being returned in a search
	 */
	public final class OpenOffer {
		/**
		 * sequence number ID of offer
		 */
		public int		m_OfferId;

		/**
		 * asset type being held in escrow
		 */
		public String	m_EscrowAsset;

		/**
		 * amount of the asset being held in escrow
		 */
		public double	m_Quantity;

		/**
		 * wanted assets, colon-separated list
		 */
		public String	m_WantedAssets;

		/**
		 * price offered, ratio of m_EscrowAsset:m_WantedAssets (if != 1.0,
		 * implies m_WantedAssets contains a single asset)
		 */
		public float	m_Price;

		/**
		 * expiry date of offer
		 */
		public Date		m_Expiry;

		/**
		 * constructor
		 * @param oId the DB sequence Id of the offer
		 */
		public OpenOffer(int oId) {
			m_OfferId = oId;
			m_EscrowAsset = m_WantedAssets = "";
		}
	}

	/**
	 * list of matching open offers (used with opcode REP_offers)
	 */
	private ArrayList<OpenOffer>	m_MatchingOffers;

	/**
	 * inner class to describe an offer belonging to a particular VSC/wallet
	 */
	public final class OfferStatus {
		/**
		 * sequence number ID of offer
		 */
		public int			m_OfferId;

		/**
		 * current status value (will be open, funded, or filled)
		 */
		public String		m_Status;

		/**
		 * description of what is offered here
		 */
		public OpenOffer	m_Offered;

		/**
		 * the wallet Id on the SVX designated to hold escrow
		 */
		public String		m_EscrowWallet;

		/**
		 * for a filled offer, what was received in exchange
		 */
		public String		m_ExAsset;

		/**
		 * for a filled offer, amount of m_ExAsset received in exchange
		 */
		public double		m_ExQuantity;

		/**
		 * for a filled offer, timestamp at which escrow was paid out
		 */
		public Date			m_ExTimestamp;

		/**
		 * constructor
		 * @param oId the offer ID sequence number
		 */
		public OfferStatus(int oId) {
			m_OfferId = oId;
			m_Offered = new OpenOffer(oId);
			m_Status = m_ExAsset = m_EscrowWallet = "";
		}
	}

	/**
	 * list of offers belonging to a particular wallet/user (used with opcode
	 * REP_offer_status)
	 */
	private ArrayList<OfferStatus>	m_WalletOffers;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param isErr true if this block represents a returned error message
	 */
	public SVXBlock(boolean isErr) {
		m_IsError = isErr;
		m_Opcode = m_ErrMsg = m_Signature = m_WalletId = m_ExtWallet = "";
		m_SVXId = m_SpendId = m_Asset = "";
		m_TradeAssets = new ArrayList<TradeAsset>();
		m_SaleAssets = new ArrayList<SaleAsset>();
		m_PurchAssets = new ArrayList<PurchAsset>();
		m_OrderId = new Integer(0);
		m_Quantity = new Double(0.0);
		m_Commission = new Double(0.0);
		m_Total = new Double(0.0);
		m_OfferId = new Integer(0);
		m_MatchingOffers = new ArrayList<OpenOffer>(100);
		m_WalletOffers = new ArrayList<OfferStatus>();
	}


	// get accessor methods
	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the error code, if any
	 * @return the code, or 0 if we're not an error message
	 */
	public int getErrCode() {
		if (m_IsError) {
			return m_ErrCode;
		}
		return 0;
	}

	/**
	 * obtain the error message, if any
	 * @return the error text, or the empty string if we're not an error message
	 */
	public String getErrMsg() {
		if (m_IsError) {
			return m_ErrMsg;
		}
		return "";
	}

	/**
	 * obtain whether this block is representing an error
	 * @return true if error
	 */
	public boolean isError() { return m_IsError; }

	/**
	 * obtain the (pubkey) ID of the SVX
	 * @return the ID which can be used to find the pubkey on a PKS
	 */
	public String getSVXId() { return m_SVXId; }

	/**
	 * obtain the server's signature
	 * @return the digital privkey signature provided on other fields
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the list of trade assets
	 * @return list describing all assets exchangeable at this SVX
	 */
	public ArrayList<TradeAsset> getTradeAssets() { return m_TradeAssets; }

	/**
	 * obtain the list of sale assets
	 * @return list describing all assets sold for external payments by this SVX
	 */
	public ArrayList<SaleAsset> getSaleAssets() { return m_SaleAssets; }

	/**
	 * obtain the list of repurchased assets
	 * @return list describing all assets bought back (redeemed) for external
	 * payouts by this SVX
	 */
	public ArrayList<PurchAsset> getPurchAssets() { return m_PurchAssets; }

	/**
	 * obtain the order Id assigned
	 * @return the id
	 */
	public int getOrderId() { return m_OrderId.intValue(); }

	/**
	 * obtain the external wallet Id
	 * @return the ID of the external wallet belonging to the SVX
	 */
	public String getExtWallet() { return m_ExtWallet; }

	/**
	 * obtain the quantity
	 * @return the amount involved
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * obtain the commission to be paid
	 * @return the amount involved (not the percentage)
	 */
	public double getCommission() { return m_Commission.doubleValue(); }

	/**
	 * obtain the total amount to be paid
	 * @return the amount involved (quantity + commission)
	 */
	public double getTotal() { return m_Total.doubleValue(); }

	/**
	 * obtain the wallet Id used in a transfer by the SVX
	 * @return the id (VS#) of the wallet
	 */
	public String getWalletId() { return m_WalletId; }

	/**
	 * obtain the external spend Id
	 * @return the Id of a spend made, if any (otherwise echo of orderId)
	 */
	public String getSpendId() { return m_SpendId; }

	/**
	 * obtain the offer Id
	 * @return the database unique sequence number of the offer
	 */
	public int getOfferId() { return m_OfferId.intValue(); }

	/**
	 * obtain the asset type
	 * @return echo confirmation of asset type picked up
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * obtain the date/time of settlement or transaction completion
	 * @return the timestamp
	 */
	public Date getTimestamp() { return m_Timestamp; }

	/**
	 * obtain matching open offers
	 * @return the list of offers matching client's search criteria
	 */
	public ArrayList<OpenOffer> getMatchingOffers() { return m_MatchingOffers; }

	/**
	 * obtain the list of offers for a particular wallet
	 * @return the list of offers entered by the signing wallet Id
	 */
	public ArrayList<OfferStatus> getWalletOffers() { return m_WalletOffers; }


	// set accessor methods
	/**
	 * set the opcode returned in the client reply
	 * @param op the opcode to set (assumed to be valid)
	 */
	public void setOpcode(String op) {
		if (op != null && op.startsWith("REP_")) {
			m_Opcode = new String(op);
		}
		else {
			Log.error("Improper outgoing opcode, " + op);
		}
	}

	/**
	 * set the error code (also sets m_IsError to true)
	 * @param code the error code value to set
	 */
	public void setErrCode(int code) {
		if (code > 0) {
			m_ErrCode = code;
			m_IsError = true;
		}
	}

	/**
	 * set the error text (also sets m_IsError to true)
	 * @param text the error message text
	 */
	public void setErrMsg(String text) {
		if (text != null && !text.isEmpty()) {
			m_ErrMsg = new String(text);
			m_IsError = true;
		}
	}

	/**
	 * set the SVX's Id
	 * @param id the pubkey identifier of the SVX on a PKS
	 */
	public void setSVXId(String id) {
		if (id != null && id.indexOf('@') != -1) {
			m_SVXId = new String(id);
		}
	}

	/**
	 * set the server's signature
	 * @param sig the privkey signature of the SVX on other fields
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_Signature = new String(sig);
		}
	}

	/**
	 * add a trade asset to the list to be returned
	 * @param tAsset the description of an asset exchangeable on this SVX
	 */
	public void addTradeAsset(TradeAsset tAsset) {
		if (tAsset != null) {
			m_TradeAssets.add(tAsset);
		}
	}

	/**
	 * add a sale asset to the list to be returned
	 * @param sAsset the description of an asset sold by this SVX
	 */
	public void addSaleAsset(SaleAsset sAsset) {
		if (sAsset != null) {
			m_SaleAssets.add(sAsset);
		}
	}

	/**
	 * add a purchase (buy-back) asset to the list to be returned
	 * @param pAsset the description of an asset redeemable on this SVX
	 */
	public void addPurchAsset(PurchAsset pAsset) {
		if (pAsset != null) {
			m_PurchAssets.add(pAsset);
		}
	}

	/**
	 * set the order Id
	 * @param oId the id assigned
	 */
	public void setOrderId(int oId) {
		if (oId > 0) {
			m_OrderId = new Integer(oId);
		}
	}

	/**
	 * set the external wallet Id belonging to the SVX
	 * @param extWallet the Id to set, such as a bitcoin hash address
	 */
	public void setExtWallet(String extWallet) {
		if (extWallet != null && !extWallet.isEmpty()) {
			m_ExtWallet = new String(extWallet);
		}
	}

	/**
	 * set the quantity
	 * @param amt the amount involved in an order or offer
	 */
	public void setQuantity(double amt) {
		if (amt > 0.0) {
			m_Quantity = new Double(amt);
		}
	}

	/**
	 * set the commission
	 * @param comm the amount being charged by the SVX for a transaction
	 */
	public void setCommission(double comm) {
		if (comm >= 0.0) {
			m_Commission = new Double(comm);
		}
	}

	/**
	 * set the total amount
	 * @param total the total, quantity + commission
	 */
	public void setTotal(double total) {
		if (total > 0.0) {
			m_Total = new Double(total);
		}
	}

	/**
	 * set the wallet Id used in an operation
	 * @param wId the id (VS#) of the wallet used by the SVX
	 */
	public void setWalletId(String wId) {
		if (wId != null && wId.indexOf('-') != -1) {
			m_WalletId = new String(wId);
		}
	}

	/**
	 * set the spendId of an external spend
	 * @param id the Id to set, which may be an echo of m_OrderId
	 */
	public void setSpendId(String id) {
		if (id != null) {
			m_SpendId = new String(id);
		}
	}

	/**
	 * set the offer Id
	 * @param oId the id assigned
	 */
	public void setOfferId(int oId) {
		if (oId > 0) {
			m_OfferId = new Integer(oId);
		}
	}

	/**
	 * set the asset type echo
	 * @param asset the asset type to set
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * set the clearing timestamp
	 * @param tstamp the date/time of settlement or clearing
	 */
	public void setTimestamp(Date tstamp) {
		if (tstamp != null) {
			m_Timestamp = tstamp;
		}
	}

	/**
	 * add a matching offer to the list
	 * @param offer the description of an open offer matching search criteria
	 */
	public void addMatchingOffer(OpenOffer offer) {
		if (offer != null) {
			m_MatchingOffers.add(offer);
		}
	}

	/**
	 * add an offer belonging to a wallet to the list
	 * @param offer the description of an offer belonging to the specific wallet
	 */
	public void addWalletOffer(OfferStatus offer) {
		if (offer != null) {
			m_WalletOffers.add(offer);
		}
	}

	/**
	 * method to parse a decrypted block from the SVX, and set field values
	 * @param svxBlock the decrypted text, including delimiters
	 * @return true on successful parse, false otherwise
	 */
	public boolean parseBlock(String svxBlock) {
		if (svxBlock == null || !svxBlock.startsWith("<svxBlock>")) {
			Log.error("Missing or improper SSM block, cannot parse");
			return false;
		}
		Element svxElt = EncodingUtils.buildElement(svxBlock);
		if (svxElt == null) {
			Log.error("<svxBlock/> parse failure");
			return false;
		}

		// <opcode/> must always be present
		Element opElt = svxElt.element("opcode");
		if (opElt == null) {
			//Log.error("Missing svxBlock opcode");
			return false;
		}
		String opcode = opElt.getTextTrim();
		setOpcode(opcode);

		// try to determine if we're an error (will have errcode and/or errmsg)
		Element eCodeElt = svxElt.element("errcode");
		if (eCodeElt != null) {
			setErrCode(Integer.parseInt(eCodeElt.getText()));
		}
		else if (m_IsError) {
			Log.error("Missing errcode for error " + opcode);
			return false;
		}

		Element eMsgElt = svxElt.element("errmsg");
		if (eMsgElt != null) {
			setErrMsg(eMsgElt.getTextTrim());
		}
		else if (m_IsError) {
			Log.error("Missing errmsg for error " + opcode);
			return false;
		}
		if (m_IsError) {
			// done after these fields
			return true;
		}

		// otherwise examine the opcode for further fields
		if (opcode.equalsIgnoreCase("REP_config")) {
			// need an SVX Id
			Element idElt = svxElt.element("svxId");
			if (idElt == null) {
				Log.error("Missing svxId for " + opcode);
				return false;
			}
			setSVXId(idElt.getTextTrim());

			// need traded asset list
			Element tradeElt = svxElt.element("tradeAssets");
			if (tradeElt == null) {
				Log.error("Missing tradeAssets for " + opcode);
				return false;
			}
			// obtain all listed assets
			Iterator trAssetIt = tradeElt.elementIterator("asset");
			while (trAssetIt.hasNext()) {
				Element trAssetElt = (Element) trAssetIt.next();
				// get voucher asset name
				String asset = trAssetElt.attributeValue("name");
				if (asset == null || asset.isEmpty()) {
					Log.error("Missing asset name for tradeAsset");
					return false;
				}
				TradeAsset tAsset = new TradeAsset(asset);
				// get minimum tradeable amount
				Element minElt = trAssetElt.element("minTrade");
				if (minElt == null) {
					Log.error("Missing minTrade for tradeAsset");
					return false;
				}
				tAsset.m_MinTrade = Double.parseDouble(minElt.getText());
				// get commission rate
				Element commElt = trAssetElt.element("commission");
				if (commElt == null) {
					Log.error("Missing commission for tradeAsset");
					return false;
				}
				tAsset.m_Commission = Float.parseFloat(commElt.getText());
				addTradeAsset(tAsset);
			}

			// need sold asset list
			Element saleElt = svxElt.element("saleAssets");
			if (saleElt == null) {
				Log.error("Missing saleAssets for " + opcode);
				return false;
			}
			// obtain all listed assets
			Iterator saleAssIt = saleElt.elementIterator("asset");
			while (saleAssIt.hasNext()) {
				Element saleAssElt = (Element) saleAssIt.next();
				// get voucher asset name
				String asset = saleAssElt.attributeValue("name");
				if (asset == null || asset.isEmpty()) {
					Log.error("Missing asset name for saleAsset");
					return false;
				}
				// get payment currency list
				Element currElt = saleAssElt.element("currencies");
				if (currElt == null) {
					Log.error("Missing currencies for saleAsset");
					return false;
				}
				SaleAsset saleAsset = new SaleAsset(asset, currElt.getTextTrim());
				// get minimum tradeable amount
				Element minElt = saleAssElt.element("minTrade");
				if (minElt == null) {
					Log.error("Missing minTrade for saleAsset");
					return false;
				}
				saleAsset.m_MinTrade = Double.parseDouble(minElt.getText());
				// get commission rate
				Element commElt = saleAssElt.element("commission");
				if (commElt == null) {
					Log.error("Missing commission for saleAsset");
					return false;
				}
				saleAsset.m_Commission = Float.parseFloat(commElt.getText());
				addSaleAsset(saleAsset);
			}

			// need purchased asset list
			Element purchElt = svxElt.element("buyAssets");
			if (purchElt == null) {
				Log.error("Missing buyAssets for " + opcode);
				return false;
			}
			// obtain all listed assets
			Iterator buyAssIt = purchElt.elementIterator("asset");
			while (buyAssIt.hasNext()) {
				Element buyAssElt = (Element) buyAssIt.next();
				// get voucher asset name
				String asset = buyAssElt.attributeValue("name");
				if (asset == null || asset.isEmpty()) {
					Log.error("Missing asset name for buyAsset");
					return false;
				}
				// get payment currency list
				Element currElt = buyAssElt.element("currencies");
				if (currElt == null) {
					Log.error("Missing currencies for buyAsset");
					return false;
				}
				PurchAsset buyAsset = new PurchAsset(asset, currElt.getTextTrim());
				// get minimum tradeable amount
				Element minElt = buyAssElt.element("minTrade");
				if (minElt == null) {
					Log.error("Missing minTrade for buyAsset");
					return false;
				}
				buyAsset.m_MinTrade = Double.parseDouble(minElt.getText());
				// get commission rate
				Element commElt = buyAssElt.element("commission");
				if (commElt == null) {
					Log.error("Missing commission for buyAsset");
					return false;
				}
				buyAsset.m_Commission = Float.parseFloat(commElt.getText());
				addPurchAsset(buyAsset);
			}

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_purchase_terms")) {
			// need orderId
			Element oIdElt = svxElt.element("orderId");
			if (oIdElt == null) {
				Log.error("Missing orderId for " + opcode);
				return false;
			}
			setOrderId(Integer.parseInt(oIdElt.getText()));

			// need external wallet address
			Element wallElt = svxElt.element("extWallet");
			if (wallElt == null) {
				Log.error("Missing extWallet for " + opcode);
				return false;
			}
			setExtWallet(wallElt.getTextTrim());

			// need quantity echo
			Element quanElt = svxElt.element("quantity");
			if (quanElt == null) {
				Log.error("Missing quantity for " + opcode);
				return false;
			}
			setQuantity(Double.parseDouble(quanElt.getText()));

			// need computed commission
			Element commElt = svxElt.element("commission");
			if (commElt == null) {
				Log.error("Missing commission for " + opcode);
				return false;
			}
			setCommission(Double.parseDouble(commElt.getText()));

			// need order total (including commission)
			Element totElt = svxElt.element("total");
			if (totElt == null) {
				Log.error("Missing total for " + opcode);
				return false;
			}
			setTotal(Double.parseDouble(totElt.getText()));

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_sale_terms")) {
			// need orderId
			Element oIdElt = svxElt.element("orderId");
			if (oIdElt == null) {
				Log.error("Missing orderId for " + opcode);
				return false;
			}
			setOrderId(Integer.parseInt(oIdElt.getText()));

			// need (internal) sale wallet address
			Element wallElt = svxElt.element("walletId");
			if (wallElt == null) {
				Log.error("Missing walletId for " + opcode);
				return false;
			}
			setWalletId(wallElt.getTextTrim());

			// need quantity echo
			Element quanElt = svxElt.element("quantity");
			if (quanElt == null) {
				Log.error("Missing quantity for " + opcode);
				return false;
			}
			setQuantity(Double.parseDouble(quanElt.getText()));

			// need computed commission
			Element commElt = svxElt.element("commission");
			if (commElt == null) {
				Log.error("Missing commission for " + opcode);
				return false;
			}
			setCommission(Double.parseDouble(commElt.getText()));

			// need order total (including commission)
			Element totElt = svxElt.element("total");
			if (totElt == null) {
				Log.error("Missing total for " + opcode);
				return false;
			}
			setTotal(Double.parseDouble(totElt.getText()));

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_voucher_sold")) {
			// need orderId
			Element oIdElt = svxElt.element("orderId");
			if (oIdElt == null) {
				Log.error("Missing orderId for " + opcode);
				return false;
			}
			setOrderId(Integer.parseInt(oIdElt.getText()));

			// need quantity of external spend
			Element quanElt = svxElt.element("quantity");
			if (quanElt == null) {
				Log.error("Missing quantity for " + opcode);
				return false;
			}
			setQuantity(Double.parseDouble(quanElt.getText()));

			// need spendId (txid from external network)
			Element spendElt = svxElt.element("spendId");
			if (spendElt == null) {
				Log.error("Missing spendId for " + opcode);
				return false;
			}
			setSpendId(spendElt.getTextTrim());

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equals("REP_offer_entered")
					|| opcode.equals("REP_accept_entered"))
		{
			// need offerId
			Element oIdElt = svxElt.element("offerId");
			if (oIdElt == null) {
				Log.error("Missing offerId for " + opcode);
				return false;
			}
			setOfferId(Integer.parseInt(oIdElt.getText()));

			// need offer total augmented by commission fee for asset type
			Element totElt = svxElt.element("total");
			if (totElt == null) {
				Log.error("Missing total for " + opcode);
				return false;
			}
			setTotal(Double.parseDouble(totElt.getText()));

			// need (internal) escrow wallet address
			Element wallElt = svxElt.element("walletId");
			if (wallElt == null) {
				Log.error("Missing walletId for " + opcode);
				return false;
			}
			setWalletId(wallElt.getTextTrim());

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_offer_funded")) {
			// need offerId
			Element oIdElt = svxElt.element("offerId");
			if (oIdElt == null) {
				Log.error("Missing offerId for " + opcode);
				return false;
			}
			setOfferId(Integer.parseInt(oIdElt.getText()));

			// need asset type echo
			Element assetElt = svxElt.element("asset");
			if (assetElt == null) {
				Log.error("Missing asset for " + opcode);
				return false;
			}
			setAsset(assetElt.getTextTrim());

			// need quantity of escrow spend (echo)
			Element quanElt = svxElt.element("quantity");
			if (quanElt == null) {
				Log.error("Missing quantity for " + opcode);
				return false;
			}
			setQuantity(Double.parseDouble(quanElt.getText()));

			// need timestamp of offer activation
			Element tsElt = svxElt.element("timestamp");
			if (tsElt == null) {
				Log.error("Missing timestamp for " + opcode);
				return false;
			}
			setTimestamp(new Date(Long.parseLong(tsElt.getText())));

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_offers")) {
			// need offer list
			Iterator offersIt = svxElt.elementIterator("offer");
			while (offersIt.hasNext()) {
				Element offerElt = (Element) offersIt.next();
				// get offer ID in DB
				String oId = offerElt.attributeValue("offerId");
				if (oId == null || oId.isEmpty()) {
					Log.error("Missing offerId for offer in " + opcode);
					return false;
				}
				OpenOffer opOff = new OpenOffer(Integer.parseInt(oId));
				// get asset type in escrow
				Element assetElt = offerElt.element("asset");
				if (assetElt == null) {
					Log.error("Missing asset for offer in " + opcode);
					return false;
				}
				opOff.m_EscrowAsset = assetElt.getTextTrim();
				// get quantity held in escrow
				Element quanElt = offerElt.element("quantity");
				if (quanElt == null) {
					Log.error("Missing quantity for offer in " + opcode);
					return false;
				}
				opOff.m_Quantity = Double.parseDouble(quanElt.getText());
				// get list of assets wanted in exchange
				Element wantElt = offerElt.element("wantedAssets");
				if (wantElt == null) {
					Log.error("Missing wantedAssets for offer in " + opcode);
					return false;
				}
				opOff.m_WantedAssets = wantElt.getTextTrim();
				// get price ratio (will be 1.0 if wanted assets are multiple)
				Element priceElt = offerElt.element("price");
				if (priceElt == null) {
					Log.error("Missing price for offer in " + opcode);
					return false;
				}
				opOff.m_Price = Float.parseFloat(priceElt.getText());
				// get expiration timestamp
				Element expElt = offerElt.element("expiry");
				if (expElt == null) {
					Log.error("Missing expiry for offer in " + opcode);
					return false;
				}
				opOff.m_Expiry = new Date(Long.parseLong(expElt.getText()));
				addMatchingOffer(opOff);
			}

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_offer_status")) {
			// need offer list
			Iterator offersIt = svxElt.elementIterator("offer");
			while (offersIt.hasNext()) {
				Element offerElt = (Element) offersIt.next();
				// get offer ID in DB
				String oId = offerElt.attributeValue("offerId");
				if (oId == null || oId.isEmpty()) {
					Log.error("Missing offerId for offer in " + opcode);
					return false;
				}
				OfferStatus oStat = new OfferStatus(Integer.parseInt(oId));
				// get status of offer (open, funded, canceled, or filled)
				Element statElt = offerElt.element("status");
				if (statElt == null) {
					Log.error("Missing status for offer in " + opcode);
					return false;
				}
				oStat.m_Status = statElt.getTextTrim();

				// obtain what is offered (OpenOffer data)
				Element offeredElt = offerElt.element("offered");
				if (offeredElt == null) {
					Log.error("Missing offered for offer in " + opcode);
					return false;
				}

				// get asset type in escrow
				Element assetElt = offeredElt.element("asset");
				if (assetElt == null) {
					Log.error("Missing asset for offer in " + opcode);
					return false;
				}
				oStat.m_Offered.m_EscrowAsset = assetElt.getTextTrim();

				// get quantity held in escrow
				Element quanElt = offeredElt.element("quantity");
				if (quanElt == null) {
					Log.error("Missing quantity for offer in " + opcode);
					return false;
				}
				oStat.m_Offered.m_Quantity = Double.parseDouble(quanElt.getText());

				// get list of assets wanted in exchange
				Element wantElt = offeredElt.element("wantedAssets");
				if (wantElt == null) {
					Log.error("Missing wantedAssets for offer in " + opcode);
					return false;
				}
				oStat.m_Offered.m_WantedAssets = wantElt.getTextTrim();

				// get price ratio (will be 1.0 if wanted assets are multiple)
				Element priceElt = offeredElt.element("price");
				if (priceElt == null) {
					Log.error("Missing price for offer in " + opcode);
					return false;
				}
				oStat.m_Offered.m_Price = Float.parseFloat(priceElt.getText());

				// get escrow wallet Id
				Element escElt = offeredElt.element("walletId");
				if (escElt == null) {
					Log.error("Missing walletId for offer in " + opcode);
					return false;
				}
				oStat.m_EscrowWallet = escElt.getTextTrim();

				// get expiration timestamp (from offer element, not offered)
				Element expElt = offerElt.element("expiry");
				if (expElt == null) {
					Log.error("Missing expiry for offer in " + opcode);
					return false;
				}
				oStat.m_Offered.m_Expiry
					= new Date(Long.parseLong(expElt.getText()));

				Element rcvElt = offerElt.element("received");
				if (rcvElt == null) {
					Log.error("Missing received for offer in " + opcode);
					return false;
				}

				// if status = filled|expired, then there are 3 additional fields
				if (oStat.m_Status.equals(TradeOffer.M_StatusFilled)
					 || oStat.m_Status.equals(TradeOffer.M_StatusExpired))
				{
					// get asset received in exchange
					Element assElt = rcvElt.element("asset");
					if (assElt == null) {
						Log.error("Missing received asset for offer in " + opcode);
						return false;
					}
					oStat.m_ExAsset = assElt.getTextTrim();
					// get quantity of this asset
					Element rQuanElt = rcvElt.element("quantity");
					if (rQuanElt == null) {
						Log.error("Missing received quantity for offer in " + opcode);
						return false;
					}
					oStat.m_ExQuantity = Double.parseDouble(rQuanElt.getText());
					// get timestamp at which escrow payout occurred
					Element tsElt = rcvElt.element("timestamp");
					if (tsElt == null) {
						Log.error("Missing escrow timestamp for offer in " + opcode);
						return false;
					}
					oStat.m_ExTimestamp = new Date(Long.parseLong(tsElt.getText()));
				}
				addWalletOffer(oStat);
			}

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equals("REP_offer_pickup")
					|| opcode.equals("REP_offer_canceled"))
		{
			// need offerId
			Element oIdElt = svxElt.element("offerId");
			if (oIdElt == null) {
				Log.error("Missing offerId for " + opcode);
				return false;
			}
			setOfferId(Integer.parseInt(oIdElt.getText()));

			// need quantity of voucher sent
			Element quanElt = svxElt.element("quantity");
			if (quanElt == null) {
				Log.error("Missing quantity for " + opcode);
				return false;
			}
			setQuantity(Double.parseDouble(quanElt.getText()));

			// need SVX's signature
			Element sigElt = svxElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else {
			Log.error("Unrecognized reply opcode, " + opcode);
			return false;
		}
		return true;
	}

	/**
	 * method to construct the svxBlock XML, based on opcode and member values.
	 * Any signature should be built using the string returned by this method.
	 * Missing required output fields generate log errors, but do not stop the
	 * generation of output.
	 * @return the complete block for the indicated reply, absent the delimiters
	 * and the SVX's signature (if required)
	 */
	public String getXML() {
		StringBuilder xml = new StringBuilder(2048);

		// if we're an error message, this is just the code and error text
		xml.append("<opcode>" + m_Opcode + "</opcode>");
		if (m_IsError) {
			xml.append("<errcode>" + m_ErrCode + "</errcode>");
			xml.append("<errmsg>" + m_ErrMsg + "</errmsg>");
			return xml.toString();
		}

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// base our output on the opcode
		if (m_Opcode.equals("REP_config")) {
			if (m_SVXId.isEmpty()) {
				Log.error("Missing svxId for " + m_Opcode);
			}
			xml.append("<svxId>" + m_SVXId + "</svxId>");

			// add assets available for trade
			if (m_TradeAssets.isEmpty()) {
				Log.error("Missing tradeAssets for " + m_Opcode);
				xml.append("<tradeAssets/>");
			}
			else {
				xml.append("<tradeAssets>");
				for (TradeAsset tAsset : m_TradeAssets) {
					xml.append("<asset name=\"" + tAsset.m_AssetName + "\">");
					xml.append("<minTrade>" + df.format(tAsset.m_MinTrade));
					xml.append("</minTrade>");
					xml.append("<commission>" + df.format(tAsset.m_Commission));
					xml.append("</commission>");
					xml.append("</asset>");
				}
				xml.append("</tradeAssets>");
			}

			// add voucher assets sold for external payments
			if (m_SaleAssets.isEmpty()) {
				xml.append("<saleAssets/>");
			}
			else {
				xml.append("<saleAssets>");
				for (SaleAsset sAsset : m_SaleAssets) {
					xml.append("<asset name=\"" + sAsset.m_AssetName + "\">");
					xml.append("<currencies>" + sAsset.m_Currencies);
					xml.append("</currencies>");
					xml.append("<minTrade>" + df.format(sAsset.m_MinTrade));
					xml.append("</minTrade>");
					xml.append("<commission>" + df.format(sAsset.m_Commission));
					xml.append("</commission>");
					xml.append("</asset>");
				}
				xml.append("</saleAssets>");
			}

			// add voucher assets repurchaseable for external payments
			if (m_PurchAssets.isEmpty()) {
				xml.append("<buyAssets/>");
			}
			else {
				xml.append("<buyAssets>");
				for (PurchAsset pAsset : m_PurchAssets) {
					xml.append("<asset name=\"" + pAsset.m_AssetName + "\">");
					xml.append("<currencies>" + pAsset.m_Currencies);
					xml.append("</currencies>");
					xml.append("<minTrade>" + df.format(pAsset.m_MinTrade));
					xml.append("</minTrade>");
					xml.append("<commission>" + df.format(pAsset.m_Commission));
					xml.append("</commission>");
					xml.append("</asset>");
				}
				xml.append("</buyAssets>");
			}
		}
		else if (m_Opcode.equals("REP_purchase_terms")) {
			xml.append("<orderId>" + m_OrderId + "</orderId>");
			if (m_ExtWallet.isEmpty()) {
				Log.error("Missing extWallet for " + m_Opcode);
			}
			xml.append("<extWallet>" + m_ExtWallet + "</extWallet>");
			xml.append("<quantity>" + df.format(getQuantity()) + "</quantity>");
			xml.append("<commission>" + df.format(getCommission()));
			xml.append("</commission>");
			xml.append("<total>" + df.format(getTotal()) + "</total>");
		}
		else if (m_Opcode.equals("REP_sale_terms")) {
			xml.append("<orderId>" + m_OrderId + "</orderId>");
			if (m_WalletId.isEmpty()) {
				Log.error("Missing walletId for " + m_Opcode);
			}
			xml.append("<walletId>" + m_WalletId + "</walletId>");
			xml.append("<quantity>" + df.format(getQuantity()) + "</quantity>");
			xml.append("<commission>" + df.format(getCommission()));
			xml.append("</commission>");
			xml.append("<total>" + df.format(getTotal()) + "</total>");
		}
		else if (m_Opcode.equals("REP_voucher_sold")) {
			xml.append("<orderId>" + m_OrderId + "</orderId>");
			xml.append("<quantity>" + df.format(getQuantity()) + "</quantity>");
			if (m_SpendId.isEmpty()) {
				Log.warning("Missing spendId for opcode " + m_Opcode);
				xml.append("<spendId>" + m_OrderId + "</spendId>");
			}
			else {
				xml.append("<spendId>" + m_SpendId + "</spendId>");
			}
		}
		else if (m_Opcode.equals("REP_offer_entered")
					|| m_Opcode.equals("REP_accept_entered"))
		{
			xml.append("<offerId>" + m_OfferId + "</offerId>");
			xml.append("<total>" + df.format(getTotal()) + "</total>");
			if (m_WalletId.isEmpty()) {
				Log.error("Missing walletId for " + m_Opcode);
			}
			xml.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equals("REP_offer_funded")) {
			xml.append("<offerId>" + m_OfferId + "</offerId>");
			xml.append("<asset>" + m_Asset + "</asset>");
			xml.append("<quantity>" + df.format(getQuantity()) + "</quantity>");
			if (m_Timestamp != null) {
				xml.append("<timestamp>" + m_Timestamp.getTime() + "</timestamp>");
			}
			else {
				Log.error("Missing timestamp for " + m_Opcode);
				xml.append("<timestamp/>");
			}
		}
		else if (m_Opcode.equals("REP_offers")) {
			for (OpenOffer off : m_MatchingOffers) {
				xml.append("<offer offerId=\"" + off.m_OfferId + "\">");
				xml.append("<asset>" + off.m_EscrowAsset + "</asset>");
				xml.append("<quantity>" + df.format(off.m_Quantity));
				xml.append("</quantity>");
				xml.append("<wantedAssets>" + off.m_WantedAssets);
				xml.append("</wantedAssets>");
				xml.append("<price>" + df.format(off.m_Price) + "</price>");
				xml.append("<expiry>" + off.m_Expiry.getTime() + "</expiry>");
				xml.append("</offer>");
			}
		}
		else if (m_Opcode.equals("REP_offer_status")) {
			for (OfferStatus off : m_WalletOffers) {
				xml.append("<offer offerId=\"" + off.m_OfferId + "\">");
				xml.append("<status>" + off.m_Status + "</status>");
				xml.append("<offered>");
				OpenOffer offered = off.m_Offered;
				xml.append("<asset>" + offered.m_EscrowAsset + "</asset>");
				xml.append("<quantity>" + df.format(offered.m_Quantity));
				xml.append("</quantity>");
				xml.append("<wantedAssets>" + offered.m_WantedAssets);
				xml.append("</wantedAssets>");
				xml.append("<price>" + df.format(offered.m_Price) + "</price>");
				xml.append("<walletId>" + off.m_EscrowWallet + "</walletId>");
				xml.append("</offered>");
				if (offered.m_Expiry != null) {
					xml.append("<expiry>" + offered.m_Expiry.getTime());
					xml.append("</expiry>");
				}
				else {
					Log.error("Missing expiry date for offerId " + off.m_OfferId);
					xml.append("<expiry/>");
				}
				if (off.m_Status.equals(TradeOffer.M_StatusFilled)
					 || off.m_Status.equals(TradeOffer.M_StatusExpired))
				{
					// add additional fields
					xml.append("<received>");
					xml.append("<asset>" + off.m_ExAsset + "</asset>");
					xml.append("<quantity>" + df.format(off.m_ExQuantity));
					xml.append("</quantity>");
					if (off.m_ExTimestamp == null) {
						Log.error("Missing fulfillment tstamp for offerId "
									+ off.m_OfferId);
						xml.append("<timestamp/>");
					}
					else {
						xml.append("<timestamp>" + off.m_ExTimestamp.getTime());
						xml.append("</timestamp>");
					}
					xml.append("</received>");
				}
				else {
					xml.append("<received/>");
				}
				xml.append("</offer>");
			}
		}
		else if (m_Opcode.equals("REP_offer_pickup")
					|| m_Opcode.equals("REP_offer_canceled"))
		{
			xml.append("<offerId>" + m_OfferId + "</offerId>");
			xml.append("<quantity>" + df.format(getQuantity()) + "</quantity>");
		}
		else {
			Log.error("Unrecognized reply opcode, " + m_Opcode);
		}

		return xml.toString();
	}

	/**
	 * validate the signature on the block, given the SVX's public key
	 * @param pubKey the public key required to validate the signature
	 * @return true if the signature validates, false otherwise
	 */
	public boolean sigVerify(PublicKey pubKey) {
		String blockXML = getXML();
		if (blockXML.isEmpty() || m_Signature.isEmpty()) {
			Log.error("Missing block data or signature for sig check");
			return false;
		}
		if (pubKey == null) {
			Log.error("No key passed for sig check");
			return false;
		}
		//Log.debug("Checking SVX sig using: " + blockXML);

		Signature vSig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initVerify(pubKey);
			vSig.update(blockXML.getBytes());
			return vSig.verify(Base64.decode(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate client signature", se);
		}
		return false;
	}

	// END methods
}
