/*
 * last modified---
 * 	07-17-15/jet/ new
 *
 * purpose---
 * 	provide static storage for a VP's configuration of assets/units/tokens
 */

package org.silentvault.client;

import java.util.HashMap;
import java.util.ArrayList;


/**
 * Class to record the configuration for the voucher assets, voucher units,
 * token values, display units, price lookup units, and unit conversions as
 * downloaded from a VP via the OFS.  This data is static in order to support
 * static methods in other classes.  It needs to be reset (cleared) upon
 * wallet logout, and is populated during handling of a REP_network_details.
 * All methods are static.
 *
 * @author Kevin Wilkerson
 */
public final class VPAssetConfig {
	// data members
	/**
	 * inner class to describe a configuration for a particular asset at the VP
	 */
	public static final class AssetConfig {
		/**
		 * name of the asset, as used in vouchers of this type
		 */
		public String								m_Asset;

		/**
		 * inner class to describe a configuration for supported units
		 */
		public final class UnitConfig {
			/**
			 * name of the unit
			 */
			public String		m_Unit;

			/**
			 * ratio for conversion of this unit to base units (if == 1.0 then
			 * this unit is base)
			 */
			public double		m_Conversion;

			/**
			 * unit name for user display purposes (may differ from m_Unit)
			 */
			public String		m_Display;

			/**
			 * unit name for price table lookups (may equal m_Asset:m_Unit)
			 */
			public String		m_Lookup;

			/**
			 * value of one token in this asset/unit (if == 0.0 then tokens
			 * do not exist in this unit type)
			 */
			public double		m_TokenValue;

			/**
			 * constructor
			 */
			public UnitConfig() {
				m_Unit = m_Display = m_Lookup = "";
				m_Conversion = 1.0;
			}
		}

		/**
		 * list of units supported for this asset, keyed by unit name
		 */
		public HashMap<String, UnitConfig>	m_Units;

		/**
		 * constructor
		 */
		public AssetConfig() {
			m_Asset = "";
			m_Units = new HashMap<String, UnitConfig>();
		}
	}

	/**
	 * static storage for the configuration (only holds one VP's data at a time),
	 * keyed by asset name
	 */
	public static HashMap<String, AssetConfig>	m_AssetConfig
		= new HashMap<String, AssetConfig>();

	// methods
	/**
	 * obtain the config for a given asset
	 * @param asset the asset whose config we want
	 * @return the asset's config, or null if not found (error)
	 */
	public static AssetConfig getAssetConfig(String asset) {
		if (asset == null || asset.isEmpty()) {
			return null;
		}
		return m_AssetConfig.get(asset);
	}

	/**
	 * obtain the base units for a given asset
	 * @param asset the asset whose base units we want
	 * @return the unit defined as base, or the empty string if not found (error)
	 */
	public static String getBaseUnits(String asset) {
		String ret = "";
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			return ret;
		}

		// walk unit list looking for first one with conversion equal to 1.0
		for (AssetConfig.UnitConfig uConf : aConf.m_Units.values()) {
			if (uConf.m_Conversion == 1.0) {
				ret = uConf.m_Unit;
				break;
			}
		}
		return ret;
	}

	/**
	 * obtain the conversion factor between a given unit and base units
	 * @param asset the asset we're dealing with
	 * @param unit the unit whose conversion factor we want
	 * @return the ratio at which the unit can be converted to base (note this
	 * will be 1.0 if the unit is in fact itself the base unit type, or on error)
	 */
	public static double getBaseUnitConversion(String asset, String unit) {
		double conv = 1.0;
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			 return conv;
		}
		if (unit == null || unit.isEmpty()) {
			return conv;
		}
		AssetConfig.UnitConfig uConf = aConf.m_Units.get(unit);
		if (uConf == null) {
			return conv;
		}
		conv = uConf.m_Conversion;
		return conv;
	}

	/**
	 * obtain the display units for a given unit
	 * @param asset the asset we're dealing with
	 * @param unit the unit whose display code we want
	 * @return the display units for the asset and unit, or empty string on error
	 */
	public static String getDisplayUnits(String asset, String unit) {
		String disp = "";
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			 return disp;
		}
		if (unit == null || unit.isEmpty()) {
			return disp;
		}
		disp = unit;
		AssetConfig.UnitConfig uConf = aConf.m_Units.get(unit);
		if (uConf == null) {
			return disp;
		}
		disp = uConf.m_Display;
		return disp;
	}

	/**
	 * obtain the price lookup units for a given unit
	 * @param asset the asset we're dealing with
	 * @param unit the unit whose price lookup code we want
	 * @return the price table units for the asset and unit
	 */
	public static String getPriceLookupUnits(String asset, String unit) {
		String look = "";
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			 return look;
		}
		if (unit == null || unit.isEmpty()) {
			return look;
		}
		look = unit;
		AssetConfig.UnitConfig uConf = aConf.m_Units.get(unit);
		if (uConf == null) {
			return look;
		}
		look = uConf.m_Lookup;
		return look;
	}

	/**
	 * obtain the price lookup units for base units
	 * @param asset the asset we're dealing with
	 * @return the price table units for the asset and the base unit
	 */
	public static String getBasePriceLookupUnits(String asset) {
		return getPriceLookupUnits(asset, getBaseUnits(asset));
	}

	/**
	 * obtain a list of all configured voucher assets
	 * @return the list, or an empty list if not found
	 */
	public static ArrayList<String> getAllAssets() {
		ArrayList<String> aList = new ArrayList<String>();
		aList.addAll(m_AssetConfig.keySet());
		return aList;
	}

	/**
	 * obtain a list of all units for an asset
	 * @param asset the asset we're dealing with
	 * @return the list, or an empty list if not found
	 */
	public static ArrayList<String> getAllAssetUnits(String asset) {
		ArrayList<String> uList = new ArrayList<String>();
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			return uList;
		}
		// add all UnitConfigS names to the list
		uList.addAll(aConf.m_Units.keySet());
		return uList;
	}

	/**
	 * obtain a list of all voucher units, whether or not used for tokens
	 * @return the list, empty if none configured
	 */
	public static ArrayList<String> getAllUnits() {
		ArrayList<String> uList = new ArrayList<String>(20);
		for (AssetConfig aConf : m_AssetConfig.values()) {
			// NB: this may result in a given unit being on the list multiple times
			uList.addAll(aConf.m_Units.keySet());
		}
		return uList;
	}

	/**
	 * obtain the units used for tokens for an asset
	 * @param asset the asset we're dealing with
	 * @return the token units, or an empty string on error
	 */
	public static String getTokenUnits(String asset) {
		String tUnit = "";
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			return tUnit;
		}
		// loop through all UnitConfigS
		for (AssetConfig.UnitConfig uConf : aConf.m_Units.values()) {
			if (uConf.m_TokenValue != 0.0) {
				tUnit = uConf.m_Unit;
				break;
			}
		}
		return tUnit;
	}

	/**
	 * obtain all units used for tokens, across all assets
	 * @return the list, or an empty list if none found
	 */
	public static ArrayList<String> getAllTokenUnits() {
		ArrayList<String> uList = new ArrayList<String>();
		// NB: this may result in units being on the list more than once
		for (AssetConfig aConf : m_AssetConfig.values()) {
			for (AssetConfig.UnitConfig uConf : aConf.m_Units.values()) {
				if (uConf.m_TokenValue != 0.0) {
					uList.add(uConf.m_Unit);
					break;
				}
			}
		}
		return uList;
	}

	/**
	 * obtain the value of one token for a given asset
	 * @param asset the asset we're dealing with
	 * @param unit the unit whose token value we want (optional param)
	 * @return the value of a token in this asset, or 0.0 if not found
	 */
	public static double getValueOfOneToken(String asset, String unit) {
		double token = 0.0;
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			 return token;
		}
		if (unit == null || unit.isEmpty()) {
			// walk through all UnitConfigS looking for the one with non-zero value
			for (AssetConfig.UnitConfig uConf : aConf.m_Units.values()) {
				if (uConf.m_TokenValue != 0.0) {
					token = uConf.m_TokenValue;
					break;
				}
			}
		}
		else {
			// get the specific asset:unit's specification (might be 0.0)
			AssetConfig.UnitConfig uConf = aConf.m_Units.get(unit);
			if (uConf == null) {
				return token;
			}
			token = uConf.m_TokenValue;
		}
		return token;
	}

	/**
	 * obtain the value of 25 tokens worth of a particular asset
	 * @param asset the base asset the token represents
	 * @return the value of 25 tokens of that type, zero if not found
	 */
	public static double get25TokensWorth(String asset) {
		return 25.0 * getValueOfOneToken(asset, "");
	}

	/**
	 * determine whether a token value is recognized for some asset
	 * @param tokVal the token value to be checked
	 * @return true if at least one asset:unit corresponds to the value given
	 */
	public static boolean isTokenValueRecognized(double tokVal) {
		if (tokVal <= 0.0) {
			return false;
		}
		for (AssetConfig aConf : m_AssetConfig.values()) {
			for (AssetConfig.UnitConfig uConf : aConf.m_Units.values()) {
				if (uConf.m_TokenValue == tokVal) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * add a token value definition
	 * @param asset the asset the token is defined in
	 * @param unit the unit the token is defined in
	 * @param tokVal the token value
	 * @return true if a change was made
	 */
	public static boolean addTokenDefinition(String asset,
														  String unit,
														  double tokVal)
	{
		if (asset == null || asset.isEmpty() || unit == null || unit.isEmpty()
			 || tokVal <= 0.0)
		{
			return false;
		}
		AssetConfig aConf = getAssetConfig(asset);
		if (aConf == null) {
			return false;
		}
		AssetConfig.UnitConfig uConf = aConf.m_Units.get(unit);
		if (uConf == null) {
			return false;
		}
		if (uConf.m_TokenValue == tokVal) {
			// nothing to change
			return false;
		}
		uConf.m_TokenValue = tokVal;
		return true;
	}

	// end methods
}
