/*
 * last modified---
 * 	07-17-15/jet/ use VPAssetConfig methods instead of static Vector
 * 	06-05-13/jet/ SilentVault client version
 * 	06-28-11/kbw/ take m_ValueUnits from network config (see OFSListener)
 * 	06-08-11/kbw/ add USV and EUV as output units
 * 	02-28-11/kbw/ allow GLO and SIL as output units
 * 	12-30-10/kbw/ contrain Locale to Locale.US, so we don't confuse the VP
 * 	08-15-10/kbw/ add M_PaymentCost, M_ReceiveCost, M_ReceiptCost
 * 	07-16-09/kbw/ add m_PayerVS field and its use for picking up payments
 * 	05-08-09/kbw/ allow encryption without checking tokens (for reclaims)
 * 	04-20-09/kbw/ add setAction() and addPaymentToken(); make costs public
 * 	02-10-09/kbw/ add VS number to encrypted <details/>; use HashMap for
 * 					  storing vouchers
 * 	12-03-08/kbw/ new
 *
 * purpose---
 * 	present a class to represent a <voucherRequest/> element, used in certain
 * 	messages of the REQ_send_payment family
 *
 * notes---
 * 	see underlying message specifications in vp_ofs.x and ofs_vp.x, which
 * 	use <voucherRequest/> elements -- this class is from the VSC's point
 * 	of view
 *
 * 	this class must be kept in sync with the class of the same name in the
 * 	Openfire OFS and VP plugins, i.e.:
 * 	org.vouchersafe.openfire.VoucherRequest
 * 	org.vouchersafe.publisher.VoucherRequest
 */

package org.silentvault.client;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import java.util.Vector;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Class to instantiate a voucherRequest element (message component) as an
 * XML object.  The class knows how to store, parse, and emit itself in XML.
 * A similar class is utilized in the Openfire plugins for VP and OFS.
 *
 * @author Kevin Wilkerson
 */
public final class VoucherRequest implements Cloneable {
	// BEGIN data members

	/**
	 * all possible values for the <action/> element
	 */
	private static Vector<String> m_ActionTypes = getActionTypes();

	// token costs for each type of action
	/**
	 * token cost to validate a single voucher
	 */
	public static final int			M_ValidateCost = 4;
	/**
	 * token cost to merge multiple vouchers together
	 */
	public static final int			M_MergeCost = 12;
	/**
	 * token cost to split a voucher into smaller vouchers
	 */
	public static final int			M_SplitCost = 12;
	/**
	 * token cost to do a combined merge and split
	 */
	public static final int			M_BothCost = 12;
	/**
	 * token cost to perform a payment, plus storing a payment record.
	 * (NB: the two receipts are stored independently, and tokens will be bought
	 * in a separate step at that point if required.)
	 */
	public static final int			M_PaymentCost = 13;
	/**
	 * token cost to close out a safe
	 */
	public static final int			M_CloseCost = 25;
	/**
	 * token cost to receive a payment plus storage of two receipts
	 */
	public static final int			M_ReceiveCost = 6;
	/**
	 * token cost to store a receipt
	 */
	public static final int			M_ReceiptCost = 1;

	/**
	 * precision to use on a <value/> quantity
	 */
	private static final int		M_ValuePrecision = 4;

	/**
	 * the action element (determines contents of input and output)
	 * must always be a member of m_ActionTypes
	 */
	private String						m_Action;

	/**
	 * the details element, sent base64-encoded encrypted to the VP's pubkey
	 */
	private String						m_Details;

	/**
	 * whether or not the m_Details field has been encrypted yet
	 */
	private boolean					m_DetailsEncrypted;

	/**
	 * the list of input vouchers, in the input section of the details
	 */
	private HashMap<String, XMLVoucher>	m_Vouchers;

	/**
	 * our VS number, corresponding to our signing key
	 */
	private String						m_SigningVS;

	/**
	 * our voucher/token folder hash of the VS on the SDS, base64-encoded
	 */
	private String						m_FolderHash;

	/**
	 * our voucher/token folder read-write cap of the VS on the SDS, encrypted
	 * with the VP's public key and base64-encoded
	 */
	private String						m_FolderCap;

	/**
	 * our private key signature on the details block, base64-encoded
	 */
	private String						m_Signature;

	/**
	 * the tokens we're offering as payment for the request
	 */
	private ArrayList<XMLToken>   m_PaymentTokens;

	/**
	 * the contents of the output element, under construction
	 */
	private StringBuffer				m_Values;

	/**
	 * the payer VS#, used only when picking up a payment
	 */
	private String						m_PayerVS;

	/**
	 * flag whether this object is fully initialized and ready to emit as XML
	 */
	private boolean					m_Initialized;

	// END data members
	
	// BEGIN methods
	
	/**
	 * initialization method to set all possible values for the action element
	 * @return a Vector of all legal values
	 */
	private static Vector<String> getActionTypes() {
		Vector<String> vActions = new Vector<String>(3);
		vActions.add(new String("validate"));
		vActions.add(new String("split"));
		vActions.add(new String("merge"));
		return vActions;
	}

	/**
	 * basic constructor
	 */
	public VoucherRequest() {
		m_Action = m_Details = m_FolderHash = m_FolderCap = m_Signature = "";
		m_SigningVS = m_PayerVS = "";
		m_Vouchers = new HashMap<String, XMLVoucher>();
		m_PaymentTokens = new ArrayList<XMLToken>(M_BothCost);
		m_Values = new StringBuffer(1024);
	}

	// get accessor methods
	/**
	 * get the details element
	 * @return the details, encrypted and base64-encoded or not depending on
	 * the value of m_DetailsEncrypted
	 */
	public String getDetails() { return m_Details; }

	/**
	 * determine whether m_Details is decrypted yet
	 * @return true if encrypted, false if still stored in clear
	 */
	public boolean isEncrypted() { return m_DetailsEncrypted; }

	/**
	 * get the action value
	 * @return the action
	 */
	public String getAction() { return m_Action; }

	/**
	 * get list of input vouchers
	 * @return the list
	 */
	public HashMap<String, XMLVoucher> getVouchers() { return m_Vouchers; }

	/**
	 * get our VS number
	 */
	public String getSigningVS() { return m_SigningVS; }

	/**
	 * get our voucher/token folder index id hash
	 * @return the hash, base64-encoded
	 */
	public String getFolderHash() { return m_FolderHash; }

	/**
	 * get our voucher/token folder read-write cap
	 * @return the cap, encrypted and base64-encoded (opaque to us)
	 */
	public String getFolderCap() { return m_FolderCap; }

	/**
	 * get our private key signature on the <details/> block
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * get the payer VS# (used only when picking up a payment)
	 */
	public String getPayerVS() { return m_PayerVS; }

	/**
	 * determine whether object is fully built/initialized
	 * @return true if initialized, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }


	// set methods
	/**
	 * set our VS number
	 * @param vsNum the number of the voucher safe issuing the request
	 */
	public void setSigningVS(String vsNum) {
		if (vsNum != null && !vsNum.isEmpty()) {
			m_SigningVS = new String(vsNum);
		}
	}

	/**
	 * set the action type
	 * @param action the action, which must be valid
	 */
	public boolean setAction(String action) {
		if (action == null || !m_ActionTypes.contains(action.toLowerCase())) {
			Log.error("Illegal voucher request action type, " + action);
			return false;
		}
		m_Action = action;
		return true;
	}

	/**
	 * set the folder index hash
	 * @param hash the index of our SDS voucher/token folder, base64-encoded
	 */
	public void setFolderHash(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_FolderHash = new String(hash);
		}
	}

	/**
	 * set the folder read-write capability
	 * @param cap the cap, encrypted with VP's public key and base64-encoded
	 */
	public void setFolderCap(String cap) {
		if (cap != null && !cap.isEmpty()) {
			m_FolderCap = new String(cap);
		}
	}

	/**
	 * add a desired output value to the list
	 * @param val the value
	 * @param units the units (must be in VPAssetConfig's list)
	 * @return true if successfully added to <output/>, false otherwise
	 */
	public boolean addOutputValue(double val, String units) {
		if (val <= 0.0) {
			Log.error(val + " is an illegal VR output value");
			return false;
		}
		if (!VPAssetConfig.getAllUnits().contains(units.toUpperCase())) {
			Log.error(units + " is an illegal VR value unit");
			return false;
		}
		m_Values.append("<value units=\"" + units.toUpperCase() + "\">");
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_ValuePrecision);
		df.setMinimumFractionDigits(M_ValuePrecision);
		df.setGroupingSize(0);
		m_Values.append(df.format(val) + "</value>");
		return true;
	}

	/**
	 * add a token to the payment list
	 * @param tok the token
	 */
	public void addPaymentToken(XMLToken tok) {
		if (tok == null) {
			return;
		}
		m_PaymentTokens.add(tok);
	}

	/**
	 * set the payer VS (used only when picking up a payment)
	 * @param vsNum the VS# of the payer
	 */
	public void setPayerVS(String vsNum) {
		if (vsNum != null && !vsNum.isEmpty()) {
			m_PayerVS = new String(vsNum);
		}
	}

	/**
	 * manually set initialization status
	 */
	public void setInitialized() {
		m_Initialized = true;
	}


	/**
	 * method to sign the details block with our private key
	 * @param privKey our private key for constructing the signature
	 * @return true on success (and m_Signature set), false otherwise
	 */
	public boolean signDetails(PrivateKey privKey) {
		// check whether we've called signDetails() before this
		if (!m_Initialized) {
			Log.error("Cannot encrypt details on uninitialized VR block");
			return false;
		}

		// build the data to be signed
		StringBuilder signBuff = new StringBuilder(256 * M_BothCost);
		if (m_Vouchers.isEmpty()) {
			signBuff.append("<input/>");
		}
		else {
			signBuff.append("<input>");
			for (String fname : m_Vouchers.keySet()) {
				XMLVoucher voucher = m_Vouchers.get(fname);
				signBuff.append(voucher.toXML(fname));
			}
			signBuff.append("</input>");
		}
		if (m_Values.toString().isEmpty()) {
			signBuff.append("<output/>");
		}
		else {
			signBuff.append("<output>");
			signBuff.append(m_Values.toString());
			signBuff.append("</output>");
		}
		signBuff.append("<vs_number>" + m_SigningVS + "</vs_number>");
		signBuff.append("<sdshash>" + m_FolderHash + "</sdshash>");
		signBuff.append("<cap>" + m_FolderCap + "</cap>");
		if (m_Action.equals("validate") && !m_PayerVS.isEmpty()) {
			signBuff.append("<payer_vs>" + m_PayerVS + "</payer_vs>");
		}

		Signature dSign = null;
		byte[] sig = null;
		try {
			dSign = Signature.getInstance("SHA1withRSA");
			dSign.initSign(privKey);
			dSign.update(signBuff.toString().getBytes());
			sig = dSign.sign();
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
			return false;
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for signature creation", ike);
			return false;
		}
		catch (SignatureException se) {
			Log.error("Unable to sign voucher request", se);
			return false;
		}
		if (sig == null) {
			// shouldn't happen if no exception was thrown
			return false;
		}
		m_Signature = StringUtils.encodeBase64(sig, false);

		// append signature element to <details/>
		signBuff.append("<signature>" + m_Signature + "</signature>");
		m_Details = signBuff.toString();

		// next step is to encrypt
		m_Initialized = true;
		return true;
	}

	/**
	 * private helper method to validate the supplied tokens against the action
	 * @return true if the payment we've offered is sufficient, false if not
	 */
	private boolean checkTokens() {
		if (m_Action.isEmpty() || !m_ActionTypes.contains(m_Action)) {
			Log.error("action type not valid; cannot check token sufficiency");
			return false;
		}

		if (m_Action.equals("validate")) {
			if (m_PaymentTokens.size() < M_ValidateCost) {
				Log.error("Insufficient tokens (" + m_PaymentTokens.size()
							+ ") provided for " + m_Action);
				return false;
			}
		}
		else if (m_Action.equals("merge")) {
			if (m_PaymentTokens.size() < M_MergeCost) {
				Log.error("Insufficient tokens (" + m_PaymentTokens.size()
							+ ") provided for " + m_Action);
				return false;
			}
		}
		else if (m_Action.equals("split")) {
			if (m_PaymentTokens.size() < M_SplitCost) {
				Log.error("Insufficient tokens (" + m_PaymentTokens.size()
							+ ") provided for " + m_Action);
				return false;
			}
		}

		return true;
	}
	
	/**
	 * method to encrypt the details block, and validate the sufficiency of
	 * the tokens offered as payment
	 * @param vpPub the VP's public key, used to encrypt the block
	 * @param payReq true if check for payment is to be performed
	 * @return true if all is well, false if a problem is encountered
	 */
	public boolean encryptDetails(PublicKey vpPub, boolean payReq)
	{
		// check that we've called signDetails() before this
		if (!m_Initialized) {
			Log.error("Cannot encrypt details on uninitialized VR block");
			return false;
		}
		if (m_DetailsEncrypted) {
			Log.warning("Redundant call to VoucherRequest.encryptDetails()");
			return true;
		}

		/* We must perform the following steps, in order:
		 * 	1. Check the sufficiency of our payment tokens for what we're doing.
		 * 	2. Using the VP's public key, encrypt the <details/> block.
		 */

		// Step 1 - check tokens (m_Action must be set too)
		if (payReq && !checkTokens()) {
			Log.error("Could not validate payment tokens for VR action");
			return false;
		}

		// Step 2 - encrypt blob
		String encDet = EncodingUtils.makeBase64PubkeyEncStr(m_Details, vpPub);
		if (encDet == null) {
			Log.error("Unable to encrypt <details/> blob in <voucherRequest/>");
			return false;
		}

		// set done
		m_Details = encDet;
		m_DetailsEncrypted = true;
		return true;
	}

	/**
	 * method to emit this voucher request as valid XML
	 * @return the XML string, or the empty string if object is not ready
	 */
	public String toXML() {
		if (!m_DetailsEncrypted) {
			Log.error("VR not prepared for emitting as XML");
			return "";
		}
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<voucherRequest>");
		xml.append("<action>" + m_Action + "</action>");
		xml.append("<details>" + m_Details + "</details>");
		xml.append("<payment>");
		for (XMLToken token : m_PaymentTokens) {
			xml.append(token.toXML());
		}
		xml.append("</payment>");
		xml.append("</voucherRequest>");
		return xml.toString();
	}

	/**
	 * method to implement the Cloneable interface
	 * @return a new voucher request object equal to this one
	 */
	public Object clone() {
		try {
			VoucherRequest nReq = (VoucherRequest) super.clone();
			nReq.m_Details = new String(m_Details);
			nReq.m_FolderHash = new String(m_FolderHash);
			nReq.m_FolderCap = new String(m_FolderCap);
			nReq.m_Signature = new String(m_Signature);
			nReq.m_Vouchers = new HashMap<String, XMLVoucher>(m_Vouchers);
			nReq.m_Values = new StringBuffer(m_Values);
			nReq.m_PaymentTokens = new ArrayList<XMLToken>(m_PaymentTokens);
			return nReq;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	// END methods
}
