/*
 * last modified---
 * 	10-02-14/jet/ SSM -> SVM
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	10-11-13/jet/ support multiple endpoint paths to SDS, PKS, DHT
 * 	06-05-13/jet/ SilentVault client version
 * 	02-02-09/kbw/ add folder indices!
 * 	10-29-08/kbw/ add m_Certificate and m_VPPubKey
 * 	10-02-08/kbw/ add m_VSnumber
 * 	10-01-08/kbw/ add m_SelectedVP
 * 	09-08-08/kbw/ new
 *
 * purpose---
 * 	class to hold VSC login secrets and network endpoint blob data
 */

package org.silentvault.client;

import java.security.KeyPair;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.security.cert.X509Certificate;


/**
 * Represents the login blob and network configuration data for a VSC.
 * The data in this object will be valid only when the VS client is
 * successfully logged in to a voucher safe.
 *
 * @author Kevin Wilkerson
 */
public final class VsSecrets {
	// BEGIN data members
	
	/**
	 * the voucher safe number (resembles email address, id123@voucher-safe.org)
	 */
	private String										m_VSnumber;

	/**
	 * the public/private key pair
	 */
	private KeyPair									m_KeyPair;

	/**
	 * the certificate signed by the VP, based on the public key
	 */
	private X509Certificate							m_Certificate;

	/**
	 * the voucher/token folder hash index, stored as char array for security
	 */
	private char[]										m_VoucherIndex;

	/**
	 * the voucher/token folder read cap, stored as char array for erasure
	 */
	private char[]										m_VoucherReadCap;

	/**
	 * the receipts/pending payments folder hash index, stored as char array
	 * for erasure
	 */
	private char[]										m_ReceiptIndex;

	/**
	 * the receipts/pending payments read-write cap, stored as char array for
	 * erasure
	 */
	private char[]										m_ReceiptRWCap;

	/**
	 * secret hash to mark records read on the DHT, stored as char array for
	 * erasure
	 */
	private char[]										m_DHTMarkHash;

	/**
	 * the read-write cap for voucher/token folder, encrypted with VP's pubkey
	 */
	private String										m_EncVouchRWCap;

	/**
	 * the XMPP domain name of the SDS for the VP
	 */
	private String										m_SDSDomain;

	/**
	 * set of available SDS gateway endpoints (paths to the c2s port)
	 */
	private final ArrayList<InetSocketAddress>	m_SDSGateways;

	/**
	 * the XMPP domain name of the DHT for the VP
	 */
	private String										m_DHTDomain;

	/**
	 * set of available DHT nodes (paths to the c2s port)
	 */
	private final ArrayList<InetSocketAddress>	m_DHTNodes;

	/**
	 * set of available key servers, XMPP domain mapped to c2s port paths
	 */
	private final HashMap<String, ArrayList<InetSocketAddress>>	m_KeyServers;

	/**
	 * The voucher publisher (VP) the VS is working with
	 */
	private String										m_SelectedVP;

	/**
	 * The public key of the VP, as retrieved from the key server
	 */
	private PublicKey									m_VPPubKey;

	/**
	 * the Silent-Vault Market (SVM) the VS is working with (if any)
	 */
	private String										m_ConnectedSVM;

	/**
	 * the public key of the SVM, as retrieved from the key server
	 */
	private PublicKey									m_SVMPubKey;

	/**
	 * the Silent-Vault Exchange (SVX) the VS is working with (if any)
	 */
	private String										m_ConnectedSVX;

	/**
	 * the public key of the SVX, as retrieved from the key server
	 */
	private PublicKey									m_SVXPubKey;

	/**
	 * whether this object is initialized
	 */
	private boolean									m_IsInitialized;

	// END data members


	// BEGIN methods
	
	/**
	 * no-args constructor
	 */
	public VsSecrets() {
		m_KeyPair = null;
		m_SDSGateways = new ArrayList<InetSocketAddress>();
		m_DHTNodes = new ArrayList<InetSocketAddress>();
		m_KeyServers = new HashMap<String, ArrayList<InetSocketAddress>>();
		m_VoucherReadCap = m_ReceiptRWCap = m_DHTMarkHash = null;
		m_VSnumber = m_EncVouchRWCap = m_SelectedVP = m_ConnectedSVM = "";
		m_ConnectedSVX = m_DHTDomain = m_SDSDomain = "";
		m_IsInitialized = false;
	}


	// get accesssor methods
	/**
	 * obtain the VS number
	 */
	public String getVSnumber() { return m_VSnumber; }

	/**
	 * obtain the public key
	 * @return the key, or null if not set
	 */
	public PublicKey getPubKey() {
		if (m_KeyPair == null) {
			return null;
		}
		return m_KeyPair.getPublic();
	}

	/**
	 * obtain the private key
	 * @return the key, or null if not set
	 */
	public PrivateKey getPrivKey() {
		if (m_KeyPair == null) {
			return null;
		}
		return m_KeyPair.getPrivate();
	}

	/**
	 * obtain the signed certificate
	 * @return the cert, or null if not set
	 */
	public X509Certificate getCertificate() { return m_Certificate; }

	/**
	 * get voucher/token folder hash index
	 * @return a clone of the index, to avoid side-effects
	 */
	public char[] getVoucherIndex() { return m_VoucherIndex.clone(); }

	/**
	 * get voucher/token folder read cap
	 * @return a clone of the read cap, to avoid side-effects
	 */
	public char[] getVoucherReadCap() { return m_VoucherReadCap.clone();
	}
	
	/**
	 * get receipts/payments folder hash index
	 * @return a clone of the index, to avoid side-effects
	 */
	public char[] getReceiptIndex() { return m_ReceiptIndex.clone(); }

	/**
	 * get receipts read-write cap
	 * @return a clone of the read-write cap, to avoid side-effects
	 */
	public char[] getReceiptRWCap() { return m_ReceiptRWCap.clone(); }

	/**
	 * get DHT record marking hash
	 * @return a clone of the hash, to avoid side-effects
	 */
	public char[] getDHTMarkHash() { return m_DHTMarkHash.clone(); }

	/**
	 * get encrypted voucher read-write cap
	 * @return the read-write cap, encrypted with VP's public key
	 */
	public String getVouchRWCap() { return m_EncVouchRWCap; }

	/**
	 * obtain the XMPP domain of the SDS
	 * @return the domain of the SDS for the logged-in VP
	 */
	public String getSDSDomain() { return m_SDSDomain; }

	/**
	 * obtain the set of available SDS gateways
	 * @return a list of socket addresses connecting to the SDS c2s port
	 */
	public ArrayList<InetSocketAddress> getSDSGateways() {
		return m_SDSGateways;
	}

	/**
	 * obtain the XMPP domain of the DHT
	 * @return the domain of the DHT for the logged-in VP
	 */
	public String getDHTDomain() { return m_DHTDomain; }

	/**
	 * obtain the set of available DHT nodes
	 * @return a list of socket addresses connecting to the DHT c2s port
	 */
	public ArrayList<InetSocketAddress> getDHTNodes() { return m_DHTNodes; }

	/**
	 * obtain the set of available key servers
	 * @return a HashMap of XMPP domains mapped to socket addresses
	 */
	public HashMap<String, ArrayList<InetSocketAddress>> getKeyServers() {
		return m_KeyServers;
	}

	/**
	 * obtain the selected voucher publisher with which the user is working
	 * @return the id of the publisher
	 */
	public String getPublisher() { return m_SelectedVP; }

	/**
	 * obtain the public key of the voucher publisher
	 * @return the key, or null if it has not yet been retrieved from the PKS
	 */
	public PublicKey getVPKey() { return m_VPPubKey; }

	/**
	 * obtain the connected SVM with which the user is working
	 * @return the id of the SVM (may be the empty string)
	 */
	public String getSVMId() { return m_ConnectedSVM; }

	/**
	 * obtain the public key of the SVM
	 * @return the key, or null if it has not yet been retrieved from the PKS
	 */
	public PublicKey getSVMKey() { return m_SVMPubKey; }

	/**
	 * obtain the connected SVX with which the user is working
	 * @return the id of the SVX (may be the empty string)
	 */
	public String getSVXId() { return m_ConnectedSVX; }

	/**
	 * obtain the public key of the SVX
	 * @return the key, or null if it has not yet been retrieved from the PKS
	 */
	public PublicKey getSVXKey() { return m_SVXPubKey; }

	/**
	 * determine whether object has been initialized yet
	 * @return true if initialized, false if not
	 */
	public boolean isInit() { return m_IsInitialized; }


	// set access methods
	/**
	 * record the data values found in a successfully decrypted login blob
	 * @param VSnum the number of the voucher safe
	 * @param cert the public key certificate, signed by the VP
	 * @param privKey the private key
	 * @param vouchIndex the voucher/token folder hash index
	 * @param vouchReadCap the voucher/token folder read cap
	 * @param receiptIndex the receipt/pending payment folder hash index
	 * @param receiptRWCap the receipt/pending payment folder read-write cap
	 * @param DHThash the hash to mark records read on the DHT
	 * @param vouchRWCap the encrypted voucher/token folder read-write cap
	 * @return true if initialization was successful, false otherwise
	 */
	public boolean initialize(	String VSnum,
										X509Certificate cert,
										PrivateKey privKey,
										char[] vouchIndex,
										char[] vouchReadCap,
										char[] receiptIndex,
										char[] receiptRWCap,
										char[] DHThash,
										String vouchRWCap	)
	{
		// error-check input
		if (cert == null || privKey == null
			 || vouchIndex == null || vouchIndex.length == 0
			 || vouchReadCap == null || vouchReadCap.length == 0
			 || receiptIndex == null || receiptIndex.length == 0
			 || receiptRWCap == null || receiptRWCap.length == 0
			 || DHThash == null || DHThash.length == 0
			 || vouchRWCap == null || vouchRWCap.isEmpty()
			 || VSnum == null || VSnum.isEmpty())
		{
			return false;
		}

		// no attempt is made here to validate that the data are correct
		m_VSnumber = new String(VSnum);
		m_Certificate = cert;
		m_KeyPair = new KeyPair(cert.getPublicKey(), privKey);
		m_VoucherIndex = vouchIndex.clone();
		m_VoucherReadCap = vouchReadCap.clone();
		m_ReceiptIndex = receiptIndex.clone();
		m_ReceiptRWCap = receiptRWCap.clone();
		m_DHTMarkHash = DHThash.clone();
		m_EncVouchRWCap = new String(vouchRWCap);
		m_IsInitialized = true;
		return true;
	}

	// following login, the VSC will ask the OFS for the network endpoints
	/**
	 * initialize all set members to record network endpoints
	 * @param SDSdomain the XMPP domain for the SDS
	 * @param SDSlist the list of SDS gateway endpoints (paths)
	 * @param DHTdomain the XMPP domain for the DHT
	 * @param DHTlist the list of DHT nodes (paths)
	 * @param PKSlist the map of public key server domains to endpoints
	 * @return true if all recorded successfully, false otherwise
	 */
	public boolean initNetwork(String SDSdomain,
										List<InetSocketAddress> SDSlist,
										String DHTdomain,
										List<InetSocketAddress> DHTlist,
								HashMap<String, ArrayList<InetSocketAddress>> PKSlist)
	{
		// must be initialized with login blob data first
		if (!m_IsInitialized) {
			Log.warning("Init network conf without login blob");
		}
		m_SDSGateways.clear();
		m_DHTNodes.clear();
		m_KeyServers.clear();

		// copy and add the input lists
		boolean gotErr = false;
		m_SDSDomain = new String(SDSdomain);
		if (!m_SDSGateways.addAll(SDSlist)) {
			gotErr = true;
		}
		m_DHTDomain = new String(DHTdomain);
		if (!m_DHTNodes.addAll(DHTlist)) {
			gotErr = true;
		}

		/* This list might not change, if we already recorded the only PKS while
		 * creating a safe.
		 */
		m_KeyServers.putAll(PKSlist);
		return !gotErr;
	}

	/**
	 * de-initialize the object (called when voucher activity timeout expires, on
	 * logout, or during safe closure)
	 */
	public void deinitialize() {
		if (!m_IsInitialized) {
			// nothing to do
			return;
		}

		// toss keypair and VP key, SVM key, SVX key
		// NB: the PrivateKey::finalize() method will null out the private key's
		// 	 bytes so that they will not still be floating around on the heap.

		m_KeyPair = null;
		m_VPPubKey = null;
		m_SVMPubKey = null;
		m_SVXPubKey = null;

		// reset arrays to null so they won't
		// still be floating around on the JVM heap awaiting garbage collection.

		Arrays.fill(m_VoucherIndex, ' ');
		Arrays.fill(m_VoucherReadCap, ' ');
		Arrays.fill(m_ReceiptIndex, ' ');
		Arrays.fill(m_ReceiptRWCap, ' ');
		Arrays.fill(m_DHTMarkHash, ' ');
		m_VoucherReadCap = m_ReceiptRWCap = m_DHTMarkHash = null;
		m_EncVouchRWCap = m_SelectedVP = m_ConnectedSVM = m_ConnectedSVX = "";
		m_SDSDomain = m_DHTDomain = "";
		m_IsInitialized = false;

		// clear address sets
		m_SDSGateways.clear();
		m_DHTNodes.clear();
		m_KeyServers.clear();
	}

	/**
	 * record the voucher publisher with which this VS is correlated
	 * @param publisher the id of the voucher publisher
	 */
	public void setPublisher(String publisher) {
		if (publisher != null) {
			m_SelectedVP = new String(publisher);
		}
	}

	/**
	 * record the VP's public key when it is retrieved off the key server
	 * @param vpKey the key to be recorded
	 */
	public void setVPKey(PublicKey vpKey) {
		if (vpKey != null) {
			m_VPPubKey = vpKey;
		}
	}

	/**
	 * record the SVM Id with which this VS is working
	 * @param svmId the id of the SVM
	 */
	public void setSVMId(String svmId) {
		if (svmId != null) {
			m_ConnectedSVM = new String(svmId);
		}
	}

	/**
	 * record the SVM's public key when it is retrieve off the key server
	 * @param svmKey the key to be recorded (may be null to erase)
	 */
	public void setSVMKey(PublicKey svmKey) {
		m_SVMPubKey = svmKey;
	}

	/**
	 * record the SVX Id with which this VS is working
	 * @param svxId the id of the SVX
	 */
	public void setSVXId(String svxId) {
		if (svxId != null) {
			m_ConnectedSVX = new String(svxId);
		}
	}

	/**
	 * record the SVX's public key when it is retrieve off the key server
	 * @param svxKey the key to be recorded (may be null to erase)
	 */
	public void setSVXKey(PublicKey svxKey) {
		m_SVXPubKey = svxKey;
	}

	/**
	 * record the keypair generated, during safe creation
	 * @param kp the keypair being recorded
	 */
	public void setKeys(KeyPair kp) {
		if (kp != null && kp.getPrivate() != null && kp.getPublic() != null) {
			m_KeyPair = kp;
		}
	}

	/**
	 * record the DHT mark hash, which is computable
	 * @param hash the hash used to mark records read on the DHT
	 */
	public void setDHThash(String hash) {
		if (hash != null && !hash.isEmpty()) {
			m_DHTMarkHash = hash.toCharArray();
		}
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		if (m_IsInitialized) {
			try {
				// It is a goodness to overwrite the bytes of the cap strings
				// in order to prevent their continued presence on the heap.

				Arrays.fill(m_VoucherIndex, ' ');
				Arrays.fill(m_VoucherReadCap, ' ');
				Arrays.fill(m_ReceiptIndex, ' ');
				Arrays.fill(m_ReceiptRWCap, ' ');
				Arrays.fill(m_DHTMarkHash, ' ');
				m_VoucherReadCap = m_ReceiptRWCap = m_DHTMarkHash = null;
				m_EncVouchRWCap = m_SelectedVP = m_ConnectedSVM = "";
				m_ConnectedSVX = m_SDSDomain = m_DHTDomain = "";

				// clear address sets
				m_SDSGateways.clear();
				m_DHTNodes.clear();
				m_KeyServers.clear();

			} finally {
				m_IsInitialized = false;
				super.finalize();
			}
		}
		else {
			super.finalize();
		}
	}

	// END methods
}
