/*
 * last modified---
 * 	07-12-13/jet/ add manual override set methods
 * 	06-05-13/jet/ SilentVault client version
 * 	03-20-12/kbw/ use PKCS5PADDING
 * 	06-28-11/kbw/ impose limit of 3 total baggage fields
 * 	07-19-10/kbw/ add m_Issuer
 * 	04-19-10/kbw/ stand-alone client version
 * 	08-27-09/kbw/ new
 *
 * purpose---
 * 	class to manipulate messages sent from a web server to a VSC
 *
 * notes---
 * 	see underlying message definitions in web_vsc.x
 */

package org.silentvault.client;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.PacketExtension;
import org.jivesoftware.smack.util.StringUtils;

import java.util.Hashtable;
import java.util.Set;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.spec.InvalidKeySpecException;
import java.security.*;
import javax.crypto.*;
import javax.crypto.spec.SecretKeySpec;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;


/**
 * Holds all the pieces of a message which is part of the
 * jabber:message:voucher-safe#web-vsc protocol between a web server and a VSC
 * using the element web_vsc.
 * The class also knows how to emit itself as XML.
 * It is used as a utility class to assist in recording
 * incoming order-related messages and forming correct stanzas.
 *
 * @author Kevin Wilkerson
 */
public final class WebMessage extends Message implements PacketExtension {
	// BEGIN data members
	/**
	 * the namespace we are supplying with the extension
	 */
	private final String		M_Namespace = "jabber:message:voucher-safe#web-vsc";

	/**
	 * info contains the extension payload of the message, base64-encoded and
	 * 3DES-encrypted
	 */
	private String				m_info;

	/**
	 * whether or not info has been decoded yet
	 */
	private boolean			m_InfoDecoded;

	/**
	 * The opcode of this message, which
	 * determines which of the other fields will contain valid data.
	 * (Only one opcode is supported: NFY_order_info.)
	 */
	private String				m_opcode;

	/**
	 * website is the sub-element of info describing the URL of the website to
	 * which the order message relates
	 */
	private String				m_website;

	/**
	 * VSnumber is the sub-element of info supplying the seller website's VS#
	 */
	private String				m_VSnumber;

	/**
	 * amount is the sub-element of info giving the amount of the payment to
	 * the seller VS
	 */
	private float				m_amount;

	/**
	 * currency is the sub-element of info giving the units in which m_amount
	 * is expressed
	 */
	private String				m_currency;

	/**
	 * The Issuer of the voucher payment requested.  Must be supported by the
	 * VP, or the payment will fail.
	 */
	private String				m_Issuer;

	/**
	 * baggageFields is the sub-element of info listing the order description
	 * keys, separated by :--: delimiters
	 */
	private String				m_baggage_fields;

	/**
	 * the order data, built from sub-elements of info listed in m_baggage_fields
	 */
	private Hashtable<String, String>	m_OrderData;

	// END data members


	// BEGIN methods

	/**
	 * base constructor
	 */
	public WebMessage() {
		super();
		m_opcode = m_info = m_website = m_VSnumber = m_currency = "";
		m_baggage_fields = m_Issuer = "";
		m_OrderData = new Hashtable<String, String>();
	}


	// get accessor methods
	/**
	 * obtain the raw value of the info sub-element
	 * @return the value, still encrypted
	 */
	public String getInfo() { return m_info; }

	/**
	 * obtain the current opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_opcode; }

	/**
	 * get the order data hash table (valid when opcode = NFY_order_info)
	 * @return the order data
	 */
	public Hashtable<String, String> getOrderData() { return m_OrderData; }

	/**
	 * get the website
	 * @return the website URL
	 */
	public String getWebsite() { return m_website; }

	/**
	 * get the web payee VS#
	 * @return the voucher safe number the web vendor wants paid
	 */
	public String getVSnumber() { return m_VSnumber; }

	/**
	 * obtain the amount of the order
	 * @return the amount, expressed in getCurrency() units
	 */
	public float getAmount() { return m_amount; }

	/**
	 * obtain the currency units of the amount
	 * @return the currency, a 3-letter code (where GAU is grams of gold)
	 */
	public String getCurrency() { return m_currency; }

	/**
	 * obtain the requested voucher Issuer
	 * @return the name of the Issuer
	 */
	public String getIssuer() { return m_Issuer; }

	/**
	 * get whether the message is already decoded successfully
	 * @return true if decrypted, false otherwise
	 */
	public boolean isDecoded() { return m_InfoDecoded; }


	// set access methods
	/**
	 * set the value of the undecrypted info element
	 * @param encInfo the element text, 3DES-encrypted and base64-encoded
	 */
	public void setInfo(String encInfo) {
		if (encInfo != null && !encInfo.isEmpty()) {
			m_info = new String(encInfo);
		}
	}

	/**
	 * manual override to set amount
	 * @param amount the amount of the payment
	 */
	public void setAmount(double amount) {
		Double amt = new Double(amount);
		m_amount = amt.floatValue();
	}

	/**
	 * manual override to set currency units
	 * @param curr the currency (display) units of the payment
	 */
	public void setCurrency(String curr) {
		if (curr != null && !curr.isEmpty()) {
			m_currency = new String(curr);
		}
	}

	/**
	 * manual override to set payee VS#
	 * @param wallet the wallet Id to set as payee for the order
	 */
	public void setVSnumber(String wallet) {
		if (wallet != null && !wallet.isEmpty()) {
			m_VSnumber = new String(wallet);
		}
	}

	/**
	 * manual override to set Issuer name
	 * @param issuer the Issuer (asset type) to set
	 */
	public void setIssuer(String issuer) {
		if (issuer != null && !issuer.isEmpty()) {
			m_Issuer = new String(issuer);
		}
	}

	/**
	 * manual override to set baggage fields
	 * @param bagFields fields used as baggage, separated by :--: delimiters
	 */
	public void setBaggageFields(String bagFields) {
		if (bagFields != null) {
			m_baggage_fields = new String(bagFields);
		}
	}

	/**
	 * manual override to set decoded
	 * @param decoded whether or not message is to be deemed as "decoded"
	 */
	public void setDecoded(boolean decoded) {
		m_InfoDecoded = decoded;
	}

	/**
	 * add a record to the list of order data key/value pairs
	 * (needed when opcode = NFY_order_info)
	 * @param key the baggage field name to use as the key
	 * @param val the value supplied by the sub-element matching that key
	 */
	public void addOrderRecord(String key, String val) {
		if (key != null && !key.isEmpty() && val != null) {
			m_OrderData.put(key, val);
		}
	}

	/**
	 * decode and decrypt the info sub-element using the provided key
	 * @param key the 3DES decryption key
	 * @return true on successful parsing
	 */
	public boolean decodeInfo(String key) {
		if (key == null || key.isEmpty()) {
			Log.error("Need key to decrypt <info/>");
			return false;
		}

		// Set up cipher object for decryption using this key.  We pad it to 24
		// bytes (192 bits) with zeros, as needed.

		byte[] decKey = new byte[24];
		byte[] secKey = key.getBytes();
		int len = secKey.length > 24 ? 24 : secKey.length;
		System.arraycopy(secKey, 0, decKey, 0, len);
		SecretKeySpec keySpec = new SecretKeySpec(decKey, "DESede");
		Cipher passphraseCipher = null;
		byte[] clearBlob = null;
		boolean gotErr = false;
		try {
			// must use ECB, since we don't get an init vector
			passphraseCipher = Cipher.getInstance("DESede/ECB/PKCS5PADDING");
			passphraseCipher.init(Cipher.DECRYPT_MODE, keySpec);
			clearBlob = passphraseCipher.doFinal(StringUtils.decodeBase64(m_info));
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Invalid algorithm to decrypt order with 3DES", nsae);
			gotErr = true;
		}
		catch (NoSuchPaddingException nspe) {
			Log.error("Invalid padding to decrypt order with 3DES", nspe);
			gotErr = true;
		}
		catch (InvalidKeyException ike) {
			Log.error("Bad key for decrypting order with 3DES", ike);
			gotErr = true;
		}
		catch (IllegalBlockSizeException ibse) {
			Log.error("Bad block size 3DES decrypting order", ibse);
			gotErr = true;
		}
		catch (BadPaddingException bpe) {
			Log.error("Bad padding 3DES decrypting order", bpe);
			gotErr = true;
		}
		if (gotErr) {
			return false;
		}
		String decodedInfo = new String(clearBlob);
		//Log.debug("Decoded WebMessage is: \"" + decodedInfo + "\"");

		// now parse the XML
		Element infoElt
			= EncodingUtils.buildElement("<info>" + decodedInfo + "</info>");
		if (infoElt == null) {
			Log.error("Unable to parse order message from website");
			return false;
		}

		// examine the parts we should have
		try {
			// get opcode and validate it
			Element opElt = infoElt.element("opcode");
			if (opElt == null) {
				throw new DocumentException("Missing <opcode>");
			}
			m_opcode = opElt.getTextTrim();
			if (!m_opcode.equalsIgnoreCase("NFY_order_info")) {
				throw new DocumentException("Unexpected opcode, " + m_opcode);
			}

			// get website
			Element webElt = infoElt.element("website");
			if (webElt == null) {
				throw new DocumentException("Missing <website>");
			}
			m_website = webElt.getTextTrim();

			// get payee VS#
			Element vsElt = infoElt.element("VSnumber");
			if (vsElt == null) {
				throw new DocumentException("Missing <VSnumber>");
			}
			m_VSnumber = vsElt.getTextTrim();

			// get amount and validate it's a number
			Element amtElt = infoElt.element("amount");
			if (amtElt == null) {
				throw new DocumentException("Missing <amount>");
			}
			try {
				m_amount = Float.parseFloat(amtElt.getText());
			}
			catch (NumberFormatException nfe) {
				throw new DocumentException("Illegal amount value", nfe);
			}

			// get currency units
			Element currElt = infoElt.element("currency");
			if (currElt == null) {
				throw new DocumentException("Missing <currency>");
			}
			m_currency = currElt.getTextTrim();
			if (m_currency.isEmpty()) {
				// supply default
				m_currency = "GAU";
			}

			// get Issuer specification
			Element issElt = infoElt.element("issuer");
			if (issElt == null) {
				throw new DocumentException("Missing <issuer>");
			}
			m_Issuer = issElt.getTextTrim();

			// deal with order-describing baggage fields
			Element bagElt = infoElt.element("baggageFields");
			if (bagElt == null) {
				throw new DocumentException("Missing <baggageFields>");
			}
			m_baggage_fields = bagElt.getTextTrim();

			// baggage fields are optional, but all supplied in list must appear
			if (!m_baggage_fields.isEmpty()) {
				String[] bagKeys = null;
				if (m_baggage_fields.indexOf(":--:") == -1) {
					// there must be only one field
					bagKeys = new String[1];
					bagKeys[0] = m_baggage_fields;
				}
				else {
					bagKeys = m_baggage_fields.split(":--:");
				}
				for (int iii = 0; iii < bagKeys.length; iii++) {
					// max of 3 total baggage fields
					String bkey = bagKeys[iii];
					if (iii >= 3) {
						Log.error("Excess web order baggage field ignored, " + bkey);
						continue;
					}
					Element fldElt = infoElt.element(bkey);
					if (fldElt == null) {
						throw new DocumentException("Missing <" + bkey + "> bag elt");
					}
					String bagVal = fldElt.getTextTrim();
					addOrderRecord(bkey, bagVal);
				}
			}
		}
		catch (DocumentException de) {
			Log.error("Error parsing website message extension", de);
			return false;
		}

		m_InfoDecoded = true;
		return true;
	}


	// methods required to implement the PacketExtension interface
	/**
	 * provide the root element name
	 * @return the name
	 */
	public String getElementName() { return "web_vsc"; }

	/**
	 * provide the XML namespace of the root element
	 * @return the namespace
	 */
	public String getNamespace() { return M_Namespace; }

	/**
	 * provide the extension element as XML
	 * @return the extended child element as a XML string
	 */
	public String toXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<" + getElementName() + " xmlns=\"" + M_Namespace + "\">");
		if (!m_InfoDecoded) {
			// merely output the encoded data inside the one sub-element
			xml.append("<info>" + m_info + "</info>");
		}
		else {
			xml.append("<info>");
			xml.append("<opcode>" + m_opcode + "</opcode>");

			// based on the value of the opcode, emit additional fields
			if (m_opcode.equalsIgnoreCase("NFY_order_info")) {
				xml.append("<website>" + m_website + "</website>");
				xml.append("<VSnumber>" + m_VSnumber + "</VSnumber>");
				DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
				df.setMaximumFractionDigits(XMLVoucher.M_Precision);
				df.setMinimumFractionDigits(XMLVoucher.M_Precision);
				df.setGroupingSize(0);
				xml.append("<amount>" + df.format(m_amount) + "</amount>");
				xml.append("<currency>" + m_currency + "</currency>");
				xml.append("<issuer>" + m_Issuer + "</issuer>");
				xml.append("<baggageFields>" + m_baggage_fields);
				xml.append("</baggageFields>");
				for (String bagKey : m_OrderData.keySet()) {
					xml.append("<" + bagKey + ">" + m_OrderData.get(bagKey));
					xml.append("</" + bagKey + ">");
				}
			}
			else {
				Log.error("Improper opcode " + m_opcode);
			}
			xml.append("</info>");
		}
		xml.append("</" + getElementName() + ">");
		return xml.toString();
	}

	// END methods
}
