/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	02-22-11/kbw/ add getDisplayUnits(); remove use of Double.parseDouble()
 * 	12-29-10/kbw/ force Locale to US so we always parse properly
 * 	10-07-10/kbw/ correct problem where a TTL could not be set in the past
 * 	09-28-10/kbw/ deal with embedded newlines and angle brackets
 * 	07-25-10/kbw/ m_TTL is in usec; expiry calculation was correct without
 * 					  reference to GMT
 * 	07-10-10/kbw/ remove spurious restrictions on asset and unit types
 * 	06-09-10/kbw/ add m_Issuer field
 * 	05-16-10/kbw/ correct expiration calculation to GMT
 * 	08-17-09/kbw/ VS# must be mixed case
 * 	01-02-09/kbw/ new
 *
 * purpose---
 * 	implement a XML payment object
 *
 * notes---
 * 	this class is to be distinguished from XMLPendingPayment
 */

package org.silentvault.client;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Date;
import java.util.Locale;
import java.security.PublicKey;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Class to instantiate a payment as an XML object.  The class knows how to
 * store, parse, and emit a payment in XML.  It is used when a VSC has just
 * made a payment, and wishes to store the payment details in its folder on
 * the SDS.  This is done so that the value of the payment voucher can be
 * reclaimed if the payee does not pick it up off the DHT and validate the
 * voucher before the expiration of the TTL.
 *
 * @author Kevin Wilkerson
 */
public final class XMLPayment implements Cloneable {
	// BEGIN data members
	
	/**
	 * precision to use on quantity
	 */
	private static final int		M_Precision = 4;

	/**
	 * names of all the elements that are part of a payment
	 */
	private static Vector<String>	M_PaymentElements = getPaymentElements();

	/**
	 * the issuer of the voucher used in the payment
	 */
	private String						m_Issuer;

	/**
	 * the quantity of the underlying asset represented by the payment
	 */
	private Double						m_Quantity;

	/**
	 * the measurement units of the payment
	 */
	private String						m_Units;

	/**
	 * the underlying asset backing the payment
	 */
	private String						m_Asset;

	/**
	 * the voucher safe number of the payee for the payment
	 */
	private String						m_PayeeVS;

	/**
	 * the serial number of the voucher used to convey the payment.
	 * (issued to payee; returned from the VP in a {@link XMLConfirmPayment}
	 * DHT message)
	 */
	private String						m_SerialNumber;

	/**
	 * The TTL (time to live), representing the time in msecs since the epoch
	 * at which the offered payment will expire and become eligible to be
	 * reclaimed.  This value defaults to one week from the time the payment
	 * voucher is issued to the payee, but may be set shorter by the payer.
	 * (This value will be supplied to the DHT when the record is stored on it.)
	 */
	private long						m_TTL;

	/**
	 * The vendor baggage fields, if any, associated with the payment.  These
	 * would normally be acquired from the payee, via a website form, email, or
	 * other communications.  They are stored here so that the VS owner can
	 * readily see who the payment was made to or what it was for.  These values
	 * are always encrypted for privacy when sent in a message.
	 */
	private Hashtable<String, String>	m_BaggageFields;

	/**
	 * whether the payment has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a payment
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getPaymentElements() {
		Vector<String> pElts = new Vector<String>(9);
		pElts.add(new String("payment"));
		pElts.add(new String("amount"));
		pElts.add(new String("units"));
		pElts.add(new String("asset"));
		pElts.add(new String("issuer"));
		pElts.add(new String("payee_vs"));
		pElts.add(new String("serial"));
		pElts.add(new String("ttl"));
		pElts.add(new String("baggage"));
		return pElts;
	}

	/**
	 * basic constructor
	 */
	public XMLPayment() {
		m_Quantity = new Double(0.0);
		m_Units = m_Asset = m_SerialNumber = m_PayeeVS = m_Issuer = "";
		m_BaggageFields = new Hashtable<String, String>();
	}

	// get accessor methods
	/**
	 * get method for payment amount value
	 * @return the quantity, as a double precision value
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * get method for payment units
	 * @return the units in which payment is denominated
	 */
	public String getUnits() { return m_Units; }

	/**
	 * get method for payment asset type
	 * @return the type of asset which payment involves
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * get method for payee VS#
	 * @return the VS#, an alpha-numeric string
	 */
	public String getPayee() { return m_PayeeVS; }

	/**
	 * get method for serial number of payment voucher issued by VP to payee
	 * @return the serial number, an alpha-numeric string
	 */
	public String getSerialNumber() { return m_SerialNumber; }

	/**
	 * get method for the TTL value
	 * @return the payment expiration time, in milliseconds since the epoch
	 */
	public long getTTL() { return m_TTL; }

	/**
	 * get method for Issuer
	 * @return the Issuer
	 */
	public String getIssuer() { return m_Issuer; }

	/**
	 * determine whether the TTL has passed
	 * @return true if so, false otherwise
	 */
	public boolean isExpired() {
		if (!isInit()) {
			return false;
		}
		// see whether in the future
		Date now = new Date();
		if (m_TTL > now.getTime()) {
			return false;
		}
		return true;
	}

	/**
	 * get the hashtable of baggage fields
	 * @return the hash table (which may be empty)
	 */
	public Hashtable<String, String> getBaggageList() { return m_BaggageFields; }

	/**
	 * obtain the initialization status of the payment
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * get method for the display units which should be shown on various screens
	 * @return the units, which may be different from m_Units
	 */
	public String getDisplayUnits() {
		return XMLVoucher.getDisplayUnits(m_Asset);
	}

	/**
	 * method to construct a payment from an input XML representation
	 * @param xmlRep the XML representation to be parsed
	 * @return the payment object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLPayment getPaymentRep(String xmlRep)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element pmtElt = EncodingUtils.buildElement(xmlRep);
		if (pmtElt == null) {
			throw new DocumentException("payment parse error");
		}
		return XMLPayment.getPaymentRep(pmtElt);
	}

	/**
	 * method to construct a payment from a dom4j Element representation
	 * @param pmtElt the Element representing the payment
	 * @return the payment object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLPayment getPaymentRep(Element pmtElt)
		throws DocumentException
	{
		// prepare output
		XMLPayment payment = new XMLPayment();

		// make sure each element looks sensible
		Iterator eltIt = pmtElt.elementIterator();
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		while (eltIt.hasNext()) {

			// make sure this element is one of those we are expecting
			Element thisElt = (Element) eltIt.next();
			String eltName = thisElt.getName().toLowerCase();
			if (!M_PaymentElements.contains(eltName)) {
				throw new DocumentException("Illegal payment element: " + eltName);
			}

			// check for each item it can be and handle appropriately
			if (eltName.equals("amount")) {
				Number quan = null;
				try {
					quan = df.parse(thisElt.getText());
				}
				catch (ParseException pe) {
					Log.error("Payment amount parse error", pe);
				}
				if (quan != null && quan.doubleValue() > 0.0) {
					payment.m_Quantity = new Double(quan.doubleValue());
				}
				else {
					throw new DocumentException("Illegal amount value: "
														+ thisElt.getText());
				}
			}
			else if (eltName.equals("units")) {
				String unitNm = thisElt.getTextTrim().toUpperCase();
				payment.m_Units = unitNm;
			} 
			else if (eltName.equals("asset")) {
				String assetNm = thisElt.getTextTrim().toUpperCase();
				payment.m_Asset = assetNm;
			}
			else if (eltName.equals("issuer")) {
				String issuer = thisElt.getTextTrim();
				payment.m_Issuer = issuer;
			}
			else if (eltName.equals("payee_vs")) {
				payment.m_PayeeVS = thisElt.getTextTrim();
				if (payment.m_PayeeVS.isEmpty()) {
					throw new DocumentException("Empty payee safe number");
				}
			}
			else if (eltName.equals("serial")) {
				String serialNum = thisElt.getTextTrim();
				// TBD: perhaps some format or length scanning would be good here
				if (!serialNum.isEmpty()) {
					payment.m_SerialNumber = serialNum;
				}
				else {
					throw new DocumentException("Empty serial number");
				}
			}
			else if (eltName.equals("ttl")) {
				Long ttl = Long.parseLong(thisElt.getText());
				if (ttl.longValue() >= 0L) {
					payment.m_TTL = ttl.longValue();
				}
				else {
					Log.error("Bad TTL value in XMLPayment: \"" + pmtElt.asXML()
								+ "\"");
					throw new DocumentException("Illegal or missing TTL");
				}
			}
			else if (eltName.equals("baggage")) {
				// Add any and all fields, with any names, to the list.  Any dups
				// will be overwritten.

				Iterator bagIt = thisElt.elementIterator();
				while (bagIt.hasNext()) {
					Element bagElt = (Element) bagIt.next();

					// undo anti-XML confusion replacements (see buildSignedDetails)
					String fldVal = bagElt.getTextTrim();
					fldVal = fldVal.replaceAll(";lt;", "<");
					fldVal = fldVal.replaceAll(";gt;", ">");
					payment.m_BaggageFields.put(bagElt.getName(), fldVal);
				}
			}
			else {
				// shouldn't happen, but throw exception anyway
				throw new DocumentException("Unrecognized payment element, "
													+ eltName);
			}
		}

		// see if the payment looks complete now
		if (payment.m_Quantity.doubleValue() <= 0.0) {
			throw new DocumentException("Missing amount");
		}
		if (payment.m_Units.isEmpty()) {
			throw new DocumentException("Missing units");
		}
		if (payment.m_Asset.isEmpty()) {
			throw new DocumentException("Missing asset type");
		}
		if (payment.m_Issuer.isEmpty()) {
			throw new DocumentException("Missing issuer");
		}
		if (payment.m_PayeeVS.isEmpty()) {
			throw new DocumentException("Missing payee safe number");
		}
		if (payment.m_SerialNumber.isEmpty()) {
			throw new DocumentException("Missing serial number");
		}
		if (payment.m_TTL == 0L) {
			Log.error("Zero TTL encountered in parsed XMLPayment");
		}
		// NB: baggage fields are optional

		payment.m_Initialized = true;
		return payment;
	}

	/**
	 * method to emit the payment's values as XML
	 * @return valid XML representation of payment (assuming all values set),
	 * or the null string if not
	 */
	public String toXML() {
		if (!m_Initialized) {
			Log.error("Payment is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(1024);
		xml.append("<payment>");
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		xml.append("<amount>" + df.format(m_Quantity.doubleValue()) +"</amount>");
		xml.append("<units>" + m_Units + "</units>");
		xml.append("<asset>" + m_Asset + "</asset>");
		xml.append("<issuer>" + m_Issuer + "</issuer>");
		xml.append("<payee_vs>" + m_PayeeVS + "</payee_vs>");
		xml.append("<serial>" + m_SerialNumber + "</serial>");
		xml.append("<ttl>" + m_TTL + "</ttl>");
		if (!m_BaggageFields.isEmpty()) {
			xml.append("<baggage>");
			for (String bagFieldName : m_BaggageFields.keySet()) {
				// fix embedded newlines or HTML chars (< and >)
				String bagFieldValue
					= m_BaggageFields.get(bagFieldName).replace('\n', ' ');
				bagFieldValue = bagFieldValue.replaceAll("<", ";lt;");
				bagFieldValue = bagFieldValue.replaceAll(">", ";gt;");
				xml.append("<" + bagFieldName + ">");
				xml.append(bagFieldValue);
				xml.append("</" + bagFieldName + ">");
			}
			xml.append("</baggage>");
		}
		else {
			xml.append("<baggage/>");
		}
		xml.append("</payment>");
		return xml.toString();
	}

	/**
	 * method to set the quantity of the asset represented by the payment
	 * @param quan the amount of the payment, in the units given by m_Units
	 */
	public void setQuantity(double quan) {
		if (quan > 0.0) {
			m_Quantity = new Double(quan);
		}
	}

	/**
	 * method to set the units in which the quantity is expressed
	 * @param units the measurement units applied to m_Quantity
	 */
	public void setUnits(String units) {
		if (units != null && !units.isEmpty()) {
			m_Units = new String(units);
		}
	}

	/**
	 * method to set the type of asset backing the payment quantity
	 * @param asset the underlying asset type
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * method to set the issuer of the underlying asset
	 * @param iss the Issuer
	 */
	public void setIssuer(String iss) {
		if (iss != null && !iss.isEmpty()) {
			m_Issuer = new String(iss);
		}
	}

	/**
	 * method to set the serial number of the voucher conveying the payment
	 * @param snum the serial number to set
	 */
	public void setSerialNumber(String snum) {
		if (snum != null && !snum.isEmpty()) {
			m_SerialNumber = new String(snum);
		}
	}

	/**
	 * method to set the payee voucher safe number
	 * @param vsnum the VS# receiving the payment
	 */
	public void setPayee(String vsnum) {
		if (vsnum != null && !vsnum.isEmpty()) {
			m_PayeeVS = new String(vsnum);
		}
	}

	/**
	 * method to set the payment expiration time
	 * @param ttl the time in milliseconds since the epoch when payment expires
	 */
	public void setTTL(long ttl) {
		m_TTL = ttl;
	}

	/**
	 * method to add a baggage field and its associated value to the list
	 * @param name the hashtable key value (name of baggage field)
	 * @param val the value of the baggage field
	 */
	public void addBaggageField(String name, String val) {
		if (name == null || val == null || name.isEmpty() || val.isEmpty()) {
			return;
		}
		m_BaggageFields.put(name.trim(), val.trim());
	}

	/**
	 * method to finalize initialzation -- used if manually building a payment
	 * using the constructor, as opposed to starting with an XML representation
	 */
	public void setInit() { m_Initialized = true; }

	/**
	 * method to encrypt the object's XML representation, e.g. to place it into
	 * a message to the SDS
	 * @param pkey the public key to use to do the encryption (probably ours)
	 * @return the encrypted string, base64-encoded, or the empty string on error
	 */
	public String getEncryptedBase64(PublicKey pkey) {
		if (pkey == null) {
			return "";
		}
		
		// get the XML representation -- will come back "" if not initialized
		String xmlRep = toXML();
		if (xmlRep.isEmpty()) {
			return xmlRep;
		}

		// perform the encryption and base64 encoding -- will return null on error
		String encRep = EncodingUtils.makeBase64PubkeyEncStr(xmlRep, pkey);
		if (encRep == null) {
			return "";
		}
		return encRep;
	}

	/**
	 * method to support the Cloneable interface
	 * @return a new payment object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLPayment nPayment = (XMLPayment) super.clone();
			nPayment.m_Quantity = new Double(m_Quantity.doubleValue());
			nPayment.m_Units = new String(m_Units);
			nPayment.m_Asset = new String(m_Asset);
			nPayment.m_Issuer = new String(m_Issuer);
			nPayment.m_SerialNumber = new String(m_SerialNumber);
			nPayment.m_PayeeVS = new String(m_PayeeVS);
			nPayment.m_BaggageFields
				= new Hashtable<String, String>(m_BaggageFields);
			return nPayment;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a payment, cannot be equal
		if (!(obj instanceof XMLPayment)) {
			return false;
		}
		XMLPayment pmt = (XMLPayment) obj;

		// a payment is equal to this one if its serial number, payee, amount,
		// and TTL are the same

		if (m_Quantity.compareTo(pmt.getQuantity()) != 0) {
			return false;
		}
		if (!m_PayeeVS.equals(pmt.getPayee())) {
			return false;
		}
		if (!m_SerialNumber.equals(pmt.getSerialNumber())) {
			return false;
		}
		if (!m_Issuer.equals(pmt.getIssuer())) {
			return false;
		}
		if (m_TTL != pmt.getTTL()) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the elements referenced by equals() above
		hc += m_Quantity.hashCode();
		hc += m_PayeeVS.hashCode();
		hc += m_SerialNumber.hashCode();
		hc += m_Issuer.hashCode();
		Long ttl = new Long(m_TTL);
		hc += ttl.hashCode();
		return hc;
	}

	// END methods
}
