/*
 * last modified---
 * 	07-25-13/jet/ do not allow problem processing baggage to abort parsing
 * 	06-05-13/jet/ SilentVault client version
 * 	09-28-10/kbw/ escape baggage values for XML
 * 	03-12-10/kbw/ use TreeMap for baggage fields so that key order is constant
 * 	06-28-09/kbw/ store the encrypted baggage as well, use to validate VP sig
 * 	05-15-09/kbw/ validate payer signature on <baggage_fields/>
 * 	04-28-09/kbw/ add m_DHTid field
 * 	01-06-09/kbw/ new
 *
 * purpose---
 * 	implement a XML pending payment object
 *
 * notes---
 * 	This class is significantly different from the class of a similar name
 * 	in org.silentvault.client.XMLPayment, but closely related to the
 * 	corresponding class org.vouchersafe.publisher.XMLPendingPayment.
 */

package org.silentvault.client;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import java.util.Vector;
import java.util.Iterator;
import java.util.TreeMap;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * Class to instantiate a pending payment as an XML object.  The class knows
 * how to store, parse, and emit a payment in XML.  It is used when a voucher
 * payment has been retrieved from the DHT, decrypted, and the object needs to
 * be deciphered and examined so that a proper XMLReceipt can be generated
 * back to the payer.
 *
 * @author Kevin Wilkerson
 */
public final class XMLPendingPayment implements Cloneable {
	// BEGIN data members
	
	/**
	 * names of all the elements that are part of a payment
	 */
	private static Vector<String>	M_PaymentElements = getPaymentElements();

	/**
	 * the voucher issued to the payee (which they must pick up and validate)
	 */
	private XMLVoucher				m_Voucher;

	/**
	 * the voucher safe number of the payer (maker of the payment)
	 */
	private String						m_PayerVS;

	/**
	 * The vendor baggage fields, if any, associated with the payment.  These
	 * would normally be set by the payee, via a website form, email, or
	 * other communications.  They are stored here so that the merchant/payee
	 * can readily see who the payment was from or what it was for (such as an
	 * invoice number or shipping address).  As provided in a raw message, this
	 * field will be encrypted with the payee's pubkey, i.e.
	 * <baggage_fields>blob</baggage_fields>.
	 */
	private TreeMap<String, String>	m_BaggageFields;

	/**
	 * The vendor baggage fields, if any, associated with the payment.  These
	 * are in encrypted form, and are used only to validate the VP's signature
	 * on the record, which is naturally made with the baggage in encrypted form.
	 */
	private String						m_EncrBaggage;

	/**
	 * The TTL (time to live), representing the time in seconds since the epoch,
	 * at which the offered payment will expire and become eligible to be
	 * reclaimed by the payer.  This value defaults to one week from the time the
	 * voucher is issued to the payee, but may be set shorter by the payer.
	 * (This value was supplied to the DHT when the record is stored on it, and
	 * is now moot, since the record has already been retrieved prior to TTL.
	 * We include it because it is part of the signed data.)
	 */
	private long						m_TTL;

	/**
	 * the private key signature of the VP on the above elements
	 */
	private String						m_Signature;

	/**
	 * whether the payment has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	/**
	 * the record id at which this object is stored on the DHT, if known
	 */
	private Integer					m_DHTid;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a payment
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getPaymentElements() {
		Vector<String> vElts = new Vector<String>(6);
		vElts.add(new String("payment"));
		vElts.add(new String("voucher"));
		vElts.add(new String("payer_vs"));
		vElts.add(new String("baggage_fields"));
		vElts.add(new String("ttl"));
		vElts.add(new String("signature"));
		return vElts;
	}

	/**
	 * default constructor
	 */
	public XMLPendingPayment() {
		m_Voucher = new XMLVoucher();
		m_PayerVS = m_Signature = m_EncrBaggage = "";
		m_BaggageFields = new TreeMap<String, String>();
	}


	// get accessor methods
	/**
	 * get method for the payment voucher
	 * @return the voucher which was issued to the payee
	 */
	public XMLVoucher getVoucher() { return m_Voucher; }

	/**
	 * get method for payer VS#
	 * @return the VS#, an alpha-numeric string
	 */
	public String getPayer() { return m_PayerVS; }

	/**
	 * get the hashtable of baggage fields
	 * @return the hash table (which may be empty)
	 */
	public TreeMap<String, String> getBaggageList() { return m_BaggageFields; }

	/**
	 * get method for the TTL
	 * @return the time to live, in milliseconds since the epoch
	 */
	public long getTTL() { return m_TTL; }

	/**
	 * get method for VP's private key signature on the data
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the initialization status of the payment
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * obtain the record id at which this item is stored on the DHT
	 * @return the id, or null if not known
	 */
	public Integer getDHTid() { return m_DHTid; }


	/**
	 * method to construct a pending payment from an input XML representation
	 * (normally called by a VSC after retrieving/decrypting one off the DHT)
	 * @param xmlRep the XML representation to be parsed
	 * @param key the private key of the payee, needed to decrypt the baggage
	 * fields (if any)
	 * @param pksListener the PKSListener object needed to obtain a pubkey
	 * @return the payment object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLPendingPayment getPaymentRep(String xmlRep,
																 PrivateKey key,
																 PKSListener pksListener)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element pmtElt = EncodingUtils.buildElement(xmlRep);
		if (pmtElt == null) {
			throw new DocumentException("pending payment parse error");
		}
		return XMLPendingPayment.getPaymentRep(pmtElt, key, pksListener);
	}

	/**
	 * method to construct a pending payment from a dom4j Element representation
	 * @param pmtElt the Element representing the payment
	 * @param key the private key of the payee, needed to decrypt the baggage
	 * fields (if any)
	 * @param pksListener the PKSListener object needed to obtain a pubkey
	 * @return the payment object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLPendingPayment getPaymentRep(Element pmtElt,
																 PrivateKey key,
																 PKSListener pksListener)
		throws DocumentException
	{
		if (pmtElt == null || key == null || pksListener == null) {
			throw new DocumentException("Null input");
		}

		// prepare output
		XMLPendingPayment ppmt = new XMLPendingPayment();

		// must have a <voucher/>
		Element vouchElt = pmtElt.element("voucher");
		if (vouchElt == null) {
			throw new DocumentException("Missing <voucher/> element in "
												+ pmtElt.asXML());
		}
		// NB: voucher's signature is not validated here
		ppmt.m_Voucher = XMLVoucher.getVoucherRep(vouchElt);

		// must have a <payer_vs/>
		Element payerElt = pmtElt.element("payer_vs");
		if (payerElt == null) {
			throw new DocumentException("Missing <payer_vs/> element in "
												+ pmtElt.asXML());
		}
		ppmt.m_PayerVS = payerElt.getTextTrim();

		// must have <ttl/>
		Element ttlElt = pmtElt.element("ttl");
		if (ttlElt == null) {
			throw new DocumentException("Missing <ttl/> element in "
												+ pmtElt.asXML());
		}
		Long ttl = null;
		try {
			ttl = new Long(Long.parseLong(ttlElt.getText()));
			if (ttl.longValue() >= 0L) {
				ppmt.m_TTL = ttl.longValue();
			}
			else {
				Log.error("Bad TTL value in XMLPendingPayment: \"" + pmtElt.asXML()
							+ "\"");
				throw new DocumentException("Illegal or missing TTL");
			}
		}
		catch (NumberFormatException nfe) {
			throw new DocumentException("Bad TTL specified", nfe);
		}

		// must have a <baggage_fields/> element, though it can be blank
		Element bagElt = pmtElt.element("baggage_fields");
		if (bagElt == null) {
			throw new DocumentException("Missing <baggage_fields/> element in "
												+ pmtElt.asXML());
		}

		// must have <signature/>
		Element sigElt = pmtElt.element("signature");
		if (sigElt == null) {
			throw new DocumentException("Missing <signature/> element in "
												+ pmtElt.asXML());
		}
		ppmt.m_Signature = sigElt.getTextTrim();
		if (ppmt.m_Signature.isEmpty()) {
			throw new DocumentException("Empty <signature/> element in "
												+ pmtElt.asXML());
		}
		// NB: we don't attempt to validate the sig here (see ::validateSig())

		// Get the encrypted data in the field.  If there is none, nothing to do.
		ppmt.m_EncrBaggage = bagElt.getTextTrim();
		if (!ppmt.m_EncrBaggage.isEmpty()) {
			/* NB: any failure processing baggage must NOT throw exception or
			 * 	 return failure.  To do so would be to allow a malicious payer
			 * 	 to prevent us from being able to pick up the payment which was
			 * 	 signed by the VP.
			 */
			// decrypt this field with our private key
			String decBaggage
				= EncodingUtils.getStrFromBase64PubkeyEnc(ppmt.m_EncrBaggage, key);
			if (decBaggage == null || decBaggage.isEmpty()) {
				// payer sent us something which didn't decrypt
				Log.error("Error decrypting pmt baggage fields, ignoring baggage");
				ppmt.m_Initialized = true;
				return ppmt;
			}

			// now get this as an Element
			Element bagFields = EncodingUtils.buildElement("<baggage_fields>"
																			+ decBaggage
																			+ "</baggage_fields>");
			if (bagFields == null) {
				// payer sent us something which didn't parse
				Log.error("Unparsable payment baggage fields, ignoring");
				ppmt.m_Initialized = true;
				return ppmt;
			}

			/* Add any and all fields, with any names, to the list.  Any dups
			 * will be overwritten.  Capture the <signature/> separately.
			 */
			String payerSig = null;
			Iterator bagIt = bagFields.elementIterator();
			while (bagIt.hasNext()) {
				Element bagFld = (Element) bagIt.next();
				String fldNm = bagFld.getName();
				if (fldNm.equalsIgnoreCase("signature")) {
					payerSig = bagFld.getTextTrim();
				}
				else {
					// it's a regular baggage field
					String bagVal = bagFld.getText();
					// undo anti-XML confusion replacements
					bagVal.replaceAll(";lt;", "<");
					bagVal.replaceAll(";gt;", ">");
					ppmt.addBaggageField(bagFld.getName(), bagVal);
				}
			}
			if (payerSig == null || payerSig.isEmpty()) {
				// payer neglected to affix signature
				Log.error("Missing payer signature on baggage");
				ppmt.m_Initialized = true;
				return ppmt;
			}

			// obtain the payer's public key from PKS
			PublicKey payerKey = pksListener.obtainVSpubkey(ppmt.getPayer());
			if (payerKey == null) {
				Log.error("Unable to fetch public key for payer "
							+ ppmt.getPayer());
				ppmt.m_Initialized = true;
				return ppmt;
			}

			// get data to validate (everything before <signature>) and validate
			String signedBags
				= decBaggage.substring(0, decBaggage.indexOf("<signature"));
			Signature pSig = null;
			boolean ok = false;
			try {
				pSig = Signature.getInstance("SHA1withRSA");
				pSig.initVerify(payerKey);
				pSig.update(signedBags.getBytes());
				ok = pSig.verify(StringUtils.decodeBase64(payerSig));
			}
			catch (NoSuchAlgorithmException nsae) {
				// impossible
				Log.error("No such sig verify algorithm", nsae);
			}
			catch (InvalidKeyException ike) {
				Log.error("Invalid pubkey for signature validation", ike);
			}
			catch (SignatureException se) {
				Log.error("Unable to validate baggage signature", se);
			}
			if (!ok) {
				Log.error("Could not verify payer signature on baggage");
			}
		}

		ppmt.m_Initialized = true;
		return ppmt;
	}

	/**
	 * private utility method to build the data to be signed/validated
	 * (used for both signing and validation to guarantee consistency)
	 * @return the XML details to be signed or validated
	 */
	private String buildSignedDetails() {
		// must be initialized first
		if (!m_Initialized) {
			Log.error("Cannot build signature details on uninitialized details");
			return "";
		}

		// build string, equal to all data fields
		StringBuilder sigData = new StringBuilder(512);
		sigData.append(m_Voucher.toXML());
		sigData.append("<payer_vs>" + m_PayerVS + "</payer_vs>");
		sigData.append("<baggage_fields>" + m_EncrBaggage);
		sigData.append("</baggage_fields>");
		sigData.append("<ttl>" + m_TTL + "</ttl>");
		return sigData.toString();
	}

	/**
	 * method to validate the signature on the data in the object
	 * @param vpPub the public key of the VP, used to validate signature
	 * @return true if all is well, false if a problem is encountered
	 */
	public boolean validateSig(PublicKey vpPub) {

		// get data to validate
		String sigData = buildSignedDetails();
		Signature dSig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initVerify(vpPub);
			dSig.update(sigData.getBytes());
			ok = dSig.verify(StringUtils.decodeBase64(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("No such sig verify algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate details signature", se);
		}
		if (!ok) {
			Log.error("Could not verify signature on payment");
			return false;
		}
		return true;
	}

	/**
	 * method to emit the payment's values as XML
	 * @return valid XML representation of payment (assuming all values set),
	 * or the null string if not
	 */
	public String toXML() {
		if (!m_Initialized) {
			Log.error("Payment is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(1024);
		xml.append("<payment>");
		xml.append(buildSignedDetails());
		xml.append("<signature>" + m_Signature + "</signature>");
		xml.append("</payment>");
		return xml.toString();
	}


	// set accessor methods
	/**
	 * method to set the voucher accomplishing the payment, which the payee must
	 * pick up and validate (to be replaced by a different voucher)
	 * @param voucher the voucher (must have valid signature)
	 * @param vpKey the public key of the VP, to validate the voucher signature
	 * @return true on success, false if voucher doesn't check out
	 */
	public boolean setVoucher(XMLVoucher voucher, PublicKey vpKey) {
		if (voucher == null || vpKey == null) {
			return false;
		}
		if (voucher.validateSig(vpKey)) {
			m_Voucher = voucher;
			return true;
		}
		Log.error("Bad voucher signature; cannot use in pending payment");
		return false;
	}

	/**
	 * method to set the payer voucher safe number
	 * @param vsnum the VS# issuing the payment
	 */
	public void setPayer(String vsnum) {
		if (vsnum != null && !vsnum.isEmpty()) {
			m_PayerVS = new String(vsnum);
		}
	}

	/**
	 * method to set the TTL (time to live)
	 * @param ttl the time in seconds since the epoch when payment expires
	 */
	public void setTTL(long ttl) {
		m_TTL = ttl;
	}

	/**
	 * method to add a baggage field and its associated value to the list
	 * @param name the hashtable key value (name of baggage field)
	 * @param val the value of the baggage field
	 */
	public void addBaggageField(String name, String val) {
		if (name == null || val == null || name.isEmpty() || val.isEmpty()) {
			return;
		}
		m_BaggageFields.put(name.trim(), val.trim());
	}

	/**
	 * method to finalize initialzation -- used if manually building a payment
	 * using the constructor, as opposed to starting with an XML representation
	 */
	public void setInit() { m_Initialized = true; }

	/**
	 * method to set the id at which this record is stored on the DHT
	 * @param id the id
	 */
	public void setDHTid(Integer id) {
		if (id != null) {
			m_DHTid = id;
		}
	}


	/**
	 * method to support the Cloneable interface
	 * @return a new payment object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLPendingPayment nPayment = (XMLPendingPayment) super.clone();
			nPayment.m_Voucher = (XMLVoucher) m_Voucher.clone();
			nPayment.m_PayerVS = new String(m_PayerVS);
			nPayment.m_Signature = new String(m_Signature);
			nPayment.m_BaggageFields
				= new TreeMap<String, String>(m_BaggageFields);
			return nPayment;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a pending payment, cannot be equal
		if (!(obj instanceof XMLPendingPayment)) {
			return false;
		}
		XMLPendingPayment pmt = (XMLPendingPayment) obj;

		// a payment is equal to this one if its payer, amount (from voucher)
		// and TTL are the same

		if (m_Voucher.getQuantity() != pmt.getVoucher().getQuantity()) {
			return false;
		}
		if (!m_PayerVS.equals(pmt.getPayer())) {
			return false;
		}
		if (m_TTL != pmt.getTTL()) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the elements referenced by equals() above
		Double quan = new Double(m_Voucher.getQuantity());
		hc += quan.hashCode();
		hc += m_PayerVS.hashCode();
		Long ttl = new Long(m_TTL);
		hc += ttl.hashCode();
		return hc;
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_BaggageFields != null && !m_BaggageFields.isEmpty()) {
				m_BaggageFields.clear();
			}
			m_Voucher = null;
		} finally {
			super.finalize();
		}
	}

	// END methods
}
