/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	01-13-12/kbw/ add emitAsCSV()
 * 	02-22-11/kbw/ add getDisplayUnits(); remove use of Double.parseDouble()
 * 	12-29-10/kbw/ force Locale to US so we always parse properly
 * 	09-28-10/kbw/ remove embedded newlines and angle brackets before signature
 * 	07-10-10/kbw/ remove spurious restrictions on asset and unit types
 * 	04-22-10/kbw/ add payer_vs field, to allow storing our own receipts
 * 	03-12-10/kbw/ use TreeMap for baggage fields so that key order is constant
 * 	08-17-09/kbw/ VS# must be mixed case
 * 	06-29-09/kbw/ add m_VoucherSerial field
 * 	06-28-09/kbw/ remove separate encryption of <details/> element
 * 	05-10-09/kbw/ break out validateSig() as separate step method
 * 	04-28-09/kbw/ add m_DHTid field
 * 	01-05-09/kbw/ new
 *
 * purpose---
 * 	implement a XML receipt object
 *
 */

package org.silentvault.client;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import java.util.Vector;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Date;
import java.util.Locale;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Class to instantiate a receipt as an XML object.  The class knows how to
 * store, parse, and emit a receipt in XML.  It is used in three situations:
 * 1) where a VSC has successfully picked up a payment addressed to it from
 * the DHT, verified it with the VP, and now wishes to generate a receipt for
 * the payer; 2) in the same situation as 1 but to generate a copy of the payer
 * receipt for the payee which the latter wishes to store in its folder on the
 * SDS; and 3) when a VSC has picked up a receipt addressed to it off the DHT,
 * and elects to create a permanent record of it in its folder on the SDS.
 *
 * @author Kevin Wilkerson
 */
public final class XMLReceipt implements Cloneable {
	// BEGIN data members
	
	/**
	 * precision to use on quantity
	 */
	public static final int			M_Precision = 4;

	/**
	 * names of all the elements that are part of a receipt
	 */
	private static Vector<String>	M_ReceiptElements = getReceiptElements();

	/**
	 * the <details/> element
	 */
	private Element					m_Details;


	// sub-elements of <details/>
	/**
	 * the quantity of the underlying asset represented by the receipt
	 */
	private Double						m_Quantity;

	/**
	 * the measurement units of the receipt (always base units for the voucher)
	 */
	private String						m_Units;

	/**
	 * the underlying asset backing the receipt
	 */
	private String						m_Asset;

	/**
	 * the voucher safe number of the payer (recipient or issuer of the receipt)
	 */
	private String						m_PayerVS;

	/**
	 * the voucher safe number of the payee (issuer or recipient of the receipt)
	 */
	private String						m_PayeeVS;

	/**
	 * the serial number of the voucher issued to the payee with the payment.
	 * (this is stored here so that we can match this receipt against the
	 * corresponding pending payment record in the safe)
	 */
	private String						m_VoucherSerial;

	/**
	 * The vendor baggage fields, if any, associated with the receipt.  These
	 * would normally be acquired from the payee, via a website form, email, or
	 * other communications.  They are stored here so that the VS owner can
	 * readily see who the receipt was from or what it was for.
	 */
	private TreeMap<String, String>	m_BaggageFields;

	/**
	 * the time, in seconds since the epoch, when this receipt was issued
	 */
	private long						m_Timestamp;

	/**
	 * the private key signature of the payee (receipt issuer) on the elements
	 * of the <details/> block
	 */
	private String						m_Signature;

	/**
	 * whether the receipt has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	/**
	 * the ID at which this record was stored on the DHT, if known
	 */
	private Integer					m_DHTid;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a receipt
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getReceiptElements() {
		Vector<String> vElts = new Vector<String>(11);
		vElts.add(new String("receipt"));
		vElts.add(new String("details"));
		vElts.add(new String("amount"));
		vElts.add(new String("units"));
		vElts.add(new String("asset"));
		vElts.add(new String("payer_vs"));
		vElts.add(new String("payee_vs"));
		vElts.add(new String("serial"));
		vElts.add(new String("timestamp"));
		vElts.add(new String("baggage"));
		vElts.add(new String("signature"));
		return vElts;
	}

	/**
	 * default constructor
	 */
	public XMLReceipt() {
		m_Quantity = new Double(0.0);
		m_Units = m_Asset = "";
		m_PayeeVS = m_PayerVS = m_Signature = m_VoucherSerial = "";
		m_BaggageFields = new TreeMap<String, String>();
	}


	// get accessor methods
	/**
	 * get method for receipt amount value
	 * @return the quantity, as a double precision value
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * get method for receipt units
	 * @return the units in which receipt is denominated
	 */
	public String getUnits() { return m_Units; }

	/**
	 * get method for receipt asset type
	 * @return the type of asset which receipt involves
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * get method for payer VS#
	 * @return the VS#, an alpha-numeric string
	 */
	public String getPayer() { return m_PayerVS; }

	/**
	 * get method for payee VS#
	 * @return the VS#, an alpha-numeric string
	 */
	public String getPayee() { return m_PayeeVS; }

	/**
	 * get method for date/time of issuance
	 * @return the timestamp, in seconds since the epoch
	 */
	public long getTimestamp() { return m_Timestamp; }

	/**
	 * get method for payee's voucher serial number
	 * @return the serial number
	 */
	public String getVoucherSerial() { return m_VoucherSerial; }

	/**
	 * get the hashtable of baggage fields
	 * @return the hash table (which may be empty)
	 */
	public TreeMap<String, String> getBaggageList() { return m_BaggageFields; }

	/**
	 * get method for issuer's private key signature on the <details/> block
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the initialization status of the receipt
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * obtain the DHT storage id
	 * @return the id, if recorded, or null if not known
	 */
	public Integer getDHTid() { return m_DHTid; }

	/**
	 * get method for the display units which should be shown on various screens
	 * @return the units, which may be different from m_Units
	 */
	public String getDisplayUnits() {
		return XMLVoucher.getDisplayUnits(m_Asset);
	}


	/**
	 * method to construct a receipt from an input XML representation
	 * @param xmlRep the XML representation to be parsed
	 * @return the receipt object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLReceipt getReceiptRep(String xmlRep)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element recElt = EncodingUtils.buildElement(xmlRep);
		if (recElt == null) {
			throw new DocumentException("receipt parse error");
		}
		return XMLReceipt.getReceiptRep(recElt);
	}

	/**
	 * method to construct a receipt from a dom4j Element representation
	 * @param recElt the Element representing the receipt
	 * @return the receipt object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLReceipt getReceiptRep(Element recElt)
		throws DocumentException
	{
		// prepare output
		XMLReceipt receipt = new XMLReceipt();

		// We expect <details/> and <signature/> under the root element.

		Element detailsElt = recElt.element("details");
		if (detailsElt == null) {
			throw new DocumentException("Missing <details/> element in "
												+ recElt.asXML());
		}
		receipt.m_Details = detailsElt;
		if (!receipt.parseDetails()) {
			throw new DocumentException("Unparseable <details/> element in "
												+ recElt.asXML());
		}

		Element sigElt = recElt.element("signature");
		if (sigElt == null) {
			throw new DocumentException("Missing <signature/> element in "
												+ recElt.asXML());
		}
		receipt.m_Signature = sigElt.getTextTrim();
		if (receipt.m_Signature.isEmpty()) {
			throw new DocumentException("Empty <signature/> element in "
												+ recElt.asXML());
		}
		// NB: we don't attempt to validate the sig here (see ::validateSig())

		return receipt;
	}

	/**
	 * private utility method to build the data to be signed/validated
	 * (used for both signing and validation to guarantee consistency)
	 * @return the XML details to be signature validated
	 */
	private String buildSignedDetails() {
		// must be initialized first
		if (!m_Initialized) {
			Log.error("Cannot build signature details on uninitialized details");
			return "";
		}

		// build string, equal to the signed <details/> block
		StringBuilder sigData = new StringBuilder(1024);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		sigData.append("<amount>" + df.format(m_Quantity.doubleValue()));
		sigData.append("</amount>");
		sigData.append("<units>" + m_Units + "</units>");
		sigData.append("<asset>" + m_Asset + "</asset>");
		sigData.append("<payer_vs>" + m_PayerVS + "</payer_vs>");
		sigData.append("<payee_vs>" + m_PayeeVS + "</payee_vs>");
		sigData.append("<serial>" + m_VoucherSerial + "</serial>");
		sigData.append("<timestamp>" + m_Timestamp + "</timestamp>");
		if (!m_BaggageFields.isEmpty()) {
			sigData.append("<baggage>");
			for (String bagFieldName : m_BaggageFields.keySet()) {
				// fix embedded newlines, also < and > to avoid confusing XML parser
				String bagFieldValue
					= m_BaggageFields.get(bagFieldName).replace('\n', ' ');
				bagFieldValue = bagFieldValue.replaceAll("<", ";lt;");
				bagFieldValue = bagFieldValue.replaceAll(">", ";gt;");
				sigData.append("<" + bagFieldName + ">");
				sigData.append(bagFieldValue);
				sigData.append("</" + bagFieldName + ">");
			}
			sigData.append("</baggage>");
		}
		else {
			sigData.append("<baggage/>");
		}
		//Log.debug("Receipt details block text for sig/verify is: " + sigData);
		return sigData.toString();
	}

	/**
	 * method to decode, decrypt, and parse the <details/> element and validate
	 * the signature on it
	 * @return true if all is well, false if a problem is encountered
	 */
	private boolean parseDetails() {
		// check that this call is appropriate
		if (m_Details == null) {
			Log.error("<details/> unsuitable for parsing");
			return false;
		}

		// obtain constituent parts
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		Iterator eltIt = m_Details.elementIterator();
		while (eltIt.hasNext()) {

			// make sure this element is one of those we are expecting
			Element thisElt = (Element) eltIt.next();
			String eltName = thisElt.getName().toLowerCase();
			if (!M_ReceiptElements.contains(eltName)) {
				Log.error("Illegal receipt element: " + eltName);
				return false;
			}

			// check for each item it can be and handle appropriately
			if (eltName.equals("amount")) {
				Number quan = null;
				try {
					quan = df.parse(thisElt.getText());
				}
				catch (ParseException pe) {
					Log.error("Receipt amount parse error", pe);
				}
				if (quan != null && quan.doubleValue() > 0.0) {
					m_Quantity = new Double(quan.doubleValue());
				}
				else {
					Log.error("Illegal amount value: " + thisElt.getText());
					return false;
				}
			}
			else if (eltName.equals("units")) {
				String unitNm = thisElt.getTextTrim().toUpperCase();
				m_Units = unitNm;
			} 
			else if (eltName.equals("asset")) {
				String assetNm = thisElt.getTextTrim().toUpperCase();
				m_Asset = assetNm;
			}
			else if (eltName.equals("payer_vs")) {
				m_PayerVS = thisElt.getTextTrim();
				if (m_PayerVS.isEmpty()) {
					Log.error("Empty receipt payer safe number");
					return false;
				}
			}
			else if (eltName.equals("payee_vs")) {
				m_PayeeVS = thisElt.getTextTrim();
				if (m_PayeeVS.isEmpty()) {
					Log.error("Empty receipt payee safe number");
					return false;
				}
			}
			else if (eltName.equals("serial")) {
				m_VoucherSerial = thisElt.getTextTrim();
				if (m_VoucherSerial.isEmpty()) {
					Log.error("Empty voucher serial number");
					return false;
				}
			}
			else if (eltName.equals("timestamp")) {
				Long ts = null;
				try {
					ts = Long.parseLong(thisElt.getText());
				}
				catch (NumberFormatException nfe) {
					Log.error("Invalid creation timestamp", nfe);
					return false;
				}
				m_Timestamp = ts.longValue();
				Date now = new Date();
				if (m_Timestamp > (now.getTime() / 1000L)) {
					Log.warning("Receipt creation date in future");
					/* We have to let this go through, because it could be the
					 * result of a malicious client's action trying to deny us
					 * our receipt by setting their local clock wrong.
					 */
				}
			}
			else if (eltName.equals("baggage")) {
				// Add any and all fields, with any names, to the list.  Any dups
				// will be overwritten.

				Iterator bagIt = thisElt.elementIterator();
				while (bagIt.hasNext()) {
					Element bagElt = (Element) bagIt.next();

					// undo anti-XML confusion replacements (see buildSignedDetails)
					String fldVal = bagElt.getText();
					fldVal = fldVal.replaceAll(";lt;", "<");
					fldVal = fldVal.replaceAll(";gt;", ">");
					m_BaggageFields.put(bagElt.getName(), fldVal);
				}
			}
			else {
				// shouldn't happen, but throw exception anyway
				Log.error("Unrecognized receipt element, " + eltName);
				return false;
			}
		}

		// see if the receipt looks complete now
		boolean ok = true;
		if (m_Quantity.doubleValue() <= 0.0) {
			Log.error("Missing amount");
			ok = false;
		}
		if (m_Units.isEmpty()) {
			Log.error("Missing units");
			ok = false;
		}
		if (m_Asset.isEmpty()) {
			Log.error("Missing asset type");
			ok = false;
		}
		// NB: payer can be missing
		if (m_PayeeVS.isEmpty()) {
			Log.error("Missing payee safe number");
			ok = false;
		}
		if (m_Timestamp <= 0L) {
			Log.error("Missing receipt creation timestamp");
			ok = false;
		}
		if (m_VoucherSerial.isEmpty()) {
			Log.error("Missing payee voucher serial number");
			ok = false;
		}
		// NB: baggage fields are optional
		if (!ok) {
			return false;
		}
		m_Initialized = true;
		return ok;
	}

	/**
	 * validate the signature on the details
	 * @param pubKey the public key of the signer
	 */
	public boolean validateSig(PublicKey pubKey) {
		if (!m_Initialized || m_Details == null) {
			Log.error("Can't validate receipt signature before decrypt/parse");
			return false;
		}

		// get data to verify
		String sigData = buildSignedDetails();
		Signature dSig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initVerify(pubKey);
			dSig.update(sigData.getBytes());
			ok = dSig.verify(StringUtils.decodeBase64(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig verify algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for details signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate details signature", se);
		}
		if (!ok) {
			Log.error("Could not verify signature on receipt <details/> block");
		}
		return ok;
	}

	/**
	 * method to emit the receipt's values as XML
	 * @return valid XML representation of receipt (assuming all values set),
	 * or the null string if not
	 */
	public String toXML() {
		if (!m_Initialized) {
			Log.error("Receipt is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(1024);
		xml.append("<receipt>");
		xml.append("<details>");
		xml.append(buildSignedDetails());
		xml.append("</details>");
		xml.append("<signature>" + m_Signature + "</signature>");
		xml.append("</receipt>");
		return xml.toString();
	}

	/**
	 * method to emit the receipt's values in CSV format
	 * @return valid comma-separated value representation of receipt, or null
	 * string if improperly initialized
	 */
	public String emitAsCSV() {
		if (!m_Initialized) {
			Log.error("Receipt is not properly initialized!");
			return "";
		}

		StringBuilder csv = new StringBuilder(1024);
		// NB: Locale.US prevents commas from appearing in the amounts
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		csv.append(df.format(m_Quantity.doubleValue()) + ",");
		csv.append(m_Units + ",");
		csv.append(m_Asset + ",");
		csv.append(m_PayerVS + ",");
		csv.append(m_PayeeVS + ",");
		csv.append(m_VoucherSerial + ",");
		csv.append(m_Timestamp);
		if (!m_BaggageFields.isEmpty()) {
			for (String bagFieldName : m_BaggageFields.keySet()) {
				csv.append(",");
				// fix embedded newlines
				String bagFieldValue
					= m_BaggageFields.get(bagFieldName).replace('\n', ' ');
				String fldNm = bagFieldName;
				if (!bagFieldName.equals("Message")) {
					fldNm = bagFieldName.substring(1);
				}
				csv.append("\"" + fldNm + ":");
				csv.append(bagFieldValue);
				csv.append("\"");
			}
		}
		return csv.toString();
	}


	// set methods
	/**
	 * method to set the quantity of the asset represented by the receipt
	 * @param quan the amount of the receipt, in the units given by m_Units
	 */
	public void setQuantity(double quan) {
		if (quan > 0.0) {
			m_Quantity = new Double(quan);
		}
	}

	/**
	 * method to set the units in which the quantity is expressed
	 * @param units the measurement units applied to m_Quantity
	 */
	public void setUnits(String units) {
		if (units != null && !units.isEmpty()) {
			m_Units = new String(units);
		}
	}

	/**
	 * method to set the type of asset backing the receipt quantity
	 * @param asset the underlying asset type
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * method to set the payer voucher safe number
	 * @param vsnum the VS# of the payer
	 */
	public void setPayer(String vsnum) {
		if (vsnum != null && !vsnum.isEmpty()) {
			m_PayerVS = new String(vsnum);
		}
	}

	/**
	 * method to set the payee voucher safe number
	 * @param vsnum the VS# of the payee
	 */
	public void setPayee(String vsnum) {
		if (vsnum != null && !vsnum.isEmpty()) {
			m_PayeeVS = new String(vsnum);
		}
	}

	/**
	 * method to set the payee voucher serial number
	 * @param serial the serial number of the voucher issued to the payee
	 */
	public void setVoucherSerial(String serial) {
		if (serial != null && !serial.isEmpty()) {
			m_VoucherSerial = new String(serial);
		}
	}

	/**
	 * method to set the creation timestamp
	 * @param tstamp the time in seconds since the epoch when receipt generated
	 */
	public void setTimestamp(long tstamp) { m_Timestamp = tstamp; }

	/**
	 * method to add a baggage field and its associated value to the list
	 * @param name the hashtable key value (name of baggage field)
	 * @param val the value of the baggage field
	 */
	public void addBaggageField(String name, String val) {
		if (name == null || val == null || name.isEmpty() || val.isEmpty()) {
			return;
		}
		m_BaggageFields.put(name.trim(), val.trim());
	}

	/**
	 * method to finalize initialzation -- used if manually building a receipt
	 * using the constructor, as opposed to starting with an XML representation
	 */
	public void setInit() { m_Initialized = true; }

	/**
	 * method to set the DHT storage id
	 * @param id the id at which this record is stored on the DHT
	 */
	public void setDHTid(Integer id) {
		if (id != null) {
			m_DHTid = id;
		}
	}


	/**
	 * method to sign the <details/> block of an initialized receipt
	 * (when building a receipt for a payer, must be called before
	 * encryptDetails() is called)
	 * @param sigKey the VS private key used to make the signature
	 * @return true on success, false on error
	 */
	public boolean signDetails(PrivateKey sigKey) {
		if (!m_Initialized) {
			Log.error("Cannot sign uninitialized <details/>");
			return false;
		}
		if (sigKey == null) {
			Log.error("Missing details signature key");
			return false;
		}

		// get the data to be signed
		String details = buildSignedDetails();
		Signature dSig = null;
		byte[] sig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initSign(sigKey);
			dSig.update(details.getBytes());
			sig = dSig.sign();
			ok = true;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such signature algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for details signature", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to create details signature", se);
		}
		if (!ok) {
			Log.error("Could not place signature on receipt <details/> block");
			return false;
		}
		m_Signature = StringUtils.encodeBase64(sig, false);
		return true;
	}

	/**
	 * method to encrypt the object's XML representation, e.g. to place it into
	 * a message to the SDS.  This method encrypts the whole object, as opposed
	 * to encryptDetails() which encrypts only the signed <details/> block.
	 * @param pkey the public key to use to do the encryption (probably ours)
	 * @return the encrypted string, base64-encoded, or the empty string on error
	 */
	public String getEncryptedBase64(PublicKey pkey) {
		if (pkey == null) {
			return "";
		}
		
		// get the XML representation -- will come back "" if not initialized
		String xmlRep = toXML();
		if (xmlRep.isEmpty()) {
			return xmlRep;
		}

		// perform the encryption and base64 encoding -- will return null on error
		String encRep = EncodingUtils.makeBase64PubkeyEncStr(xmlRep, pkey);
		if (encRep == null) {
			return "";
		}
		return encRep;
	}

	/**
	 * method to support the Cloneable interface
	 * @return a new receipt object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLReceipt nReceipt = (XMLReceipt) super.clone();
			nReceipt.m_Details = (Element) m_Details.clone();
			nReceipt.m_Quantity = new Double(m_Quantity.doubleValue());
			nReceipt.m_Units = new String(m_Units);
			nReceipt.m_Asset = new String(m_Asset);
			nReceipt.m_PayeeVS = new String(m_PayeeVS);
			nReceipt.m_VoucherSerial = new String(m_VoucherSerial);
			nReceipt.m_Signature = new String(m_Signature);
			nReceipt.m_BaggageFields
				= new TreeMap<String, String>(m_BaggageFields);
			return nReceipt;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a receipt, cannot be equal
		if (!(obj instanceof XMLReceipt)) {
			return false;
		}
		XMLReceipt rec = (XMLReceipt) obj;

		// a receipt is equal to this one if its payee, amount, serial number,
		// and timestamp are the same

		if (m_Quantity.compareTo(rec.getQuantity()) != 0) {
			return false;
		}
		if (!m_PayeeVS.equals(rec.getPayee())) {
			return false;
		}
		if (!m_VoucherSerial.equals(rec.getVoucherSerial())) {
			return false;
		}
		if (m_Timestamp != rec.getTimestamp()) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the elements referenced by equals() above
		hc += m_Quantity.hashCode();
		hc += m_PayerVS.hashCode();
		hc += m_PayeeVS.hashCode();
		hc += m_VoucherSerial.hashCode();
		Long ts = new Long(m_Timestamp);
		hc += ts.hashCode();
		return hc;
	}

	// END methods
}
