/*
 * last modified---
 * 	07-09-13/jet/ new
 *
 * purpose---
 * 	provide a TableModel for the offer JTable display in the OffersPane
 */

package org.silentvault.client.ui;

import javax.swing.table.AbstractTableModel;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;

import org.silentvault.client.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of trade offers in a scrolling window.
 *
 * @author Justin Turrell
 */
public final class OfferTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names
	 */
	private final String[]	M_ColumnNames = { "",
															"ID",
															"Offered",
															"Amount",
															"Wanted",
															"Ratio",
															"Status",
															"Expiration" };
	
	// selection modes
	/**
	 * selection mode constant for when offers are not selectable
	 */
	public static final int	M_SelectNone = 0;
	/**
	 * selection mode constant for when offers are multi-selectable (unused)
	 */
	public static final int	M_SelectMulti = 1;
	/**
	 * selection mode constant for when offers are single-selectable
	 */
	public static final int	M_SelectSingle = 2;

	/**
	 * the list of offers we're representing (our data)
	 */
	private ArrayList<TradeOffer>	m_OfferList;

	/**
	 * the owning plugin/application
	 */
	private WalletClient		m_Plugin;

	/**
	 * the list of checkboxes which correspond to the offers and are shown
	 * in the first column
	 */
	private Hashtable<TradeOffer, JCheckBox>	m_SelectBoxes;

	/**
	 * the list of radio buttons which correspond to the offers and are shown
	 * in the first column
	 */
	private Hashtable<TradeOffer, JRadioButton>	m_SelectButtons;

	/**
	 * the selection mode (one of M_SelectNone, M_SelectMulti, M_SelectSingle)
	 */
	private int					m_SelectMode;

	/**
	 * the button group for the radio buttons, to ensure only one at a time can
	 * be checked
	 */
	private ButtonGroup		m_ButtonGroup;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param offers the list of offers which we are to represent
	 * @param plugin the plugin that owns our display
	 */
	public OfferTableModel(ArrayList<TradeOffer> offers, WalletClient plugin) {
		super();
		m_OfferList = offers;
		m_Plugin = plugin;
		m_SelectBoxes = new Hashtable<TradeOffer, JCheckBox>();
		m_SelectButtons = new Hashtable<TradeOffer, JRadioButton>();
		m_ButtonGroup = new ButtonGroup();
		m_SelectMode = M_SelectNone;

		// make a checkbox and a button for each offer
		for (TradeOffer vvv : m_OfferList) {
			JCheckBox cb = new JCheckBox();
			JRadioButton rb = new JRadioButton();
			cb.setEnabled(true);
			rb.setEnabled(true);
			rb.setRolloverEnabled(true);
			m_SelectBoxes.put(vvv, cb);
			m_SelectButtons.put(vvv, rb);
			m_ButtonGroup.add(rb);
		}
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of offers)
	 */
	public int getRowCount() {
		return m_OfferList.size();
	}

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (offer) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_OfferList.size() || row < 0 || col < 0) {
			// request input was invalid
			return null;
		}
		TradeOffer offer = m_OfferList.get(row);

		// this should never happen, but...
		if (offer == null) {
			return null;
		}

		Object ret = null;
		switch (col) {
			case 0:	// select checkbox or radio button
				if (m_SelectMode == M_SelectMulti) {
					ret = m_SelectBoxes.get(offer);
				}
				else if (m_SelectMode == M_SelectSingle) {
					ret = m_SelectButtons.get(offer);
				}
				// else: null is fine
				break;

			case 1:	// offer ID number
				Integer oId = new Integer(offer.getOfferId());
				ret = oId;
				break;

			case 2:	// offered asset
				ret = offer.getOffered();
				break;
			
			case 3:	// amount of offered asset
				Double amt = new Double(offer.getQuantity());
				ret = amt;
				break;

			case 4:	// wanted asset (list)
				ret = offer.getWanted();
				break;

			case 5:	// ratio of OFFERED to WANTED (price)
				Float ratio = offer.getPrice();
				ret = ratio;
				break;

			case 6:	// offer status
				ret = offer.getStatus();
				break;

			case 7:	// expiration date
				ret = offer.getExpiration();
				break;

			default:	// impossible
				Log.error("Impossible offer column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	// other get methods
	/**
	 * get the whole offer list
	 * @return a copy of the list (to avoid side-effects)
	 */
	public ArrayList<TradeOffer> getOfferList() {
		return new ArrayList<TradeOffer>(m_OfferList);
	}

	/**
	 * get the offer at a specific row
	 * @param row the row whose offer we're interested in
	 * @return the offer (null if not found or out of range)
	 */
	public TradeOffer getOfferAtRow(int row) {
		if (row >= m_OfferList.size() || row < 0) {
			return null;
		}
		return m_OfferList.get(row);
	}

	/**
	 * return all selected offers
	 * @return all offers whose associated box has been checked
	 */
	public ArrayList<TradeOffer> getSelectedOffers() {
		ArrayList<TradeOffer> retList
			= new ArrayList<TradeOffer>(m_OfferList.size());
		for (TradeOffer offer : m_OfferList) {
			if (m_SelectMode == M_SelectMulti) {
				JCheckBox chk = m_SelectBoxes.get(offer);
				if (chk != null && chk.isSelected()) {
					retList.add(offer);
				}
			}
			else if (m_SelectMode == M_SelectSingle) {
				JRadioButton rad = m_SelectButtons.get(offer);
				if (rad != null && rad.isSelected()) {
					retList.add(offer);
				}
			}
		}
		return retList;
	}

	/**
	 * obtain the current selection mode
	 * @return the mode (one of M_SelectNone, M_SelectMulti, M_SelectSingle)
	 */
	public int getSelectMode() { return m_SelectMode; }


	// write operations
	/**
	 * insert a new offer into the list
	 * @param offer the offer to be added
	 */
	public void addOffer(TradeOffer offer) {
		if (offer != null) {
			if (!m_OfferList.contains(offer)) {
				m_OfferList.add(offer);
				// this was added
				JCheckBox cb = new JCheckBox();
				JRadioButton rb = new JRadioButton();
				cb.setEnabled(true);
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectBoxes.put(offer, cb);
				m_SelectButtons.put(offer, rb);
				m_ButtonGroup.add(rb);

				// notify any listeners
				fireTableDataChanged();
			}
		}
	}

	/**
	 * destructively merge a list of offers into the list
	 * @param offerList the offers we need to merge
	 */
	public void mergeOffers(List<TradeOffer> offerList) {
		m_OfferList.clear();
		if (offerList == null || offerList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_OfferList.addAll(offerList);

		// rebuild select controls
		for (TradeOffer vvv : m_OfferList) {
			if (m_SelectBoxes.get(vvv) == null) {
				JCheckBox cb = new JCheckBox();
				JRadioButton rb = new JRadioButton();
				cb.setEnabled(true);
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectBoxes.put(vvv, cb);
				m_SelectButtons.put(vvv, rb);
				m_ButtonGroup.add(rb);
			}
			else {
				// it was already there, make sure buttons are enabled
				JCheckBox cb = m_SelectBoxes.get(vvv);
				cb.setEnabled(true);
				JRadioButton rb = m_SelectButtons.get(vvv);
				rb.setEnabled(true);
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * remove a list of offers
	 * @param offerList the offers we're to remove
	 */
	public void removeOffers(ArrayList<TradeOffer> offerList) {
		if (offerList != null && !offerList.isEmpty()) {
			for (TradeOffer offer : offerList) {
				m_OfferList.remove(offer);
				m_SelectBoxes.remove(offer);
				JRadioButton rb = m_SelectButtons.get(offer);
				m_SelectButtons.remove(offer);
				m_ButtonGroup.remove(rb);
			}

			// notify listeners
			fireTableDataChanged();
		}
	}

	/**
	 * set the selection mode: 0 == none; 1 == checkboxes, 2 == radios
	 * @param mode the offer selection mode
	 */
	public void setSelectMode(int mode) {
		if (mode >= M_SelectNone && mode <= M_SelectSingle) {
			m_SelectMode = mode;
		}
		else {
			Log.error("Invalid offer selection mode, " + mode);
		}
	}

	/**
	 * deselect everything (clear all checkboxes and radio buttons)
	 */
	public void deselectAll() {
		for (TradeOffer offer : m_SelectBoxes.keySet()) {
			JCheckBox chk = m_SelectBoxes.get(offer);
			if (chk != null) {
				chk.setSelected(false);
			}
		}
		for (TradeOffer offer : m_SelectButtons.keySet()) {
			JRadioButton rad = m_SelectButtons.get(offer);
			if (rad != null) {
				rad.setSelected(false);
			}
		}
		m_ButtonGroup.clearSelection();

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * select everything on a list (set checkboxes)
	 * @param offers the list of offers to select
	 */
	public void selectOffers(ArrayList<TradeOffer> offers) {
		for (TradeOffer offer : offers) {
			JCheckBox chk = m_SelectBoxes.get(offer);
			if (chk != null) {
				chk.setSelected(true);
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * de-select a particular offer
	 * @param unselected the offer to unselect
	 */
	public void deselectOffer(TradeOffer unselected) {
		if (unselected == null) {
			return;
		}
		JCheckBox chk = m_SelectBoxes.get(unselected);
		if (chk != null) {
			chk.setSelected(false);
		}
		JRadioButton rad = m_SelectButtons.get(unselected);
		if (rad != null) {
			rad.setSelected(false);
		}
		if (m_SelectMode == M_SelectSingle) {
			m_ButtonGroup.clearSelection();
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_OfferList != null && !m_OfferList.isEmpty()) {
				m_OfferList.clear();
			}
			if (m_SelectBoxes != null && !m_SelectBoxes.isEmpty()) {
				m_SelectBoxes.clear();
			}
			if (m_SelectButtons != null && !m_SelectButtons.isEmpty()) {
				m_SelectButtons.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
