/*
 * last modified---
 * 	06-07-13/jet/ SilentVault client version
 * 	03-18-10/kbw/ allow for leading "B" on baggage field names
 * 	05-08-09/kbw/ new
 *
 * purpose---
 * 	provide a TableModel for the payment JTable display in the WorkPane
 */

package org.silentvault.client.ui;

import javax.swing.table.AbstractTableModel;

import org.silentvault.client.Log;
import org.silentvault.client.XMLPayment;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Date;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of decoded pending payments in a scrolling window.
 *
 * @author Kevin Wilkerson
 */
public final class PaymentTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names
	 */
	private final String[]	M_ColumnNames = { "Payee",
															"Amount",
															"Serial no.",
															"Expiration",
															"Transaction info" };

	/**
	 * the list of payments we're representing (our data), mapped to their
	 * filenames on the SDS
	 */
	private Hashtable<XMLPayment, String>	m_PaymentList;

	/**
	 * list of payments, without filenames
	 */
	private ArrayList<XMLPayment>	m_Payments;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param payments the list of payments which we are to represent
	 */
	public PaymentTableModel(Hashtable<XMLPayment, String> payments) {
		super();
		m_PaymentList = payments;
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of receipts)
	 */
	public int getRowCount() { return m_PaymentList.size(); }

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (receipt) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_PaymentList.size() || row < 0 || col < 0
			 || col >= M_ColumnNames.length)
		{
			return null;
		}

		// build list out of hashtable key set
		m_Payments = new ArrayList<XMLPayment>(m_PaymentList.keySet());
		XMLPayment payment = m_Payments.get(row);
		Object ret = null;
		switch (col) {
			case 0:	// payee
				ret = payment.getPayee();
				break;

			case 1:	// amount -- express this as quantity/units/asset
				String amt = payment.getQuantity() + " " + payment.getUnits()
								+ " of " + payment.getAsset();
				ret = amt;
				break;

			case 2:	// serial number of voucher
				ret = payment.getSerialNumber();
				break;

			case 3:	// TTL expiration
				ret = new Date(payment.getTTL());
				break;

			case 4:	// baggage fields
				StringBuilder transInfo = new StringBuilder(512);
				Hashtable<String, String> baggage = payment.getBaggageList();
				if (!baggage.isEmpty()) {
					for (String field : baggage.keySet()) {
						if (field.equals("field_names")) {
							continue;
						}
						transInfo.append(field.substring(1) + ": "
											+ baggage.get(field) + " | ");
					}
				}
				ret = transInfo.toString();
				break;

			default:	// impossible
				Log.error("Impossible column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	/**
	 * get the payment list
	 * @return the list (not a copy; side-effects are intended here)
	 */
	public Hashtable<XMLPayment, String> getPaymentList() {
		return m_PaymentList;
	}

	/**
	 * get the payment at a specific row
	 * @param row the row whose payment we're interested in
	 * @return the payment
	 */
	public XMLPayment getPaymentAtRow(int row) {
		if (row >= m_PaymentList.size() || row < 0) {
			return null;
		}

		// build list out of hashtable key set
		m_Payments = new ArrayList<XMLPayment>(m_PaymentList.keySet());
		return m_Payments.get(row);
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_PaymentList != null && !m_PaymentList.isEmpty()) {
				m_PaymentList.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
