/*
 * last modified---
 * 	03-20-14/jet/ reword Auto-connect button text
 * 	06-07-13/jet/ SilentVault client version
 * 	11-12-12/kbw/ make default value of inactivity timeout 20 minutes not 10
 * 	11-18-10/kbw/ enforce valid values for inactivity, timeout, and token lot
 * 	10-26-10/kbw/ make L&F invisible in the UI; add autoHousekeeping;
 * 					  harmonize layout with other screens
 * 	08-19-10/kbw/ add lookAndFeel preference
 * 	07-31-10/kbw/ add prefPublisher preference, but invisibly in the UI
 * 	06-03-10/kbw/ add showSafeKeys preference, but invisibly in the UI
 * 	03-02-10/kbw/ add preference directory and connectOnStartup preference
 * 	06-01-09/kbw/ add session inactivity
 * 	05-26-09/kbw/ new
 *
 * purpose---
 * 	provide a UI portmanteau class for wallet preferences
 */

package org.silentvault.client.ui;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import org.silentvault.client.WalletClient;
import org.silentvault.client.AssetPriceCache;
import org.silentvault.client.Log;
import org.silentvault.client.ui.wallet.WTabManager;

import java.util.Set;
import java.util.Vector;


/**
 * This class works with VSPreferences to provide Swing UI elements and storage
 * for preference values associated with the application.
 *
 * @author Kevin Wilkerson
 */
public class PreferencesUI extends JPanel {
	// BEGIN data members
	/**
	 * all look and feels installed
	 */
	private final UIManager.LookAndFeelInfo[]	M_AllLF;

	/**
	 * the application to which we belong
	 */
	private WalletClient	m_Plugin;

	/**
	 * the number of tokens to purchase in a lot (can be overridden)
	 */
	private JTextField	m_TokenLotSize;

	/**
	 * the preferred fiat currency code for pricing voucher value
	 */
	private JComboBox<String>	m_ValueCurrency;

	/**
	 * the length of time we should wait on responses from partners (in secs)
	 */
	private JTextField	m_Timeout;

	/**
	 * the length of time we should wait before inactivity logout (in minutes)
	 */
	private JTextField	m_Inactivity;

	/**
	 * whether or not we should connect to the OFS immediately on startup
	 */
	private JRadioButton	m_ConnectAtStart;

	/**
	 * whether or not we should show safe keys along with other confirmation
	 * data at safe creation
	 */
	private JRadioButton	m_ShowKeysDisplay;

	/**
	 * submit action for changes
	 */
	private AbstractAction	m_SubmitAction;

	/**
	 * the domain of the last publisher to which we connected
	 */
	private JTextField	m_PrefPublisher;

	/**
	 * the look and feel preference
	 */
	private JComboBox<String>	m_LookAndFeel;

	/**
	 * whether or not we should perform housekeeping automatically
	 */
	private JRadioButton	m_AutoHousekeeping;

	// END data members

	// BEGIN methods
	/**
	 * create default panel (constructor)
	 * @param plugin the plugin which owns us
	 */
	public PreferencesUI(WalletClient plugin) {
		super();
		m_Plugin = plugin;
		m_ShowKeysDisplay = new JRadioButton();
		m_PrefPublisher = new JTextField();
		m_PrefPublisher.setDropTarget(null);

		// get the list of all installed L&Fs
		M_AllLF = UIManager.getInstalledLookAndFeels();
		buildUI();
	}

	/**
	 * create UI elements and resources
	 */
	public void buildUI() {
		setLayout(new GridLayout());
		removeAll();
		if (m_TokenLotSize == null) {
			m_TokenLotSize = new JTextField(5);
			m_TokenLotSize.setDropTarget(null);
			m_TokenLotSize.setMinimumSize(new Dimension(30, 20));
		}

		// obtain the set of possible fiat currencies
		AssetPriceCache priceCache = m_Plugin.getPricingModule();
		Set<String> fiatCurrCodes = priceCache.getSupportedCurrencies();
		Vector<String> fiats = new Vector<String>(fiatCurrCodes.size() + 1);
		int idx = 0;
		for (String fCode : fiatCurrCodes) {
			String item = fCode + " - " + priceCache.getCurrencyName(fCode);
			fiats.add(item);
		}

		// append grams of gold and make it default valuation
		fiats.add("GAU - grams of gold");
		if (m_ValueCurrency == null || m_ValueCurrency.getItemCount() == 1) {
			m_ValueCurrency = new JComboBox<String>(fiats);
			m_ValueCurrency.setSelectedIndex(m_ValueCurrency.getItemCount()-1);
		}

		if (m_Timeout == null) {
			m_Timeout = new JTextField(5);
			m_Timeout.setDropTarget(null);
			m_Timeout.setMinimumSize(new Dimension(20, 20));
		}
		if (m_Inactivity == null) {
			m_Inactivity = new JTextField(5);
			m_Inactivity.setDropTarget(null);
			m_Inactivity.setMinimumSize(new Dimension(20, 20));
		}

		// add preference UI widgets, with labels
		setBorder(BorderFactory.createTitledBorder(
													"Preferences for Your Wallet Client"));
		JPanel contents = new JPanel(new GridBagLayout());
		contents.setBorder(null);
		EmptyBorder eb = new EmptyBorder(4, 8, 4, 8);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		contents.add(Box.createVerticalStrut(15), gbc);

		// token lot size
		JTextArea tokLab = new JTextArea("Enter the lot size in which you wish "
												+ "to purchase tokens.  (10-200)", 1, 35);
		tokLab.setLineWrap(false);
		tokLab.setEditable(false);
		tokLab.setBorder(eb);
		tokLab.setBackground(m_Plugin.m_LabelBackColor);
		tokLab.setFocusable(false);
		m_TokenLotSize.setToolTipText("the number of tokens you want to "
												+ "buy at one time");
		contents.add(tokLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_TokenLotSize, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// fiat display currency
		String currTxt = "Select the currency in which you prefer to see voucher "
								+ "values represented.  All vouchers are actually "
								+ "denominated in metal weight, so this setting is "
								+ "merely for display purposes.";
		JTextArea currLab = new JTextArea(currTxt, 4, 35);
		currLab.setLineWrap(true);
		currLab.setWrapStyleWord(true);
		currLab.setEditable(false);
		currLab.setBorder(eb);
		currLab.setBackground(m_Plugin.m_LabelBackColor);
		currLab.setFocusable(false);
		contents.add(currLab, gbc);
		m_ValueCurrency.setToolTipText("pick a currency from the drop-down list");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_ValueCurrency, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// message reply timeout
		String timeTxt = "Enter the number of seconds which your client will "
								+ "wait for a response before timing out.  Your "
								+ "location and the quality of your internet "
								+ "connection may affect this.  For most users, the "
								+ "one minute default should be acceptable.  (10-120)";
		JTextArea timeLab = new JTextArea(timeTxt, 4, 35);
		timeLab.setLineWrap(true);
		timeLab.setWrapStyleWord(true);
		timeLab.setEditable(false);
		timeLab.setBorder(eb);
		timeLab.setBackground(m_Plugin.m_LabelBackColor);
		timeLab.setFocusable(false);
		contents.add(timeLab, gbc);
		m_Timeout.setToolTipText("seconds before a OFS/SDS/DHT request times "
										+ "out");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_Timeout, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// inactivity timeout
		String inacTxt = "Enter the number of minutes which your client will "
								+ "wait without message activity before automatically "
								+ "logging out of your wallet. (5-60)";
		JTextArea inactLab = new JTextArea(inacTxt, 3, 35);
		inactLab.setLineWrap(true);
		inactLab.setWrapStyleWord(true);
		inactLab.setEditable(false);
		inactLab.setBorder(eb);
		inactLab.setBackground(m_Plugin.m_LabelBackColor);
		inactLab.setFocusable(false);
		contents.add(inactLab, gbc);
		m_Inactivity.setToolTipText("<html>minutes before your wallet<br/>"
											+ "times out without user input</html>");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_Inactivity, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// network connect on startup
		String connTxt = "Check the button to connect to the network "
								+ "automatically, or uncheck to connect manually";
		JTextArea connLab = new JTextArea(connTxt, 2, 35);
		connLab.setLineWrap(true);
		connLab.setWrapStyleWord(true);
		connLab.setEditable(false);
		connLab.setBorder(eb);
		connLab.setBackground(m_Plugin.m_LabelBackColor);
		connLab.setFocusable(false);
		m_ConnectAtStart = new JRadioButton("Auto-connect to selected network "
														+ "gateway");
		m_ConnectAtStart.setToolTipText("<html>select to connect to the network "
												+ "<br/>gateway automatically</html>");
		contents.add(connLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_ConnectAtStart, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// automate housekeeping
		String autoTxt = "Check the button to auto-process incoming items found "
								+ "on the DHT at login or when Refresh is clicked.  "
								+ "Uncheck to process Refresh items manually.";
		JTextArea autoLab = new JTextArea(autoTxt, 3, 35);
		autoLab.setLineWrap(true);
		autoLab.setWrapStyleWord(true);
		autoLab.setEditable(false);
		autoLab.setBorder(eb);
		autoLab.setBackground(m_Plugin.m_LabelBackColor);
		autoLab.setFocusable(false);
		m_AutoHousekeeping
			= new JRadioButton("Automatic processing of Refresh items");
		m_AutoHousekeeping.setToolTipText("disable to process incoming items "
													+ "manually");
		contents.add(autoLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_AutoHousekeeping, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// build chooser for look and feels
		Vector<String> lfNames = new Vector<String>();
		lfNames.add("System");
		for (int iii = 0; iii < M_AllLF.length; iii++) {
			lfNames.add(M_AllLF[iii].getName());
		}
		m_LookAndFeel = new JComboBox<String>(lfNames);
	/*
		m_LookAndFeel.setToolTipText("Select the look and feel for your client");
		String lfTxt = "Select the look and feel which you wish to use for your "
							+ "client.  The default is the Nimbus look and feel.  "
							+ "If you change this setting while logged in to a "
							+ "wallet, please logout and log in again to see "
							+ "the full effect.";
		JTextArea lfLab = new JTextArea(lfTxt, 4, 35);
		lfLab.setLineWrap(true);
		lfLab.setWrapStyleWord(true);
		lfLab.setEditable(false);
		lfLab.setBorder(eb);
		lfLab.setBackground(m_Plugin.m_LabelBackColor);
		lfLab.setFocusable(false);
		contents.add(lfLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_LookAndFeel, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);
	 */

		// add submit button
		m_SubmitAction = new AbstractAction("Accept") {
			public void actionPerformed(ActionEvent actionEvent) {
				// update L&F
				//setLookAndFeel((String) m_LookAndFeel.getSelectedItem());

				// commit any changes and return to appropriate screen
				m_Plugin.getUserPreferences().commit();
				WTabManager tm = m_Plugin.getWalletTabManager();
				if (m_Plugin.getVscState().isLoggedIn()) {
					// go back to home pane
					tm.queueTransition(new Integer(WTabManager.M_HOME_RETURN));
					tm.makeTransition();
				}
				else
				{
					if (m_Plugin.getOFSConnection() != null) {
						// go to login pane
						tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
						tm.makeTransition();
					}
					else {
						// go to disconnected page
						tm.queueTransition(new Integer(WTabManager.M_DISCONNECTED));
						tm.makeTransition();
					}
				}
			}
		};
		JButton submitButton = new JButton(m_SubmitAction);
		submitButton.setMnemonic(KeyEvent.VK_S);
		submitButton.setFont(m_Plugin.M_ButtonFont);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(submitButton, gbc);
		contents.add(Box.createVerticalStrut(15), gbc);

		// put contents in a scrolling window in case it's too large
		JScrollPane conScroll = new JScrollPane(contents);
		conScroll.setBorder(null);
		add(conScroll);
		revalidate();
	}

	/**
	 * provide read access to token lot size
	 * @return the token lot size value
	 */
	public String getTokenLotSize() { return m_TokenLotSize.getText(); }

	/**
	 * provide read access to value currency setting
	 * @return the currently set item
	 */
	public String getValueCurrency() {
		String sel = (String) m_ValueCurrency.getSelectedItem();
		return sel;
	}

	/**
	 * provide read access to reply timeout setting
	 * @return the reply timeout
	 */
	public String getTimeout() { return m_Timeout.getText(); }

	/**
	 * provide read access to inactivity setting
	 * @return the inactivity timeout
	 */
	public String getInactivity() { return m_Inactivity.getText(); }

	/**
	 * provide read access to connect at startup setting
	 * @return true if we should connect at the start, false otherwise
	 */
	public boolean getConnectAtStart() { return m_ConnectAtStart.isSelected(); }

	/**
	 * provide read access to show keys setting
	 * @return true if we should display a newly created safe's keys, else false
	 */
	public boolean getSafeKeysDisp() { return m_ShowKeysDisplay.isSelected(); }

	/**
	 * provide read access to token lot size
	 * @return the token lot size value
	 */
	public String getPrefPublisher() { return m_PrefPublisher.getText(); }

	/**
	 * provide read access to desired look and feel
	 * @return the look and feel selected
	 */
	public String getLookAndFeel() {
		String sel = (String) m_LookAndFeel.getSelectedItem();
		return sel;
	}

	/**
	 * provide read access to auto-housekeeping setting
	 * @return true if we should automate processing, false otherwise
	 */
	public boolean getAutoHousekeeping() {
		return m_AutoHousekeeping.isSelected();
	}


	/**
	 * write access to token lot size
	 * @param lotSize the lot size to set
	 */
	public void setTokenLotSize(String lotSize) {
		// make sure it's a number from 10-200 (in case file was edited by hand)
		Integer lot = null;
		try {
			lot = Integer.parseInt(lotSize);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid token lot size, " + lotSize);
			lotSize = "50";
		}
		int tokens = lot.intValue();
		if (tokens < 10 || tokens > 200) {
			Log.error("Invalid token lot size, " + tokens
						+ ", must be between 10 and 200");
			lotSize = "50";
		}
		m_TokenLotSize.setText(lotSize);
	}

	/**
	 * write access to value currency setting
	 * @param cCode the currency code to set as default
	 */
	public void setValueCurrency(String cCode) {
		int cnt = m_ValueCurrency.getItemCount();
		boolean set = false;
		for (int iii = 0; iii < cnt; iii++) {
			String item = m_ValueCurrency.getItemAt(iii);
			if (item.startsWith(cCode)) {
				m_ValueCurrency.setSelectedIndex(iii);
				set = true;
				break;
			}
		}
		if (!set) {
			// default to GAU if cCode was not found
			m_ValueCurrency.setSelectedIndex(cnt-1);
		}
	}

	/**
	 * write access to timeout setting
	 * @param timeout the number of seconds to wait before timeouts
	 */
	public void setTimeout(String timeout) {
		// make sure it's a number from 10-120 (in case file was edited by hand)
		Integer tout = null;
		try {
			tout = Integer.parseInt(timeout);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid timeout value, " + timeout);
			timeout = "60";
		}
		int seconds = tout.intValue();
		if (seconds < 10 || seconds > 120) {
			Log.error("Invalid timeout value, " + seconds
						+ ", must be between 10 and 120 seconds");
			timeout = "60";
		}
		m_Timeout.setText(timeout);
	}

	/**
	 * write access to inactivity setting
	 * @param inactive the number of minutes to wait before auto-logout
	 */
	public void setInactivity(String inactive) {
		// make sure it's a number from 5-60 (in case file was edited by hand)
		Integer inact = null;
		try {
			inact = Integer.parseInt(inactive);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid inactivity value, " + inactive);
			inactive = "20";
		}
		int minutes = inact.intValue();
		if (minutes < 5 || minutes > 60) {
			Log.error("Invalid inactivity value, " + minutes
						+ ", must be between 5 and 60 minutes");
			inactive = "20";
		}
		m_Inactivity.setText(inactive);
	}

	/**
	 * write access to connect at startup setting
	 * @param connect true if we should connect, false otherwise
	 */
	public void setConnectAtStart(String connect) {
		Boolean conn = Boolean.parseBoolean(connect);
		m_ConnectAtStart.setSelected(conn.booleanValue());
	}

	/**
	 * write access to show keys setting
	 * @param display true if we should display newly created keys, else false
	 */
	public void setSafeKeysDisp(String display) {
		Boolean disp = Boolean.parseBoolean(display);
		m_ShowKeysDisplay.setSelected(disp.booleanValue());
	}

	/**
	 * write access to last (preferred) publisher
	 * @param pub the publisher we most recently logged into
	 */
	public void setPrefPublisher(String pub) { m_PrefPublisher.setText(pub); }

	/**
	 * write access to selected look and feel
	 * @param landf the look and feel to set
	 */
	public void setLookAndFeel(String landf) {
		String curLF = UIManager.getLookAndFeel().getName();
		m_LookAndFeel.setSelectedItem(landf);
		if (curLF.equals(landf)) {
			// no change
			return;
		}

		// find corresponding class name
		String className = "";
		if (landf.equals("System")) {
			className = UIManager.getSystemLookAndFeelClassName();
		}
		else {
			for (int jjj = 0; jjj < M_AllLF.length; jjj++) {
				if (landf.equals(M_AllLF[jjj].getName())) {
					className = M_AllLF[jjj].getClassName();
					break;
				}
			}
		}
		try {
			// now set the default L&F and redisplay
			UIManager.setLookAndFeel(className);
			//Log.debug("Setting L&F class: " + className);
			SwingUtilities.updateComponentTreeUI(m_Plugin);
			m_Plugin.pack();
		}
		catch (ClassNotFoundException cnfe) {
			Log.error("Cannot set look and feel", cnfe);
		}
		catch (InstantiationException ie) {
			Log.error("Cannot set look and feel", ie);
		}
		catch (IllegalAccessException iae) {
			Log.error("Cannot set look and feel", iae);
		}
		catch (UnsupportedLookAndFeelException ulafe) {
			Log.error("Cannot set look and feel", ulafe);
		}
		finally {
			// force double display of login pane (kludge for buttons)
			WTabManager wtMgr = m_Plugin.getWalletTabManager();
			if (wtMgr != null) {
				wtMgr.resetFirstLogin();
			}
		}
	}

	/**
	 * write access to housekeeping automation setting
	 * @param automate true if we should automate, false otherwise
	 */
	public void setAutoHousekeeping(String automate) {
		Boolean autoH = Boolean.parseBoolean(automate);
		m_AutoHousekeeping.setSelected(autoH.booleanValue());
	}

	// END methods
}
