/*
 * last modified---
 * 	06-07-13/jet/ SilentVault client version
 * 	10-13-11/kbw/ add mergeTokens() to avoid recreating table
 * 	10-19-10/kbw/ add Issuer restriction to getAvailableTokens() and *Count()
 * 	05-12-10/kbw/ show Issuer instead of units
 * 	04-28-10/kbw/ add getAvailableCount()
 * 	06-26-09/kbw/ debugging changes, update token pane title on spent tokens
 * 	05-12-09/kbw/ winnow out spent tokens using SpentTokenCache
 * 	04-22-09/kbw/ add handling for tokens offered but not confirmed spent
 * 	04-16-09/kbw/ new
 *
 * purpose---
 * 	provide a TableModel for the table JTable display in the HomePane
 */

package org.silentvault.client.ui;

import javax.swing.table.AbstractTableModel;

import org.silentvault.client.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.security.PublicKey;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of tokens in a scrolling window.
 *
 * @author Kevin Wilkerson
 */
public final class TokenTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names
	 */
	private final String[]	M_ColumnNames = { "Serial no.",
															"Amount",
															"Asset",
															"Issuer",
															"Valid" };

	/**
	 * the list of tokens we're representing (our data)
	 */
	private List<XMLToken>	m_TokenList;

	/**
	 * the owning plugin
	 */
	private WalletClient		m_Plugin;

	/**
	 * the list of tokens (if any) which we've already spent by including them
	 * in the <payment/> section of a request, but which are not yet confirmed
	 * as spent (by receiving the success reply)
	 */
	private List<XMLToken>	m_CommittedTokens;

	/**
	 * the public key (of the VP) which can be used to validate token signatures
	 */
	private PublicKey			m_SigKey;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param tokens the list of tokens which we are to represent
	 * @param plugin the Plugin which owns us
	 */
	public TokenTableModel(ArrayList<XMLToken> tokens, WalletClient plugin) {
		super();
		m_TokenList
			= Collections.synchronizedList(new ArrayList<XMLToken>(tokens));
		m_CommittedTokens = Collections.synchronizedList(
											new ArrayList<XMLToken>(m_TokenList.size()));
		m_Plugin = plugin;
		m_SigKey = m_Plugin.getLoginSecrets().getVPKey();
	}

	/**
	 * method to return count of available tokens (as opposed to getRowCount())
	 * @param issuer the Issuer whose tokens we're interested in
	 * @return the count of uncommitted tokens
	 */
	public int getAvailableCount(String issuer) {
		if (issuer == null || issuer.isEmpty() || issuer.equals("-none-")) {
			// show all, regardless of Issuer
			return getRowCount();
		}
		int rawCnt = 0;
		for (XMLToken tok : m_TokenList) {
			if (tok.getIssuer().equals(issuer)) {
				rawCnt += 1;
			}
		}
		int commitCnt = 0;
		for (XMLToken tok : m_CommittedTokens) {
			if (tok.getIssuer().equals(issuer)) {
				commitCnt += 1;
			}
		}
		return rawCnt - commitCnt;
	}


	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of tokens)
	 */
	public int getRowCount() { return m_TokenList.size(); }

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (token) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_TokenList.size() || row < 0 || col < 0
			 || col >= M_ColumnNames.length)
		{
			return null;
		}
		XMLToken tok = m_TokenList.get(row);
		Object ret = null;
		switch (col) {
			case 0:	// serial number
				ret = tok.getSerialNumber();
				break;

			case 1:	// amount
				ret = new Double(tok.getQuantity());
				break;

			case 2:	// asset
				ret = tok.getAsset();
				break;

			case 3:	// issuer
				ret = tok.getIssuer();
				break;

			case 4:	// valid (i.e. good sig)
				if (tok.validateSig(m_SigKey)) {
					ret = new Boolean(true);
				}
				else {
					ret = new Boolean(false);
				}
				break;

			default:	// impossible
				Log.error("Impossible column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	/**
	 * get the current list of tokens, available or not
	 * @return the list
	 */
	public List<XMLToken> getAllTokens() { return m_TokenList; }

	/**
	 * add a list of tokens to the commited (pending spent) list
	 * @param tokList the list of tokens we're offering to spend
	 */
	public void commitTokens(ArrayList<XMLToken> tokList) {
		if (tokList != null && !tokList.isEmpty()) {
			m_CommittedTokens.addAll(tokList);
		}
	}

	/**
	 * obtain the list of presently committed tokens
	 * @return a copy of the list, to avoid side-effects
	 */
	public ArrayList<XMLToken> getCommittedTokens() {
		return new ArrayList<XMLToken>(m_CommittedTokens);
	}
	

	// write operations
	/**
	 * set the sig key, in the event that we didn't know it at construction
	 * @param key the public key to use validating voucher signatures
	 */
	public void setSigKey(PublicKey key) {
		if (key != null) {
			m_SigKey = key;
		}
	}

	/**
	 * remove a list of tokens to the committed (pending spent) list, as when
	 * an attempted operation fails
	 * @param tokList the list of tokens offered but not actually spent
	 */
	public void uncommitTokens(ArrayList<XMLToken> tokList) {
		if (tokList != null && !tokList.isEmpty()) {
			m_CommittedTokens.removeAll(tokList);
		}
	}

	/**
	 * method to obtain a given number of available tokens (tokens on the
	 * m_TokenList but not on the m_CommittedTokens list)
	 * @param count the number of tokens wanted
	 * @param issuer the Issuer whose tokens we're interested in
	 * @return the list, or null if the requested amount is not available
	 */
	public synchronized ArrayList<XMLToken> getAvailableTokens(int count,
																				  String issuer)
	{
		if (count <= 0
			 || issuer == null || issuer.isEmpty() || issuer.equals("-none-"))
		{
			return null;
		}
		ArrayList<XMLToken> availToks = new ArrayList<XMLToken>(m_TokenList);

		// get rid of any tokens committed but not confirmed spent
		availToks.removeAll(m_CommittedTokens);

		// loop through the rest, validating signatures, until we have enough
		ArrayList<XMLToken> retToks = new ArrayList<XMLToken>(count);
		int validCnt = 0;
		for (XMLToken tok : availToks) {
			// don't use a token from the wrong Issuer
			if (!tok.getIssuer().equals(issuer)) {
				continue;
			}

			// don't use a token already spent
			if (m_Plugin.getTokenCache().spentBefore(tok)) {
				continue;
			}

			// don't use a token with an invalid sig
			if (tok.validateSig(m_SigKey)) {
				retToks.add(tok);
			}
			else {
				continue;
			}
			if (++validCnt >= count) {
				// we have enough
				break;
			}
		}
		if (retToks.size() < count) {
			return null;
		}
		return retToks;
	}

	/**
	 * method to "spend" committed tokens by removing them from both lists
	 * @param spentList the list of tokens confirmed spent, or null if we are
	 * to record as spent the entire list of previously committed tokens
	 */
	public synchronized void spendTokens(ArrayList<XMLToken> spentList) {
		if (spentList == null || spentList.isEmpty()) {
			m_TokenList.removeAll(m_CommittedTokens);
			m_CommittedTokens = Collections.synchronizedList(
											new ArrayList<XMLToken>(m_TokenList.size()));
		}
		else {
			m_CommittedTokens.removeAll(spentList);
			m_TokenList.removeAll(spentList);
		}

		// notify listener
		fireTableDataChanged();
	}

	/**
	 * check whether a single given token is available (called by renderer)
	 * @param row the token's row in the table
	 */
	public boolean isAvailable(int row) {
		XMLToken tok = m_TokenList.get(row);
		if (m_CommittedTokens.contains(tok)
			 || m_Plugin.getTokenCache().spentBefore(tok)) {
			return false;
		}
		return true;
	}

	/**
	 * destructively merge a list of tokens into the list
	 * @param tokenList the tokens we need to merge
	 */
	public void mergeTokens(List<XMLToken> tokenList) {
		if (tokenList == null || tokenList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_TokenList.clear();
		m_TokenList.addAll(tokenList);

		// notify listener
		fireTableDataChanged();
	}


	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_TokenList != null && !m_TokenList.isEmpty()) {
				m_TokenList.clear();
			}
			if (m_CommittedTokens != null && !m_CommittedTokens.isEmpty()) {
				m_CommittedTokens.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
