/*
 * last modified---
 * 	11-23-16/jet/ TokensPane now uses multi-select mode
 * 	03-28-14/jet/ javadoc fix
 * 	06-07-13/jet/ SilentVault client version
 * 	04-18-11/kbw/ use revalidate rather than paint to redraw the table
 * 	09-20-10/kbw/ also notify MergePane and ValidatePane on voucher selects
 * 	04-13-10/kbw/ new
 *
 * purpose---
 * 	provide a cell editor for the selection column in a voucher table
 */

package org.silentvault.client.ui;

import javax.swing.JCheckBox;
import javax.swing.JToggleButton;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.AbstractCellEditor;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.silentvault.client.XMLVoucher;
import org.silentvault.client.WalletClient;
import org.silentvault.client.Log;
import org.silentvault.client.ui.wallet.*;

import java.util.EventObject;


/**
 * This class is used as a specialty editor for voucher table cells which are
 * in the 0th (Select) column of the table.  It is necessary because we need
 * both radio button as well as checkbox support (and DefaultCellEditor only
 * supports the latter).
 *
 * @author Kevin Wilkerson
 */
public final class VoucherSelector extends AbstractCellEditor
		implements TableCellEditor, ActionListener {
	// data members
	/**
	 * the editor activation command
	 */
	protected static final String	SELECT = "select";

	/**
	 * the editor component
	 */
	private JToggleButton		m_CellEditor;

	/**
	 * the value (checkbox or radio button) in the table
	 */
	private JToggleButton		m_CellValue;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	/**
	 * the voucher table model in support of which we edit the cells
	 */
	private VoucherTableModel	m_VoucherModel;


	// methods
	/**
	 * constructor
	 * @param plugin the plugin or app that we're part of
	 * @param vouchModel the voucher table model we are supporting
	 */
	public VoucherSelector(WalletClient plugin, VoucherTableModel vouchModel) {
		super();
		m_Plugin = plugin;
		m_VoucherModel = vouchModel;

		// configure the buttons
		m_CellEditor = new JToggleButton();
		m_CellEditor.setVisible(false);
		m_CellEditor.setActionCommand(SELECT);
		m_CellEditor.addActionListener(this);
		m_CellEditor.setBorderPainted(false);
	}


	// method required to implement ActionListener
	/**
	 * handle an action performed (a click to toggle state)
	 * @param e the source action event
	 */
	public void actionPerformed(ActionEvent e) {
		if (SELECT.equals(e.getActionCommand())) {
			if (m_CellValue == null) {
				Log.error("Null cell value on click to start edit");
			}
			else {
				// only allow selection if corresponding cell value button enabled
				if (m_CellValue.isEnabled()) {
					m_CellEditor.setSelected(!m_CellValue.isSelected());
				}
			}

			// tell JTable to switch back to renderer for this cell
			fireEditingStopped();
		}
		else {
			Log.error("Unexpected action on select button");
		}
	}


	// method required to implement CellEditor
	/**
	 * obtain the value contained in the editor
	 * @return the value
	 */
	public Object getCellEditorValue() {
		// propogate the selected status if the corresponding button is enabled
		if (m_CellValue.isEnabled()) {
			m_CellValue.setSelected(m_CellEditor.isSelected());
		}

		// redraw the table to force repainting any now de-selected button or box
		// (the buttons are in a ButtonGroup, so side-effects are desired)

		WTabManager tm = m_Plugin.getWalletTabManager();
		final JTable vt = tm.getHomePane().getVoucherTable();

		/*
		 * There are certain panes which might want to know about the change of
		 * selection.  Notify them, according to the applicable selection mode.
		 */
		int mode = m_VoucherModel.getSelectMode();
		if (mode == VoucherTableModel.M_SelectMulti) {
			PaymentPane pp = tm.getPaymentPane();
			pp.voucherSelectChanged();
			MergePane mp = tm.getMergePane();
			mp.voucherSelectChanged();
			TokensPane tp = tm.getTokensPane();
			tp.voucherSelectChanged();
			vt.revalidate();
			vt.repaint();
		}
		else if (mode == VoucherTableModel.M_SelectSingle) {
			SplitPane sp = tm.getSplitPane();
			sp.voucherSelectChanged();
			ValidatePane vp = tm.getValidatePane();
			vp.voucherSelectChanged();

			// redraw the table to force repainting any now de-selected button
			// (the buttons are in a ButtonGroup, so side-effects are desired)

			vt.revalidate();
			vt.repaint();
		}
		return m_CellEditor;
	}


	// method required to implement TableCellEditor
	/**
	 * obtain the cell editor Component
	 * @param vouchTable the voucher table we're rendering
	 * @param colValue the value of the particular column we are editing
	 * (NB: this value is itself a JCheckBox or JRadioButton)
	 * @param isSelected whether this cell (voucher) is currently selected
	 * @param row the table row (i.e. which voucher in the model, ignored)
	 * @param column the table column, which we validate
	 * @return the Component to be used as the cell editor
	 */
	public Component getTableCellEditorComponent(JTable vouchTable,
																Object colValue,
																boolean isSelected,
																int row,
																int column)
	{
		if (column == 0) {
			m_CellValue = (JToggleButton) colValue;
		}
		else {
			Log.error("getTableCellEditorComponent() invoked on column " + column);
			return null;
		}
		m_CellEditor.setSelected(false);
		return m_CellEditor;
	}

	// optional method extending AbstractCellEditor
	/**
	 * see whether the cell can be edited
	 * @param anEvent the event which starts cell editing
	 * @return always true unless we're in no-select mode
	 */
	@Override
	public boolean isCellEditable(EventObject anEvent) {
		// check with the model
		if (m_VoucherModel.getSelectMode() == VoucherTableModel.M_SelectNone) {
			return false;
		}
		return true;
	}

	// optional method extending CellEditor
	/**
	 * state whether cell should be selected when edited
	 * @return always false
	 */
	@Override
	public boolean shouldSelectCell(EventObject anEvent) { return false; }

	// end methods
}
