/*
 * last modified---
 * 	07-17-15/jet/ use VPAssetConfig methods
 * 	06-13-13/jet/ add getAssetVouchers(), getFirstAssetVoucher(),
 * 					  getVoucherAssetSum(), autoSelectAssetVouchers()
 * 	06-07-13/jet/ SilentVault client version
 * 	01-17-13/kbw/ add flipExpSelectability()
 * 	12-18-12/kbw/ use getNPV() instead of baseUnitValue()
 * 	10-12-11/kbw/ fix getVoucherAtRow() to use same ordering as getValueAt()
 * 	05-17-11/kbw/ add deselectVoucher()
 * 	04-15-11/kbw/ make mergeVouchers() destructive, to pick up deletions
 * 	02-23-11/kbw/ do not assume voucher base units are GAU; add method
 * 					  getFirstVoucher()
 * 	10-20-10/kbw/ restrict TableModel interface to vouchers from the selected
 * 					  Issuer only; add autoSelectVouchers() and selectVouchers()
 * 	09-06-10/kbw/ move to 4 column format for "new look"
 * 	09-03-10/kbw/ add getVoucherSum()
 * 	08-11-10/kbw/ grey out select on "pending" vouchers
 * 	08-03-10/kbw/ remove redundant Issuer column; show gAu values in 4 digits
 * 	04-12-10/kbw/ add deselectAll()
 * 	03-30-10/kbw/ rework for new UI spec
 * 	06-18-09/kbw/ add getSumValue()
 * 	05-20-09/kbw/ add handling for committing/uncommitting vouchers
 * 	05-13-09/kbw/ add Value column
 * 	04-22-09/kbw/ add sig validation column
 * 	04-16-09/kbw/ new
 *
 * purpose---
 * 	provide a TableModel for the voucher JTable display in the HomePane
 */

package org.silentvault.client.ui;

import javax.swing.table.AbstractTableModel;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.event.TableModelEvent;

import org.silentvault.client.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Collections;
import java.util.Arrays;
import java.security.PublicKey;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of decoded vouchers in a scrolling window.
 *
 * @author Kevin Wilkerson
 */
public final class VoucherTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names
	 */
	private String[]	m_ColumnNames = { "",
													"Value - ",
													"Equivalent - ",
													"Expiry Date" };
	
	// selection modes
	/**
	 * selection mode constant for when vouchers are not selectable
	 */
	public static final int	M_SelectNone = 0;
	/**
	 * selection mode constant for when vouchers are multi-selectable
	 */
	public static final int	M_SelectMulti = 1;
	/**
	 * selection mode constant for when vouchers are single-selectable
	 */
	public static final int	M_SelectSingle = 2;

	/**
	 * the list of vouchers we're representing (our data), mapped to their
	 * filenames on the SDS
	 */
	private Hashtable<XMLVoucher, String>	m_VoucherList;

	/**
	 * the public key (the VP's) we need to use to validate voucher signatures
	 */
	private PublicKey			m_SigKey;

	/**
	 * the owning plugin/application
	 */
	private WalletClient		m_Plugin;

	/**
	 * the list of vouchers we have committed to an operation, which will be
	 * obsoleted on success, and uncommitted on failure
	 */
	private List<XMLVoucher>	m_CommittedVouchers;

	/**
	 * the list of checkboxes which correspond to the vouchers and are shown
	 * in the first column
	 */
	private Hashtable<XMLVoucher, JCheckBox>	m_SelectBoxes;

	/**
	 * the list of radio buttons which correspond to the vouchers and are shown
	 * in the first column
	 */
	private Hashtable<XMLVoucher, JRadioButton>	m_SelectButtons;

	/**
	 * the selection mode (one of M_SelectNone, M_SelectMulti, M_SelectSingle)
	 */
	private int					m_SelectMode;

	/**
	 * the button group for the radio buttons, to ensure only one at a time can
	 * be checked
	 */
	private ButtonGroup		m_ButtonGroup;

	/**
	 * the selected voucher Issuer
	 */
	private String				m_SelectedIssuer;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param vouchers the list of vouchers which we are to represent
	 * @param pubKey the public key needed to validate voucher sigs
	 * @param plugin the plugin that owns our display
	 */
	public VoucherTableModel(Hashtable<XMLVoucher, String> vouchers,
									 PublicKey pubKey,
									 WalletClient plugin)
	{
		super();
		m_VoucherList = vouchers;
		m_SigKey = pubKey;
		m_Plugin = plugin;
		m_SelectedIssuer = "";
		m_CommittedVouchers = Collections.synchronizedList(
										new ArrayList<XMLVoucher>(m_VoucherList.size()));
		m_SelectBoxes = new Hashtable<XMLVoucher, JCheckBox>();
		m_SelectButtons = new Hashtable<XMLVoucher, JRadioButton>();
		m_ButtonGroup = new ButtonGroup();
		m_SelectMode = M_SelectNone;

		// make a checkbox and a button for each voucher
		for (XMLVoucher vvv : m_VoucherList.keySet()) {
			JCheckBox cb = new JCheckBox();
			JRadioButton rb = new JRadioButton();
			cb.setEnabled(!vvv.isExpired());
			rb.setEnabled(!vvv.isExpired());
			rb.setRolloverEnabled(true);
			m_SelectBoxes.put(vvv, cb);
			m_SelectButtons.put(vvv, rb);
			m_ButtonGroup.add(rb);
		}
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of vouchers)
	 * NB: implied param m_SelectedIssuer
	 */
	public int getRowCount() {
		if (m_SelectedIssuer.isEmpty()) {
			// include all vouchers
			return m_VoucherList.size();
		}

		int rowCnt = 0;
		// include only vouchers belonging to the selected Issuer
		for (XMLVoucher vvv : m_VoucherList.keySet()) {
			if (vvv.getIssuer().equals(m_SelectedIssuer)) {
				rowCnt += 1;
			}
		}
		return rowCnt;
	}

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return m_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > m_ColumnNames.length) {
			return null;
		}
		return m_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (voucher) number
	 * @param col the column number
	 * NB: implied param m_SelectedIssuer
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_VoucherList.size() || row < 0 || col < 0) {
			// request input was invalid
			return null;
		}
		ArrayList<XMLVoucher> voucherList = null;
		if (m_SelectedIssuer.isEmpty()) {
			// include all vouchers
			voucherList = new ArrayList<XMLVoucher>(m_VoucherList.keySet());
		}
		else {
			// include only vouchers from the selected Issuer
			voucherList = new ArrayList<XMLVoucher>(m_VoucherList.size());
			for (XMLVoucher vvv : m_VoucherList.keySet()) {
				if (vvv.getIssuer().equals(m_SelectedIssuer)) {
					voucherList.add(vvv);
				}
			}
		}

		if (voucherList.isEmpty()) {
			return null;
		}
		XMLVoucher vouch = voucherList.get(row);
		double vAmt = vouch.getNPV();

		// this should never happen, unless the m_SelectedIssuer changed between
		// now and the time that getRowCount() was called, but...
		if (vouch == null) {
			return null;
		}

		Float gau = null;
		Object ret = null;
		switch (col) {
			case 0:	// select checkbox or radio button
				if (m_SelectMode == M_SelectMulti) {
					ret = m_SelectBoxes.get(vouch);
				}
				else if (m_SelectMode == M_SelectSingle) {
					ret = m_SelectButtons.get(vouch);
				}
				// else: null is fine
				break;

			case 1:	// amount in base units (Float)
				gau = new Float(vAmt);
				ret = gau;
				break;

			case 2:	// amount in fiat equivalent (Double)
				String prefCurrency
					= m_Plugin.getUserPreferences().getValueCurrency();
				String lookUnits
					= VPAssetConfig.getPriceLookupUnits(vouch.getAsset(),
																	vouch.getUnits());
				AssetPriceCache apc = m_Plugin.getPricingModule();
				Double assetPrice = apc.getRelativePrice(lookUnits, prefCurrency);
				if (assetPrice == null) {
					// provide default since lookup failed
					assetPrice = new Double(1.0);
				}
				double value = assetPrice.doubleValue() * vAmt;
				ret = new Double(value);
				break;

			case 3:	// expiration date
				ret = vouch.getExpiration();
				break;

			default:	// impossible
				Log.error("Impossible voucher column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	// other get methods
	/**
	 * get the whole voucher list
	 * @return a copy of the list (to avoid side-effects)
	 */
	public Hashtable<XMLVoucher, String> getVoucherList() {
		return new Hashtable<XMLVoucher, String>(m_VoucherList);
	}

	/**
	 * get only those vouchers belonging to the selected Issuer.
	 * implied param: m_SelectedIssuer
	 * @return a list of vouchers, mapped to their filenames (may be empty).
	 * NB: the Hashtable returned is a subset of m_VoucherList, but its keySet()
	 * may list vouchers in a different order.
	 */
	public Hashtable<XMLVoucher, String> getIssuerVouchers() {
		Hashtable<XMLVoucher, String> ret = new Hashtable<XMLVoucher, String>();
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getIssuer().equals(m_SelectedIssuer)) {
				String file = m_VoucherList.get(vouch);
				ret.put(vouch, file);
			}
		}
		return ret;
	}

	/**
	 * get only those vouchers representing a certain asset type
	 * @param asset the asset type wanted
	 * @return a list of vouchers, mapped to their filenames (may be empty).
	 * NB: the Hashtable returned is a subset of m_VoucherList, but its keySet()
	 * may list vouchers in a different order.
	 */
	public Hashtable<XMLVoucher, String> getAssetVouchers(String asset) {
		Hashtable<XMLVoucher, String> ret = new Hashtable<XMLVoucher, String>();
		if (asset == null || !VPAssetConfig.getAllAssets().contains(asset)) {
			Log.error("Invalid asset type for listing vouchers, " + asset);
			return ret;
		}
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getAsset().equals(asset)) {
				String file = m_VoucherList.get(vouch);
				ret.put(vouch, file);
			}
		}
		return ret;
	}

	/**
	 * return the first voucher belonging to the selected Issuer.
	 * implied param: m_SelectedIssuer
	 * @return the first voucher found belonging to the given Issuer, or null
	 */
	public XMLVoucher getFirstVoucher() {
		XMLVoucher retV = null;
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getIssuer().equals(m_SelectedIssuer)) {
				retV = vouch;
				break;
			}
		}
		return retV;
	}

	/**
	 * return the first voucher of a certain asset type
	 * @param asset the asset type wanted
	 * @return the first such voucher found, or null
	 */
	public XMLVoucher getFirstAssetVoucher(String asset) {
		XMLVoucher retV = null;
		if (asset == null || !VPAssetConfig.getAllAssets().contains(asset)) {
			Log.error("Invalid asset for obtaining first voucher, " + asset);
			return retV;
		}
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getAsset().equals(asset)) {
				retV = vouch;
				break;
			}
		}
		return retV;
	}

	/**
	 * get the voucher at a specific row
	 * @param row the row whose voucher we're interested in
	 * implied param: m_SelectedIssuer
	 * @return the voucher (null if not found or out of range)
	 */
	public XMLVoucher getVoucherAtRow(int row) {
		if (row >= m_VoucherList.size() || row < 0) {
			return null;
		}
		ArrayList<XMLVoucher> voucherList = null;
		if (m_SelectedIssuer.isEmpty()) {
			// include all vouchers
			voucherList = new ArrayList<XMLVoucher>(m_VoucherList.keySet());
		}
		else {
			// include only vouchers from the selected Issuer
			voucherList = new ArrayList<XMLVoucher>(m_VoucherList.size());
			for (XMLVoucher vvv : m_VoucherList.keySet()) {
				if (vvv.getIssuer().equals(m_SelectedIssuer)) {
					voucherList.add(vvv);
				}
			}
		}

		if (voucherList.isEmpty() || row >= voucherList.size()) {
			return null;
		}
		return voucherList.get(row);
	}

	/**
	 * return all selected vouchers (for the currently selected Issuer)
	 * @return all current Issuer vouchers whose associated box has been checked
	 */
	public ArrayList<XMLVoucher> getSelectedVouchers() {
		ArrayList<XMLVoucher> retList
			= new ArrayList<XMLVoucher>(m_VoucherList.size());
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (!m_SelectedIssuer.isEmpty()
				 && !m_SelectedIssuer.equals(vouch.getIssuer()))
			{
				// voucher is for different Issuer and cannot be selected
				continue;
			}
			if (m_SelectMode == M_SelectMulti) {
				JCheckBox chk = m_SelectBoxes.get(vouch);
				if (chk != null && chk.isSelected()) {
					retList.add(vouch);
				}
			}
			else if (m_SelectMode == M_SelectSingle) {
				JRadioButton rad = m_SelectButtons.get(vouch);
				if (rad != null && rad.isSelected()) {
					retList.add(vouch);
				}
			}
		}
		return retList;
	}

	/**
	 * Sum the value of the vouchers for the selected Issuer.
	 * implied param: m_SelectedIssuer
	 * @return the value in base units of the vouchers present
	 */
	public double getVoucherSum() {
		double total = 0.0;
		for (XMLVoucher vvv : m_VoucherList.keySet()) {
			// if issuer is unset, or if it matches the voucher, include the value
			if (m_SelectedIssuer.isEmpty()
				 || m_SelectedIssuer.equals(vvv.getIssuer()))
			{
				// do not include expired vouchers in total
				if (!vvv.isExpired()) {
					total += vvv.getNPV();
				}
			}
		}
		return total;
	}

	/**
	 * sum the value of the vouchers for a selected asset
	 * @param asset the asset type wanted
	 * @return the value in base units of the vouchers representing this asset
	 */
	public double getVoucherAssetSum(String asset) {
		double total = 0.0;
		if (asset == null || !VPAssetConfig.getAllAssets().contains(asset)) {
			Log.error("Invalid asset for voucher sum, " + asset);
			return total;
		}
		for (XMLVoucher vvv : m_VoucherList.keySet()) {
			if (vvv.getAsset().equals(asset) && !vvv.isExpired()) {
				total += vvv.getNPV();
			}
		}
		return total;
	}

	/**
	 * obtain the list of vouchers we previously committed to an operation
	 * @return a copy of the list (to avoid side-effects)
	 */
	public ArrayList<XMLVoucher> getCommittedVouchers() {
		return new ArrayList<XMLVoucher>(m_CommittedVouchers);
	}

	/**
	 * obtain the current selection mode
	 * @return the mode (one of M_SelectNone, M_SelectMulti, M_SelectSingle)
	 */
	public int getSelectMode() { return m_SelectMode; }

	/**
	 * obtain the currently selected Issuer
	 * @return the Issuer, or the empty string if none set
	 */
	public String getSelectedIssuer() { return m_SelectedIssuer; }

	/**
	 * obtain the signature key
	 * @return the key of the signing authority
	 */
	public PublicKey getSigKey() { return m_SigKey; }


	// write operations
	/**
	 * set the sig key, in the event that we didn't know it at construction
	 * @param key the public key to use validating voucher signatures
	 */
	public void setSigKey(PublicKey key) {
		if (key != null) {
			m_SigKey = key;
		}
	}

	/**
	 * insert a new voucher into the list
	 * @param voucher the voucher
	 * @param file the filename where it's stored on the SDS
	 */
	public void addVoucher(XMLVoucher voucher, String file) {
		if (voucher != null && file != null) {
			String newFile = m_VoucherList.put(voucher, file);
			if (newFile == null) {
				// this was added
				JCheckBox cb = new JCheckBox();
				JRadioButton rb = new JRadioButton();
				// do not allow selection on a voucher from an incoming confirmation
				if (file.equals("pending")) {
					cb.setEnabled(false);
					rb.setEnabled(false);
				}
				else {
					cb.setEnabled(!voucher.isExpired());
					rb.setEnabled(!voucher.isExpired());
				}
				rb.setRolloverEnabled(true);
				m_SelectBoxes.put(voucher, cb);
				m_SelectButtons.put(voucher, rb);
				m_ButtonGroup.add(rb);

				// notify listeners
				fireTableDataChanged();
			}
		}
	}

	/**
	 * destructively merge a list (map) of vouchers into the list
	 * @param vouchList the vouchers we need to merge
	 */
	public void mergeVouchers(Map<XMLVoucher, String> vouchList) {
		m_VoucherList.clear();
		if (vouchList == null || vouchList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_VoucherList.putAll(vouchList);

		// rebuild select controls
		for (XMLVoucher vvv : m_VoucherList.keySet()) {
			if (m_SelectBoxes.get(vvv) == null) {
				JCheckBox cb = new JCheckBox();
				JRadioButton rb = new JRadioButton();
				cb.setEnabled(!vvv.isExpired());
				rb.setEnabled(!vvv.isExpired());
				rb.setRolloverEnabled(true);
				m_SelectBoxes.put(vvv, cb);
				m_SelectButtons.put(vvv, rb);
				m_ButtonGroup.add(rb);
			}
			else {
				// it was already there, make sure buttons are enabled
				JCheckBox cb = m_SelectBoxes.get(vvv);
				cb.setEnabled(!vvv.isExpired());
				JRadioButton rb = m_SelectButtons.get(vvv);
				rb.setEnabled(!vvv.isExpired());
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * add vouchers to the committed (pending used) list
	 * @param vouchList the vouchers we're offering in an operation
	 */
	public void commitVouchers(ArrayList<XMLVoucher> vouchList) {
		if (vouchList != null && !vouchList.isEmpty()) {
			m_CommittedVouchers.addAll(vouchList);
		}
	}

	/**
	 * remove vouchers from the committed (pending used) list
	 * @param vouchList the vouchers we're no longer using in an operation
	 */
	public void uncommitVouchers(ArrayList<XMLVoucher> vouchList) {
		if (vouchList != null && !vouchList.isEmpty()) {
			m_CommittedVouchers.removeAll(vouchList);
		}
	}

	/**
	 * "spend" a list of vouchers, removing them from both lists
	 * @param vouchList the vouchers we've just used/replaced in an operation
	 */
	public void removeVouchers(ArrayList<XMLVoucher> vouchList) {
		if (vouchList != null && !vouchList.isEmpty()) {
			m_CommittedVouchers.removeAll(vouchList);
			for (XMLVoucher vouch : vouchList) {
				m_VoucherList.remove(vouch);
				m_SelectBoxes.remove(vouch);
				JRadioButton rb = m_SelectButtons.get(vouch);
				m_SelectButtons.remove(vouch);
				m_ButtonGroup.remove(rb);
			}

			// notify listeners
			fireTableDataChanged();
		}
	}

	/**
	 * set the desired Issuer
	 * @param issuer the selected Issuer
	 */
	public void selectIssuer(String issuer) {
		if (issuer != null && !issuer.equals("-none-")) {
			m_SelectedIssuer = issuer;
		}
	}

	/**
	 * set the selection mode: 0 == none; 1 == checkboxes, 2 == radios
	 * @param mode the voucher selection mode
	 */
	public void setSelectMode(int mode) {
		if (mode >= M_SelectNone && mode <= M_SelectSingle) {
			m_SelectMode = mode;
		}
		else {
			Log.error("Invalid voucher selection mode, " + mode);
		}
	}

	/**
	 * deselect everything (clear all checkboxes and radio buttons)
	 */
	public void deselectAll() {
		for (XMLVoucher vouch : m_SelectBoxes.keySet()) {
			JCheckBox chk = m_SelectBoxes.get(vouch);
			if (chk != null) {
				chk.setSelected(false);
			}
		}
		for (XMLVoucher vouch : m_SelectButtons.keySet()) {
			JRadioButton rad = m_SelectButtons.get(vouch);
			if (rad != null) {
				rad.setSelected(false);
			}
		}
		m_ButtonGroup.clearSelection();

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * select everything on a list (set checkboxes)
	 * @param vouchers the list of vouchers to select
	 * NB: normally called after an invocation of autoSelectVouchers()
	 */
	public void selectVouchers(ArrayList<XMLVoucher> vouchers) {
		for (XMLVoucher vouch : vouchers) {
			JCheckBox chk = m_SelectBoxes.get(vouch);
			if (chk != null) {
				chk.setSelected(true);
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * de-select a particular voucher
	 * @param unselected the voucher to unselect
	 */
	public void deselectVoucher(XMLVoucher unselected) {
		JCheckBox chk = m_SelectBoxes.get(unselected);
		if (chk != null) {
			chk.setSelected(false);
		}
		JRadioButton rad = m_SelectButtons.get(unselected);
		if (rad != null) {
			rad.setSelected(false);
		}
		if (m_SelectMode == M_SelectSingle) {
			m_ButtonGroup.clearSelection();
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * set the value of the fiat equivalent header, to show user preference
	 * @param currPref the user's preferred currency code
	 */
	public void setEquivHeader(String currPref) {
		if (currPref == null || currPref.isEmpty()) {
			Log.error("Null currency pref for voucher table header");
			return;
		}
		m_ColumnNames[2] = "Equivalent - " + currPref;

		// need the asset type for vouchers of the selected Issuer
		XMLVoucher first = getFirstVoucher();
		if (first != null) {
			m_ColumnNames[1]
				= "Value - " + XMLVoucher.getDisplayUnits(first.getAsset());
		}
		else {
			m_ColumnNames[1] = "Value - asset";
		}

		// tell listeners header changed
		TableModelEvent hdEvent
			= new TableModelEvent(this, TableModelEvent.HEADER_ROW);
		fireTableChanged(hdEvent);
	}

	/**
	 * Obtain a list of vouchers, from the selected Issuer, which total at
	 * least the indicated weight.  Vouchers are returned oldest-first.
	 * @param gauWt the weight in base units
	 * NB: implied param: m_SelectedIssuer
	 * @return a list of vouchers, null if the safe does not contain enough value
	 */
	public ArrayList<XMLVoucher> autoSelectVouchers(double gauWt) {
		if (gauWt <= 0.0) {
			Log.error("Illegal requested voucher total, " + gauWt);
			return null;
		}
		if (m_SelectedIssuer.isEmpty()) {
			return null;
		}

		// compute the available sum of unexpired vouchers
		double total = 0.0;
		ArrayList<XMLVoucher> candidates
			= new ArrayList<XMLVoucher>(m_VoucherList.size());
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getIssuer().equals(m_SelectedIssuer)
				 && !vouch.isExpired())
			{
				candidates.add(vouch);
				total += vouch.getNPV();
			}
		}

		if (total < gauWt) {
			Log.debug("Too much voucher value requested, " + gauWt + " vs. "
						+ total + " available");
			return null;
		}

		// now sort the vouchers according to date (the natural ordering)
		XMLVoucher[] vouchers = new XMLVoucher[candidates.size()];
		vouchers = candidates.toArray(vouchers);
		Arrays.sort(vouchers, null);
		ArrayList<XMLVoucher> retList
			= new ArrayList<XMLVoucher>(candidates.size());
		total = 0.0;
		for (int iii = 0; iii < candidates.size(); iii++) {
			retList.add(vouchers[iii]);
			total += vouchers[iii].getNPV();
			if (total >= gauWt) {
				break;
			}
		}
		return retList;
	}

	/**
	 * Obtain a list of vouchers, representing the indicated asset, which total
	 * least the indicated quantity.  Vouchers are returned oldest-first.
	 * @param asset the asset type wanted
	 * @param amount the request amount in base units
	 * @return a list of vouchers, null if the safe does not contain enough value
	 */
	public ArrayList<XMLVoucher> autoSelectAssetVouchers(String asset,
																		  double amount)
	{
		if (amount <= 0.0) {
			Log.error("Illegal requested voucher total, " + amount);
			return null;
		}
		if (asset == null || !VPAssetConfig.getAllAssets().contains(asset)) {
			Log.error("Illegal asset type, " + asset);
			return null;
		}

		// compute the available sum of unexpired vouchers
		double total = 0.0;
		ArrayList<XMLVoucher> candidates
			= new ArrayList<XMLVoucher>(m_VoucherList.size());
		for (XMLVoucher vouch : m_VoucherList.keySet()) {
			if (vouch.getAsset().equals(asset) && !vouch.isExpired()) {
				candidates.add(vouch);
				total += vouch.getNPV();
			}
		}

		if (total < amount) {
			Log.debug("Too much voucher value requested, " + amount + " vs. "
						+ total + " available");
			return null;
		}

		// now sort the vouchers according to date (the natural ordering)
		XMLVoucher[] vouchers = new XMLVoucher[candidates.size()];
		vouchers = candidates.toArray(vouchers);
		Arrays.sort(vouchers, null);
		ArrayList<XMLVoucher> retList
			= new ArrayList<XMLVoucher>(candidates.size());
		total = 0.0;
		for (int iii = 0; iii < candidates.size(); iii++) {
			retList.add(vouchers[iii]);
			total += vouchers[iii].getNPV();
			if (total >= amount) {
				break;
			}
		}
		return retList;
	}

	/**
	 * flip selectability to select or not select expired vouchers.
	 *	NB: expired vouchers can only be selected one at a time, so only buttons
	 * but not checkboxes are manipulated.
	 * @param selExp if true, expired vouchers should be selectable
	 */
	public void flipExpSelectability(boolean selExp) {
		for (XMLVoucher vouch : m_SelectButtons.keySet()) {
			JRadioButton selectBut = m_SelectButtons.get(vouch);
			if (selExp) {
				selectBut.setEnabled(vouch.isExpired());
			}
			else {
				selectBut.setEnabled(!vouch.isExpired());
			}
		}
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_VoucherList != null && !m_VoucherList.isEmpty()) {
				m_VoucherList.clear();
			}
			if (m_CommittedVouchers != null && !m_CommittedVouchers.isEmpty()) {
				m_CommittedVouchers.clear();
			}
			if (m_SelectBoxes != null && !m_SelectBoxes.isEmpty()) {
				m_SelectBoxes.clear();
			}
			if (m_SelectButtons != null && !m_SelectButtons.isEmpty()) {
				m_SelectButtons.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
