/*
 * last modified---
 * 	07-17-15/jet/ use lookup units for relative prices
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	03-12-14/jet/ show amount to be paid as max of commission or 25 tokens
 * 	09-11-13/jet/ present single revised popup on escrow funding success
 * 	07-23-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to allow acceptance of trade offers by other users
 */

package org.silentvault.client.ui.svx;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;
import org.silentvault.client.ui.wallet.HomePane;
import org.silentvault.client.ui.wallet.WTabManager;
import org.silentvault.client.ui.wallet.PaymentPane;
import org.silentvault.client.ui.wallet.ConfirmPane;

import java.util.Vector;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Date;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends WorkPane to provide a screen for registering acceptance of a funded
 * offer by another user.
 *
 * @author Justin Turrell
 */
public final class AcceptPane extends WorkPane {
	// BEGIN data members
	/**
	 * action to initiate an acceptance of the indicated trade offer
	 */
	private AbstractAction				m_AcceptAction;

	/**
	 * button to go with accept action
	 */
	private JButton						m_AcceptButton;

	/**
	 * action to return to the Search screen results
	 */
	private AbstractAction				m_BackAction;

	/**
	 * button to go with back action
	 */
	private JButton						m_BackButton;

	/**
	 * action to acknowledge offer acceptance language and reset
	 */
	private AbstractAction				m_AckAction;

	/**
	 * button to go with acknowledge action
	 */
	private JButton						m_AckButton;

	/**
	 * mode toggle: if true, display accept entry form; if false show confirm
	 */
	private boolean						m_AcceptMode;

	/**
	 * control to select wanted asset being provided
	 */
	private JComboBox<String>			m_ProvidedUI;

	/**
	 * the currently selected asset to be provided with acceptance
	 */
	private String							m_Provided;

	/**
	 * the offer being accepted
	 */
	private TradeOffer					m_TradeOffer;

	/**
	 * label showing amount which must be tendered to accept for delivered asset
	 */
	private JLabel							m_Required;

	/**
	 * panel to contain accept entry form
	 */
	private JPanel							m_FormPane;

	/**
	 * escrow wallet designated by the SVX for payment
	 */
	private String							m_EscrowWalletId;

	/**
	 * total payment to be made to fund escrow (includes commission)
	 */
	private double							m_EscrowAmount;

	/**
	 * timestamp of escrow payout
	 */
	private Date							m_EscrowTimestamp;

	/**
	 * flag to indicate (to WTabManager) that we're making an escrow payment
	 */
	private boolean						m_DoingEscrow;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public AcceptPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);
		m_EscrowWalletId = m_Provided = "";
		m_ProvidedUI = new JComboBox<String>();
		m_ProvidedUI.addActionListener(this);
		m_ProvidedUI.setPreferredSize(new Dimension(50, 25));
		m_ProvidedUI.setToolTipText(
									"Select the asset you wish to offer in exchange");
		m_AcceptMode = true;
		m_UserInstructs = "To accept this offer, select the asset you will "
								+ "provide in exchange for the offered asset.<br/><br/>"
								+ "The amount you will need to pay into escrow is "
								+ "shown below, and includes the commission owed to "
								+ "the SVX.<br/><br/>"
								+ "Note that you cannot accept your own offers.  "
								+ "(Use cancel from the My Offers screen instead.)";

		// define Accept confirmation action
		m_AcceptAction = new AbstractAction("Confirm Acceptance") {
			public void actionPerformed(ActionEvent ae) {
				if (m_TradeOffer == null) {
					Log.error("No offer to accept selected");
					showError("cannot accept offer", "none selected",
								"select a trade offer to accept");
					return;
				}

				// validate that selected delivered asset is in wanted list
				String delivAsset = (String) m_ProvidedUI.getSelectedItem();
				if (delivAsset == null
					 || !m_TradeOffer.getWanted().contains(delivAsset)) 
				{
					Log.error("No provided asset picked");
					showError("cannot accept offer",
								"selected provided asset (" + delivAsset + ") "
								+ "is not accepted for this offer",
								"select a wanted asset type to provide");
					return;
				}
				m_Provided = delivAsset;

				// compute amount of provided asset required, including commission
				double reqAmt = computeReqAmount(m_Provided);
				if (reqAmt == 0.0) {
					showError("cannot accept offer",
								"relative market prices of assets not available");
					return;
				}
				m_EscrowAmount = reqAmt;

				// check that user has this much
				HomePane whp = m_ParentHome.getWalletHome();
				VoucherTableModel vtm = whp.getVoucherModel();
				double avail = vtm.getVoucherAssetSum(m_Provided);
				if (avail < m_EscrowAmount) {
					showError("cannot accept offer",
								"you do not have enough " + m_Provided
								+ " available in your wallet");
					return;
				}

				/* Send the SVX our acceptance.  This will create a record in its
				 * database, and produce a reply which will confirm the amount of
				 * the escrow payment and give us the escrow wallet to use.  We
				 * will then send the actual payment, followed by a fund offer req.
				 */
				sendAcceptOffer();
			}
		};
		// define associated button
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_AcceptButton = new JButton(m_AcceptAction);
		m_AcceptButton.setFont(buttonFont);
		m_AcceptButton.setToolTipText("Accept this trade offer");

		// define action to return to Search results
		m_BackAction = new AbstractAction("Back") {
			public void actionPerformed(ActionEvent ae) {
				// go back to Search results display, without clearing entries
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SEARCH));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_BackButton = new JButton(m_BackAction);
		m_BackButton.setFont(buttonFont);
		m_BackButton.setToolTipText("Return to search results");

		// define funding acknowledgement action
		m_AckAction = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent ae) {
				// attempt to pick up payment via housekeeping
				WTabManager wtm = m_Plugin.getWalletTabManager();
				m_Plugin.setSelectedTab(wtm);
				wtm.queryDHTrecords(false);

				// then in this tab, reset for next offer acceptance
				m_AcceptButton.setEnabled(true);
				m_AcceptButton.setToolTipText("Accept this trade offer");
				m_Provided = m_EscrowWalletId = "";
				m_EscrowAmount = 0.0;
				m_AcceptMode = true;
				OfferTableModel otm = m_ParentHome.getOfferModel();
				ArrayList<TradeOffer> dels = new ArrayList<TradeOffer>(1);
				dels.add(m_TradeOffer);
				otm.removeOffers(dels);
				m_TradeOffer = null;

				// go back to redisplay searched offers (results)
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SEARCH));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_AckButton = new JButton(m_AckAction);
		m_AckButton.setFont(buttonFont);
		m_AckButton.setToolTipText("Return to search results screen");

		// init label showing amount required to accept an offer in an asset type
		m_Required = new JLabel("0.0", SwingConstants.CENTER);
		m_Required.setFont(m_ParentHome.getWalletHome().M_LabelFont);
		m_Required.setForeground(UIManager.getColor("standardGray"));

		// create offer accept form and its layout parameters
		m_FormPane = new JPanel(new GridBagLayout());
		m_FormPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_WorkArea.setBorder(null);
	}

	/**
	 * record which offer the user wants to accept
	 * @param offer the offer to be accepted
	 */
	public void setOffer(TradeOffer offer) {
		m_TradeOffer = offer;
		if (m_TradeOffer != null) {
			m_AcceptMode = true;
		}
	}

	/**
	 * perform a (blocking) query to the SVX accepting the offer.
	 * implied param: m_Provided, the asset type we'll be delivering
	 */
	private void sendAcceptOffer() {
		// send a REQ_accept_offer to register acceptance and obtain escrow wallet
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock acceptOffer = new SVXClientBlock();
		acceptOffer.setOpcode("REQ_accept_offer");
		acceptOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		acceptOffer.setAsset(m_Provided);
		acceptOffer.setOfferId(m_TradeOffer.getOfferId());
		setCursor(M_WaitCursor);
		svxIface.sendRequest(acceptOffer);
		setCursor(null);
		// NB: reply comes synchronously to processAcceptReply() below
	}

	/**
	 * handle the reply to our offer to accept
	 * @param reply the reply received from the SVX
	 */
	public void processAcceptReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX config reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_accept_entered")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			showError("SVX offer accept failed", reply.getErrMsg());
			return;
		}

		// confirm that offer Id has not changed
		if (reply.getOfferId() != m_TradeOffer.getOfferId()) {
			Log.error("Offer Id being accepted changed, was "
						+ m_TradeOffer.getOfferId() + " now " + reply.getOfferId());
			showError("cannot accept offer", "internal inconsistency");
			return;
		}

		// double-check amount
		if (reply.getTotal() != m_EscrowAmount) {
			// this is normal, due to rounding
			m_EscrowAmount = reply.getTotal();
		}

		// record the wallet Id where we should send the escrow payment
		m_EscrowWalletId = reply.getWalletId();
		if (m_EscrowWalletId.isEmpty()) {
			Log.error("No valid escrow wallet Id specified by SVX");
			showError("cannot accept offer", "missing escrow wallet Id");
			return;
		}

		// build a (decoded) WebMessage object to pass data to PaymentPane
		WebMessage webPass = new WebMessage();
		webPass.setAmount(m_EscrowAmount);
		webPass.setVSnumber(m_EscrowWalletId);
		// determine issuer for this asset by examining a voucher
		HomePane whp = m_ParentHome.getWalletHome();
		VoucherTableModel vtm = whp.getVoucherModel();
		XMLVoucher firstV = vtm.getFirstAssetVoucher(m_Provided);
		if (firstV == null) {
			// should not happen, since we vetted asset balance already
			showError("cannot accept offer",
						"no vouchers found for asset " + m_Provided);
			Log.error("No vouchers found to accept offer "
						+ m_TradeOffer.getOfferId() + " with " + m_Provided);
			return;
		}
		// record expected results in the offer
		m_TradeOffer.setExAsset(m_Provided);
		m_TradeOffer.setExQuantity(m_EscrowAmount);

		// prepare payment screen to make this payment
		vtm.selectIssuer(firstV.getIssuer());
		webPass.setIssuer(firstV.getIssuer());
		webPass.setCurrency(XMLVoucher.getDisplayUnits(m_Provided));
		webPass.setBaggageFields("offerId");
		webPass.addOrderRecord("offerId",
									new Integer(m_TradeOffer.getOfferId()).toString());
		webPass.setDecoded(true);
		m_Plugin.getWebListener().addWebMessage(webPass);
		final WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		pp.setSelectedSiteMsg(webPass);

		// disable button to prevent multiple entry
		m_AcceptButton.setEnabled(false);
		m_AcceptButton.setToolTipText("Please make your escrow "
												+ "payment in the Wallet tab.");

		// send user to payment screen in WTabManager, in another thread
		m_DoingEscrow = true;
		Runnable goPay = new Runnable() {
			public void run() {
				wtm.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
				m_Plugin.setSelectedTab(wtm);
				wtm.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goPay);
	}

	/**
	 * register confirmation that we have successfully completed escrow payment
	 */
	public void confirmEscrowFunding() {
		if (m_TradeOffer == null) {
			Log.error("No offer selected");
			return;
		}

		// send a REQ_fund_offer to ask SVX to pickup the escrow payment
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock fundOffer = new SVXClientBlock();
		fundOffer.setOpcode("REQ_fund_offer");
		fundOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		fundOffer.setOfferId(m_TradeOffer.getOfferId());
		fundOffer.setAsset(m_Provided);

		// return focus to exchange tab
		m_Plugin.setSelectedTab(m_TabManager);

		// put up info saying what we are doing (on SVX tab)
		WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		final StringBuilder popTxt = new StringBuilder(256);
		// obtain last confirmation recorded by ConfirmPane off the DHT
		XMLConfirmPayment conf = pp.getLatestConfirmation();
		if (conf != null) {
			/* we deferred showing payment success, and so must recapitulate
			 * what the ConfirmPane would have displayed as popup text.
			 */
			popTxt.append("Your payment of " + conf.getQuantity());
			popTxt.append(" " + conf.getUnits() + " of " + conf.getAsset());
			if (!conf.getAsset().equals(conf.getIssuer())) {
				popTxt.append(" at " + conf.getIssuer());
			}
			popTxt.append("\nwas sent to ");
			popTxt.append(conf.getPayee() + " successfully.");
			// show that change voucher was received if there was one
			if (conf.getVoucher() != null) {
				popTxt.append("\nYou have received a voucher back as change.");
			}

		}
		else {
			Log.error("Unexpected: no XMLConfirmPayment found in PaymentPane");
			// show something generic
			popTxt.append("Your escrow payment to accept offer ID ");
			popTxt.append(m_TradeOffer.getOfferId() + " succeeded.");
		}

		// add generic indication that we're now telling the SVX to pick it up
		popTxt.append("\n\nThe SVX is now processing the escrow payment.  ");
		popTxt.append("This may take a few moments.\nPlease wait for the ");
		popTxt.append("confirmation screen, read it carefully, and acknowledge.");

		// for safety, do in separate thread while allowing message to proceed
		Runnable notice = new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
														"Escrow Payment Completed",
														JOptionPane.INFORMATION_MESSAGE);
			}
		};
		SwingUtilities.invokeLater(notice);

		// reset ConfirmPane to default behavior
		ConfirmPane cp = wtm.getConfirmPane();
		cp.setDeferPopup(false);

		// send message to SVX to trigger pickup
		setCursor(M_WaitCursor);
		svxIface.sendRequest(fundOffer);
		setCursor(null);
		// NB: reply comes synchronously to processFundingReply() below
	}

	/**
	 * record the results of an offer funding attempt (post-payment)
	 * @param reply the reply received from the SVX
	 */
	public void processFundingReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer funding reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_funded")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		m_DoingEscrow = false;
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (reply.isError()) {
			showError("SVX offer funding failed", reply.getErrMsg());
			// in any case, reset for next
			m_AcceptButton.setEnabled(true);
			m_AcceptButton.setToolTipText("Accept this trade offer");
			m_TradeOffer = null;
			m_Provided = m_EscrowWalletId = "";
			m_EscrowAmount = 0.0;
			m_AcceptMode = true;

			/* go back to redisplay searched offers, in a separate thread since
			 * we are an event
			 */
			Runnable goBack = new Runnable() {
				public void run() {
					m_TabManager.queueTransition(
														new Integer(XTabManager.M_GO_SEARCH));
					m_TabManager.makeTransition();
				}
			};
			SwingUtilities.invokeLater(goBack);
		}
		else {
			if (m_TradeOffer != null) {
				// double-check SVX seems to be referring to the same offer
				boolean same = true;
				if (m_TradeOffer.getOfferId() != reply.getOfferId()) {
					Log.error("AP: funding oId inconsistent, "
								+ m_TradeOffer.getOfferId() + " vs. "
								+ reply.getOfferId());
					same = false;
				}
				if (!m_TradeOffer.getExAsset().equals(reply.getAsset())) {
					Log.error("AP: funding asset inconsistent, "
								+ m_TradeOffer.getExAsset() + " vs. "
								+ reply.getAsset());
					same = false;
				}
				if (m_TradeOffer.getExQuantity() != reply.getQuantity()) {
					Log.error("AP: funding amount inconsistent, "
								+ m_TradeOffer.getExQuantity() + " vs. "
								+ reply.getQuantity());
					same = false;
				}
				if (same) {
					// update status on offer
					m_TradeOffer.setStatus(TradeOffer.M_StatusFilled);
				}

				// record timestamp of escrow payout so it can be shown
				m_EscrowTimestamp = reply.getTimestamp();

				// de-select in offer table
				otm.deselectOffer(m_TradeOffer);

				// go to phase 2 to show post-acceptance verbiage
				m_AcceptMode = false;
				// use separate thread since we are an event thread here
				Runnable goPhase2 = new Runnable() {
					public void run() {
						m_TabManager.queueTransition(
														new Integer(XTabManager.M_GO_ACCEPT));
						m_TabManager.makeTransition();
					}
				};
				SwingUtilities.invokeLater(goPhase2);
			}
			else {
				Log.error("Successful accept funding but no trade offer set!");
			}
		}
	}

	/**
	 * method to determine whether this pane is presently doing an escrow payment
	 * @return true if yes, false if no
	 */
	public boolean doingEscrow() { return m_DoingEscrow; }

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Accept an existing funded trade offer");

		HomePane whp = m_ParentHome.getWalletHome();
		if (m_AcceptMode) {
			// abandon any uncompleted escrow payment for a previous offer
			m_DoingEscrow = false;

			// make sure we have a trade offer indicated, or we can't build screen
			if (m_TradeOffer == null) {
				Log.error("No trade offer set, cannot prep display form");
				showError("cannot accept offer", "none selected",
							"select an offer from the Search results screen");
				return;
			}

			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(20));     // spacer
			m_ButtonPane.add(m_BackButton);
			m_ButtonPane.add(Box.createHorizontalStrut(235));    // spacer
			m_ButtonPane.add(m_AcceptButton);
			m_AckButton.setVisible(false);
			m_BackButton.setVisible(true);
			m_AcceptButton.setVisible(true);
			m_AcceptButton.setToolTipText("Accept this trade offer");

			// build accept entry form
			m_FormPane.removeAll();
			m_GBC.gridheight = 15;
			m_FormPane.add(Box.createVerticalGlue(), m_GBC);
			final Color labColor = UIManager.getColor("standardGray");
			final Dimension rowSize = new Dimension(450, 25);
			
			// header label
			JLabel headLab = new JLabel("Please indicate how you would like to "
											+ "accept this offer:", SwingConstants.CENTER);
			headLab.setFont(whp.M_LabelFont);
			headLab.setForeground(labColor);
			m_FormPane.add(headLab, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(10), m_GBC);

			// offer Id
			JPanel oIdPane = new JPanel(new GridLayout(1, 2));
			oIdPane.setPreferredSize(rowSize);
			oIdPane.setBorder(null);
			JLabel oIdLab = new JLabel("Offer Id: ", SwingConstants.LEFT);
			oIdLab.setFont(whp.M_LabelFont);
			oIdLab.setForeground(labColor);
			oIdPane.add(oIdLab);
			JLabel oId = new JLabel(m_TradeOffer.getOfferId() + " ",
											SwingConstants.CENTER);
			oId.setFont(whp.M_LabelFont);
			oId.setForeground(labColor);
			oIdPane.add(oId);
			m_FormPane.add(oIdPane, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

			// offered asset
			JPanel offeredPane = new JPanel(new GridLayout(1, 2));
			offeredPane.setPreferredSize(rowSize);
			offeredPane.setBorder(null);
			JLabel offLab = new JLabel("Offered asset: ", SwingConstants.LEFT);
			offLab.setFont(whp.M_LabelFont);
			offLab.setForeground(labColor);
			offeredPane.add(offLab);
			JLabel offeredLab = new JLabel(m_TradeOffer.getOffered(),
													SwingConstants.CENTER);
			offeredLab.setFont(whp.M_LabelFont);
			offeredLab.setForeground(labColor);
			offeredLab.setToolTipText("Held in escrow by the SVX");
			offeredPane.add(offeredLab);
			m_FormPane.add(offeredPane, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

			// offered amount
			JPanel offAmtPane = new JPanel(new GridLayout(1, 2));
			offAmtPane.setBorder(null);
			offAmtPane.setPreferredSize(rowSize);
			JLabel amtLab = new JLabel("Amount offered: ", SwingConstants.LEFT);
			amtLab.setFont(whp.M_LabelFont);
			amtLab.setForeground(labColor);
			offAmtPane.add(amtLab);
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(0);
			JLabel quanLab = new JLabel(df.format(m_TradeOffer.getQuantity()),
												SwingConstants.CENTER);
			quanLab.setFont(whp.M_LabelFont);
			quanLab.setForeground(labColor);
			quanLab.setToolTipText("Amount you will receive");
			offAmtPane.add(quanLab);
			m_FormPane.add(offAmtPane, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

			// price ratio
			JPanel ratioPane = new JPanel(new GridLayout(1, 2));
			ratioPane.setPreferredSize(rowSize);
			ratioPane.setBorder(null);
			JLabel ratioLab = new JLabel("Price ratio: ", SwingConstants.LEFT);
			ratioLab.setFont(whp.M_LabelFont);
			ratioLab.setForeground(labColor);
			ratioPane.add(ratioLab);
			JLabel priceLab = new JLabel(df.format(m_TradeOffer.getPrice()),
												SwingConstants.CENTER);
			priceLab.setFont(whp.M_LabelFont);
			priceLab.setForeground(labColor);
			priceLab.setToolTipText("the ratio of Offered to Provided");
			ratioPane.add(priceLab);
			m_FormPane.add(ratioPane, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

			// wanted assets, in drop-down list
			JPanel wantedPane = new JPanel(new GridLayout(1, 3));
			wantedPane.setPreferredSize(rowSize);
			wantedPane.setBorder(null);
			JLabel wantLab = new JLabel("You provide: ", SwingConstants.LEFT);
			wantLab.setFont(whp.M_LabelFont);
			wantLab.setForeground(labColor);
			wantedPane.add(wantLab);
			wantedPane.add(Box.createHorizontalStrut(75));
			// rebuild available assets based on wallet contents
			VoucherTableModel vtm = whp.getVoucherModel();
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			m_ProvidedUI.removeAllItems();
			String wanted = m_TradeOffer.getWanted();
			String[] allWanted = wanted.split(":");
			for (int www = 0; www < allWanted.length; www++) {
				String wAsset = allWanted[www];
				m_ProvidedUI.addItem(wAsset);
			}
			wantedPane.add(m_ProvidedUI);
			m_FormPane.add(wantedPane, m_GBC);
			m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

			// acceptance amount
			JPanel acceptPane = new JPanel(new GridLayout(1, 2));
			acceptPane.setBorder(null);
			acceptPane.setPreferredSize(rowSize);
			JLabel pAmtLab = new JLabel("Amount you will pay: ",
												SwingConstants.LEFT);
			pAmtLab.setFont(whp.M_LabelFont);
			pAmtLab.setForeground(labColor);
			acceptPane.add(pAmtLab);
			// this has the side-effect of calling actionPerformed() and setting
			// the text of m_Required to the requisite amount
			m_ProvidedUI.setSelectedIndex(0);
			acceptPane.add(m_Required);
			m_FormPane.add(acceptPane, m_GBC);
			m_FormPane.add(Box.createVerticalGlue(), m_GBC);

			// place entire quasi-table UI into the main scroll area
			m_WorkArea.setViewportView(m_FormPane);
		}
		else {
			if (m_TradeOffer == null) {
				Log.error("Accept conf display with no trade offer!");
				showError("cannot display confirmation",
							"trade offer accepted, but error encountered displaying "
							+ "results");
				m_AcceptMode = true;
				return;
			}

			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(180));     // spacer
			m_ButtonPane.add(m_AckButton);
			m_AckButton.setVisible(true);
			m_AcceptButton.setVisible(false);
			m_BackButton.setVisible(false);

			// display language about offer handling, place in m_WorkArea
			JPanel confPane = new JPanel(new GridBagLayout());
			confPane.setBorder(null);
			confPane.setPreferredSize(new Dimension(470, 250));
			m_GBC.gridheight = 5;
			confPane.add(Box.createVerticalGlue(), m_GBC);
			JLabel confLab = new JLabel("Trade Order Now Filled and Settled",
												SwingConstants.CENTER);
			final Color labColor = UIManager.getColor("standardGray");
			confLab.setFont(whp.M_LabelFont);
			confLab.setForeground(labColor);
			confPane.add(confLab, m_GBC);
			confPane.add(Box.createVerticalStrut(5), m_GBC);
			StringBuilder confTxt = new StringBuilder(128);
			confTxt.append("Trade offer ID # " + m_TradeOffer.getOfferId());
			confTxt.append(" has been filled at the SVX, and your side of the ");
			confTxt.append("escrow\nhas now been funded.\n\n");
			confTxt.append("Now we will go pick up the incoming voucher payment.");
			if (m_EscrowTimestamp != null) {
				confTxt.append("\nThis was sent by the SVX at:\n\n");
				confTxt.append(m_EscrowTimestamp.toString() + ".\n");
			}
			else {
				Log.error("No escrow payout timestamp found");
			}
			JTextArea confBlurb = new JTextArea(confTxt.toString(), 8, 40);
			confBlurb.setEditable(false);
			confPane.add(confBlurb, m_GBC);
			confPane.add(Box.createVerticalGlue(), m_GBC);
			m_WorkArea.setViewportView(confPane);
		}
		revalidate();
	}

	// implement ActionListener
	/**
	 * method to update required amount based on selection of provided asset
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		// when m_ProvidedUI changes, update m_Required appropriately
		Object evSrc = ae.getSource();
		if (evSrc instanceof JComboBox) {
			String provAsset = (String) m_ProvidedUI.getSelectedItem();
			if (provAsset == null || provAsset.isEmpty()) {
				//Log.error("No wanted asset items to select");
				return;
			}
			if (m_TradeOffer == null) {
				Log.error("No selected offer for computing accept amount");
				return;
			}
			m_Provided = provAsset;

			// compute and show amount required to accept this
			double reqAmt = computeReqAmount(provAsset);
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(3);
			m_Required.setText(df.format(reqAmt));

			// check that user has this much
			HomePane whp = m_ParentHome.getWalletHome();
			VoucherTableModel vtm = whp.getVoucherModel();
			double avail = vtm.getVoucherAssetSum(m_Provided);
			String unitLab = "In " + XMLVoucher.getDisplayUnits(m_Provided);
			if (avail < reqAmt) {
				m_Required.setToolTipText(unitLab
												+ " (you do not have this much available)");
				m_AcceptButton.setEnabled(false);
			}
			else {
				m_Required.setToolTipText(unitLab + " (you can accept this offer)");
				m_AcceptButton.setEnabled(true);
			}
		}
	}

	/**
	 * utility method to compute the amount required to accept an offer given
	 * a chosen asset type, including the SVX commission.
	 * implied param: m_TradeOffer, the offer to be accepted
	 * @param asset the asset to be provided
	 * @return the amount which would be required, or 0 on failure
	 */
	private double computeReqAmount(String asset) {
		// compute amount of provided asset required, including commission
		double ret = 0.0;
		AssetPriceCache pricer = m_Plugin.getPricingModule();
		String offeredLookup
			= VPAssetConfig.getBasePriceLookupUnits(m_TradeOffer.getOffered());
		String providedLookup = VPAssetConfig.getBasePriceLookupUnits(asset);
		Double relPrice = pricer.getRelativePrice(offeredLookup, providedLookup);
		if (relPrice == null || relPrice.doubleValue() == 0.0) {
			return ret;
		}
		double rawAmt = m_TradeOffer.getQuantity() * relPrice.doubleValue();
		// adjust by ratio, if this isn't 1.0
		rawAmt /= m_TradeOffer.getPrice();

		/* Tack on commission according to SVX's asset config.  This will be
		 * the greater of 25 tokens for this asset or the indicated percentage.
		 */
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		SVXAsset delivConf = svxConfig.get(asset);
		double commission = rawAmt * (delivConf.m_ExchangeCommission / 100.0);
		double twentyFiveTokens = VPAssetConfig.get25TokensWorth(asset);
		ret = rawAmt + Math.max(twentyFiveTokens, commission);
		return ret;
	}

	// END methods
}
