/*
 * last modified---
 * 	03-22-17/jet/ parsing correction and better prompts (SVSpark merge)
 * 	07-17-15/jet/ use lookup units for relative prices
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	03-12-14/jet/ compute and show commissions either as max of % or 25 tokens
 * 	09-16-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to enter orders to sell vouchers for external payments
 */

package org.silentvault.client.ui.svx;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;
import org.silentvault.client.ui.wallet.HomePane;
import org.silentvault.client.ui.wallet.WTabManager;
import org.silentvault.client.ui.wallet.PaymentPane;
import org.silentvault.client.ui.wallet.ConfirmPane;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Extends WorkPane to provide a screen for entering sell orders, sending
 * them to the SVX, and then funding them using a voucher payment.
 *
 * @author Justin Turrell
 */
public final class SellPane extends WorkPane {
	// BEGIN data members
	/**
	 * action to initiate an entry of the indicated sell offer
	 */
	private AbstractAction			m_EnterAction;

	/**
	 * button to go with enter action
	 */
	private JButton					m_EnterButton;

	/**
	 * action to acknowledge order resolution language and reset
	 */
	private AbstractAction			m_AckAction;

	/**
	 * button to go with acknowledge action
	 */
	private JButton					m_AckButton;

	/**
	 * mode toggle: if true, display order entry form; if false show confirmation
	 */
	private boolean					m_EnterMode;

	/**
	 * control to select asset being sold by user
	 */
	private JComboBox<String>		m_OrderedUI;

	/**
	 * control to select external payment mechanism selected by user
	 */
	private JComboBox<String>		m_ExtPaymentUI;

	/**
	 * box for entering desired quantity
	 */
	private JTextField				m_AmountUI;

	/**
	 * box for entering external wallet address for payments
	 */
	private JTextField				m_PaymentUI;

	/**
	 * value specified via m_AmountUI
	 */
	private double						m_Amount;

	/**
	 * calculated commission amount
	 */
	private double						m_Commission;

	/**
	 * label describing commission rate (adjusted by m_OrderedUI)
	 */
	private JLabel						m_CommRate;

	/**
	 * total voucher payment to be made to fund order (includes commission)
	 */
	private double						m_PaymentAmount;

	/**
	 * label describing the total payment amount (m_PaymentAmount)
	 */
	private JLabel						m_TotalOwed;

	/**
	 * panel to contain order entry form
	 */
	private JPanel						m_FormPane;

	/**
	 * order ID entered and assigned by SVX
	 */
	private int							m_OrderId;

	/**
	 * (escrow) wallet identifier where SVX wants voucher payment sent
	 */
	private String						m_WalletId;

	/**
	 * SVX pickup reply giving external spend data
	 */
	private SVXBlock					m_SaleReply;

	/**
	 * formatter for amounts
	 */
	private DecimalFormat	m_AmtFormatter;

	/**
	 * inner class to provide a DocumentListener for the amount field
	 */
	final class AmtDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// interpret a '.' or ',' as 0 if it's the first character added
			int curLen = doc.getLength();
			if (curLen == 1) {
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad insert doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_AmountUI, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_AmountUI, "0,");
					return;
				}
			}

			// disallow letters etc. altogether
			int changeAt = de.getOffset();
			try {
				curTxt = doc.getText(0, doc.getLength());
			}
			catch (BadLocationException ble) {
				Log.error("Cannot undo bad value edit", ble);
				return;
			}
			char chg = curTxt.charAt(changeAt);
			if (!Character.isDigit(chg) && chg != '.' && chg != ',') {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_AmountUI, modTxt);
				return;
			}

			// otherwise just parse
			if (parseValue()) {
				updateValueDisplays();
			}
			else {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_AmountUI, modTxt);
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// if the only thing left is a '.' or ',', force it to a 0
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 1) {
				String curTxt = "";
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad remove doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_AmountUI, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_AmountUI, "0,");
					return;
				}
			}

			// otherwise just parse
			if (parseValue()) {
				updateValueDisplays();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_AmountUI
		 *	@return true on success, false on failure
		 */
		private boolean parseValue() {
			// examine what's in the value entry field
			String enteredVal = m_AmountUI.getText();
			if (enteredVal.isEmpty() || enteredVal.equals("0.00")
				 || enteredVal.equals("0") || enteredVal.equals("0,00"))
			{
				return true;
			}

			// attempt to parse
			double vvv = 0.0;
			try {
				Number val = m_AmtFormatter.parse(enteredVal);
				vvv = val.doubleValue();
			}
			catch (ParseException pe) {
				return false;
			}
			if (vvv < 0.0) {
				return false;
			}
			return true;
		}
	}

	/**
	 * document listener for dynamic updates of m_AmountUI
	 */
	private AmtDocumentListener	m_AmountListener;

	/**
	 * whether an undo mod is ongoing
	 */
	private boolean					m_DoingUndo;

	/**
	 * inner class to modify a text component from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		/**
		 * the text component to be modified
		 */
		JTextComponent			m_ModComponent;

		/**
		 * its new contents
		 */
		String					m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());

			// update value displays
			updateValueDisplays();
		}
	}

	/**
	 * flag to indicate (to WTabManager) that we're making an escrow payment
	 */
	private boolean					m_DoingEscrow;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public SellPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);
		m_EnterMode = true;
		m_WalletId = "";
		m_AmtFormatter = new DecimalFormat("#,###,##0.0###");

		// define Enter action
		m_EnterAction = new AbstractAction("Enter Sale") {
			public void actionPerformed(ActionEvent ae) {
				// error check specified inputs
				String ordered = (String) m_OrderedUI.getSelectedItem();
				if (ordered == null || ordered.isEmpty()) {
					showError("no asset to be sold selected",
								"select a voucher asset type from the drop-down list");
					return;
				}
				String amtText = m_AmountUI.getText();
				SVXAsset offConf = m_ParentHome.getSVXConfig().get(ordered);
				if (m_Amount < offConf.m_MinPurchase) {
					showError("illegal offer amount",
								"enter " + offConf.m_MinPurchase + " or more");
					return;
				}

				/* Recompute commission based on selection and amount.  This will be
				 * 25 tokens' worth or the percentage, whichever is greater.
				 */
				double commission = (offConf.m_PurchCommission / 100.0) * m_Amount;
				double twentyFiveTokens = VPAssetConfig.get25TokensWorth(ordered);
				m_Commission = Math.max(commission, twentyFiveTokens);
				AssetPriceCache apc = m_Plugin.getPricingModule();
				String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
				String lUnits = VPAssetConfig.getBasePriceLookupUnits(ordered);
				Double relPrice = apc.getRelativePrice(lUnits, asPayment);
				if (relPrice == null) {
					relPrice = new Double(1.0);
				}
				m_PaymentAmount
					= (m_Amount + m_Commission) * relPrice.doubleValue();
				// make sure user has this much
				HomePane whp = m_ParentHome.getWalletHome();
				VoucherTableModel vtm = whp.getVoucherModel();
				double avail = vtm.getVoucherAssetSum(ordered);
				if (m_PaymentAmount > avail) {
					showError("illegal order amount", "you do not have this much "
								+ ordered + " in your wallet");
					return;
				}

				// make sure they've specified a refund address
				String refHash = m_PaymentUI.getText();
				if (refHash.isEmpty()) {
					showError("missing payment address",
								"enter your " + asPayment + " wallet address");
					return;
				}

				// send REQ_voucher_sale to SVX
				sendEnterSale();
			}
		};
		// define associated button
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_EnterButton = new JButton(m_EnterAction);
		m_EnterButton.setFont(buttonFont);
		m_EnterButton.setToolTipText("Submit the sale order");

		// define order acknowledgement action
		m_AckAction = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent ae) {
				// reset for next order entry
				m_EnterMode = true;
				m_OrderId = 0;
				m_Amount = m_Commission = m_PaymentAmount = 0.0;
				m_AmountUI.setText("");
				m_WalletId = "";
				m_DoingEscrow = false;
				m_EnterButton.setEnabled(true);
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SELL));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_AckButton = new JButton(m_AckAction);
		m_AckButton.setFont(buttonFont);
		m_AckButton.setToolTipText("Return to sale order entry screen");

		// populate button bar (replaced in prepDisplay())
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(150));		// spacer
		m_ButtonPane.add(m_EnterButton);
		m_ButtonPane.add(Box.createHorizontalStrut(30));		// spacer
		m_ButtonPane.add(m_AckButton);

		// create controls for order entry form
		m_FormPane = new JPanel(new GridBagLayout());
		m_FormPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.gridheight = 16;
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");
		HomePane whp = m_ParentHome.getWalletHome();
		JLabel headLab = new JLabel("Enter an order to sell a voucher as "
											+ "follows:", SwingConstants.LEFT);
		headLab.setFont(whp.M_LabelFont);
		headLab.setForeground(labColor);
		m_FormPane.add(headLab, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(25), m_GBC);

		// build list of repuchased assets selectable
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		Vector<String> trAssets = new Vector<String>(svxConfig.size());
		for (String asset : svxConfig.keySet()) {
			SVXAsset aConf = svxConfig.get(asset);
			if (aConf.m_MinPurchase > 0.0) {
				trAssets.add(aConf.m_AssetName);
			}
		}

		// based on the list, create label and drop-down for ordered assets
		final Dimension rowSize = new Dimension(450, 25);
		final Dimension boxSz = new Dimension(50, 25);
		JPanel offPane = new JPanel(new GridLayout(1, 3));
		offPane.setBorder(null);
		offPane.setPreferredSize(rowSize);
		// ordered asset list: label
		JLabel ordLab = new JLabel("Selling: ", SwingConstants.CENTER);
		ordLab.setFont(whp.M_LabelFont);
		ordLab.setForeground(labColor);
		ordLab.setToolTipText("Voucher asset you wish to sell");
		offPane.add(ordLab);
		// NB: prepDisplay() will update the asset list to match wallet contents
		m_OrderedUI = new JComboBox<String>(trAssets);
		m_OrderedUI.addActionListener(this);
		m_OrderedUI.setPreferredSize(boxSz);
		m_OrderedUI.setToolTipText("Select the asset you wish to sell");
		ordLab.setLabelFor(m_OrderedUI);
		offPane.add(m_OrderedUI);
		offPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(offPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		/* init controls for available external currencies; note this is populated
		 * dynamically in the listener based on the selected asset being purchased
		 */
		JPanel extPayPane = new JPanel(new GridLayout(1, 3));
		extPayPane.setBorder(null);
		extPayPane.setPreferredSize(rowSize);
		JLabel payLab = new JLabel("For: ", SwingConstants.CENTER);
		payLab.setFont(whp.M_LabelFont);
		payLab.setForeground(labColor);
		payLab.setToolTipText("External currency you want to be paid in");
		extPayPane.add(payLab);
		m_ExtPaymentUI = new JComboBox<String>();
		m_ExtPaymentUI.setPreferredSize(boxSz);
		m_ExtPaymentUI.setToolTipText("Select the currency you wish to "
												+ "receive as payment for your sale");
		m_ExtPaymentUI.addActionListener(this);
		payLab.setLabelFor(m_ExtPaymentUI);
		extPayPane.add(m_ExtPaymentUI);
		extPayPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(extPayPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add control for entering the amount on order
		JPanel amtPane = new JPanel(new GridLayout(1, 3));
		amtPane.setPreferredSize(rowSize);
		m_AmountUI = new JTextField(6);
		m_AmountUI.setDropTarget(null);
		String selAsset = (String) m_OrderedUI.getSelectedItem();
		String dUnits = XMLVoucher.getDisplayUnits(selAsset);
		// update tooltip with max amount
		VoucherTableModel vtm = whp.getVoucherModel();
		double avail = vtm.getVoucherAssetSum(selAsset);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_AmountUI.setToolTipText("<html>Enter the amount (in " + dUnits
											+ ") of " + selAsset + " you wish to sell."
											+ "<br/>You have up to " + df.format(avail)
											+ " available.</html>");
		m_AmountUI.setPreferredSize(boxSz);
		m_AmountUI.setText("0.0");
		m_AmountListener = new AmtDocumentListener();
		m_AmountUI.getDocument().addDocumentListener(m_AmountListener);

		JLabel amtLab = new JLabel("Amount: ", SwingConstants.CENTER);
		amtLab.setFont(whp.M_LabelFont);
		amtLab.setForeground(labColor);
		amtLab.setLabelFor(m_AmountUI);
		amtLab.setToolTipText("Quantity you wish to sell");
		amtPane.add(amtLab);
		amtPane.add(m_AmountUI);
		amtPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(amtPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add labels to show commission which SVX will charge
		JPanel commPane = new JPanel(new GridLayout(1, 3));
		commPane.setBorder(null);
		commPane.setPreferredSize(rowSize);
		JLabel comLab = new JLabel("Commission: ", SwingConstants.CENTER);
		comLab.setFont(whp.M_LabelFont);
		comLab.setForeground(labColor);
		comLab.setToolTipText("<html>The repurchase commission charged by<br/> "
									+ "the SVX for this voucher asset</html>");
		commPane.add(comLab);
		m_CommRate = new JLabel("0.0", SwingConstants.LEFT);
		m_CommRate.setFont(whp.M_LabelFont);
		m_CommRate.setForeground(labColor);
		m_CommRate.setToolTipText("The commission amount");
		// NB: m_CommRate will be updated with changes in m_OrderedUI
		commPane.add(m_CommRate);
		commPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(commPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add labels to show the calculated external payment amount
		JPanel totPane = new JPanel(new GridLayout(1, 3));
		totPane.setBorder(null);
		totPane.setPreferredSize(rowSize);
		JLabel costLab = new JLabel("Net sale: ", SwingConstants.CENTER);
		costLab.setFont(whp.M_LabelFont);
		costLab.setForeground(labColor);
		costLab.setToolTipText("<html>The total cost of your order in<br/>"
									+ "the selected voucher asset type</html>");
		totPane.add(costLab);
		// NB: m_TotalOwed will be updated with changes in m_OrderedUI
		m_TotalOwed = new JLabel("0.0", SwingConstants.LEFT);
		m_TotalOwed.setFont(whp.M_LabelFont);
		m_TotalOwed.setForeground(labColor);
		m_TotalOwed.setToolTipText("The total voucher cost of your order");
		totPane.add(m_TotalOwed);
		totPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(totPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// wallet hash entry
		JPanel refPane = new JPanel();
		BoxLayout rbl = new BoxLayout(refPane, BoxLayout.X_AXIS);
		refPane.setLayout(rbl);
		refPane.setBorder(null);
		refPane.setPreferredSize(rowSize);
		JLabel refLab = new JLabel("Wallet hash: ", SwingConstants.CENTER);
		refLab.setFont(whp.M_LabelFont);
		refLab.setForeground(labColor);
		refLab.setToolTipText("The external wallet address (hash) where you "
									+ "want your payment to be sent");
		// add control for entering payment address
		m_PaymentUI = new JTextField(40);
		m_PaymentUI.setDropTarget(null);
		m_PaymentUI.setPreferredSize(new Dimension(150, 25));
		m_PaymentUI.setToolTipText("<html>The <i>external</i> wallet address "
											+ "hash where "
											+ "you want your sale proceeds sent.<br/>"
											+ "This address will be paid by the SVX once "
											+ "your voucher payment has been received."
											+ "<br/>Note this is <b>not</b> your "
											+ "SilentVault wallet ID.</html>");
		refLab.setLabelFor(m_PaymentUI);
		refPane.add(refLab);
		refPane.add(Box.createHorizontalStrut(30));
		refPane.add(m_PaymentUI);
		m_FormPane.add(refPane, m_GBC);
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);

		// set the user instructions
		m_UserInstructs = "<html>Enter your sale order below.<br/><br/>Note "
								+ "that you must fulfill the indicated<br/>voucher "
								+ "payment, and this must be<br/>confirmed, before "
								+ "your proceeds will be<br/>sent to your external "
								+ "wallet.<br/><br/>"
								+ "Unfulfilled sale orders expire 24 hours<br/>after "
								+ "entry.</html>";
		m_WorkArea.setBorder(null);
	}

	/**
	 * helper method to queue up a modification to a text component from within a
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	/**
	 * perform a (blocking) query to the SVX entering the order in the database
	 */
	private void sendEnterSale() {
		// send a REQ_voucher_sale to register the sale order at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock mkOffer = new SVXClientBlock();
		mkOffer.setOpcode("REQ_voucher_sale");
		mkOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());

		// set purchase data based on user entries
		String ordered = (String) m_OrderedUI.getSelectedItem();
		mkOffer.setAsset(ordered);
		mkOffer.setQuantity(m_Amount);
		String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
		mkOffer.setCurrency(asPayment);
		mkOffer.setExtWallet(m_PaymentUI.getText());

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(mkOffer);
		setCursor(null);
		// NB: reply comes synchronously to processSaleTerms() below
	}

	/**
	 * record the reply to our query of extant offers
	 * @param reply the reply received from the SVX
	 */
	public void processSaleTerms(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer search reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_sale_terms")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			m_TabManager.showError("SVX sale entry failed", reply.getErrMsg());
			return;
		}

		// record values returned from SVX
		m_OrderId = reply.getOrderId();
		m_WalletId = reply.getWalletId();

		// check echoes
		if (reply.getQuantity() != m_Amount) {
			m_Amount = reply.getQuantity();
		}
		if (m_Commission != reply.getCommission()) {
			m_Commission = reply.getCommission();
		}
		// round to 4 digits before comparing
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);
		double expected = m_PaymentAmount;
		String amtStr = df.format(m_PaymentAmount);
		try {
			Number quan = df.parse(amtStr);
			expected = quan.doubleValue();
		}
		catch (ParseException pe) {
			Log.error("Error parsing 4-digit total, " + amtStr, pe);
		}
		if (expected != reply.getTotal()) {
			// popup?
		/*
			Log.warning("Total payment amount changed, " + reply.getTotal()
							+ " vs. " + expected + " expected");
		 */
			m_PaymentAmount = reply.getTotal();
		}

		// build a (decoded) WebMessage object to pass data to the PaymentPane
		WebMessage webPass = new WebMessage();
		webPass.setAmount(m_PaymentAmount);
		webPass.setVSnumber(m_WalletId);
		String asset = (String) m_OrderedUI.getSelectedItem();
		// determine issuer for this asset by examining a voucher of the type
		HomePane whp = m_ParentHome.getWalletHome();
		VoucherTableModel vtm = whp.getVoucherModel();
		XMLVoucher firstV = vtm.getFirstAssetVoucher(asset);
		if (firstV == null) {
			// should not happen, since we vetted asset balance already
			m_TabManager.showError("Cannot fund sale order",
											"no vouchers found for asset " + asset);
			Log.error("No vouchers found to fund order " + m_OrderId + " with "
						+ asset);
			return;
		}
		vtm.selectIssuer(firstV.getIssuer());
		webPass.setIssuer(firstV.getIssuer());
		webPass.setCurrency(XMLVoucher.getDisplayUnits(asset));
		webPass.setBaggageFields("orderId");
		webPass.addOrderRecord("orderId", new Integer(m_OrderId).toString());
		webPass.setDecoded(true);
		m_Plugin.getWebListener().addWebMessage(webPass);
		final WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		pp.setSelectedSiteMsg(webPass);

		// go to prefab wallet tab payment screen
		m_EnterMode = false;
		Runnable goPay = new Runnable() {
			public void run() {
				// disable button to prevent multiple entry
				m_EnterButton.setEnabled(false);
				m_EnterButton.setToolTipText("<html>Your trade has been entered."
													+ "<br/>Please make your voucher "
													+ "payment in the Wallet tab.</html>");
				revalidate();
				wtm.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
				m_Plugin.setSelectedTab(wtm);
				wtm.makeTransition();
			}
		};
		m_DoingEscrow = true;
		SwingUtilities.invokeLater(goPay);
	}

	/**
	 * query whether we're doing an escrow payment or not
	 * @return true if in the middle of conducting a payment to escrow
	 */
	public boolean doingEscrow() { return m_DoingEscrow; }

	/**
	 * make funding confirmation request
	 */
	public void confirmEscrowFunding() {
		// send a REQ_voucher_sold to pickup the escrow payment at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock fundOrder = new SVXClientBlock();
		fundOrder.setOpcode("REQ_voucher_sold");
		fundOrder.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		fundOrder.setOrderId(m_OrderId);

		// return focus to exchange tab
		m_Plugin.setSelectedTab(m_TabManager);

		// put up info saying what we are doing (on SVX tab)
		WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		final StringBuilder popTxt = new StringBuilder(256);
		// obtain last confirmation recorded by ConfirmPane off the DHT
		XMLConfirmPayment conf = pp.getLatestConfirmation();
		if (conf != null) {
			/* we deferred showing payment success, and so must recapitulate
			 * what the ConfirmPane would have displayed as popup text.
			 */
			popTxt.append("Your payment of " + conf.getQuantity());
			popTxt.append(" " + conf.getUnits() + " of " + conf.getAsset());
			if (!conf.getAsset().equals(conf.getIssuer())) {
				popTxt.append(" at " + conf.getIssuer());
			}
			popTxt.append("\nwas sent to ");
			popTxt.append(conf.getPayee() + " successfully.");
			// show that change voucher was received if there was one
			if (conf.getVoucher() != null) {
				popTxt.append("\nYou have received a voucher back as change.");
			}
		}
		else {
			Log.error("Unexpected: no XMLConfirmPayment found in PaymentPane");
			// show something generic
			popTxt.append("Your escrow payment to fund order ID ");
			popTxt.append(fundOrder.getOrderId() + " succeeded.");
		}

		// add generic indication that we're now telling the SVX to pick it up
		popTxt.append("\n\nThe SVX is now processing the voucher payment.  ");
		popTxt.append("This may take a few moments.\nPlease wait for the ");
		popTxt.append("confirmation screen, read it carefully, and acknowledge.");

		// for safety, do in separate thread while allowing message to proceed
		Runnable notice = new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
														"Sale Payment Completed",
														JOptionPane.INFORMATION_MESSAGE);
			}
		};
		SwingUtilities.invokeLater(notice);

		// reset ConfirmPane to default behavior
		ConfirmPane cp = wtm.getConfirmPane();
		cp.setDeferPopup(false);

		// send message to SVX to trigger pickup
		setCursor(M_WaitCursor);
		m_SaleReply = null;
		svxIface.sendRequest(fundOrder);
		setCursor(null);
		// NB: reply comes synchronously to processFundingReply() below
	}

	/**
	 * record the results of an order funding attempt (post-payment)
	 * @param reply the reply received from the SVX
	 */
	public void processFundingReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX order entry reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_voucher_sold")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		m_DoingEscrow = false;
		if (reply.isError()) {
			m_TabManager.showError("SVX order funding failed", reply.getErrMsg());
			// reset
			m_AckButton.doClick();
			return;
		}

		// record for use in showing confirmation
		m_SaleReply = reply;

		// go to confirmation
		m_EnterMode = false;
		Runnable goConf = new Runnable() {
			public void run() {
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SELL));
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goConf);
	}

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Enter a voucher asset sale order");

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		HomePane whp = m_ParentHome.getWalletHome();

		if (m_EnterMode) {
			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(185));		// spacer
			m_ButtonPane.add(m_EnterButton);
			m_AckButton.setVisible(false);
			m_EnterButton.setVisible(true);
			m_EnterButton.setToolTipText("Submit the sale order");

			// rebuild available assets based on wallet contents
			VoucherTableModel vtm = whp.getVoucherModel();
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			m_OrderedUI.removeAllItems();
			for (String tAsset : svxConfig.keySet()) {
				SVXAsset tConf = svxConfig.get(tAsset);
				double sum = vtm.getVoucherAssetSum(tAsset);
				if (sum > 0.0 && tConf.m_MinPurchase > 0.0) {
					m_OrderedUI.addItem(tAsset);
				}
			}

			// now take account of XMenuPane's drop-down buy asset selection
			XMenuPane xmp = m_TabManager.getMenuPane();
			String asset = xmp.getBuyAsset();
			String dUnits = "";
			if (vtm.getVoucherAssetSum(asset) > 0.0) {
				// NB: does nothing if this asset is not in our allowed list
				m_OrderedUI.setSelectedItem(asset);
			}
			else if (m_OrderedUI.getItemCount() > 0) {
				// default to first
				m_OrderedUI.setSelectedIndex(0);
				asset = m_OrderedUI.getItemAt(0);
			}
			dUnits = XMLVoucher.getDisplayUnits(asset);

			// based on the selected asset, obtain list of external payment types
			SVXAsset aConf = svxConfig.get(asset);
			String[] accepted = aConf.m_PurchCurrencies.split(":");
			m_ExtPaymentUI.removeActionListener(this);
			m_ExtPaymentUI.removeAllItems();
			for (int aaa = 0; aaa < accepted.length; aaa++) {
				m_ExtPaymentUI.addItem(accepted[aaa]);
			}
			String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
			m_ExtPaymentUI.addActionListener(this);

			/* Recompute commission based on selection and amount.  This will be
			 * 25 tokens' worth or the percentage, whichever is greater.
			 */
			double commission = (aConf.m_PurchCommission / 100.0) * m_Amount;
			double twentyFiveTokens = VPAssetConfig.get25TokensWorth(asset);
			String rate = "";
			if (commission > twentyFiveTokens) {
				m_Commission = commission;
				rate = "@ " + df.format(aConf.m_PurchCommission) + " %)";
			}
			else {
				if (m_Amount <= 0.0) {
					m_Commission = 0.0;
					rate = "none)";
				}
				else {
					m_Commission = twentyFiveTokens;
					rate = " = 25 tokens)";
				}
			}
			m_CommRate.setText(df.format(m_Commission));
			m_CommRate.setToolTipText("The commission amount (" + rate);
			AssetPriceCache apc = m_Plugin.getPricingModule();
			String lUnits = VPAssetConfig.getBasePriceLookupUnits(asset);
			Double relPrice = apc.getRelativePrice(lUnits, asPayment);
			if (relPrice == null) {
				relPrice = new Double(1.0);
			}
			m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
			m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
			m_TotalOwed.setToolTipText("<html>The total cost of your order in "
												+ asset + " at "
												+ df.format(relPrice.doubleValue())
												+ " per " + dUnits + "</html>");

			// place entire quasi-table UI into the main scroll area
			m_WorkArea.setViewportView(m_FormPane);
		}
		else {
			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(185));		// spacer
			m_ButtonPane.add(m_AckButton);
			m_AckButton.setVisible(true);
			m_EnterButton.setVisible(false);

			// display language about offer handling, place in m_WorkArea
			JPanel confPane = new JPanel(new GridBagLayout());
			confPane.setBorder(null);
			confPane.setPreferredSize(new Dimension(470, 250));
			m_GBC.gridheight = 5;
			confPane.add(Box.createVerticalGlue(), m_GBC);
			JLabel confLab = new JLabel("Sale Order Now Fulfilled",
												SwingConstants.CENTER);
			final Color labColor = UIManager.getColor("standardGray");
			confLab.setFont(whp.M_LabelFont);
			confLab.setForeground(labColor);
			confPane.add(confLab, m_GBC);
			confPane.add(Box.createVerticalStrut(5), m_GBC);
			StringBuilder confTxt = new StringBuilder(512);
			String vehicle = (String) m_ExtPaymentUI.getSelectedItem();
			confTxt.append("Thank you for your sale order, ID # " + m_OrderId);
			confTxt.append(".\nYour voucher payment for this order has been ");
			confTxt.append("received.\n\n");
			confTxt.append("A payment to your " + vehicle + " address hash:\n\n");
			confTxt.append(m_PaymentUI.getText() + "\n\nhas been made in the ");
			confTxt.append("amount of " + m_SaleReply.getQuantity() + " ");
			confTxt.append(vehicle + ".\n\n");
			confTxt.append("The transaction ID (txid) was:\n");
			/* This txid will be too long to fit on one line without truncation.
			 * Therefore, we break it up for display and show a continuation.
			 */
			String txid = m_SaleReply.getSpendId();
			int txidLen = txid.length();
			if (txidLen > 40) {
				int half = txidLen / 2;
				confTxt.append(txid.substring(0, half) + "\n");
				confTxt.append(txid.substring(half));
			}
			else {
				confTxt.append(txid);
			}
			JTextArea confBlurb = new JTextArea(confTxt.toString(), 10, 40);
			confBlurb.setDropTarget(null);
			confBlurb.setEditable(false);
			confPane.add(confBlurb, m_GBC);
			confPane.add(Box.createVerticalGlue(), m_GBC);
			m_WorkArea.setViewportView(confPane);

			if (m_Plugin.isTestNetwork()) {
				final StringBuilder popTxt = new StringBuilder(128);
				popTxt.append("Because this is a test environment, you did not \n");
				popTxt.append("get an actual bitcoin/litecoin payment.\n\n");
				popTxt.append("Your voucher was simply de-issued automatically.\n");
				popTxt.append("If you'd like to replace it, use the Buy screen.\n");
				Runnable notice = new Runnable() {
					public void run() {
						JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
																"Voucher Sale Completed",
															JOptionPane.INFORMATION_MESSAGE);
					}
				};
				SwingUtilities.invokeLater(notice);
			}
		}
		revalidate();
	}

	/**
	 * update the text of the labels which show values dependent on inputs
	 */
	private void updateValueDisplays() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		String txtAmt = m_AmountUI.getText();
		if (txtAmt.isEmpty()) {
			txtAmt = "0.0";
		}

		double vvv = 0.0;
		try {
			Number val = m_AmtFormatter.parse(txtAmt);
			vvv = val.doubleValue();
		}
		catch (ParseException pe) {
			Log.error("Pre-parsed value did not parse, " + txtAmt, pe);
			return;
		}
		m_Amount = vvv;
		String trAsset = (String) m_OrderedUI.getSelectedItem();
		if (trAsset == null || trAsset.isEmpty()) {
			return;
		}
		SVXAsset aConf = m_ParentHome.getSVXConfig().get(trAsset);
		String dUnits = XMLVoucher.getDisplayUnits(aConf.m_AssetName);
		String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		String lUnits = VPAssetConfig.getBasePriceLookupUnits(aConf.m_AssetName);
		Double relPrice = apc.getRelativePrice(lUnits, asPayment);
		if (relPrice == null) {
			relPrice = new Double(1.0);
		}
		/* Recompute commission based on selection and amount.  This will be
		 * 25 tokens' worth or the percentage, whichever is greater.
		 */
		double commission = (aConf.m_PurchCommission / 100.0) * m_Amount;
		double twentyFiveTokens = VPAssetConfig.get25TokensWorth(trAsset);
		String rate = "";
		if (commission > twentyFiveTokens) {
			m_Commission = commission;
			rate = "@ " + df.format(aConf.m_PurchCommission) + " %)";
		}
		else {
			if (m_Amount <= 0.0) {
				m_Commission = 0.0;
				rate = "none)";
			}
			else {
				m_Commission = twentyFiveTokens;
				rate = " = 25 tokens)";
			}
		}
		m_CommRate.setText(df.format(m_Commission));
		m_CommRate.setToolTipText("The commission amount (" + rate);
		m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
		m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
		m_TotalOwed.setToolTipText("<html>The total cost of your order in "
											+ trAsset + " at "
											+ df.format(relPrice.doubleValue()) + " per "
											+ dUnits + "</html>");
	}

	// implement ActionListener
	/**
	 * method to update displayed commission rate and max amount tooltips
	 * when an offered asset is selected
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		// when m_OrderedUI changes, update m_CommRate appropriately
		Object evSrc = ae.getSource();
		if (evSrc instanceof JComboBox) {
			JComboBox src = (JComboBox) evSrc;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(3);
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			String trAsset = (String) m_OrderedUI.getSelectedItem();
			if (trAsset == null || trAsset.isEmpty()) {
				return;
			}
			SVXAsset aConf = svxConfig.get(trAsset);
			if (aConf == null) {
				Log.error("Config for sold asset " + trAsset + " not found");
				return;
			}
			String dUnits = XMLVoucher.getDisplayUnits(trAsset);
			// update tooltip with max amount
			HomePane whp = m_ParentHome.getWalletHome();
			VoucherTableModel vtm = whp.getVoucherModel();
			double avail = vtm.getVoucherAssetSum(trAsset);
			m_AmountUI.setToolTipText("<html>Enter the amount (in " + dUnits
											+ ") of " + trAsset + " you wish to sell."
											+ "<br/>You have up to " + df.format(avail)
											+ " available.</html>");
			if (src == m_OrderedUI) {
				// recompute accepted payment assets
				String[] accepted = aConf.m_PurchCurrencies.split(":");
				// prevent recursion
				m_ExtPaymentUI.removeActionListener(this);
				m_ExtPaymentUI.removeAllItems();
				for (int aaa = 0; aaa < accepted.length; aaa++) {
					m_ExtPaymentUI.addItem(accepted[aaa]);
				}
				m_ExtPaymentUI.addActionListener(this);
			}
			String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
			AssetPriceCache apc = m_Plugin.getPricingModule();
			String lUnits = VPAssetConfig.getBasePriceLookupUnits(trAsset);
			Double relPrice = apc.getRelativePrice(lUnits, asPayment);
			if (relPrice == null) {
				relPrice = new Double(1.0);
			}

			/* Recompute commission based on selection and amount.  This will be
			 * 25 tokens' worth or the percentage, whichever is greater.
			 */
			double commission = (aConf.m_PurchCommission / 100.0) * m_Amount;
			double twentyFiveTokens = VPAssetConfig.get25TokensWorth(trAsset);
			String rate = "";
			if (commission > twentyFiveTokens) {
				m_Commission = commission;
				rate = "@ " + df.format(aConf.m_PurchCommission) + " %)";
			}
			else {
				if (m_Amount <= 0.0) {
					m_Commission = 0.0;
					rate = "none)";
				}
				else {
					m_Commission = twentyFiveTokens;
					rate = " = 25 tokens)";
				}
			}
			m_CommRate.setText(df.format(m_Commission));
			m_CommRate.setToolTipText("The commission amount (" + rate);
			m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
			m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
			m_TotalOwed.setToolTipText("<html>The total cost of your order in "
												+ trAsset + " at "
												+ df.format(relPrice.doubleValue())
												+ " per " + dUnits + "</html>");
		}
	}

	// END methods
}
