/*
 * last modified---
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	08-09-13/jet/ abort if fetching vouchers, tokens, or payments fails
 * 	06-05-13/jet/ SilentVault client version
 * 	01-23-13/kbw/ abort operation if safe contains any expired vouchers
 * 	12-18-12/kbw/ use getNPV() instead of baseUnitValue()
 * 	02-25-11/kbw/ allow for non-GRAMS vouchers
 * 	10-20-10/kbw/ make buttons conform to new layout guidelines
 * 	07-09-10/kbw/ disallow closure if vouchers from >1 Issuer are present
 * 	04-19-10/kbw/ integrate by doing pre-checks here
 * 	02-24-10/kbw/ stand-alone version extending JPanel
 * 	09-24-09/kbw/ further debugging of error paths
 * 	07-10-09/kbw/ debug
 * 	05-28-09/kbw/ new
 *
 * purpose---
 * 	provide a UI widget for closing a wallet
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.event.ActionEvent;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.security.*;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Arrays;
import java.util.concurrent.ConcurrentHashMap;
import java.text.DecimalFormat;
import java.text.NumberFormat;


/**
 * Extends JPanel to provide a screen for closing out a wallet.
 * There are two main phases to the closure: 1) initialization with data from
 * the old (closing) wallet, and determination of whether we need a successor VS
 * into which the stored value must be transferred; and 2) obtaining the user's
 * concurrence and signature on the transfer/closure.
 *
 * @author Kevin Wilkerson
 */
public final class CloseSafePane extends JPanel {
	// BEGIN data members

	// states/phases we can be in (possible values of m_CurrentPhase):
	/**
	 * obtaining existing safe data while creating new safe
	 */
	private final static byte	M_Phase1 = (byte)1;
	/**
	 * obtaining user consent
	 */
	private final static byte	M_Phase2 = (byte)2;
	/**
	 * performing the actual closure/value transfer
	 */
	private final static byte	M_Phase3 = (byte)3;

	/**
	 * the plugin/application to which we belong
	 */
	private WalletClient		m_Plugin;

	/**
	 * the secrets pertaining to the old safe
	 */
	private VsSecrets			m_OldVSsecrets;

	/**
	 * the new (successor) voucher safe number
	 */
	private String				m_NewVSnumber;

	/**
	 * the new (successor) VS hash index for its voucher/token folder on the SDS
	 */
	private char[]				m_NewVSindex;

	/**
	 * the new (successor) VS read-write voucher/token folder cap (encrypted)
	 */
	private String				m_NewVSrwCap;

	/**
	 * storage for the list of vouchers to be transferred out of the closing VS
	 */
	private Hashtable<XMLVoucher, String>	m_XferVouchers;

	/**
	 * the total grams of voucher value being transferred, if any
	 */
	private double				m_XferGrams;

	/**
	 * the text of the user's digitally signed statement authorizing the
	 * safe closure
	 */
	private StringBuilder	m_Statement;

	/**
	 * the current title bar text
	 */
	private String				m_TitlebarText;

	/**
	 * the submit button
	 */
	private JButton			m_SubmitButton;

	/**
	 * the action for handling submission of closure confirmation
	 */
	private AbstractAction	m_SubmitAction;

	/**
	 * the scroll area used to ensure that this tab always displays everything
	 * regardless of window size
	 */
	private JScrollPane		m_ScrollPane;

	/**
	 * which phase we're in (see constants above)
	 */
	private byte				m_CurrentPhase;

	/**
	 * the tab manager which manages us
	 */
	private WTabManager		m_TabManager;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param mgr the tab manager which created us
	 */
	public CloseSafePane(WalletClient plugin, WTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_TabManager = mgr;
		m_CurrentPhase = M_Phase1;
		m_NewVSnumber = m_NewVSrwCap = m_TitlebarText = "";
		m_Statement = new StringBuilder(2048);

		// use grid layout
		removeAll();
		setLayout(new GridLayout());
	}

	/**
	 * initialize for phase 1
	 * @param secrets the login secrets for the safe being closed
	 * @return true if we can go on and close, false otherwise
	 */
	public boolean initPhase1(VsSecrets secrets) {
		m_CurrentPhase = M_Phase1;

		// make a copy of the secrets we were passed
		m_OldVSsecrets = new VsSecrets();
		m_OldVSsecrets.setPublisher(secrets.getPublisher());
		m_OldVSsecrets.setVPKey(secrets.getVPKey());
		if (!m_OldVSsecrets.initialize(secrets.getVSnumber(),
												secrets.getCertificate(),
												secrets.getPrivKey(),
												secrets.getVoucherIndex(),
												secrets.getVoucherReadCap(),
												secrets.getReceiptIndex(),
												secrets.getReceiptRWCap(),
												secrets.getDHTMarkHash(),
												secrets.getVouchRWCap()))
		{
			Log.error("Could not make copy of closing VS secrets!");
			return false;
		}
		ArrayList<InetSocketAddress> sdsList
			= new ArrayList<InetSocketAddress>(secrets.getSDSGateways());
		ArrayList<InetSocketAddress> dhtList
			= new ArrayList<InetSocketAddress>(secrets.getDHTNodes());
		HashMap<String, ArrayList<InetSocketAddress>> pksList
			= new HashMap<String, ArrayList<InetSocketAddress>>(
																		secrets.getKeyServers());
		if (!m_OldVSsecrets.initNetwork(secrets.getSDSDomain(),
												  sdsList,
												  secrets.getDHTDomain(),
												  dhtList,
												  pksList))
		{
			Log.error("Could not make copy of closing VS network config!");
			return false;
		}

		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

		// make sure we sync (purge) any tokens spent this session
		if (!m_Plugin.getTokenCache().purgeSpentTokens()) {
			Log.error("Could not purge spent tokens in VS about to close!");
			// but go on
		}

		// we'll need an up-to-date list of vouchers and tokens in the safe
		// (NB: this call obtains the data synchronously)

		if (!m_TabManager.querySafeContents(true)) {
			Log.error("Unable to update wallet contents prior to closeout");
			return false;
		}

		// also ping the DHT serially
		hkp.refreshDHT(true);

		// lastly, query synchronously for outstanding outgoing payments
		boolean mustAbort = false;
		SDSMessage pmtRep = hkp.queryPayments(true);
		if (pmtRep != null) {
			hkp.recordPayments(pmtRep.getPayment_list());
			hkp.updatePaymentRecords();
		}
		else {
			mustAbort = true;
			showError("cannot close wallet",
						"could not obtain outstanding payment list",
						"please try again later");
		}
		setCursor(null);

		// query the housekeeping pane to see if there were any outstanding pmts
		PaymentTableModel ptm = hkp.getPaymentModel();
		if (ptm.getRowCount() > 0) {
			mustAbort = true;
			showError("cannot close wallet",
						"outstanding payments to other wallets exist",
						"wait for all payments to be picked up or to be reclaimed");
		}

		// confirm that all DHT items have been done
		ConcurrentHashMap<Integer, String> dhtRawRecs
			= m_Plugin.getDHTListener().getDHTrecs();
		if (!dhtRawRecs.isEmpty()) {
			mustAbort = true;
			showError("cannot close wallet",
						"housekeeping items exist",
						"please go to Housekeeping and take care of all items");
		}

		// check to be certain that no more than one Issuer is represented,
		// and that no vouchers in the safe are expired

		HomePane hp = m_TabManager.getHomePane();
		VoucherTableModel vtm = hp.getVoucherModel();
		ArrayList<XMLVoucher> allVouchers
			= new ArrayList<XMLVoucher>(vtm.getVoucherList().keySet());
		ArrayList<String> allIssuers = new ArrayList<String>();
		boolean foundExp = false;
		for (XMLVoucher vvv : allVouchers) {
			if (vvv.isExpired()) {
				foundExp = true;
			}
			String iss = vvv.getIssuer();
			if (!allIssuers.contains(iss)) {
				allIssuers.add(iss);
			}
		}
		if (allIssuers.size() > 1) {
			showError("cannot close wallet",
						"this wallet contains vouchers from more than one Issuer",
						"please spend the vouchers from all but one Issuer to other "
						+ "wallets manually, then close this wallet");
			mustAbort = true;
		}
		if (foundExp) {
			showError("cannot close wallet",
						"this wallet contains one or more expired vouchers",
						"please use Advanced/Expired Vouchers to deal with all "
						+ "expired vouchers first");
			mustAbort = true;
		}

		// go back to main menu if anything is wrong
		if (mustAbort) {
			m_TabManager.queueTransition(new Integer(WTabManager.M_GO_MAIN_MENU));
			m_TabManager.makeTransition();
			return false;
		}

		// now make sure we have enough tokens on hand
		TokenTableModel ttm = hp.getTokenModel();
		if (ttm.getAvailableCount(vtm.getSelectedIssuer())
			 < VoucherRequest.M_CloseCost)
		{
			// this is still okay if there are no vouchers
			if (vtm.getRowCount() > 0) {
				// user must buy tokens before closing
				m_TabManager.setPendingTokens(this);
				m_TabManager.queueTransition(new Integer(
																WTabManager.M_PURCHASE_TOKENS));
				m_TabManager.makeTransition();
				return false;
			}
		}
		return true;
	}

	/**
	 * initialize for phase 2
	 */
	public void initPhase2() {
		m_TitlebarText = "Voucher Safe Closure";
		m_CurrentPhase = M_Phase2;

		// layout setup
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel closePane = new JPanel(gbl);
		closePane.setBorder(null);

		HomePane hp = m_TabManager.getHomePane();
		VoucherTableModel vtm = hp.getVoucherModel();
		if (m_XferVouchers == null) {
			m_XferVouchers = vtm.getVoucherList();
		}
		TokenTableModel ttm = hp.getTokenModel();
		final Font taFont = new Font(Font.SANS_SERIF, Font.PLAIN, 12);
		final EmptyBorder taBord = new EmptyBorder(4, 8, 4, 8);

		// If we have a successor recorded, then we have some extra steps.
		if (!m_NewVSnumber.isEmpty()) {
			// total up the weight of all the vouchers in the safe
			m_XferGrams = 0.0;
			String asset = "";
			String units = "";
			Iterator<XMLVoucher> vouchIt = m_XferVouchers.keySet().iterator();
			while (vouchIt.hasNext()) {
				XMLVoucher vvv = vouchIt.next();
				// don't tally any expired vouchers
				if (vvv.isExpired()) {
					vouchIt.remove();
					continue;
				}
				if (asset.isEmpty()) {
					asset = vvv.getAsset();
				}
				if (units.isEmpty()) {
					units = vvv.getUnits();
				}
				m_XferGrams += vvv.getNPV();
			}

			// calculate number of surplus tokens
			int extraToks = ttm.getAvailableCount(vtm.getSelectedIssuer())
								- VoucherRequest.M_CloseCost;

			// display summary of transfer actions to be taken
			StringBuilder summaryTxt = new StringBuilder(512);
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			summaryTxt.append("Value Transfer Terms:\n\n");
			summaryTxt.append("Closing a wallet costs ");
			summaryTxt.append(VoucherRequest.M_CloseCost + " tokens.\n\n");
			summaryTxt.append("This closing wallet, ID \"");
			summaryTxt.append(m_OldVSsecrets.getVSnumber());
			summaryTxt.append("\", contains vouchers totaling ");
			summaryTxt.append(df.format(m_XferGrams) + " " + units + " of ");
			summaryTxt.append(asset + ".\n\n");
			summaryTxt.append("This value, together with " + extraToks);
			summaryTxt.append(" surplus tokens, will be transferred to your new ");
			summaryTxt.append("wallet, ID \"" + m_NewVSnumber + "\".\n");
			summaryTxt.append("\nThis transfer will be irrevocable.\n");
			m_Statement.append(summaryTxt.toString());

			JTextArea summary = new JTextArea(summaryTxt.toString());
			summary.setEditable(false);
			summary.setLineWrap(true);
			summary.setWrapStyleWord(true);
			summary.setFont(taFont);
			summary.setBorder(taBord);
			closePane.add(summary, gbc);
		}
		else {
			String txt = "Value Transfer Terms:\n\n"
							+ "Closing a wallet costs " + VoucherRequest.M_CloseCost
							+ " tokens, or the number in your wallet (whichever "
							+ "is less).\n\nThis closing wallet, ID \""
							+ m_OldVSsecrets.getVSnumber() + "\", does not contain "
							+ "any unexpired vouchers or excess tokens.  Therefore no "
							+ "transfer of value to a successor wallet will occur.\n";
			m_Statement.append(txt.toString());
			JTextArea summary = new JTextArea(txt);
			summary.setEditable(false);
			summary.setLineWrap(true);
			summary.setWrapStyleWord(true);
			summary.setFont(taFont);
			summary.setBorder(taBord);
			closePane.add(summary, gbc);
		}
		closePane.add(Box.createVerticalStrut(10), gbc);

		// now add boiler plate legal language
		StringBuilder agreeTxt = new StringBuilder(1024);
		agreeTxt.append("After closing this wallet, you will still be able to ");
		agreeTxt.append("log in, but it will no longer be able to receive ");
		agreeTxt.append("voucher payments.  Payment history records (receipts) ");
		agreeTxt.append("will be retained indefinitely, unless you purge them ");
		agreeTxt.append("manually.\n\n");
		agreeTxt.append("By clicking on the Agree button below, you warrant ");
		agreeTxt.append("and acknowledge that:\n\n");
		agreeTxt.append("\t1. You are the owner of this wallet, and as ");
		agreeTxt.append("such entitled to perform this operation.\n");
		agreeTxt.append("\t2. You agree to the value transfer terms above.\n");
		agreeTxt.append("\t3. You understand that any transfer of value ");
		agreeTxt.append("to a new successor wallet is irrevocable.\n");
		agreeTxt.append("\t4. Your closed wallet will no longer be able to ");
		agreeTxt.append("receive voucher payments.\n");
		agreeTxt.append("\t5. Your closed wallet's public key certificate on ");
		agreeTxt.append("the publisher's public key server will be marked ");
		agreeTxt.append("as revoked.\n\n");
		agreeTxt.append("Your private key will be used to attest to this ");
		agreeTxt.append("agreement with your digital signature.");
		m_Statement.append(agreeTxt.toString());
		JTextArea agreement = new JTextArea(agreeTxt.toString());
		agreement.setEditable(false);
		agreement.setLineWrap(true);
		agreement.setWrapStyleWord(true);
		agreement.setFont(taFont);
		agreement.setBorder(taBord);
		closePane.add(agreement, gbc);
		closePane.add(Box.createVerticalStrut(15), gbc);

		// make a borderless panel for the buttons, one beside the other
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setPreferredSize(new Dimension(450, 30));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalGlue());

		// define action for handling submission of rejection
		AbstractAction reject = new AbstractAction("Rejected") {
			public void actionPerformed(ActionEvent actionEvent) {
				// clean up and go back to login
				removeAll();
				m_Plugin.getVscState().setLoggedIn(false);
				m_OldVSsecrets.deinitialize();
				m_XferVouchers.clear();
				m_Plugin.resetSecrets();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();
			}
		};
		JButton rejButton = new JButton(reject);
		rejButton.setToolTipText("<html>Click here to reject closing this wallet "
										+ "<br/>and return to login screen</html>");
		rejButton.setFont(m_Plugin.M_ButtonFont);
		buttonPane.add(rejButton);
		buttonPane.add(Box.createVerticalStrut(40));		// spacer

		// define the action for handling submission of closure
		m_SubmitAction = new AbstractAction("Agreed") {
			public void actionPerformed(ActionEvent actionEvent) {
				// send the safe closure message to the OFS for processing
				m_CurrentPhase = M_Phase3;
				sendCloseSafe();
			}
		};
		m_SubmitButton = new JButton(m_SubmitAction);
		m_SubmitButton.setToolTipText("<html>Click here to confirm that you want "
												+ "<br/>to close this wallet as "
												+ "specified</html>");
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		buttonPane.add(m_SubmitButton);
		buttonPane.add(Box.createHorizontalGlue());
		gbc.fill = GridBagConstraints.NONE;
		closePane.add(buttonPane, gbc);
		closePane.add(Box.createVerticalStrut(20), gbc);

		// put everything in a grid bag with border, inside the scroll area
		m_ScrollPane = new JScrollPane(closePane);
		m_ScrollPane.setBorder(null);
		m_ScrollPane.setPreferredSize(new Dimension(450, 450));
		add(m_ScrollPane);
	}


	// get methods
	/**
	 * Determine whether a new successor safe is required to close this safe.
	 * This will be true if:
	 * 	1. There are any unexpired vouchers, of any value, in the safe.
	 * 	2. There are more tokens in the safe than the cost of the closure (25).
	 * @return true if we need to transfer stored value to a successor VS
	 */
	public boolean needSuccessor() {
		HomePane hp = m_TabManager.getHomePane();
		// NB: by this point we have obtained updated voucher/token lists
		TokenTableModel ttm = hp.getTokenModel();
		VoucherTableModel vtm = hp.getVoucherModel();
		if (ttm.getAvailableCount(vtm.getSelectedIssuer())
			 > VoucherRequest.M_CloseCost)
		{
			// this alone is enough to force successor safe
			return true;
		}
		m_XferVouchers = vtm.getVoucherList();
		for (XMLVoucher vvv : m_XferVouchers.keySet()) {
			// one unexpired voucher is enough to do it, regardless of value
			if (!vvv.isExpired()) {
				return true;
			}
		}
		return false;
	}


	// set methods
	/**
	 * record critical data for the new safe (called by NewSafePane)
	 * @param newVS the VS# of the successor safe
	 * @param newVSindex the voucher/token folder hash of the successor safe
	 * @param newVScap the encrypted RW cap of the voucher/token folder
	 */
	public void setSuccessorSafeData(String newVS, char[] newVSindex,
												String newVScap)
	{
		m_NewVSnumber = new String(newVS);
		m_NewVSindex = newVSindex;
		m_NewVSrwCap = new String(newVScap);
	}


	// outbound messaging methods
	/**
	 * send a REQ_close_safe to the OFS
	 */
	private void sendCloseSafe() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		final String err = "could not close wallet";

		// build message to send
		OFSMessage vsClos = new OFSMessage();
		vsClos.setType(IQ.Type.GET);
		vsClos.setFrom(conn.getUser());
		vsClos.setTo(conn.getServiceName());
		vsClos.setPacketID("close_" + stData.getNextOFSid());
		vsClos.setOpcode("REQ_close_safe");
		vsClos.setVoucher_publisher(m_OldVSsecrets.getPublisher());

		// build the <voucherRequest/>
		VoucherRequest vr = new VoucherRequest();
		vr.setAction("merge");
		vr.setSigningVS(m_OldVSsecrets.getVSnumber());
		vr.setFolderHash(new String(m_OldVSsecrets.getVoucherIndex()));
		vr.setFolderCap(m_OldVSsecrets.getVouchRWCap());
		HashMap<String, XMLVoucher> inpList = vr.getVouchers();
		String units = "";
		for (XMLVoucher vvv : m_XferVouchers.keySet()) {
			inpList.put(m_XferVouchers.get(vvv), vvv);
			if (units.isEmpty()) {
				units = vvv.getUnits();
			}
		}
		if (m_XferGrams > 0.0) {
			vr.addOutputValue(m_XferGrams, units);
		}
		vr.setInitialized();
		// sign and encrypt it
		if (!vr.signDetails(m_OldVSsecrets.getPrivKey())) {
			Log.error("Could not sign closing voucher request");
			showError(err, "error signing voucher request");
			return;
		}
		if (!vr.encryptDetails(m_OldVSsecrets.getVPKey(), false)) {
			Log.error("Could not encrypt closing voucher request");
			showError("could not close wallet",
						"error encrypting voucher request");
			return;
		}
		vsClos.setVouchReq(vr);

		// build the <closure_block/>
		StringBuilder closBlk = new StringBuilder(10240);
		closBlk.append(vr.toXML());
		closBlk.append("<new_VS>" + m_NewVSnumber + "</new_VS>");
		if (m_NewVSindex != null) {
			closBlk.append("<new_VS_index>" + new String(m_NewVSindex));
			closBlk.append("</new_VS_index>");
		}
		else {
			closBlk.append("<new_VS_index/>");
		}
		closBlk.append("<newVS_rwcap>" + m_NewVSrwCap + "</newVS_rwcap>");
		closBlk.append("<statement>" + m_Statement.toString() + "</statement>");

		// encrypt it to VP
		String encClos = EncodingUtils.makeBase64PubkeyEncStr(closBlk.toString(),
																	m_OldVSsecrets.getVPKey());
		if (encClos == null || encClos.isEmpty()) {
			Log.error("Could not encrypt closing block");
			showError(err, "error encrypting closure block");
			return;
		}
		vsClos.setClosure_block(encClos);

		// sign the clear text version, including delimiters
		String encTxt
			= "<closure_block>" + closBlk.toString() + "</closure_block>";
		Signature cSign = null;
		byte[] sig = null;
		try {
			cSign = Signature.getInstance("SHA1withRSA");
			cSign.initSign(m_OldVSsecrets.getPrivKey());
			cSign.update(encTxt.getBytes());
			sig = cSign.sign();
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
			return;
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for signature creation", ike);
			showError(err, "invalid sig private key: " + ike.getMessage());
			return;
		}
		catch (SignatureException se) { 
			Log.error("Unable to sign closure block", se);
			showError(err, "error signing closure block: " + se.getMessage());
			return;
		}
		vsClos.setSignature(StringUtils.encodeBase64(sig, false));

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(vsClos.getPacketID());
		m_Plugin.getOFSListener().addFilter(vsClos.getPacketID(), pFilter);

		// set wait cursor
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

		// send it to the server
		conn.sendPacket(vsClos);
		//Log.debug(vsClos.getOpcode() + " sent");

		// record outgoing packet
		stData.setLastActivity(vsClos);
	}

	/**
	 * method to handle reply from OFS to close safe (invoked by callback)
	 * @param reply the message from the OFS, which could be an error
	 */
	public void processCloseReply(OFSMessage reply) {
		if (reply == null ||
			 !reply.getOpcode().equalsIgnoreCase("REP_safe_closed"))
		{
			Log.error("Bad closure reply received");
			setCursor(null);
			return;
		}
		if (m_CurrentPhase != M_Phase3) {
			Log.error("Bad state handling closure reply");
		}

		// see whether this reply is an error
		if (reply.getType() == IQ.Type.ERROR) {
			setCursor(null);
			showError("error closing wallet",
						"code " + reply.getErrcode() + ": " + reply.getErrmsg());
			return;
		}

		// display success and provide button to go to login screen
		m_TitlebarText = "Voucher Safe Closure Succeeded";
		m_ScrollPane.removeAll();
		removeAll();
		setCursor(null);
		JPanel successPane = new JPanel();
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		successPane.setLayout(gbl);

		StringBuilder successTxt = new StringBuilder(256);
		successTxt.append("\nYour wallet closure succeeded.\n");
		if (reply.syncPending()) {
			successTxt.append("\nDue to network latency, any value transferred ");
			successTxt.append("may not appear in your new wallet immediately.\n");
		}
		successTxt.append("\nClick on the button below to log in to your new ");
		successTxt.append("wallet.\n");
		JTextArea conf = new JTextArea(successTxt.toString(), 5, 30);
		conf.setEditable(false);
		conf.setLineWrap(true);
		conf.setWrapStyleWord(true);
		successPane.add(conf, gbc);
		successPane.add(Box.createVerticalStrut(10), gbc);

		m_SubmitAction = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent actionEvent) {
				// clean up and log out of both safes; close connections
				m_Plugin.getVscState().setLoggedIn(false);
				m_OldVSsecrets.deinitialize();
				m_XferVouchers.clear();
				m_Plugin.resetSecrets();
				m_Plugin.getDHTListener().closeDHTConnection();
				m_Plugin.getDHTListener().getDHTrecs().clear();
				m_Plugin.getSDSListener().closeSDSConnection();

				// transition back to login screen
				removeAll();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();

				// explicitly garbage collect
				m_SubmitAction = null;
				System.gc();
			}
		};
		m_SubmitButton = new JButton(m_SubmitAction);
		m_SubmitButton.setToolTipText("Click here to go to the login screen");
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		gbc.fill = GridBagConstraints.NONE;
		successPane.add(m_SubmitButton, gbc);

		m_ScrollPane = new JScrollPane(successPane);
		add(m_ScrollPane);
	}


	// error handlers which invoke analogs in WTabManager
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	private void showError(String errMsg) {
		m_TabManager.showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	private void showError(String errMsg, String cause) {
		m_TabManager.showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	private void showError(String errMsg, String cause, String action) {
		m_TabManager.showError(errMsg, cause, action);
	}


	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_OldVSsecrets != null) {
				m_OldVSsecrets.deinitialize();
			}
			if (m_XferVouchers != null && !m_XferVouchers.isEmpty()) {
				m_XferVouchers.clear();
			}
			if (m_NewVSindex != null && m_NewVSindex.length != 0) {
				Arrays.fill(m_NewVSindex, ' ');
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
