/*
 * last modified---
 * 	03-22-17/jet/ reference Resync button (SVSpark mod)
 * 	06-05-13/jet/ SilentVault client version
 * 	01-18-13/kbw/ rework as an Advanced option instead of a part of
 * 					  housekeeping; add support for redemption options
 * 	02-22-11/kbw/ remove assumption that vouchers are in GAU by default
 * 	11-05-10/kbw/ force manual processing on delete errors, to avoid looping
 * 	10-28-10/kbw/ provide automated mode
 * 	10-18-10/kbw/ vertical compression of layout
 * 	09-30-10/kbw/ layout changes
 * 	09-02-10/kbw/ overhaul for Nimbus L&F
 * 	08-04-10/kbw/ use standard fonts
 * 	04-29-10/kbw/ flesh out
 * 	04-07-10/kbw/ new (stub)
 *
 * purpose---
 * 	provide a UI widget to handle display and deletion of expired vouchers
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Font;
import java.awt.Color;
import java.awt.event.ActionEvent;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.security.*;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for handling expired vouchers.
 *
 * @author Kevin Wilkerson
 */
public final class ExpiredPane extends WorkPane {
	// BEGIN data members
	/**
	 * storage for a voucher we are attempting to delete via the VP
	 */
	private XMLVoucher			m_ExpiredVouch;

	/**
	 * storage for the SDS filename for this voucher
	 */
	private String					m_ExpVouchFile;

	/**
	 * action to cause deletion of expired voucher
	 */
	private AbstractAction		m_DeleteAction;

	/**
	 * associated button for expired voucher deletion
	 */
	private JButton				m_DelButton;

	/**
	 * action to cause attempted expired voucher renewal
	 */
	private AbstractAction		m_RenewAction;

	/**
	 * associated button for expired voucher renewal
	 */
	private JButton				m_RenewButton;

	/**
	 * action to cause attempted expired voucher redemption (outbail)
	 */
	private AbstractAction		m_RedeemAction;

	/**
	 * associated button for expired voucher redemption
	 */
	private JButton				m_RedeemButton;

	/**
	 * special button to return to main menu (used in place of m_MainButton)
	 */
	private JButton				m_CustomMainButton;

	/**
	 * flag to set when doing renewal
	 */
	private boolean				m_DoingRenew;

	/**
	 * flag for whether renewal has been attempted and failed
	 */
	private boolean				m_RenewFailed;

	/**
	 * flag for whether redemption has been attempted and failed
	 */
	private boolean				m_RedeemFailed;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which created us
	 */
	public ExpiredPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_ExpVouchFile = "";

		String usrTxt = "Once vouchers are past their expiration date, they "
							+ "cannot be spent or re-validated.<br/><br/>"
							+ "Depending on the policy of the voucher Issuer, you "
							+ "may be able to renew the voucher, or redeem its value "
							+ "to an account you have established at the Issuer which "
							+ "is associated with your wallet ID.  "
							+ "Note that some Issuers may not support either of these "
							+ "options, in which "
							+ "case expired voucher value is lost and you will need "
							+ "to delete such vouchers from your wallet.<br/>"
							+ "<br/>Select an expired voucher and click on the "
							+ "appropriate action button.  You will not be able "
							+ "to select unexpired vouchers, or to delete expired "
							+ "vouchers until the other actions have been tried and "
							+ "have failed.<br/><br/>To re-validate an "
							+ "unexpired voucher, use the Advanced/Validate option.";
		m_UserInstructs = usrTxt;

		// define action buttons
		setDeleteAction();
		m_DelButton = new JButton(m_DeleteAction);
		m_DelButton.setFont(plugin.M_ButtonFont);
		m_DelButton.setToolTipText("Click here to delete this expired voucher "
											+ "from your wallet");
		m_DelButton.setEnabled(false);
		setRenewAction();
		m_RenewButton = new JButton(m_RenewAction);
		m_RenewButton.setFont(plugin.M_ButtonFont);
		m_RenewButton.setToolTipText("Click here to attempt to renew this expired"
											+ " voucher at the Issuer");
		setRedeemAction();
		m_RedeemButton = new JButton(m_RedeemAction);
		m_RedeemButton.setFont(plugin.M_ButtonFont);
		m_RedeemButton.setToolTipText("Click here to attempt to redeem this "
												+ "expired voucher to your account at the "
												+ "Issuer");
		AbstractAction goMain = new AbstractAction("Back to Overview") {
			public void actionPerformed(ActionEvent actionEvent) {
				// clear special select condition
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				vtm.flipExpSelectability(false);
				// switch back to main overview screen
				m_TabManager.queueTransition(
													new Integer(WTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
		m_CustomMainButton = new JButton(goMain);
		m_CustomMainButton.setFont(m_Plugin.M_ButtonFont);
		m_CustomMainButton.setToolTipText("Return to the main menu");

		// eliminate border around the whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to define the action to delete an expired voucher
	 */
	private void setDeleteAction() {
		m_DeleteAction = new AbstractAction("Delete") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need exactly one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() < 1) {
					showError("no vouchers to delete", "none selected",
								"please select exactly one expired voucher to delete");
					return;
				}
				// make sure we have the expired voucher, and a filename
				m_ExpiredVouch = selVouchers.get(0);
				if (!m_ExpiredVouch.isExpired()) {
					showError("cannot delete voucher",
								"voucher is not yet expired",
								"please select an expired voucher");
					return;
				}
				Hashtable<XMLVoucher, String> allVouchers = vtm.getVoucherList();
				m_ExpVouchFile = allVouchers.get(m_ExpiredVouch);
				if (m_ExpVouchFile.isEmpty()) {
					showError("cannot delete voucher", "filename not found");
					return;
				}

				// ensure we've tried the other two options first
				if (!m_RenewFailed) {
					showError("cannot delete voucher",
								"renewal at Issuer has not yet been attempted",
								"please try a Renew operation first");
					return;
				}
				if (!m_RedeemFailed) {
					showError("cannot delete voucher",
								"redemption at Issuer has not yet been attempted",
								"please try a Redeem operation first");
					return;
				}

				// send the message to the SDS
				sendDeleteVoucher();
			}
		};
	}

	/**
	 * utility method to define the action to renew an expired voucher
	 */
	private void setRenewAction() {
		m_RenewAction = new AbstractAction("Renew") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need exactly one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() < 1) {
					showError("no vouchers to renew", "none selected",
								"please select exactly one expired voucher to renew");
					return;
				}
				// make sure we have the expired voucher, and a filename
				m_ExpiredVouch = selVouchers.get(0);
				if (!m_ExpiredVouch.isExpired()) {
					showError("cannot renew voucher",
								"voucher is not yet expired",
								"please select an expired voucher");
					return;
				}
				Hashtable<XMLVoucher, String> allVouchers = vtm.getVoucherList();
				m_ExpVouchFile = allVouchers.get(m_ExpiredVouch);
				if (m_ExpVouchFile.isEmpty()) {
					showError("cannot renew voucher", "filename not found");
					return;
				}

				// send the message to the SDS
				sendRedeemVoucher(true);
			}
		};
	}

	/**
	 * utility method to define the action to redeem an expired voucher
	 */
	private void setRedeemAction() {
		m_RedeemAction = new AbstractAction("Redeem") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need exactly one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() < 1) {
					showError("no vouchers to redeem", "none selected",
								"please select exactly one expired voucher to redeem");
					return;
				}
				// make sure we have the expired voucher, and a filename
				m_ExpiredVouch = selVouchers.get(0);
				if (!m_ExpiredVouch.isExpired()) {
					showError("cannot redeem voucher",
								"voucher is not yet expired",
								"please select an expired voucher");
					return;
				}
				Hashtable<XMLVoucher, String> allVouchers = vtm.getVoucherList();
				m_ExpVouchFile = allVouchers.get(m_ExpiredVouch);
				if (m_ExpVouchFile.isEmpty()) {
					showError("cannot redeem voucher", "filename not found");
					return;
				}

				// send the message to the SDS
				sendRedeemVoucher(false);
			}
		};
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// add custom return-to-main button as sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_CustomMainButton);
		m_WorkBar.setVisible(true);

		// set heading pane
		buildValueHeader();

		// set screen label
		buildScreenLabel("Handle Expired Vouchers");

		// set that expired vouchers should be selectable rather than unexpired
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.flipExpSelectability(true);

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, the button to renew this voucher
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer
		m_ButtonPane.add(m_RenewButton);
		m_ButtonPane.add(Box.createHorizontalStrut(100));	 // spacer
		// next, the button the redeem this voucher
		m_ButtonPane.add(m_RedeemButton);
		m_ButtonPane.add(Box.createHorizontalStrut(100));	 // spacer
		// last, the button to delete this voucher
		m_ButtonPane.add(m_DelButton);

		// adjust enabled status of delete action button based on previous results
		if (m_RenewFailed && m_RedeemFailed) {
			m_DelButton.setEnabled(true);
		}
		else {
			m_DelButton.setEnabled(false);
		}

		// place the voucher table into work area, set for radio select mode
		vtm.deselectAll();
		vtm.setSelectMode(VoucherTableModel.M_SelectSingle);
		JTable vt = m_ParentHome.getVoucherTable();
		m_WorkArea.setViewportView(vt);
	}

	/**
	 * private utility method to tell the SDS to delete a voucher.
	 * implied param: m_ExpVouchFile holds the filename to delete
	 */
	private void sendDeleteVoucher() {
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot delete vouchers while not logged in!");
			showError("cannot delete vouchers", "not logged in", "log in first");
			return;
		}

		// wait cursor and disable buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// build message to send
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		OFSMessage delVoucher = new OFSMessage();
		delVoucher.setType(IQ.Type.SET);
		delVoucher.setFrom(conn.getUser());
		delVoucher.setTo(conn.getServiceName());
		delVoucher.setPacketID("vdel_" + state.getNextOFSid());
		delVoucher.setOpcode("REQ_delete_voucher");
		delVoucher.setVoucher_publisher(secrets.getPublisher());

		// create <deleteDetails/> element
		StringBuilder dDets = new StringBuilder(1024);
		dDets.append("<vs_number>" + secrets.getVSnumber() + "</vs_number>");
		String vIdx = new String(secrets.getVoucherIndex());
		dDets.append("<sdshash>" + vIdx + "</sdshash>");
		dDets.append("<cap>" + secrets.getVouchRWCap() + "</cap>");
		String vouchXML = m_ExpiredVouch.toXML();
		dDets.append("<voucher filename=\"" + m_ExpVouchFile + "\"");
		dDets.append(vouchXML.substring(8));

		// sign it with our private key
		Signature dSig = null;
		byte[] sig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initSign(secrets.getPrivKey());
			ok = true;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such signature algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for details signature", ike);
		}
		if (!ok) {
			showError("cannot delete voucher",
						"internal error signing with private key");
			return;
		}
		try {
			dSig.update(dDets.toString().getBytes());
			sig = dSig.sign();
		}
		catch (SignatureException se) {
			Log.error("Unable to sign deleteDetails", se);
			showError("cannot delete voucher", "error signing delete details");
			setButtonState(true);
			return;
		}
		dDets.append("<signature>" + StringUtils.encodeBase64(sig, false));
		dDets.append("</signature>");

		// now encrypt it with the VP's public key so the OFS cannot read it
		String delDetails = "<deleteDetails>" + dDets.toString()
									+ "</deleteDetails>";
		String encDel = EncodingUtils.makeBase64PubkeyEncStr(delDetails,
																			  secrets.getVPKey());
		if (encDel == null || encDel.isEmpty()) {
			showError("cannot delete voucher",
						"error encrypting delete details for VP");
			return;
		}
		delVoucher.setDeleteDetails(encDel);

		// register async reply listener
		PacketIDFilter pFilter = new PacketIDFilter(delVoucher.getPacketID());
		m_Plugin.getOFSListener().addFilter(delVoucher.getPacketID(), pFilter);

		// send request to the OFS
		conn.sendPacket(delVoucher);

		// record outgoing packet
		state.setLastActivity(delVoucher);

		// NB: when reply is received it will be routed to processDeleteReply()
	}

	/**
	 * method to process a response to our voucher delete request
	 * @param ofsRep the reply from the OFS (could be error or result)
	 */
	public void processDeleteReply(OFSMessage ofsRep) {
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (ofsRep.getType() == IQ.Type.ERROR) {
			showError("unable to delete voucher", ofsRep.getErrmsg());
			prepDisplay();
			return;
		}

		// record voucher deletion and reset
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> removed = new ArrayList<XMLVoucher>(1);
		removed.add(m_ExpiredVouch);
		vtm.removeVouchers(removed);
		m_ExpiredVouch = null;
		m_ExpVouchFile = "";
		m_RenewFailed = m_RedeemFailed = false;
		prepDisplay();
	}

	/**
	 * private utility method to tell the SDS to redeem a voucher.
	 * @param renew if true, try a renew, else try a redeem
	 * implied param: m_ExpVouchFile holds the filename to delete
	 */
	private void sendRedeemVoucher(boolean renew) {
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot redeem vouchers while not logged in!");
			showError("cannot redeem vouchers", "not logged in", "log in first");
			return;
		}

		// wait cursor and disable buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// build message to send
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		OFSMessage redVoucher = new OFSMessage();
		redVoucher.setType(IQ.Type.SET);
		redVoucher.setFrom(conn.getUser());
		redVoucher.setTo(conn.getServiceName());
		redVoucher.setPacketID("vredeem_" + state.getNextOFSid());
		redVoucher.setOpcode("REQ_redeem_voucher");
		redVoucher.setVoucher_publisher(secrets.getPublisher());

		// create <redeemDetails/> element
		StringBuilder rDets = new StringBuilder(1024);
		rDets.append("<vs_number>" + secrets.getVSnumber() + "</vs_number>");
		String vIdx = new String(secrets.getVoucherIndex());
		rDets.append("<sdshash>" + vIdx + "</sdshash>");
		rDets.append("<cap>" + secrets.getVouchRWCap() + "</cap>");
		String vouchXML = m_ExpiredVouch.toXML();
		rDets.append("<voucher filename=\"" + m_ExpVouchFile + "\"");
		rDets.append(vouchXML.substring(8));
		rDets.append("<disposition>");
		if (renew) {
			rDets.append("renew");
			m_DoingRenew = true;
		}
		else {
			rDets.append("redeem");
			m_DoingRenew = false;
		}
		rDets.append("</disposition>");

		// sign it with our private key
		Signature dSig = null;
		byte[] sig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initSign(secrets.getPrivKey());
			ok = true;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such signature algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for details signature", ike);
		}
		if (!ok) {
			showError("cannot redeem voucher",
						"internal error signing with private key");
			return;
		}
		try {
			dSig.update(rDets.toString().getBytes());
			sig = dSig.sign();
		}
		catch (SignatureException se) {
			Log.error("Unable to sign redeemDetails", se);
			showError("cannot redeem voucher", "error signing redeem details");
			setButtonState(true);
			return;
		}
		rDets.append("<signature>" + StringUtils.encodeBase64(sig, false));
		rDets.append("</signature>");

		// now encrypt it with the VP's public key so the OFS cannot read it
		String redDetails = "<redeemDetails>" + rDets.toString()
									+ "</redeemDetails>";
		String encRed = EncodingUtils.makeBase64PubkeyEncStr(redDetails,
																			  secrets.getVPKey());
		if (encRed == null || encRed.isEmpty()) {
			showError("cannot redeem voucher",
						"error encrypting redeem details for VP");
			return;
		}
		redVoucher.setRedeemDetails(encRed);

		// register async reply listener
		PacketIDFilter pFilter = new PacketIDFilter(redVoucher.getPacketID());
		m_Plugin.getOFSListener().addFilter(redVoucher.getPacketID(), pFilter);

		// send request to the OFS
		conn.sendPacket(redVoucher);

		// record outgoing packet
		state.setLastActivity(redVoucher);

		// NB: when reply is received it will be routed to processRedeemReply()
	}

	/**
	 * method to process a response to our voucher redeem request
	 * @param ofsRep the reply from the OFS (could be error or result)
	 */
	public void processRedeemReply(OFSMessage ofsRep) {
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (ofsRep.getType() == IQ.Type.ERROR) {
			if (ofsRep.getErrmsg().contains("404")) {
				showError("unable to redeem voucher", ofsRep.getErrmsg(),
							"you may need to open an account at the Issuer");
			}
			else {
				showError("unable to " + (m_DoingRenew ? "renew" : "redeem")
							+ " voucher",
							ofsRep.getErrmsg());
			}
			if (ofsRep.getErrcode() == HttpURLConnection.HTTP_BAD_METHOD) {
				// this means Issuer doesn't support the disposition attempted
				if (m_DoingRenew) {
					m_RenewFailed = true;
				}
				else {
					m_RedeemFailed = true;
				}
			}
			prepDisplay();
			return;
		}

		// record voucher deletion and reset
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> removed = new ArrayList<XMLVoucher>(1);
		removed.add(m_ExpiredVouch);
		vtm.removeVouchers(removed);
		m_ExpiredVouch = null;
		m_ExpVouchFile = "";
		m_RenewFailed = m_RedeemFailed = false;

		// parse the <confirmation> block and validate the VP's signature on it
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String decConf
			= EncodingUtils.getStrFromBase64PubkeyEnc(ofsRep.getConfirmation(),
																	secrets.getPrivKey());
		final String err = "error processing voucher redemption reply";
		if (decConf == null || decConf.isEmpty()) {
			Log.error("Unable to decode <confirmation/> of " + ofsRep.getOpcode());
			showError(err, "could not decode");
			return;
		}
		Element confElt = EncodingUtils.buildElement(decConf);
		if (confElt == null) {
			Log.error("Unable to parse <confirmation/> of " + ofsRep.getOpcode());
			showError(err, "could not parse confirmation details");
			return;
		}

		// get the <voucher/>, which may be empty but must be present
		Element vouchElt = confElt.element("voucher");
		if (vouchElt == null) {
			Log.error("Missing <voucher/> element in <confirmation/>");
			showError(err, "missing voucher element");
			return;
		}

		// get voucher's filename
		Element vouchFileElt = confElt.element("voucherFile");
		if (vouchFileElt == null) {
			Log.error("Missing <voucherFile/> element in <confirmation/>");
			showError(err, "missing voucher filename element");
			return;
		}
		String vouchFile = vouchFileElt.getTextTrim();

		// get VP's signature, and verify it
		Element sigElt = confElt.element("signature");
		if (sigElt == null) {
			Log.error("Missing <signature/> element in <confirmation/>");
			showError(err, "missing signature");
			return;
		}
		StringBuilder cSigData = new StringBuilder(1024);
		cSigData.append("<confirmation>");
		String vouchText = vouchElt.asXML();
		cSigData.append(vouchText);
		if (vouchFile.isEmpty()) {
			cSigData.append("<voucherFile/>");
		}
		else {
			cSigData.append("<voucherFile>" + vouchFile + "</voucherFile>");
		}
		Signature cSig = null;
		boolean ok = false;
		try {
			cSig = Signature.getInstance("SHA1withRSA");
			cSig.initVerify(secrets.getVPKey());
			cSig.update(cSigData.toString().getBytes());
			ok = cSig.verify(StringUtils.decodeBase64(sigElt.getTextTrim()));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate <confirmation/> signature", se);
		}
		if (!ok) {
			Log.error("publisher's sig on redeem confirmation did not verify");
			showError(err, "publisher's signature on confirmation did not verify");
			return;
		}

		// if we have one, parse the voucher and validate its signature
		if (!vouchText.equals("<voucher/>")) {
			XMLVoucher newVouch = null;
			try {
				newVouch = XMLVoucher.getVoucherRep(vouchElt);
			}
			catch (DocumentException de) {
				Log.error("Unparsable <voucher/> element in <confirmation/>");
				showError(err, "voucher element did not parse");
				return;
			}

			if (!newVouch.validateSig(secrets.getVPKey())) {
				Log.error("Bad sig on <voucher/> element in <confirmation/>");
				showError(err, "voucher signature did not verify");
				return;
			}

			// show warning message if we're pending
			if (ofsRep.syncPending()) {
				showError("warning: synchronization pending",
							"the voucher redemption succeeded, but the publisher has "
							+ "queued storing the new voucher in your wallet",
							"wait a few minutes and then use the Resync button to "
							+ "see the new voucher");
			}
			else {
				// tell parent about any new voucher so it can update display
				m_ParentHome.addVoucher(newVouch, vouchFile);
			}
		}

		// return to screen display
		prepDisplay();
	}

	// END methods
}
