/*
 * last modified---
 * 	11-30-16/jet/ improve wording on tokens used and reclaimed vouchers
 * 	07-17-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine values
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	09-06-13/jet/ also show voucher asset type on outstanding payments and
 * 					  incoming receipts
 * 	08-09-13/jet/ show incoming voucher asset type on picked-up payments
 * 	06-05-13/jet/ SilentVault client version
 * 	01-17-13/kbw/ remove processing of expired vouchers
 * 	12-18-12/kbw/ use getNPV() instead of baseUnitValue()
 * 	11-12-12/kbw/ change Refresh label to Incoming items, to avoid verb
 * 	04-18-11/kbw/ use a ConcurrentHashMap for m_DecodedDHTRecs
 * 	04-12-11/kbw/ provide separate thread to update HKP progress bar
 * 	03-28-11/kbw/ invoke M_CONFIRMATION in separate thread so display appears
 * 	02-22-11/kbw/ remove assumption that all prices are for GAU
 * 	11-05-10/kbw/ add manualOverride()
 * 	11-03-10/kbw/ show failed pickups in auto report in gray background
 * 	10-28-10/kbw/ convert to building report after processing, for auto mode
 * 	10-12-10/kbw/ use StringTokenizer to fix memory leak
 * 	09-23-10/kbw/ do show pending payments, but skip over and do not count
 * 	09-20-10/kbw/ place baggage into a table for tooltips
 * 	09-13-10/kbw/ entirely new more compact look
 * 	09-02-10/kbw/ overhaul for Nimbus L&F
 * 	08-16-10/kbw/ force silent ConfirmPane invocations when needed
 * 	08-04-10/kbw/ use standardized fonts; simplify color scheme
 * 	04-24-10/kbw/ debug
 * 	04-06-10/kbw/ flesh out
 * 	03-29-10/kbw/ new (stub)
 *
 * purpose---
 * 	provide a UI widget to handle wallet housekeeping tasks
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.Dimension;

import org.dom4j.DocumentException;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.Map;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.concurrent.ConcurrentHashMap;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;


/**
 * Extends WorkPane to provide a screen for doing safe-related housekeeping.
 *
 * @author Kevin Wilkerson
 */
public final class HousekeepingPane extends WorkPane {
	// BEGIN data members
	/**
	 * storage for a pending payment record on the DHT which we're processing
	 */
	private XMLPendingPayment	m_PendPayRecord;

	/**
	 * storage for a pending payment record we are placing on the SDS, or
	 * attempting to reclaim or delete from the SDS
	 */
	private XMLPayment			m_PaymentRecord;

	/**
	 * storage for a receipt record we are retrieving or placing on the DHT
	 */
	private XMLReceipt			m_ReceiptRecord;

	/**
	 * storage for a payment confirmation we are processing
	 */
	private XMLConfirmPayment	m_PaymentConfirm;

	/**
	 * action to process the next housekeeping record
	 */
	private AbstractAction		m_StartAction;

	/**
	 * button to go with m_StartAction
	 */
	private JButton				m_StartButton;

	/**
	 * storage for tokens used in an operation storing on the SDS or DHT
	 */
	private ArrayList<XMLToken>   m_PaymentTokens;

	/**
	 * the table model for the pending payment records table (on SDS, not DHT)
	 */
	private PaymentTableModel	m_PaymentTableModel;

	/**
	 * the table model for the receipt records table (on SDS, not DHT)
	 */
	private ReceiptTableModel	m_ReceiptTableModel;

	/**
	 * the list of raw outstanding payments, received from the SDS encrypted
	 */
	private Hashtable<String, String>	m_OutstandingPayments;

	/**
	 * the list of incoming payments made to this safe and found on the DHT
	 * (decoded)
	 */
	private ArrayList<XMLPendingPayment>	m_IncomingPayments;

	/**
	 * the list of incoming receipts found on the DHT (decoded)
	 */
	private ArrayList<XMLReceipt>		m_IncomingReceipts;

	/**
	 * the list of incoming payment confirmations found on the DHT (decoded)
	 */
	private ArrayList<XMLConfirmPayment>	m_IncomingConfirmations;

	/**
	 * The list of decrypted raw records from the DHT, in XML form, keyed by
	 * their record IDs on the DHT.  Every item in this table will also
	 * (assuming it parsed successfully) be found in one of m_IncomingPayments,
	 * m_IncomingReceipts, or m_IncomingConfirmations.
	 */
	private ConcurrentHashMap<Integer, String>	m_DecodedDHTRecs;

	/**
	 * whether or not we've recorded a reply from the DHT
	 */
	private boolean				m_GotDHTreply;

	/**
	 * whether or not we've recorded a reply from the SDS (pendpay list)
	 */
	private boolean				m_GotSDSreply;

	/**
	 * the list of actions we are working on
	 */
	private ArrayList<Object>	m_ProcessingList;

	/**
	 * the list of actions we've already performed (in auto mode)
	 */
	private ArrayList<Object>	m_ProcessedList;

	/**
	 * the index of the item in m_ProcessingList currently selected to work on
	 */
	private int						m_CurrItemIdx;

	/**
	 * whether housekeeping is to be automated (controlled by user preference)
	 */
	private boolean				m_Automation;

	/**
	 * the aggregate token count we used to process items
	 */
	private int						m_TokensUsed;

	/**
	 * initialization flag, indicating we're at the start of processing, used
	 * to control the state of the progress bar in automation mode
	 */
	private boolean				m_ResetInit;

	/**
	 * the original number of items to process which we had when m_ResetInit
	 * was set
	 */
	private int						m_OrgItemCnt;

	/**
	 * the progress bar used in automating mode
	 */
	private JProgressBar			m_AutoProgress;

	/**
	 * the list of XMLPayments which were reclaimed successfully
	 */
	private ArrayList<XMLPayment>	m_SuccessfulReclaims;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public HousekeepingPane(WalletClient plugin, HomePane home, WTabManager mgr)
	{
		super(plugin, home, mgr);
		m_HeadingPane.setLayout(new GridLayout(1, 3));
		m_Automation = true;
		m_AutoProgress = new JProgressBar();
		m_AutoProgress.setFont(m_ParentHome.M_DataFont);
		m_AutoProgress.setStringPainted(true);
		m_AutoProgress.setBackground(Color.LIGHT_GRAY);
		m_AutoProgress.setOpaque(true);
		m_AutoProgress.setMinimum(0);

		// provide user textual instructions
		String housekeepTxt = "Housekeeping keeps your wallet up to "
									+ "date.<br/><br/>You need to do these tasks "
									+ "regularly or they may expire and will no longer "
									+ "be available to you.";
		m_UserInstructs = housekeepTxt;

		// init various containers
		m_PaymentTokens = new ArrayList<XMLToken>();
		m_PaymentTableModel
			= new PaymentTableModel(new Hashtable<XMLPayment, String>());
		m_ReceiptTableModel
			= new ReceiptTableModel(new Hashtable<XMLReceipt, String>(), plugin);
		m_OutstandingPayments = new Hashtable<String, String>();
		m_IncomingPayments = new ArrayList<XMLPendingPayment>();
		m_IncomingReceipts = new ArrayList<XMLReceipt>();
		m_IncomingConfirmations = new ArrayList<XMLConfirmPayment>();
		m_DecodedDHTRecs = new ConcurrentHashMap<Integer, String>();
		m_ProcessingList = new ArrayList<Object>();
		m_ProcessedList = new ArrayList<Object>();
		m_SuccessfulReclaims = new ArrayList<XMLPayment>();

		// define button action
		setStartAction();
		AbstractAction mainReturn = new AbstractAction("Back to Overview") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we only want to reset this when user clicks overview
				m_TokensUsed = 0;
				m_ProcessedList.clear();
				// now go back to main, as normal
				m_TabManager.queueTransition(
													new Integer(WTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
		// override WorkPane's default action for the m_MainButton
		m_MainButton.setAction(mainReturn);

		// no border around whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to define the action to do the next housekeeping item
	 */
	private void setStartAction() {
		m_StartAction = new AbstractAction("Start Housekeeping") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_ProcessingList.isEmpty()) {
					showError("cannot do housekeeping",
								"there are no housekeeping items",
								"use Back to Overview to return to main menu");
					return;
				}
				// start at the first item
				processNextItem(0);
			}
		};
	}

	/**
	 * method to process the nth item in the list of housekeeping items
	 * @param idx the index of the item to process; if given as -1 the next one
	 * will be selected
	 */
	public void processNextItem(int idx) {
		int item = idx;
		if (item == -1) {
			item = ++m_CurrItemIdx;
		}
		else {
			m_CurrItemIdx = item;
		}
		if (m_CurrItemIdx >= m_ProcessingList.size()) {
			// wrap back to beginning
			m_CurrItemIdx = 0;

			// if there is only one item, go back to menu
			if (m_ProcessingList.size() == 1) {
				m_TabManager.queueTransition(
										new Integer(WTabManager.M_HOUSEKEEPING_RETURN));
				m_TabManager.makeTransition();
			}
		}

		// take the next item in the list
		Object item1 = m_ProcessingList.get(m_CurrItemIdx);
		if (item1 == null) {
			Log.error("Bad housekeeping list index, " + m_CurrItemIdx);
			if (m_ProcessingList.isEmpty()) {
				showError("cannot do housekeeping", "no housekeeping items exist",
							"skip Housekeeping and return to main menu");
			}
			else {
				showError("error processing housekeeping item " + (m_CurrItemIdx+1),
							"item not found", "skip Housekeeping and try again");
			}
			return;
		}

		// branch to the appropriate sub-class screen based on type
		int tmState = 0;
		if (item1 instanceof XMLPendingPayment) {
			m_PendPayRecord = (XMLPendingPayment) item1;
			tmState = WTabManager.M_INCOMING_PAYMENT;
		}
		else if (item1 instanceof XMLReceipt) {
			m_ReceiptRecord = (XMLReceipt) item1;
			tmState = WTabManager.M_INCOMING_RECEIPT;
		}
		else if (item1 instanceof XMLPayment) {
			m_PaymentRecord = (XMLPayment) item1;
			tmState = WTabManager.M_UNCLAIMED;
		}
		else if (item1 instanceof XMLConfirmPayment) {
			// NB: present only for completeness; these are handled upfront
			Log.warning("XMLConfirmPayment not expected in general housekeeping");
			m_PaymentConfirm = (XMLConfirmPayment) item1;
			tmState = WTabManager.M_CONFIRMATION;
		}
		else {
			showError("error processing housekeeping item",
						"unexpected item type");
			return;
		}

		// go to the indicated screen, via a separate thread
		m_TabManager.queueTransition(new Integer(tmState));
		Runnable doTrans = new Runnable() {
			public void run() {
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(doTrans);
	}

	/**
	 * Check whether there is any housekeeping to be done.
	 * (invoked from WTabManager, usually when replies show up)
	 */
	public void doHousekeeping() {
		// make sure we have received replies from SDS and DHT before this
		if (!m_GotDHTreply) {
			// this means we called refreshDHT() but not updateDHTrecords() yet
			showError("cannot do housekeeping", "no reply received from DHT yet",
						"wait a few moments and click Refresh again");
			return;
		}
		if (!m_GotSDSreply) {
			showError("cannot do housekeeping", "no reply received from SDS yet",
						"wait a few moments and click Refresh again");
			return;
		}

		// set up our display
		m_Plugin.setDefaultStatusBar();
		m_ParentHome.setWorkPane(this);
	}

	/**
	 * process the received updates to the DHT record list
	 */
	public void updateDHTrecords() {
		// check the DHTListener's raw record list
		ConcurrentHashMap<Integer, String> dhtList
			= m_Plugin.getDHTListener().getDHTrecs();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!dhtList.isEmpty()) {
			// decrypt and parse all the records returned by the DHT
			m_ParentHome.setCursor(M_WaitCursor);
			MenuPane mp = m_Plugin.getWalletTabManager().getMenuPane();
			mp.disableMenu();
			JProgressBar decodeProg = new JProgressBar(0, dhtList.size());
			decodeProg.setString("Decoding " + dhtList.size()
										+ " records from DHT, please wait");
			decodeProg.setStringPainted(true);
			decodeProg.setOpaque(true);
			decodeProg.setBackground(Color.LIGHT_GRAY);
			decodeProg.setFont(m_ParentHome.M_DataFont);
			JPanel statBar = m_Plugin.getStatusBar();
			statBar.removeAll();
			statBar.add(decodeProg);
			statBar.revalidate();
			statBar.repaint();
			int idx = 0, dCnt = 0;

			for (Integer dhtId : dhtList.keySet()) {
				decodeProg.setValue(++dCnt);
				// see if we have it in our decoded list already
				if (m_DecodedDHTRecs.containsKey(dhtId)) {
					// no need to decrypt it again
					idx += 1;
					continue;
				}

				// decrypt and figure out what kind of record it is
				String encRec = dhtList.get(dhtId);
				String decRec = EncodingUtils.getStrFromBase64PubkeyEnc(encRec,
																			secrets.getPrivKey());
				if (decRec == null || decRec.isEmpty()) {
					Log.error("Could not decrypt a DHT record, id " + dhtId
								+ ", raw = " + encRec);
					continue;
				}
				m_DecodedDHTRecs.put(dhtId, decRec);

				// peek at the string to see which type it is
				XMLPendingPayment pendPay = null;
				XMLReceipt receipt = null;
				XMLConfirmPayment confPay = null;
				if (decRec.startsWith("<payment")) {
					try {
						pendPay = XMLPendingPayment.getPaymentRep(decRec,
																			secrets.getPrivKey(),
																	m_Plugin.getPKSListener());
					}
					catch (DocumentException de) {
						Log.error("Unable to parse pending payment", de);
						continue;
					}
					pendPay.setDHTid(dhtId);
					m_IncomingPayments.add(pendPay);
				}
				else if (decRec.startsWith("<receipt")) {
					try {
						receipt = XMLReceipt.getReceiptRep(decRec);
					}
					catch (DocumentException de) {
						Log.error("Unable to parse receipt", de);
						continue;
					}
					receipt.setDHTid(dhtId);
					m_IncomingReceipts.add(receipt);
				}
				else if (decRec.startsWith("<confirmDetails")) {
					try {
						confPay = XMLConfirmPayment.getConfirmRep(decRec);
					}
					catch (DocumentException de) {
						Log.error("Unable to parse payment confirmation", de);
						continue;
					}
					confPay.setDHTid(dhtId);
					m_IncomingConfirmations.add(confPay);
				}
				else {
					Log.error("Unrecognized DHT record type, decrypted record was: "
								+ decRec);
					continue;
				}
				idx += 1;
			}
			mp.enableMenu();
			m_ParentHome.setCursor(null);
			statBar.remove(decodeProg);
			m_Plugin.setDefaultStatusBar();
			if (dhtList.size() > idx) {
				Log.error((dhtList.size() - idx)
							+ " records on DHT did not parse");
				showError("warning: " + (dhtList.size() - idx) + " records on "
							+ "DHT would not decrypt or did not parse");
			}
		}
		m_GotDHTreply = true;

		/* We are also called when a record has been marked read on the DHT.
		 * Make certain that we throw away any record no longer on the list.
		 */
		for (Integer dhtId : m_DecodedDHTRecs.keySet()) {
			if (!dhtList.containsKey(dhtId)) {
				m_DecodedDHTRecs.remove(dhtId);
			}
		}
	}

	/**
	 * record the results of a pending payments query from the SDS
	 * @param payments contains retrieved pending payments records (encrypted)
	 */
	public void recordPayments(Hashtable<String, String> payments) {
		if (payments == null) {
			Log.error("Null payments hashtable from callback");
			return;
		}
		m_OutstandingPayments.clear();
		m_OutstandingPayments.putAll(payments);
		m_GotSDSreply = true;
	}

	/**
	 * rebuild the list of pending payment records (called by WTabManager, or by
	 * MenuPane or CloseSafePane)
	 */
	public void updatePaymentRecords() {
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		Hashtable<XMLPayment, String> existingPayments
			= m_PaymentTableModel.getPaymentList();
		int pCnt = 0;

		// Examine returned records.  Use a progress bar as we decrypt each.
		JProgressBar decodeProg = new JProgressBar();
		decodeProg.setString("Decoding " + m_OutstandingPayments.size()
									+ " pending payment records in wallet, please wait");
		decodeProg.setStringPainted(true);
		decodeProg.setMinimum(0);
		decodeProg.setMaximum(m_OutstandingPayments.size());
		decodeProg.setOpaque(true);
		decodeProg.setBackground(Color.LIGHT_GRAY);
		decodeProg.setFont(m_ParentHome.M_DataFont);
		JPanel statBar = m_Plugin.getStatusBar();
		MenuPane mp = m_Plugin.getWalletTabManager().getMenuPane();
		if (!m_OutstandingPayments.isEmpty()) {
			m_ParentHome.setCursor(M_WaitCursor);
			mp.disableMenu();
			statBar.removeAll();
			statBar.add(decodeProg);
			statBar.revalidate();
			statBar.repaint();
		}
		for (String fname : m_OutstandingPayments.keySet()) {
			decodeProg.setValue(pCnt+1);

			// see whether we already have this payment decoded in our list
			if (existingPayments.containsValue(fname)) {
				// no need to decrypt this file again
				pCnt += 1;
				continue;
			}

			String encRec = m_OutstandingPayments.get(fname);
			XMLPayment pmt = null;

			// attempt decrypt and parse
			String decRec = EncodingUtils.getStrFromBase64PubkeyEnc(encRec,
																			secrets.getPrivKey());
			if (decRec == null || decRec.isEmpty()) {
				Log.error("Could not decrypt a pending payment, filename " + fname);
			}
			else {
				try {
					pmt = XMLPayment.getPaymentRep(decRec);
					pCnt += 1;
					existingPayments.put(pmt, fname);
				}
				catch (DocumentException de) {
					Log.error("Unable to parse pending payment", de);
				}
			}
		}
		if (!m_OutstandingPayments.isEmpty()) {
			mp.enableMenu();
			m_ParentHome.setCursor(null);
			statBar.remove(decodeProg);
			m_Plugin.setDefaultStatusBar();
		}

		// double-check that all worked
		int diff = m_OutstandingPayments.size() - pCnt;
		if (diff > 0) {
			Log.error("Error parsing " + diff + " payments");
			showError("warning: " + diff + " pending payments "
						+ "would not decrypt or did not parse");
		}
		m_OutstandingPayments.clear();
	}

	/**
	 * process a reply on a request to remove a pending payment record
	 * @param success true if it worked
	 * implied param: payment record will be stored in m_PaymentRecord
	 */
	public void confirmPaymentDeletion(boolean success) {
		m_ParentHome.setCursor(null);
		if (m_PaymentRecord == null) {
			Log.error("conf del cannot find saved XMLPayment record");
			showError("deletion of pending payment record may have failed",
						"internal error, missing data",
						"try clicking on Refresh again to update the payment list");
			return;
		}
		
		if (success) {
			// update the table data, using actual copy
			Hashtable<XMLPayment, String> tData
				= m_PaymentTableModel.getPaymentList();
			tData.remove(m_PaymentRecord);
			if (m_Automation) {
				// add to done list
				m_ProcessedList.add(m_PaymentRecord);
			}
		}
		else {
			showError("warning: deletion of pending payment record failed");
		}
		m_PaymentRecord = null;
	}

	/**
	 * discard a receipt (should be the current item being worked on)
	 * @param receipt the receipt record to throw away
	 */
	public void discardReceipt(XMLReceipt receipt) {
		if (receipt != m_ReceiptRecord) {
			// something is amiss
			Log.error("Receipt being deleted is not the one being worked on");
		}
		m_IncomingReceipts.remove(receipt);
		if (m_Automation) {
			// add to done list
			m_ProcessedList.add(receipt);
			m_TokensUsed += VoucherRequest.M_ReceiptCost;
		}
		m_ReceiptRecord = null;
	}

	/**
	 * discard a confirmation (should be the current item being worked on)
	 * @param conf the confirmation record to throw away
	 */
	public void discardConfirmation(XMLConfirmPayment conf) {
		if (conf != m_PaymentConfirm) {
			// something is amiss
			Log.error("Confirmation being deleted is not the one being worked on");
		}
		m_IncomingConfirmations.remove(conf);
		m_PaymentConfirm = null;
	}

	/**
	 * discard an incoming payment (should be the current item being worked on)
	 * @param payment the pending payment record to throw away
	 * @param tokCost the number of tokens it cost to pick up this payment
	 */
	public void discardPayment(XMLPendingPayment payment, int tokCost) {
		if (payment != m_PendPayRecord) {
			// something is amiss
			Log.error("Pendpay being deleted is not the one being worked on");
		}
		m_IncomingPayments.remove(payment);
		if (m_Automation) {
			// add to done list
			m_ProcessedList.add(payment);
			m_TokensUsed += tokCost;
		}
		m_PendPayRecord = null;
	}


	// get methods
	/**
	 * obtain the list of decoded receipts (on DHT, not stored)
	 * @return the receipt list
	 */
	public ArrayList<XMLReceipt> getReceiptList() { return m_IncomingReceipts; }

	/**
	 * obtain the list of decoded incoming payments (on DHT)
	 * @return the payment list
	 */
	public ArrayList<XMLPendingPayment> getPaymentList() {
		return m_IncomingPayments;
	}

	/**
	 * obtain the list of decoded incoming confirmations (on DHT)
	 * @return the confirmation list
	 */
	public ArrayList<XMLConfirmPayment> getConfirmationList() {
		return m_IncomingConfirmations;
	}

	/**
	 * obtain the table model keeping track of receipts on the SDS
	 * @return the data model
	 */
	public ReceiptTableModel getReceiptModel() { return m_ReceiptTableModel; }

	/**
	 * obtain the table model keeping track of pending payments on the SDS
	 * @return the data model
	 */
	public PaymentTableModel getPaymentModel() { return m_PaymentTableModel; }

	/**
	 * obtain the pending payment record being worked on
	 * @return the payment record
	 */
	public XMLPendingPayment getPendPayRecord() { return m_PendPayRecord; }

	/**
	 * obtain the receipt record being worked on
	 * @return the receipt record
	 */
	public XMLReceipt getReceiptRecord() { return m_ReceiptRecord; }

	/**
	 * obtain the outstanding payment record being worked on
	 * @return the (outgoing) payment record
	 */
	public XMLPayment getPaymentRecord() { return m_PaymentRecord; }

	/**
	 * obtain the payment confirmation record being worked on
	 * @return the payment confirmation
	 */
	public XMLConfirmPayment getConfRecord() { return m_PaymentConfirm; }

	/**
	 * determine whether we're in automated mode
	 * @return true if automating, false if manual
	 */
	public boolean automating() { return m_Automation; }


	// message send methods
	/**
	 * utility method to request a list of pending payments from the SDS
	 * @param synch whether or not we must perform the operation synchronously
	 * @return the reply message from the SDS, in sync mode, or null otherwise
	 */
	public SDSMessage queryPayments(boolean synch) {
		SDSMessage rep = null;
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot list payments while not logged in!");
			showError("cannot list payments", "not logged in", "log in first");
			return rep;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot list payments without SDS connection!");
			showError("cannot list payments", "SDS unavailable",
						"try again later");
			return rep;
		}

		// build message to send
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		SDSMessage getPmts = new SDSMessage();
		getPmts.setType(IQ.Type.GET);
		getPmts.setFrom(sdsConn.getUser());
		getPmts.setTo(sdsConn.getServiceName());
		getPmts.setPacketID("plist_" + state.getNextSDSid());
		getPmts.setOpcode("REQ_list_pending_pmts");
		String rIdx = new String(secrets.getReceiptIndex());
		getPmts.setFolder_index(rIdx); 
		String rCap = new String(secrets.getReceiptRWCap());
		getPmts.setReadcap(rCap);

		// create a packet filter which can look for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getPmts.getPacketID());
		PacketCollector repColl = null;
		if (synch) {
			repColl = sdsConn.createPacketCollector(pFilter);
		}
		else {
			m_Plugin.getSDSListener().addFilter(getPmts.getPacketID(), pFilter);
		}

		// send request to the SDS
		sdsConn.sendPacket(getPmts); 
		//Log.debug(getPmts.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(getPmts);

		// wait for reply if asked to do so
		if (synch) {
			// wait for reply or timeout
			IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
			if (reply == null) {
				// got timeout; build error
				rep = new SDSMessage();
				rep.setOpcode("REP_payment_list");
				rep.setFrom(getPmts.getTo());
				rep.setTo(getPmts.getFrom());
				rep.setType(IQ.Type.ERROR);
				rep.setPacketID(getPmts.getPacketID());
				rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
				rep.setErrmsg("timed out");
			}
			else {
				rep = (SDSMessage) reply;
			}
		}
		return rep;
	}

	/**
	 * utility method to tell the SDS to delete a pending payment
	 * @param pmt the payment record to purge
	 * @param synch whether or not we must execute this operation synchronously
	 * @return the reply from the SDS, in sync mode, otherwise null
	 */
	public SDSMessage sendDelPendPmt(XMLPayment pmt, boolean synch) {
		SDSMessage rep = null;
		if (pmt == null) {
			Log.error("Null payment supplied for deletion");
			return rep;
		}
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot purge pending payments while not logged in!");
			showError("cannot delete pending payments", "not logged in",
						"please log in first");
			return rep;
		}

		// find the filename where this pending payment is presently stored
		Hashtable<XMLPayment, String> tData
			= m_PaymentTableModel.getPaymentList();
		String fName = tData.get(pmt);
		if (fName == null || fName.isEmpty()) {
			Log.error("Cannot find filename for pend payment to be deleted");
			showError("cannot delete pending payments", "filename not found");
			return rep;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot purge pending payments without SDS connection!");
			showError("cannot delete pending payments", "SDS unavailable",
						"try again later");
			return rep;
		}

		// build message to send
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		SDSMessage delPmt = new SDSMessage();
		delPmt.setType(IQ.Type.SET);
		delPmt.setFrom(sdsConn.getUser());
		delPmt.setTo(sdsConn.getServiceName());
		delPmt.setPacketID("pdel_" + state.getNextSDSid());
		delPmt.setOpcode("REQ_delete_pending_pmt");
		String rIdx = new String(secrets.getReceiptIndex());
		delPmt.setFolder_index(rIdx);
		String rwCap = new String(secrets.getReceiptRWCap());
		delPmt.setReadwritecap(rwCap);
		delPmt.setFilename(fName);

		// register reply listener of appropriate type
		PacketIDFilter pFilter = new PacketIDFilter(delPmt.getPacketID());
		PacketCollector repColl = null;
		if (synch) {
			repColl = sdsConn.createPacketCollector(pFilter);
		}
		else {
			m_Plugin.getSDSListener().addFilter(delPmt.getPacketID(), pFilter);
		}

		// send request to the SDS
		sdsConn.sendPacket(delPmt);
		//Log.debug(delPmt.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(delPmt);

		// wait for reply if asked to do so
		if (synch) {
			// wait for reply or timeout
			IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
			if (reply == null) {
				// got timeout; build error
				rep = new SDSMessage();
				rep.setOpcode("REP_payment_deleted");
				rep.setFrom(delPmt.getTo());
				rep.setTo(delPmt.getFrom());
				rep.setPacketID(delPmt.getPacketID());
				rep.setType(IQ.Type.ERROR);
				rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
				rep.setErrmsg("timed out");
			}
			else {
				rep = (SDSMessage) reply;
			}
		}
		return rep;
	}

	/**
	 * method to re-query the DHT for records to process.
	 * @param noDisplay if set, do not shift display until doHousekeeping()
	 * (when reply is received, we will doHousekeeping() again)
	 */
	public void refreshDHT(boolean noDisplay) {
		m_GotDHTreply = false;

		// ask the WTabManager to do it for us
		m_Plugin.getWalletTabManager().queryDHTrecords(noDisplay);
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class (called by WTabManager)
	 */
	public void prepDisplay() {
		// determine whether we're automating
		m_Automation = m_Plugin.getUserPreferences().getAutoHousekeeping();

		// if we have any incoming confirmations, do these directly
		if (!m_IncomingConfirmations.isEmpty()) {
			m_PaymentConfirm = m_IncomingConfirmations.get(0);
			ConfirmPane cp = m_TabManager.getConfirmPane();
			// if we're doing initial login processing, set visibility override
			if (m_TabManager.getLoginInitState()) {
				// actually show the screen to the user
				cp.setDisplayOverride(true);
			}
			else {
				// do processing invisibly
				cp.setDisplayOverride(false);
			}
			m_TabManager.queueTransition(new Integer(WTabManager.M_CONFIRMATION));

			// use another thread to avoid display problems
			Runnable doConf = new Runnable() {
				public void run() {
					m_TabManager.makeTransition();
				}
			};
			SwingUtilities.invokeLater(doConf);
			return;
		}

		// set heading and ensure it's visible
		buildValueHeader();

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);
		m_ButtonPane.removeAll();

		// set screen label
		buildScreenLabel("Incoming items");
		m_TokenCost.setText("Variable tokens required");

		// build lists of outstanding payments which have expired and not expired
		Hashtable<XMLPayment, String> outPayments
			= m_PaymentTableModel.getPaymentList();
		ArrayList<XMLPayment> expPayments
			= new ArrayList<XMLPayment>(outPayments.size());
		ArrayList<XMLPayment> unexpPayments
			= new ArrayList<XMLPayment>(outPayments.size());
		for (XMLPayment pmt : outPayments.keySet()) {
			if (pmt.isExpired()) {
				expPayments.add(pmt);
			}
			else {
				unexpPayments.add(pmt);
			}
		}

		// build ordered processing list
		m_ProcessingList.clear();
		m_ProcessingList.addAll(m_IncomingPayments);
		m_ProcessingList.addAll(m_IncomingReceipts);
		m_ProcessingList.addAll(expPayments);

		// if we're starting processing, record the initial number of items
		final JPanel statBar = m_Plugin.getStatusBar();
		if (m_ResetInit) {
			m_OrgItemCnt = m_ProcessingList.size();
			m_ResetInit = false;

			// set progress bar if we're automating
			if (m_Automation && m_OrgItemCnt > 0) {
				m_AutoProgress.setMaximum(m_OrgItemCnt);
				m_AutoProgress.setString("Processing " + m_OrgItemCnt + " new "
												+ "incoming items");
				// reset bar panel completely, since a token purchase might
				// have overwritten it for us
				statBar.removeAll();
				statBar.add(m_AutoProgress);
				statBar.invalidate();
				statBar.revalidate();
				statBar.repaint();
			}
		}
		else {
			// in auto-mode, set the appropriate value on the status bar
			if (m_Automation) {
				int itemNum = m_OrgItemCnt - m_ProcessingList.size();
				m_AutoProgress.setValue(itemNum);
				// reset bar panel completely, since a token purchase might
				// have overwritten it for us
				statBar.removeAll();
				statBar.add(m_AutoProgress);
				statBar.invalidate();
				statBar.revalidate();
				statBar.repaint();
			}
		}

		if (!m_ProcessingList.isEmpty() && m_Automation) {
			// process the first item in the list
			processNextItem(0);
			return;
		}
		ArrayList<Object> displayList = m_ProcessingList;
		boolean doingReport = false;
		if (m_Automation && !m_ProcessedList.isEmpty()) {
			// go through done items to build summary
			displayList = m_ProcessedList;
			doingReport = true;
			m_OrgItemCnt = 0;
			m_Plugin.setDefaultStatusBar();
		}

		// add the do housekeeping button iff we have any items
		m_StartButton = new JButton(m_StartAction);
		m_StartButton.setToolTipText("Click here to process the next item in "
											+ "your refresh list");
		m_StartButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(Box.createHorizontalStrut(290));		// spacer
		if (!m_ProcessingList.isEmpty()) {
			m_ButtonPane.add(m_StartButton);
		}
		m_ButtonPane.revalidate();
		m_ButtonPane.repaint();

		/* Now provide a scrolling list of all items, placing each into a
		 * JTextField with suitable formatting and background.  We also need
		 * to construct a toolbar recapitulating details and supplying more.
		 */
		JPanel hkDisp = new JPanel();
		hkDisp.setLayout(new GridBagLayout());
		hkDisp.setOpaque(true);
		hkDisp.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;

		// display the list (and headers) in a panel
		if (displayList.isEmpty() && unexpPayments.isEmpty()) {
			// there is nothing, say so
			hkDisp.setLayout(new GridLayout());
			ConcurrentHashMap<Integer, String> dhtList
				= m_Plugin.getDHTListener().getDHTrecs();
			JTextArea noItems
				= new JTextArea("\nThere are no incoming items.", 2, 30);
			noItems.setEditable(false);
			noItems.setBorder(m_InsetBorder);
			noItems.setOpaque(false);
			noItems.setBackground(m_Plugin.m_LabelBackColor);
			noItems.setFocusable(false);
			hkDisp.add(noItems);
		}
		else {
			// clarify vague "Variable tokens required" default
			m_TokenCost.setText(m_TokensUsed + " tokens required");

			// put all items into the display
			JPanel itemPane = new JPanel();
			itemPane.setLayout(new BoxLayout(itemPane, BoxLayout.Y_AXIS));
			itemPane.setOpaque(false);
			boolean firstPend = true, firstRct = true, firstPmt = true;
			boolean firstExp = true, firstConf = true;
			final Dimension itemDim = new Dimension(450, 25);
			String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
			AssetPriceCache apc = m_Plugin.getPricingModule();
			Double price = null;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMinimumFractionDigits(0);
			df.setGroupingSize(3);
			SimpleDateFormat sdf = new SimpleDateFormat(XMLVoucher.M_DateFmt);

			// add a title label, if we're giving a report after the fact
			if (doingReport) {
				installLeftLabel("The following new items were processed:",
										itemPane);
			}

			for (Object hkObj : displayList) {
				JTextField hkObjFld = new JTextField();
				hkObjFld.setEditable(false);
				hkObjFld.setFont(m_ParentHome.M_DataFont);
				hkObjFld.setOpaque(true);
				hkObjFld.setBorder(m_InsetBorder);
				StringBuilder itemTxt = new StringBuilder(128);
				StringBuilder tipTxt = new StringBuilder(256);
				tipTxt.append("<html>");
				df.setMaximumFractionDigits(XMLVoucher.M_Precision);

				// fill in text and colors based on the type of object to process
				if (hkObj instanceof XMLPendingPayment) {
					// put up section header on first one
					if (firstPend) {
						if (!doingReport) {
							installLeftLabel("Vouchers sent to me: "
													+ m_IncomingPayments.size(), itemPane);
						}
						else {
							installLeftLabel("Vouchers sent to me:", itemPane);
						}
						firstPend = false;
					}

					XMLPendingPayment ppay = (XMLPendingPayment) hkObj;
					XMLVoucher incVouch = ppay.getVoucher();

					// from
					itemTxt.append("From: ");
					tipTxt.append("From:  ");
					boolean newMintage = ppay.getPayer().equals(
												m_Plugin.getLoginSecrets().getPublisher());
					if (newMintage) {
						itemTxt.append(ppay.getPayer() + " (VP).  ");
						tipTxt.append(ppay.getPayer() + " (VP).<br/>");
					}
					else {
						itemTxt.append(ppay.getPayer() + ".  ");
						tipTxt.append(ppay.getPayer() + ".<br/>");
					}

					// amount
					itemTxt.append("Amount: ");
					tipTxt.append("Amount:   ");
					String asset = incVouch.getAsset();
					String dispUnits = XMLVoucher.getDisplayUnits(asset);
					double vAmt = incVouch.getNPV();
					itemTxt.append(df.format(vAmt) + " ");
					itemTxt.append(dispUnits + " " + asset + ". ");
					// tooltip only, fiat equiv value
					tipTxt.append(df.format(vAmt) + " ");
					tipTxt.append(dispUnits + " of " + asset + " = ");
					df.setMaximumFractionDigits(2);
					String lookUnits = VPAssetConfig.getPriceLookupUnits(asset,
																			incVouch.getUnits());
					price = apc.getRelativePrice(lookUnits, prefCurrency);
					tipTxt.append(df.format(vAmt * price.doubleValue()));
					tipTxt.append(" " + prefCurrency + ".<br/>");

					if (!doingReport) {
						// tooltip only, pickup date
						Date picDt = new Date(ppay.getTTL());
						tipTxt.append("Pick up by: " + sdf.format(picDt) + ".<br/>");
					}

					// on tooltip only, add baggage details
					TreeMap<String, String> bags = ppay.getBaggageList();
					tipTxt.append(buildBaggageTable(bags, false));
				}
				else if (hkObj instanceof XMLReceipt) {
					// put up section header on first one
					if (firstRct) {
						if (!doingReport) {
							installLeftLabel("Receipts sent to me: "
													+ m_IncomingReceipts.size(), itemPane);
						}
						else {
							installLeftLabel("Receipts sent to me:", itemPane);
						}
						firstRct = false;
					}
					XMLReceipt receipt = (XMLReceipt) hkObj;

					// from
					itemTxt.append("From: " + receipt.getPayee() +".  ");
					tipTxt.append("From:  " + receipt.getPayee() +".<br/>");

					// amount
					double quan = receipt.getQuantity();
					String dispUnits = receipt.getDisplayUnits();
					String asset = receipt.getAsset();
					itemTxt.append("Amount: " + df.format(quan) + " ");
					itemTxt.append(dispUnits + " " + asset + ". ");
					// tooltip only, fiat equiv value
					tipTxt.append("Amount:   " + df.format(quan) + " ");
					tipTxt.append(dispUnits + " of " + asset + " = ");
					df.setMaximumFractionDigits(2);
					String lookUnits = VPAssetConfig.getPriceLookupUnits(asset,
																				receipt.getUnits());
					price = apc.getRelativePrice(lookUnits, prefCurrency);
					tipTxt.append(df.format(quan * price.doubleValue()));
					tipTxt.append(" " + prefCurrency + ".<br/>");

					// on tooltip only, add baggage details
					TreeMap<String, String> bags = receipt.getBaggageList();
					tipTxt.append(buildBaggageTable(bags, true));
				}
				else if (hkObj instanceof XMLPayment) {
					// put up section header on first one
					if (firstPmt) {
						if (doingReport) {
							installLeftLabel("Reclaimed Vouchers I'd sent:", itemPane);
							installLeftLabel("(white = success, gray = failed, see "
													+ "hover text)", itemPane);
						}
						else {
							installLeftLabel("Unclaimed Vouchers I've sent: "
													+ expPayments.size(), itemPane);
						}
						firstPmt = false;
					}

					// to
					XMLPayment pmt = (XMLPayment) hkObj;
					itemTxt.append("To: " + pmt.getPayee() + ".  ");
					tipTxt.append("To:  " + pmt.getPayee() + ".<br/>");
					double quan = pmt.getQuantity();

					// amount
					String dispUnits = pmt.getDisplayUnits();
					itemTxt.append("Amount: " + df.format(quan) + " ");
					String asset = pmt.getAsset();
					itemTxt.append(dispUnits + " " + asset + ". ");
					// tooltip only, add fiat equiv value
					tipTxt.append("Amount:  " + df.format(quan) + " ");
					tipTxt.append(dispUnits + " of " + asset + " = ");
					df.setMaximumFractionDigits(2);
					String lookUnits = VPAssetConfig.getPriceLookupUnits(asset,
																					pmt.getUnits());
					price = apc.getRelativePrice(lookUnits, prefCurrency);
					tipTxt.append(df.format(quan * price.doubleValue()));
					tipTxt.append(" " + prefCurrency + ".<br/>");

					if (!doingReport) {
						// on tooltip only, pickup date
						Date pickupDate = new Date(pmt.getTTL());
						tipTxt.append("Pick up by:  " + sdf.format(pickupDate)
											+ " = ");
						tipTxt.append("Unclaimed.<br/>");
						hkObjFld.setBackground(Color.PINK);
					}
					else {
						// see whether this one was reclaimed successfully
						if (m_SuccessfulReclaims.contains(pmt)) {
							tipTxt.append("Successfully reclaimed.<br/>");
						}
						else {
							hkObjFld.setBackground(Color.LIGHT_GRAY);
							tipTxt.append("This voucher was picked up by the payee, "
												+ "and therefore was not reclaimed.<br/>");
						}
					}

					// on tooltip only, add baggage details
					Hashtable<String, String> bags = pmt.getBaggageList();
					tipTxt.append(buildBaggageTable(bags, false));
				}
				else {
					Log.error("Unexpected object type in refresh list");
					itemTxt.append("Unexpected object type!");
					hkObjFld.setBackground(Color.RED);
				}
				tipTxt.append("</html>");
				hkObjFld.setText(itemTxt.toString());
				hkObjFld.setMaximumSize(itemDim);
				hkObjFld.setToolTipText(tipTxt.toString());
				itemPane.add(hkObjFld);
				itemPane.add(Box.createVerticalStrut(3));
			}

			// add a label showing total tokens expended
			if (doingReport) {
				installLeftLabel("These operations cost a total of "
										+ m_TokensUsed + " tokens.", itemPane);
			}

			// if we have any, show unexpired payments last
			firstPmt = true;
			for (XMLPayment pmt : unexpPayments) {
				JTextField pmtObjFld = new JTextField();
				pmtObjFld.setEditable(false);
				pmtObjFld.setFont(m_ParentHome.M_DataFont);
				pmtObjFld.setOpaque(true);
				pmtObjFld.setBorder(m_InsetBorder);
				StringBuilder itemTxt = new StringBuilder(128);
				StringBuilder tipTxt = new StringBuilder(256);
				tipTxt.append("<html>");
				df.setMaximumFractionDigits(XMLVoucher.M_Precision);

				// put up section header on first one
				if (firstPmt) {
					installLeftLabel("Outstanding Vouchers I've sent: "
											+ unexpPayments.size(), itemPane);
					firstPmt = false;
				}

				// to
				itemTxt.append("To: " + pmt.getPayee() + ".  ");
				tipTxt.append("To:  " + pmt.getPayee() + ".<br/>");
				double quan = pmt.getQuantity();

				// amount
				String dispUnits = pmt.getDisplayUnits();
				itemTxt.append("Amount: " + df.format(quan) + " ");
				String asset = pmt.getAsset();
				itemTxt.append(dispUnits + " " + asset + ". ");
				// tooltip only, add fiat equiv value
				tipTxt.append("Amount:  " + df.format(quan) + " ");
				tipTxt.append(dispUnits + " of " + asset + " = ");
				df.setMaximumFractionDigits(2);
				String lookUnits
					= VPAssetConfig.getPriceLookupUnits(asset, pmt.getUnits());
				price = apc.getRelativePrice(lookUnits, prefCurrency);
				tipTxt.append(df.format(quan * price.doubleValue()));
				tipTxt.append(" " + prefCurrency + ".<br/>");

				// on tooltip only, pickup date
				Date pickupDate = new Date(pmt.getTTL());
				tipTxt.append("Pick up by:  " + sdf.format(pickupDate) + " = ");
				tipTxt.append("Awaiting Pickup<br/>");
				pmtObjFld.setBackground(Color.GREEN);

				// on tooltip only, add baggage details
				Hashtable<String, String> bags = pmt.getBaggageList();
				tipTxt.append(buildBaggageTable(bags, false));

				// complete entry
				tipTxt.append("</html>");
				pmtObjFld.setText(itemTxt.toString());
				pmtObjFld.setMaximumSize(itemDim);
				pmtObjFld.setToolTipText(tipTxt.toString());
				itemPane.add(pmtObjFld);
				itemPane.add(Box.createVerticalStrut(3));
			}

			// place the item list into a scroll pane
			JScrollPane hkSummary = new JScrollPane(itemPane);
			hkSummary.setOpaque(true);
			hkSummary.setPreferredSize(new Dimension(450, 235));
			hkSummary.setBorder(null);
			hkDisp.add(hkSummary, m_GBC);
		}

		// install everything into the main WorkPane scroll area
		m_WorkArea.setViewportView(hkDisp);
	}

	/**
	 * helper method to install a label into a pane with left justification
	 * @param labelTxt the title of the label
	 * @param pane the panel into which to install the label
	 */
	private void installLeftLabel(String labelTxt, JPanel pane) {
		if (labelTxt == null || pane == null) {
			Log.error("Missing arguments");
			return;
		}
		final Dimension labelDim = new Dimension(450, 20);
		JLabel secTitle = new JLabel(labelTxt);
		secTitle.setFont(m_ParentHome.M_LabelFont);
		secTitle.setBorder(m_InsetBorder);
		secTitle.setMinimumSize(labelDim);
		JPanel tPane = new JPanel();
		tPane.setLayout(new GridLayout(1, 0));
		tPane.add(secTitle);
		tPane.setMaximumSize(labelDim);
		pane.add(tPane);
	}

	/**
	 * helper method to build an HTML table out of baggage fields
	 * @param baggage the baggage map
	 * @param multiMsg whether multiple Message fields can be expected
	 * @return the table as text, or the null string if there is no baggage
	 * NB: extremely long fields (over 100 chars) will be broken up
	 */
	public String buildBaggageTable(Map<String, String> bags, boolean multiMsg) {
		if (bags == null || bags.isEmpty()) {
			return "";
		}
		final int lineLen = 30;
		StringBuilder tabTxt = new StringBuilder(1024);
		tabTxt.append("<table width=\"350\" border=\"1\">");
		for (String field : bags.keySet()) {
			String val = bags.get(field);
			// undo any anti-XML-confusion replacements
			val = val.replaceAll(";lt;", "<");
			val = val.replaceAll(";gt;", ">");

			// if the value is >35 chars, break it up at whitespace
			StringBuilder valTxt = new StringBuilder(val.length() + 64);
			if (val.length() <= lineLen) {
				valTxt.append(val);
			}
			else {
				// add tokens until the line length is reached, then add a break
				int cnt = 0;
				StringTokenizer st = new StringTokenizer(val);
				while (st.hasMoreTokens()) {
					String tok = st.nextToken();
					if (cnt+tok.length() <= lineLen) {
						valTxt.append(tok + " ");
						cnt = cnt + tok.length() + 1;
					}
					else {
						valTxt.append("<br/>");
						valTxt.append(tok + " ");
						cnt = tok.length() + 1;
					}
				}
			}

			// the first message is the payee's message, BMessage is the payer's
			tabTxt.append("<tr align=\"left\" valign=\"middle\">");
			if (field.equals("Message")) {
				if (multiMsg) {
					tabTxt.append("<th align=\"left\">Payee " + field + ":</th>");
				}
				else {
					tabTxt.append("<th align=\"left\">" + field + ":</th>");
				}
				tabTxt.append("<td align=\"left\">" + valTxt.toString() + "</td>");
			}
			else if (multiMsg && field.equals("BMessage")) {
				tabTxt.append("<th align=\"left\">Payer Message:</th>");
				tabTxt.append("<td align=\"left\">" + valTxt.toString() + "</td>");
			}
			else if (!field.equals("field_names")) {
				// strip leading B from field name
				tabTxt.append("<th align=\"left\">" +field.substring(1) + ":</th>");
				tabTxt.append("<td align=\"left\">" + valTxt.toString() + "</td>");
			}
			tabTxt.append("</tr>");
		}
		tabTxt.append("</table>");
		return tabTxt.toString();
	}

	/**
	 * set that we are initializing (beginning) processing
	 */
	public void initProcessing() {
		m_ResetInit = true;
		// toss old data if there was any
		m_TokensUsed = 0;
		m_ProcessedList.clear();
		m_SuccessfulReclaims.clear();
	}

	/**
	 * append a reclaimed payment to the list of successfully reclaimed vouchers
	 * @param payment the payment which was successfully reverted
	 */
	public void recordReclaimSuccess(XMLPayment payment) {
		if (payment != null) {
			m_SuccessfulReclaims.add(payment);
		}
	}

	/**
	 * switch to manual processing and continue with next item
	 */
	public void manualOverride() {
		// force preference value to manual housekeeping
		PreferencesUI pui
			= (PreferencesUI) m_Plugin.getUserPreferences().getGUI();
		pui.setAutoHousekeeping("false");

		m_TabManager.queueTransition(
										new Integer(WTabManager.M_HOUSEKEEPING_RETURN));
		Runnable doHKReturn = new Runnable() {
			public void run() {
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(doHKReturn);
	}

	// END methods
}
