/*
 * last modified---
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	03-19-14/jet/ font spacing fixes for OpenJDK-7
 * 	06-05-13/jet/ SilentVault client version
 * 	01-10-11/kbw/ correction to JComboBox width so various wide letters fit
 * 	11-24-10/kbw/ if no last publisher, always present @voucher-safe.org last
 * 	10-13-10/kbw/ compress layout to fit in new reduced height
 * 	10-12-10/kbw/ make enter in text inputs transfer focus
 * 	10-05-10/kbw/ rework for layout; add field listeners
 * 	09-02-10/kbw/ changes for Nimbus L&F
 * 	08-17-10/kbw/ place controls in JPanelS
 * 	08-12-10/kbw/ do not grey out long-phrase pick controls
 * 	07-30-10/kbw/ use preferences to remember the last publisher logged into;
 * 					  add whitespace between login controls and other buttons
 * 	07-22-10/kbw/ sort the publishers alphabetically; supply pub with login
 * 	04-27-10/kbw/ interface changes for improved design and multiple VPs
 * 	03-19-10/kbw/ PINs need to be String, not Integer
 * 	03-17-10/kbw/ back login messages with a timeout, in case we lost
 * 					  network connectivity after starting the app
 * 	02-24-10/kbw/ stand-alone version extending JPanel
 * 	07-22-09/kbw/ use SHA512 instead of SHA1
 * 	06-01-09/kbw/ add initPhase0() for accommodating auto-logout
 * 	05-25-09/kbw/ add button to vector to safe recovery
 * 	04-10-09/kbw/ new
 *
 * purpose---
 * 	provide a UI widget for logging into a voucher safe
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Dimension;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Color;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;
import org.silentvault.client.ui.svm.MTabManager;
import org.silentvault.client.ui.svx.XTabManager;

import java.security.*;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.net.URL;


/**
 * Extends JPanel to provide a screen for logging into a voucher safe.
 * There are two main phases to the login: 1) collection of VS# and PIN; and
 * 2) collection of the three requested chars from the long-phrase, and the
 * passphrase.  Buttons are also provided to vector to new safe creation
 * {@link NewSafePane} and lost safe recovery {@link RecoveryPane}.
 *
 * @author Kevin Wilkerson
 */
public final class LoginPane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * possible chars that can be in a long-phrase
	 */
	private final static String[]	M_Longphrase_Chars = {
		" ", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
		"A", "a", "B", "b", "C", "c", "D", "d", "E", "e",
		"F", "f", "G", "g", "H", "h", "I", "i", "J", "j",
		"K", "k", "L", "l", "M", "m", "N", "n", "O", "o",
		"P", "p", "Q", "q", "R", "r", "S", "s", "T", "t",
		"U", "u", "V", "v", "W", "w", "X", "x", "Y", "y",
		"Z", "z" };

	// states/phases we can be in (possible values of m_CurrentPhase):
	/**
	 * user auto-logged out, prepare to go log in again
	 */
	private final static byte	M_Phase0 = (byte)0;
	/**
	 * obtaining VS# and PIN, building hashkey
	 */
	private final static byte	M_Phase1 = (byte)1;
	/**
	 * obtaining long-phrase excerpt chars and passphrase
	 */
	private final static byte	M_Phase2 = (byte)2;
	/**
	 * decrypting login secrets using passphrase
	 */
	private final static byte	M_Phase3 = (byte)3;

	/**
	 * wait cursor
	 */
	private final Cursor		M_WaitCursor
										= Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);

	/**
	 * the font to use for prompts
	 */
	private final Font		M_PromptFont;

	/**
	 * path to project icon
	 */
	private final static String	M_ProjIconName = "images/SV_Logo.png";


	/**
	 * the plugin/application to which we belong
	 */
	private WalletClient		m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private WTabManager		m_TabManager;

	/**
	 * the VS number entered by the user
	 */
	private String				m_VSnumber;

	/**
	 * the PIN entered by the user
	 */
	private String				m_PIN;

	/**
	 * the passphrase entered by the user
	 */
	private byte[]				m_Passphrase;

	/**
	 * the login-secrets blob, encrypted and base64-encoded
	 */
	private String				m_LoginBlob;

	/**
	 * the message from the OFS specifying the long-phrase challenge chars
	 */
	private OFSMessage		m_LPchallenge;

	/**
	 * the first long-phrase char selected by the user
	 */
	private byte				m_lp_char1;
	/**
	 * the second long-phrase char selected by the user
	 */
	private byte				m_lp_char2;
	/**
	 * the third long-phrase char selected by the user
	 */
	private byte				m_lp_char3;

	/**
	 * the UI object for the first long-phrase character entered by the user
	 */
	private JComboBox<String>	m_longphrase_ch1;
	/**
	 * the UI object for the second long-phrase character entered by the user
	 */
	private JComboBox<String>	m_longphrase_ch2;
	/**
	 * the UI object for the third long-phrase character entered by the user
	 */
	private JComboBox<String>	m_longphrase_ch3;

	/**
	 * which phase we're in (see constants above)
	 */
	private byte				m_CurrentPhase;

	/**
	 * the current title bar text
	 */
	private String				m_TitlebarText;

	/**
	 * the prompt for entering the safe number
	 */
	private JLabel				m_VSprompt;

	/**
	 * the UI object for the safe number
	 */
	private JTextField		m_VSnumberUI;

	/**
	 * prompt for specifying the VP
	 */
	private JComboBox<String>	m_VPselector;

	/**
	 * the prompt for entering the PIN
	 */
	private JLabel				m_PINprompt;

	/**
	 * the UI object for the PIN
	 */
	private JPasswordField	m_PINUI;

	/**
	 * the prompt for entering the passphrase
	 */
	private JLabel				m_PphPrompt;

	/**
	 * the UI object for the passphrase
	 */
	private JPasswordField	m_PPhUI;

	/**
	 * the submit button
	 */
	private JButton			m_SubmitButton;

	/**
	 * the button to vector to new safe creation (NewSafePane)
	 */
	private JButton			m_NewSafeButton;

	/**
	 * the button to vector to safe recovery (RecoveryPane)
	 */
	private JButton			m_RecoverSafeButton;

	/**
	 * the action for handling submission of VS# and PIN
	 */
	private AbstractAction	m_Phase1Submit;

	/**
	 * the action for handling submission of 3 long-phrase chars and passphrase
	 */
	private AbstractAction	m_Phase2Submit;

	/**
	 * the action for handling of click to vector to new safe creation
	 */
	private AbstractAction	m_GoNewSafe;

	/**
	 * the action for handling the click to vector to lost safe recovery
	 */
	private AbstractAction	m_GoRecovSafe;

	/**
	 * the scroll area used to ensure that this tab always displays everything
	 * regardless of window size
	 */
	private JScrollPane		m_ScrollPane;

	/**
	 * all discovered publishers
	 */
	private Vector<String>	m_PublisherList;

	/**
	 * whether a VS# has been entered yet
	 */
	private boolean			m_HaveVSnumber;

	/**
	 * whether a PIN has been entered yet
	 */
	private boolean			m_HavePIN;

	/**
	 * whether a passphrase has been entered yet
	 */
	private boolean			m_HavePassphrase;

	/**
	 * inner class to provide a DocumentListener for the m_VSnumberUI field
	 */
	final class SafeDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = "";
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			if (curLen >= 7 && curTxt.indexOf('-') != -1) {
				m_HaveVSnumber = true;
			}
			else {
				m_HaveVSnumber = false;
			}
			updateContinueButtonStatus();
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = "";
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			if (curLen >= 7 && curTxt.indexOf('-') != -1) {
				m_HaveVSnumber = true;
			}
			else {
				m_HaveVSnumber = false;
			}
			updateContinueButtonStatus();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * document listener for dynamic updates of m_VSnumberUI
	 */
	private SafeDocumentListener	m_VSnumberListener;

	/**
	 * track whether a field entry undo is in progress
	 */
	private boolean					m_DoingUndo;

	/**
	 * inner class to provide a DocumentListener for the m_PINUI field
	 */
	final class PINDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// see if value parses
			if (parseValue()) {
				m_HavePIN = true;
			}
			else {
				m_HavePIN = false;

				// undo this edit
				int changeAt = de.getOffset();
				try {
					curTxt = doc.getText(0, doc.getLength());
					String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
					queueTextMod(m_PINUI, modTxt);
				}
				catch (BadLocationException ble) {
					Log.error("Cannot undo bad PIN edit", ble);
				}
			}
			updateContinueButtonStatus();
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// see if value parses
			m_HavePIN = parseValue();
			updateContinueButtonStatus();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_PINUI as a valid PIN
		 */
		private boolean parseValue() {
			// examine what's in the PIN entry field
			char[] pin = m_PINUI.getPassword();
			String enteredPIN = new String(pin).trim();
			if (enteredPIN.isEmpty()) {
				return false;
			}

			// format allows l to specify long, but we don't
			if (enteredPIN.indexOf('l') != -1 || enteredPIN.indexOf('L') != -1) {
				return false;
			}

			// disallow +-
			if (enteredPIN.startsWith("-") || enteredPIN.startsWith("+")) {
				return false;
			}

			// attempt to parse, very strictly
			int vvv = 0;
			try {
				Integer val = Integer.parseInt(enteredPIN);
				vvv = val.intValue();
			}
			catch (NumberFormatException nfe) {
				return false;
			}

			// NB: we cannot check length, because it can vary with the VP
			return true;
		}
	}

	/**
	 * document listener for the m_PINUI entry field
	 */
	private PINDocumentListener	m_PINListener;

	/**
	 * inner class to provide a DocumentListener for the m_PPhUI field
	 */
	final class PhraseDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 8 chars, max is 24
			if (curLen >= 8 && curLen <= 24) {
				m_HavePassphrase = true;
			}
			else {
				if (curLen > 24) {
					// undo this edit
					int changeAt = de.getOffset();
					try {
						String curTxt = doc.getText(0, doc.getLength());
						String modTxt = curTxt.substring(0, changeAt)
											+ curTxt.substring(changeAt+1);
						queueTextMod(m_PPhUI, modTxt);
					}
					catch (BadLocationException ble) {
						Log.error("Cannot undo bad passphrase edit", ble);
					}
					// NB: this will fix the state of m_HavePassphrase
				}
				else {
					m_HavePassphrase = false;
				}
			}
			updateLoginButtonStatus();
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 8 chars, max is 24
			if (curLen >= 8 && curLen <= 24) {
				m_HavePassphrase = true;
			}
			else {
				m_HavePassphrase = false;
			}
			updateLoginButtonStatus();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * document listener for dynamic updates of the m_PPhUI field
	 */
	private PhraseDocumentListener	m_PassphraseListener;

	/**
	 * inner class to modify a text field from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		JTextComponent	m_ModComponent;
		String			m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());
			if (m_CurrentPhase == M_Phase1) {
				updateContinueButtonStatus();
			}
			else if (m_CurrentPhase == M_Phase2) {
				updateLoginButtonStatus();
			}
		}
	}

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param mgr the tab manager which creates us
	 */
	public LoginPane(WalletClient plugin, WTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_TabManager = mgr;
		m_CurrentPhase = M_Phase1;
		m_VSnumber = m_LoginBlob = m_TitlebarText = m_PIN = "";
		M_PromptFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		m_PublisherList = new Vector<String>();
		// NB: will be replaced in getPublisherList()
		m_VPselector = new JComboBox<String>();

		// use grid layout
		setLayout(new GridLayout());
		setBorder(null);

		// set up to acquire the VS number: prompt and text field
		m_VSprompt = new JLabel("Enter your wallet ID");
		m_VSprompt.setFont(M_PromptFont);
		m_VSnumberListener = new SafeDocumentListener();
		m_VSnumberUI = new JTextField(22);
		m_VSnumberUI.setDropTarget(null);
		m_VSnumberUI.setHorizontalAlignment(JTextField.LEFT);
		m_VSnumberUI.setToolTipText("<html>Your wallet ID resembles an email<br/>"
										+ "address, and is case-insensitive</html>");
		m_VSnumberUI.getDocument().addDocumentListener(m_VSnumberListener);
		m_VSnumberUI.addActionListener(this);
		m_VSprompt.setLabelFor(m_VSnumberUI);

		// set up to acquire the PIN: prompt and text field, with Login button
		m_PINprompt = new JLabel("Enter your PIN");
		m_PINprompt.setFont(M_PromptFont);
		m_PINListener = new PINDocumentListener();
		m_PINUI = new JPasswordField(5);
		m_PINUI.setDropTarget(null);
		m_PINUI.setHorizontalAlignment(JTextField.LEFT);
		m_PINUI.setToolTipText("Your PIN is all digits");
		m_PINUI.setMaximumSize(new Dimension(50, 30));
		m_PINUI.getDocument().addDocumentListener(m_PINListener);
		m_PINUI.addActionListener(this);
		m_PINprompt.setLabelFor(m_PINUI);

		// define the action for going to new safe creation
		m_GoNewSafe = new AbstractAction("Create a new wallet login") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_TabManager.queueTransition(new Integer(WTabManager.M_NEW_SAFE1));
				// NB: ok to do in this thread
				m_TabManager.makeTransition();
			}
		};
		// define the corresponding button
		m_NewSafeButton = new JButton(m_GoNewSafe);
		m_NewSafeButton.setToolTipText("Click here to create a new wallet login");

		// define the action for going to safe recovery
		m_GoRecovSafe = new AbstractAction("Go to lost wallet recovery") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_TabManager.queueTransition(new Integer(WTabManager.M_RECOV1));
				// NB: ok to do in this thread
				m_TabManager.makeTransition();
			}
		};
		// define the corresponding button
		m_RecoverSafeButton = new JButton(m_GoRecovSafe);
		m_RecoverSafeButton.setToolTipText("<html>Click here to recover your<br/>"
													+ "wallet login information</html>");

		// define phase 2 UI widgets
		m_PphPrompt = new JLabel("Enter your complete passphrase");
		m_PphPrompt.setFont(M_PromptFont);
		m_PassphraseListener = new PhraseDocumentListener();
		m_PPhUI = new JPasswordField(12);
		m_PPhUI.setDropTarget(null);
		m_PPhUI.setHorizontalAlignment(JTextField.LEFT);
		m_PPhUI.setFont(M_PromptFont);
		m_PPhUI.setToolTipText("Your passphrase is alphanumeric");
		m_PPhUI.getDocument().addDocumentListener(m_PassphraseListener);
		m_PPhUI.addActionListener(this);
		m_PphPrompt.setLabelFor(m_PPhUI);

		// define the action for handling submission of long-phrase and passphrase
		m_Phase2Submit = new AbstractAction("Login") {
			public void actionPerformed(ActionEvent actionEvent) {
				// retrieve the 3 chars from the combo boxes, which can't be spaces
				String ch = (String) m_longphrase_ch1.getSelectedItem();
				if (ch.charAt(0) == ' ') {
					showError("missing long-phrase", 
								"character " + m_LPchallenge.getLong_phrase_index1()
								+ " not selected", "please make selection");
					return;
				}
				m_lp_char1 = ch.getBytes()[0];
				ch = (String) m_longphrase_ch2.getSelectedItem();
				if (ch.charAt(0) == ' ') {
					showError("missing long-phrase", 
								"character " + m_LPchallenge.getLong_phrase_index2()
								+ " not selected", "please make selection");
					return;
				}
				m_lp_char2 = ch.getBytes()[0];
				ch = (String) m_longphrase_ch3.getSelectedItem();
				if (ch.charAt(0) == ' ') {
					showError("missing long-phrase", 
								"character " + m_LPchallenge.getLong_phrase_index3()
								+ " not selected", "please make selection");
					return;
				}
				m_lp_char3 = ch.getBytes()[0];

				// retrieve the passphrase
				String pPh = new String(m_PPhUI.getPassword()).trim();
				if (!setPassphrase(pPh)) {
					showError("missing passphrase", "none entered",
								"please enter your passphrase");
					return;
				}

				// send the next login message to the OFS
				sendPerformLogin();
			}
		};
	}

	/**
	 * initialize for phase 0, auto-logged out
	 */
	public void initPhase0() {
		m_TitlebarText = "Wallet Logged Out";
		m_CurrentPhase = M_Phase0;
		removeAll();

		// put up message about timeout, and button to go to re-login
		JLabel tout = new JLabel("Your wallet session timed out.");
		m_Phase1Submit = new AbstractAction("Login") {
			public void actionPerformed(ActionEvent actionEvent) {
				// re-establish OFS connection first
				AbstractAction connAction = m_Plugin.getConnectAction();
				connAction.actionPerformed(actionEvent);

				// now go to actual login screen
				removeAll();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();
			}
		};
		m_SubmitButton = new JButton(m_Phase1Submit);
		m_SubmitButton.setToolTipText("Click here to log in again");
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		tout.setLabelFor(m_SubmitButton);

		// put everything in a border, inside the scroll area
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel loginPane = new JPanel(new FlowLayout(FlowLayout.CENTER, 50, 100));
		loginPane.setPreferredSize(new Dimension(450, 50));
		loginPane.add(tout);
		loginPane.add(m_SubmitButton);
		m_ScrollPane = new JScrollPane(loginPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}
	}

	/**
	 * initialize for phase 1
	 */
	public void initPhase1() {
		m_TitlebarText = "Wallet Login (Step 1)";
		m_CurrentPhase = M_Phase1;
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();

		// define the action for handling submission of VS# and PIN
		m_Phase1Submit = new AbstractAction("Continue") {
			public void actionPerformed(ActionEvent actionEvent) {
				// retrieve VS#
				String vsNum = m_VSnumberUI.getText().trim();
				if (vsNum == null || vsNum.isEmpty()) {
					showError("missing wallet ID", "none entered",
								"please enter a wallet ID");
					return;
				}
				// strip it if they entered the VP manually
				int atIdx = vsNum.indexOf('@');
				if (atIdx != -1) {
					vsNum = vsNum.substring(0, atIdx);
				}
				String selectedVP = (String) m_VPselector.getSelectedItem();
				if (!setVSnumber(vsNum + selectedVP)) {
					showError("missing wallet ID",
								"wallet ID entered was improper",
								"please try again");
					return;
				}

				// retrieve PIN
				char[] pin = m_PINUI.getPassword();
				if (pin == null || pin.length == 0) {
					showError("missing PIN", "none entered",
								"please enter your PIN");
					return;
				}
				if (!setPIN(new String(pin).trim())) {
					showError("missing PIN", "PIN entered was improper",
								"please try again");
					return;
				}

				// send initiate login to OFS
				sendInitLogin();
			}
		};

		// this panel contains the VS# entry and the VP drop-down list
		JPanel vsGrp = new JPanel();
		vsGrp.setLayout(new BoxLayout(vsGrp, BoxLayout.X_AXIS));
		vsGrp.add(m_VSnumberUI);
		vsGrp.add(m_VPselector);

		// this panel contains the entry field and login button
		JPanel pinGrp = new JPanel();
		pinGrp.setLayout(new BoxLayout(pinGrp, BoxLayout.X_AXIS));
		pinGrp.add(m_PINUI);
		m_SubmitButton = new JButton(m_Phase1Submit);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		pinGrp.add(m_SubmitButton);
		updateContinueButtonStatus();

		// make a borderless panel for the last two buttons, one beside the other
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setPreferredSize(new Dimension(450, 30));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalGlue());
		m_NewSafeButton.setFont(m_Plugin.M_ButtonFont);
		buttonPane.add(m_NewSafeButton);
		buttonPane.add(Box.createHorizontalStrut(30));	// spacer
		m_RecoverSafeButton.setFont(m_Plugin.M_ButtonFont);
		buttonPane.add(m_RecoverSafeButton);
		buttonPane.add(Box.createHorizontalGlue());

		// obtain the project icon and place it into a label
		URL imgURL = m_Plugin.getClass().getResource(M_ProjIconName);
		ImageIcon projIcon = null;
		if (imgURL == null) {
			Log.error("Could not find " + M_ProjIconName);
		}
		else {
			projIcon = new ImageIcon(imgURL, "silentvault.com icon");
		}
		JLabel pIcon = new JLabel(projIcon);

		// now put everything into a grid bag inside the scroll area
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel loginPane = new JPanel();
		loginPane.setBorder(null);
		GridBagLayout gbl = new GridBagLayout();
		loginPane.setLayout(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 13;
		gbc.anchor = GridBagConstraints.NORTH;
		loginPane.add(Box.createVerticalGlue(), gbc);
		loginPane.add(m_VSprompt, gbc);
		loginPane.add(Box.createVerticalStrut(5), gbc);
		loginPane.add(vsGrp, gbc);
		loginPane.add(Box.createVerticalStrut(30), gbc);
		loginPane.add(m_PINprompt, gbc);
		loginPane.add(Box.createVerticalStrut(5), gbc);
		loginPane.add(pinGrp, gbc);
		loginPane.add(Box.createVerticalStrut(40), gbc);
		loginPane.add(pIcon, gbc);
		loginPane.add(Box.createVerticalStrut(30), gbc);
		loginPane.add(buttonPane, gbc);
		loginPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(loginPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}

		// set focus in VS# input
		m_VSnumberUI.requestFocusInWindow();
	}

	/**
	 * initialize for phase 2
	 */
	public void initPhase2() {
		if (m_LPchallenge == null) {
			Log.error("Null long-phrase challenge, cannot init phase 2 login");
			return;
		}
		m_TitlebarText = "Wallet Login (Step 2)";
		m_CurrentPhase = M_Phase2;
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();

		// record data from input message (NB: these indices are 1-based)
		int idx1 = m_LPchallenge.getLong_phrase_index1();
		int idx2 = m_LPchallenge.getLong_phrase_index2();
		int idx3 = m_LPchallenge.getLong_phrase_index3();
		int longphrase_len = m_LPchallenge.getLong_phrase_len();
		if (longphrase_len > NewSafePane.M_Longphrase_Max
			 || longphrase_len < NewSafePane.M_Longphrase_Min)
		{
			Log.error("Invalid long-phrase length indicated, " + longphrase_len);
		}

		/* Set up to acquire the 3 indicated chars from the long-phrase.  We do
		 * this by presenting an X for the characters that are not wanted, with
		 * a drop-down list for the characters that are wanted.
		 * At the same time, we build a prompt bar containing non-selectable
		 * buttons which describe the character number wanted.  This goes above
		 * the actual selection tool.
		 */
		JLabel lpPrompt = new JLabel("Select the indicated characters from "
											+ "within your long-phrase");
		lpPrompt.setFont(M_PromptFont);
		JPanel promptBar = new JPanel();
		promptBar.setLayout(new BoxLayout(promptBar, BoxLayout.X_AXIS));

		// define pixel height for labels, and distinct fonts
		final int labelHeight = 30;
		final Font labFnt = new Font(Font.SANS_SERIF, Font.ITALIC, 14);
		final Font fillFnt = new Font(Font.SANS_SERIF, Font.PLAIN, 13);

		// make separate pane group for chars to get them all in a line
		JPanel charGroup = new JPanel();
		charGroup.setLayout(new BoxLayout(charGroup, BoxLayout.LINE_AXIS));

		// add a number of Xs that represents the leading chars not used
		charGroup.add(Box.createHorizontalStrut(2));
		StringBuilder fill = new StringBuilder(32);
		if (idx1 > 1) {
			final int diff = idx1 - 1;
			for (int iii = 0; iii < diff; iii++) {
				fill.append(" X ");
			}
			JTextField fill1 = new JTextField(fill.length() / 2);
			fill1.setDropTarget(null);
			fill1.setHorizontalAlignment(JTextField.CENTER);
			fill1.setFont(fillFnt);
			fill1.setText(fill.toString());
			fill1.setEditable(false);
			fill1.setFocusable(false);
			charGroup.add(fill1);
			charGroup.add(Box.createHorizontalStrut(2));

			// place a strut of equal width into the prompt bar
			Dimension fillSz = fill1.getPreferredSize();
			int pWid = (int)fillSz.getWidth() + 4;
			promptBar.add(Box.createHorizontalStrut(pWid));
		}
		else {
			promptBar.add(Box.createHorizontalStrut(3));
		}

		m_longphrase_ch1 = new JComboBox<String>(M_Longphrase_Chars);
		m_longphrase_ch1.setToolTipText("Pick long-phrase character #" + idx1);
		m_longphrase_ch1.addActionListener(this);

		/* The act of querying the JComboBox for its preferred size causes it
		 * to shrink that size to a point where certain wide letters, such as
		 * U, D, G, H, m, N, O, Q, w, X, Z, and especially W and M, will no longer
		 * fit in the box and will be displayed as '...'.  To prevent this, we
		 * now need to add 3 pixels back to the width.  This is almost certainly
		 * a bug in Java-6, and may need revisiting in Java-7.
		 */
		Dimension boxDim = m_longphrase_ch1.getPreferredSize();
		int dimx = (int) boxDim.getWidth() + 3;
		int dimy = (int) boxDim.getHeight();
		boxDim.setSize(dimx, dimy);
		m_longphrase_ch1.setPreferredSize(boxDim);

		charGroup.add(m_longphrase_ch1);
		charGroup.add(Box.createHorizontalStrut(2));

		// prompt label
		JLabel ch1L = new JLabel(getNumberingStr(idx1));
		ch1L.setFont(labFnt);
		ch1L.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY.darker()));
		ch1L.setMinimumSize(new Dimension((int)boxDim.getWidth(), labelHeight));
		ch1L.setSize(ch1L.getMinimumSize());
		ch1L.setMaximumSize(ch1L.getMinimumSize());
		promptBar.add(ch1L);
		promptBar.add(Box.createHorizontalStrut(2));

		// add a number of Xs that represents the chars between index 1 and 2
		if (idx2-idx1 > 1) {
			final int diff = idx2 - idx1 - 1;
			fill = new StringBuilder(32);
			for (int iii = 0; iii < diff; iii++) {
				fill.append(" X ");
			}
			JTextField fill2 = new JTextField(fill.length() / 2);
			fill2.setDropTarget(null);
			fill2.setHorizontalAlignment(JTextField.CENTER);
			fill2.setFont(fillFnt);
			fill2.setText(fill.toString());
			fill2.setEditable(false);
			fill2.setFocusable(false);
			charGroup.add(fill2);
			charGroup.add(Box.createHorizontalStrut(2));

			// place a strut of equal width into the prompt bar
			Dimension fillSz = fill2.getPreferredSize();
			int pWid = (int)fillSz.getWidth() + 4;
			promptBar.add(Box.createHorizontalStrut(pWid));
		}
		m_longphrase_ch2 = new JComboBox<String>(M_Longphrase_Chars);
		m_longphrase_ch2.setToolTipText("Pick long-phrase character #" + idx2);
		m_longphrase_ch2.addActionListener(this);
		charGroup.add(m_longphrase_ch2);
		charGroup.add(Box.createHorizontalStrut(2));

		// prompt label
		JLabel ch2L = new JLabel(getNumberingStr(idx2));
		ch2L.setFont(labFnt);
		ch2L.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY.darker()));
		ch2L.setMinimumSize(new Dimension((int)boxDim.getWidth(), labelHeight));
		ch2L.setMaximumSize(ch2L.getMinimumSize());
		ch2L.setSize(ch2L.getMinimumSize());
		promptBar.add(ch2L);
		promptBar.add(Box.createHorizontalStrut(2));

		// add a number of Xs that represents the chars between index 2 and 3
		if (idx3-idx2 > 1) {
			final int diff = idx3 - idx2 - 1;
			fill = new StringBuilder(32);
			for (int iii = 0; iii < diff; iii++) {
				fill.append(" X ");
			}
			JTextField fill3 = new JTextField(fill.length() / 2);
			fill3.setDropTarget(null);
			fill3.setHorizontalAlignment(JTextField.CENTER);
			fill3.setFont(fillFnt);
			fill3.setText(fill.toString());
			fill3.setEditable(false);
			fill3.setFocusable(false);
			charGroup.add(fill3);
			charGroup.add(Box.createHorizontalStrut(2));

			// place a strut of equal width into the prompt bar
			Dimension fillSz = fill3.getPreferredSize();
			int pWid = (int)fillSz.getWidth() + 4;
			promptBar.add(Box.createHorizontalStrut(pWid));
		}
		m_longphrase_ch3 = new JComboBox<String>(M_Longphrase_Chars);
		m_longphrase_ch3.setToolTipText("Pick long-phrase character #" + idx3);
		m_longphrase_ch3.addActionListener(this);
		charGroup.add(m_longphrase_ch3);
		charGroup.add(Box.createHorizontalStrut(2));

		// prompt label
		JLabel ch3L = new JLabel(getNumberingStr(idx3));
		ch3L.setFont(labFnt);
		ch3L.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY.darker()));
		ch3L.setMinimumSize(new Dimension((int)boxDim.getWidth(), labelHeight));
		ch3L.setMaximumSize(ch3L.getMinimumSize());
		ch3L.setSize(ch3L.getMinimumSize());
		promptBar.add(ch3L);

		// add a number of Xs that represents any trailing chars unused
		if (longphrase_len > idx3) {
			final int diff = longphrase_len - idx3;
			fill = new StringBuilder(32);
			for (int iii = 0; iii < diff; iii++) {
				fill.append(" X ");
			}
			JTextField fill4 = new JTextField(fill.length() / 2);
			fill4.setDropTarget(null);
			fill4.setHorizontalAlignment(JTextField.CENTER);
			fill4.setFont(fillFnt);
			fill4.setText(fill.toString());
			fill4.setEditable(false);
			fill4.setFocusable(false);
			charGroup.add(fill4);
			charGroup.add(Box.createHorizontalStrut(2));
		}

		// enclosing pane to get the prompt bar close to the char picking line
		JPanel charDisp = new JPanel(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		charDisp.add(promptBar, gbc);
		charDisp.add(Box.createVerticalStrut(10), gbc);
		charDisp.add(charGroup, gbc);

		// set up pane to acquire the passphrase
		JPanel ppPane = new JPanel(new GridBagLayout());
		gbc.fill = GridBagConstraints.NONE;
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		ppPane.add(m_PPhUI, gbc);
		m_SubmitButton = new JButton(m_Phase2Submit);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		updateLoginButtonStatus();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		ppPane.add(m_SubmitButton, gbc);

		// put everything in a grid bag with border, inside a scroll area
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel loginPane = new JPanel(new GridBagLayout());
		gbc.fill = GridBagConstraints.NONE;
		loginPane.add(Box.createVerticalGlue(), gbc);
		loginPane.add(Box.createVerticalStrut(20), gbc);
		loginPane.add(lpPrompt, gbc);
		loginPane.add(Box.createVerticalStrut(15), gbc);
		loginPane.add(charDisp, gbc);
		loginPane.add(Box.createVerticalStrut(45), gbc);
		loginPane.add(m_PphPrompt, gbc);
		loginPane.add(Box.createVerticalStrut(5), gbc);
		loginPane.add(ppPane, gbc);
		loginPane.add(Box.createVerticalStrut(100), gbc);
		loginPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(loginPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}

		// set focus to first pick control
		m_longphrase_ch1.requestFocusInWindow();
	}


	// get methods
	/**
	 * obtain VS number
	 * @return the voucher safe number, or the empty string if not yet set
	 */
	public String getVSnumber() { return m_VSnumber; }

	/**
	 * obtain PIN
	 * @return the PIN, or empty string if not yet set
	 */
	public String getPIN() { return m_PIN; }


	// set methods
	/**
	 * set VS number
	 * @param vs the number of the voucher safe
	 * @return true if VS# looked legal, false otherwise
	 */
	public boolean setVSnumber(String vs) {
		if (vs != null && !vs.isEmpty()) {
			if (vs.indexOf('@') == -1 || vs.indexOf('.') == -1) {
				return false;
			}
			m_VSnumber = new String(vs);
			return true;
		}
		return false;
	}

	/**
	 * set the PIN number
	 * @param pinStr the string containing the PIN
	 * @return true if PIN was legal (all numbers), false otherwise
	 */
	public boolean setPIN(String pinStr) {
		if (pinStr != null && !pinStr.isEmpty()) {
			try {
				Integer pin = Integer.parseInt(pinStr);
			}
			catch (NumberFormatException nfe) {
				return false;
			}
			m_PIN = pinStr;
			return true;
		}
		return false;
	}

	/**
	 * set the passphrase
	 * @param pph the passphrase
	 */
	public boolean setPassphrase(String pph) {
		if (pph != null && !pph.isEmpty()) {
			m_Passphrase = pph.getBytes();
			return true;
		}
		return false;
	}

	/**
	 * set (record) the login blob (invoked by OFS callback)
	 * @param secrets the encrypted secrets returned from the SDS
	 */
	public void setLoginblob(String secrets) {
		if (secrets != null && !secrets.isEmpty()) {
			m_LoginBlob = new String(secrets);

			// we are now in the final phase, decoding using the passphrase
			m_CurrentPhase = M_Phase3;
		}
	}

	/**
	 * set (record) the long-phrase challenge chars (invoked by OFS callback)
	 */
	public void setChallenge(OFSMessage lpChallenge) {
		if (lpChallenge != null) {
			m_LPchallenge = lpChallenge;
		}
	}


	// message-sending methods
	/**
	 * send a synchronous REQ_create_safe to the OFS, to obtain the VP list
	 */
	public void getPublisherList() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("No OFS connection, cannot get voucher publisher list");
			return;
		}
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage getVPs = new OFSMessage();
		getVPs.setType(IQ.Type.GET);
		getVPs.setFrom(conn.getUser());
		getVPs.setTo(conn.getServiceName());
		getVPs.setPacketID("create_" + stData.getNextOFSid());
		getVPs.setOpcode("REQ_create_safe");

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getVPs.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(getVPs);

		// record outgoing packet
		stData.setLastActivity(getVPs);

		// wait for reply or timeout
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; show error
			showError("cannot log in", "timeout obtaining publisher list",
						"please try again later");
			return;
		}
		rep = (OFSMessage) reply;

		// record publishers listed
		List<String> vps = rep.getVoucherPublishers();
		m_PublisherList.clear();
		m_PublisherList.addAll(vps);
		if (vps.isEmpty()) {
			showError("cannot log in", "no voucher publishers found",
						"please try again later");
			return;
		}

		// load the UI element, which for display purposes deals with domains only
		Object[] vpArray = new Object[vps.size()];
		int idx = 0;
		for (String vp : vps) {
			int atIdx = vp.indexOf('@');
			vpArray[idx++] = vp.substring(atIdx);
		}
		Arrays.sort(vpArray);

		VSPreferences prefs = m_Plugin.getUserPreferences();
		String prefPub = prefs.getPrefPublisher();
		Vector<String> vpList = new Vector<String>(vps.size());

		// build the list, in order: preference first, default network last
		int defIdx = 0;
		boolean defFound = false;
		final String def = "@voucher-safe.org";
		for (int iii = 0; iii < vpArray.length; iii++) {
			String pub = (String) vpArray[iii];
			if (pub.equals(prefPub)) {
				// if we have a recorded preference for our publisher, make it first
				vpList.add(0, pub);
			}
			else if (pub.equals(def)) {
				defFound = true;
				vpList.add(pub);
			}
			else {
				vpList.add(pub);
			}
		}
		if (defFound && vpArray.length > 1) {
			// move the default test network to last
			defIdx = vpList.indexOf(def);
			int last = vpArray.length - 1;
			if (defIdx != last) {
				String other = vpList.set(last, def);
				vpList.set(defIdx, other);
			}
		}
		m_VPselector = new JComboBox<String>(vpList);
		m_VPselector.setFocusable(false);
		m_VPselector.setToolTipText("<html>Select the voucher publisher<br/>"
											+ "where your wallet was created</html>");
	}

	/**
	 * send a REQ_initiate_login to the OFS (synchronous)
	 */
	private void sendInitLogin() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("No OFS connection, cannot initiate login");
			return;
		}
		VscState stData = m_Plugin.getVscState();

		// wait cursor
		m_ScrollPane.setCursor(M_WaitCursor);

		// build message to send
		OFSMessage logSt = new OFSMessage();
		logSt.setType(IQ.Type.GET);
		logSt.setFrom(conn.getUser());
		logSt.setTo(conn.getServiceName());
		logSt.setPacketID("login_" + stData.getNextOFSid());
		logSt.setOpcode("REQ_initiate_login");

		// build the hash of VS# + PIN
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Can't build SHA-512 hash", nsae);
		}
		md.update(m_VSnumber.getBytes());
		md.update(m_PIN.getBytes());
		String hashKey = StringUtils.encodeBase64(md.digest(), false);
		logSt.setHashkey(hashKey);

		// add the publisher, in full, based on the @domain selected
		String selectedVP = (String) m_VPselector.getSelectedItem();
		for (String pub : m_PublisherList) {
			if (pub.lastIndexOf(selectedVP) != -1) {
				logSt.setVoucher_publisher(pub);
				break;
			}
		}

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(logSt.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(logSt);
		//Log.debug("REQ_initiate_login sent with hashkey: " + hashKey);

		// record outgoing packet
		stData.setLastActivity(logSt);

		// wait for reply or timeout
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ScrollPane.setCursor(null);
		if (reply == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_long-phrase_challenge");
			rep.setFrom(logSt.getTo());
			rep.setTo(logSt.getFrom());
			rep.setPacketID(logSt.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (OFSMessage) reply;
		}

		// invoke callback with result
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * send a REQ_perform_login to the OFS (synchronous)
	 */
	private void sendPerformLogin() {
		// we don't need to do this if we're merely resubmitting the passphrase
		if (m_CurrentPhase == M_Phase3) {
			if (decodeLoginblob()) {
				m_Plugin.getVscState().setLoggedIn(true);
				if (!m_Plugin.getTokenCache().initSpentTokens()) {
					Log.debug("No list of previously spent tokens found");
				}

				// transition to wallet home
				m_TabManager.queueTransition(new Integer(WTabManager.M_HOME_INIT));

				// reconfigure the SVX and SSM connect panes
				final XTabManager xtm = m_Plugin.getExchangeTabManager();
				xtm.queueTransition(new Integer(XTabManager.M_CONNECT));
				final MTabManager mtm = m_Plugin.getMarketplaceTabManager();
				mtm.queueTransition(new Integer(MTabManager.M_LOGIN));

				// not okay to do in this thread, because it might exit
				Runnable goHome = new Runnable() {
					public void run() {
						m_TabManager.makeTransition();
						xtm.makeTransition();
						mtm.makeTransition();
					}
				};
				new Thread(goHome).start();
			}
			return;
		}

		// wait cursor
		m_ScrollPane.setCursor(M_WaitCursor);

		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// Double-check we previously sent REQ_initiate_login and got reply.
		// If not, recompute hash key.

		String hashKey = null;
		IQ lastOFS = stData.getLastSentOFS();
		OFSMessage lastToOfs = null;
		if (lastOFS instanceof OFSMessage) {
			lastToOfs = (OFSMessage) lastOFS;
		}
		if (lastToOfs == null
			|| !lastToOfs.getOpcode().equalsIgnoreCase("REQ_initiate_login"))
		{
			// build the hash of VS# + PIN
			MessageDigest md = null;
			try {
				md = MessageDigest.getInstance("SHA-512");
			}
			catch (NoSuchAlgorithmException nsae) {
				Log.error("Can't build SHA-512 hash", nsae);
			}
			md.update(m_VSnumber.getBytes());
			md.update(m_PIN.getBytes());
			hashKey = StringUtils.encodeBase64(md.digest(), false);
		}
		else {
			// do not recompute hashkey
			hashKey = lastToOfs.getHashkey();
		}

		// build the message to send
		OFSMessage perfLogin = new OFSMessage();
		perfLogin.setType(IQ.Type.GET);
		perfLogin.setFrom(conn.getUser());
		perfLogin.setTo(conn.getServiceName());
		perfLogin.setPacketID("login_" + stData.getNextOFSid());
		perfLogin.setOpcode("REQ_perform_login");
		perfLogin.setHashkey(hashKey);
		perfLogin.setLong_phrase_char1(m_lp_char1);
		perfLogin.setLong_phrase_char2(m_lp_char2);
		perfLogin.setLong_phrase_char3(m_lp_char3);

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(perfLogin.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(perfLogin);

		// record outgoing packet
		stData.setLastActivity(perfLogin);

		// wait for reply or timeout
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ScrollPane.setCursor(null);
		if (reply == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_login-blob");
			rep.setFrom(perfLogin.getTo());
			rep.setTo(perfLogin.getFrom());
			rep.setPacketID(perfLogin.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (OFSMessage) reply;
		}

		// invoke callback with result
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * decrypt and parse the login blob
	 * NB: implicit input is a previously set value of m_LoginBlob
	 * @return true on successful decode, false otherwise
	 */
	public boolean decodeLoginblob() {
		if (m_CurrentPhase != M_Phase3 || m_LoginBlob.isEmpty()) {
			Log.error("Cannot decode empty login secrets blob");
			return false;
		}

		// attempt decrypt
		String decode
			= EncodingUtils.getStrFromBase64DESEnc(m_LoginBlob, m_Passphrase);
		if (decode == null || decode.isEmpty()) {
			showError("could not log in",
						"no match using entered long-phrase and passphrase",
						"please try again");
			return false;
		}

		// add the XML tags
		String clearBlob = "<secret_details>\n" + decode + "</secret_details>\n";

		// now parse the blob
		Element detElt = EncodingUtils.buildElement(clearBlob);
		if (detElt == null) {
			Log.error("Could not parse login blob details");
			showError("unable to parse login secret details", "corrupt data");
			return false;
		}

		String publisher = null;
		X509Certificate cert = null;
		PrivateKey privKey = null;
		String vouchIndex = null;
		String vouchRdStr = null;
		String receiptIndex = null;
		String recRwStr = null;
		String vtRwStr = null;
		try {
			// get voucher publisher
			Element pubElt = detElt.element("voucher_publisher");
			if (pubElt == null) {
				throw new DocumentException("Missing voucher_publisher element");
			}
			else {
				publisher = pubElt.getTextTrim();
				if (publisher.isEmpty()) {
					throw new DocumentException("Empty voucher_publisher field");
				}
			}

			// get certificate
			Element certElt = detElt.element("certificate");
			if (certElt == null) {
				throw new DocumentException(
										"Could not find certificate in login blob");
			}
			String certRawBase64 = certElt.getTextTrim().replace(' ', '\n');
			String certStr = EncodingUtils.M_CertSt + "\n" + certRawBase64
								  + "\n" + EncodingUtils.M_CertEnd;
			cert = EncodingUtils.getCertFromBase64Str(certStr);
			if (cert == null) {
				throw new DocumentException(
									"Could not parse certificate in login blob");
			}

			// get private key
			Element keyElt = detElt.element("private_key");
			if (keyElt == null) {
				throw new DocumentException(
									"Could not find private key in login blob");
			}
			privKey
				= EncodingUtils.getPrivkeyFromBase64Str(keyElt.getTextTrim());
			if (privKey == null) {
				throw new DocumentException(
									"Could not parse private key in login blob");
			}

			// get voucher/token folder index
			Element vtIdxElt = detElt.element("voucher_index");
			if (vtIdxElt == null) {
				throw new DocumentException(
									"Missing voucher/token index in login blob");
			}
			vouchIndex = vtIdxElt.getTextTrim();

			// get voucher/token folder read cap
			Element vtReadElt = detElt.element("voucher_rd_cap");
			if (vtReadElt == null) {
				throw new DocumentException(
						"Could not find voucher/token read cap in login blob");
			}
			vouchRdStr = vtReadElt.getTextTrim();
			if (vouchRdStr.isEmpty()) {
				throw new DocumentException(
									"Empty voucher/token read cap in login blob");
			}
			
			// get receipt/payment folder index
			Element rpIdxElt = detElt.element("receipts_index");
			if (rpIdxElt == null) {
				throw new DocumentException(
								"Missing receipts/payments index in login blob");
			}
			receiptIndex = rpIdxElt.getTextTrim();

			// get receipt/payment folder read-write cap
			Element rpRwElt = detElt.element("receipts_rw_cap");
			if (rpRwElt == null) {
				throw new DocumentException(
							"Could not find receipts/payments read-write cap "
							+ "in login blob");
			}
			recRwStr = rpRwElt.getTextTrim();
			if (recRwStr.isEmpty()) {
				throw new DocumentException(
						"Null receipts/payments read-write cap in login blob");
			}

			// get voucher/token folder read-write cap (encrypted and opaque)
			Element vtRwElt = detElt.element("voucher_rw_cap");
			if (vtRwElt == null) {
				throw new DocumentException(
								"Could not find voucher/token read-write cap in "											+ "login blob");
			}
			vtRwStr = vtRwElt.getTextTrim();
			if (vtRwStr.isEmpty()) {
				throw new DocumentException(
							"Null voucher/token read-write cap in login blob");
			}
		}
		catch (DocumentException de) {
			Log.error("Login blob parse error: ", de);
			showError("could not parse login secrets!", "corrupt data");
			return false;
		}

		// compute the DHT marking hash, hash(VS#+PIN+passphrase)
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			return false;
		}
		md.reset();
		md.update(m_VSnumber.getBytes());
		md.update(m_PIN.getBytes());
		md.update(m_Passphrase);
		byte[] dhtHash = md.digest();
		String hashStr = StringUtils.encodeBase64(dhtHash, false);

		// initialize the secrets
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!secrets.initialize(m_VSnumber,
										cert,
										privKey,
										vouchIndex.toCharArray(),
										vouchRdStr.toCharArray(),
										receiptIndex.toCharArray(),
										recRwStr.toCharArray(),
										hashStr.toCharArray(),
										vtRwStr))
		{
			Log.error("Error initializing login secrets block");
			showError("could not initialize login secrets block",
						"internal error");
			return false;
		}

		// if the VS# is on @voucher-safe.org, then override the test net setting
		if (m_VSnumber.endsWith("@voucher-safe.org")) {
			m_Plugin.setTestNetwork(true);
		}

		// record the publisher we're working with
		secrets.setPublisher(publisher);

		// record the publisher domain name in preferences
		VSPreferences vsp = m_Plugin.getUserPreferences();
		Object prefUI = vsp.getGUI();
		if (prefUI instanceof PreferencesUI) {
			PreferencesUI ui = (PreferencesUI) prefUI;
			int atIdx = publisher.indexOf('@');
			ui.setPrefPublisher(publisher.substring(atIdx));
			vsp.commit();
		}
		return true;
	}

	/**
	 * utility method to determine the proper string suffix for a number
	 * @param num the number we're interested in
	 * @return the string, 1st, 2nd, 3rd, 4th, etc.
	 */
	private String getNumberingStr(int num) {
		String ret = null;
		switch (num) {
			case 1:
				ret = " 1st";
				break;
			case 2:
				ret = " 2nd";
				break;
			case 3:
				ret = " 3rd";
				break;
			default:
				ret = " " + num + "th";
				break;
		}
		return ret;
	}

	/**
	 * utility method to update the status of the phase 1 Continue button
	 */
	private void updateContinueButtonStatus() {
		if (m_HaveVSnumber) {
			if (m_HavePIN) {
				m_SubmitButton.setEnabled(true);
				m_SubmitButton.setToolTipText("Click here to submit your wallet ID "
														+ "and PIN");
			}
			else {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("Please enter your PIN");
			}
		}
		else {
			if (m_HavePIN) {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("Please enter your wallet ID");
			}
			else {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("<html>Please enter your<br/>"
														+ "wallet ID and PIN</html>");
			}
		}
	}

	/**
	 * utility method to update the status of the phase 2 Login button
	 */
	private void updateLoginButtonStatus() {
		String ch1 = (String) m_longphrase_ch1.getSelectedItem();
		String ch2 = (String) m_longphrase_ch2.getSelectedItem();
		String ch3 = (String) m_longphrase_ch3.getSelectedItem();

		if (m_HavePassphrase) {
			// check long-phrase chars
			if (ch1.equals(" ") || ch2.equals(" ") || ch3.equals(" ")) {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("Please select your long-phrase "
														+ "characters");
			}
			else {
				m_SubmitButton.setEnabled(true);
				m_SubmitButton.setToolTipText("Click here to submit your "
														+ "long-phrase and passphrase");
			}
		}
		else {
			m_SubmitButton.setEnabled(false);

			// check long-phrase chars
			if (ch1.equals(" ") || ch2.equals(" ") || ch3.equals(" ")) {
				m_SubmitButton.setToolTipText("Please select your long-phrase "
												+ "characters and enter your passphrase");
			}
			else {
				m_SubmitButton.setToolTipText("Please enter your passphrase");
			}
		}
	}

	/**
	 * method to queue up a modification to a text component from within its
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}


	// implement ActionListener
	/**
	 * method to process the event when a long-phrase character is chosen, or
	 * when enter is pressed in a text input
	 * @param ae the event from one of the long-phrase selection JComboBoxS or
	 * a text entry field
	 */
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JComboBox) {
			updateLoginButtonStatus();
		}
		else if (src instanceof JTextField) {
			JTextField input = (JTextField) src;
			// move to the next focus item
			input.transferFocus();
		}
	}


	// error handlers which invoke analogs in WTabManager
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	private void showError(String errMsg) {
		m_TabManager.showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	private void showError(String errMsg, String cause) {
		m_TabManager.showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	private void showError(String errMsg, String cause, String action) {
		m_TabManager.showError(errMsg, cause, action);
	}


	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_Passphrase != null) {
				Arrays.fill(m_Passphrase, (byte)0);
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
