/*
 * last modified---
 * 	03-22-17/jet/ cosmetic (SVSpark merge)
 * 	07-17-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine values
 * 	03-19-14/jet/ improve spacing with OpenJDK-7 fonts
 * 	06-05-13/jet/ SilentVault client version
 * 	03-02-11/kbw/ obtain available tokens from an Issuer which corresponds to
 * 					  the asset type of the incoming receipt
 * 	02-23-11/kbw/ remove assumption that all vouchers are in GAU
 * 	11-18-10/kbw/ on SDS token rejects, grab next token and retry
 * 	11-05-10/kbw/ suggest going to manual on receipt store errors
 * 	10-28-10/kbw/ provide automated mode
 * 	10-18-10/kbw/ vertical compression of layout
 * 	09-28-10/kbw/ rework for layout
 * 	09-02-10/kbw/ overhaul for Nimbus L&F
 * 	08-15-10/kbw/ rework for style
 * 	05-03-10/kbw/ grey out buttons during messaging operations
 * 	04-08-10/kbw/ flesh out
 * 	03-29-10/kbw/ new
 *
 * purpose---
 * 	provide a UI widget to handle processing and storage of new receipts
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.Font;
import java.awt.Color;
import java.awt.GridLayout;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.TreeMap;
import java.util.Date;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.security.PublicKey;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for handling incoming receipts.
 *
 * @author Kevin Wilkerson
 */
public final class NewReceiptPane extends WorkPane {
	// BEGIN data members
	/**
	 * storage for tokens used in an operation storing something on the SDS
	 */
	private ArrayList<XMLToken>   m_PaymentTokens;

	/**
	 * storage for a receipt record we are placing on the SDS
	 */
	private XMLReceipt				m_ReceiptRecord;

	/**
	 * the action to store the receipt
	 */
	private AbstractAction			m_StoreAction;

	/**
	 * the action to discard the receipt
	 */
	private AbstractAction			m_DiscardAction;

	/**
	 * the action to skip over this item and move to the next
	 */
	private AbstractAction			m_SkipAction;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public NewReceiptPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_PaymentTokens = new ArrayList<XMLToken>();

		// provide user textual instructions
		String receiptStore = "You can store this receipt for your records, "
										+ "or discard it.<br/><br/>If you store it, you "
										+ "can discard it later if you wish.";
		m_UserInstructs = receiptStore;

		// show token cost
		m_TokenCost.setText("To store uses " + VoucherRequest.M_ReceiptCost
									+ " token");

		// define button actions
		setStoreAction();
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_DiscardAction = new AbstractAction("Discard") {
			public void actionPerformed(ActionEvent actionEvent) {
				// kill it before we go back to housekeeping
				hkp.discardReceipt(m_ReceiptRecord);

				// get rid of the pending payment nonetheless
				removeCorrespondingPendingPayment(m_ReceiptRecord);

				// mark the record read on the DHT
				m_ParentHome.setCursor(M_WaitCursor);
				m_TabManager.markReadOnDHT(m_ReceiptRecord.getDHTid(), false);
				m_ReceiptRecord = null;

				// NB: when the DHT's reply is received, we'll go back to the
				// 	 housekeeping main screen

				// clear so they don't click on Discard again by mistake
				m_ButtonPane.removeAll();
			}
		};
		m_SkipAction = new AbstractAction("Do this one later") {
			public void actionPerformed(ActionEvent actionEvent) {
				hkp.processNextItem(-1);
			}
		};
		m_HeadingPane.setLayout(new GridLayout(1, 3));

		// we don't want a border around the whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * process a reply received from SDS to a receipt store
	 * @param storeRep the reply message from the SDS
	 */
	public void processRctStoreReply(SDSMessage storeRep) {
		if (storeRep == null
			|| !storeRep.getOpcode().equalsIgnoreCase("REP_receipt_stored"))
		{
			Log.error("Bad receipt store reply received");
			return;
		}
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		m_ParentHome.setCursor(null);
		setButtonState(true);
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		SpentTokenCache stc = m_Plugin.getTokenCache();
		String vsNo = m_Plugin.getLoginSecrets().getVSnumber();

		// see whether this reply is an error
		if (storeRep.getType() == IQ.Type.ERROR) {
			/* If the error here is 402 (payment required) then the implication is
			 * that our token was no good, i.e. we are attempting to double-spend
			 * it somehow and the SDS has rejected it.  In this case we need to
			 * record the token as spent, get another one, and retry the store.
			 */
			if (storeRep.getErrcode() == HttpURLConnection.HTTP_PAYMENT_REQUIRED) {
				// purge the bad token and record with cache
				ttm.spendTokens(m_PaymentTokens);
				stc.recordToken(m_PaymentTokens.get(0), vsNo);
				m_PaymentTokens.clear();

				// recurse, in another thread to avoid blocking this event thread
				Runnable retry = new Runnable() {
					public void run() {
						sendStoreReceipt();
					}
				};
				new Thread(retry).start();
				return;
			}

			// uncommit the token and display error message
			ttm.uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();

			/* About the only way this can happen is if we lost connectivity to
			 * the SDS.  Suggest manual processing to prevent recursion.
			 */
			if (hkp.automating()) {
				showError("error storing receipt", "code " + storeRep.getErrcode()
							+ ": " + storeRep.getErrmsg(),
							"switch to manual processing");
			}
			else {
				showError("error storing receipt", "code " + storeRep.getErrcode()
							+ ": " + storeRep.getErrmsg());
			}
			return;
		}

		// record token as spent, and report to spent cache
		ttm.spendTokens(m_PaymentTokens);
		stc.recordToken(m_PaymentTokens.get(0), vsNo);
		m_PaymentTokens.clear();

		// tell the receipt table about the new data and filename where stored
		Integer dhtId = null;
		if (m_ReceiptRecord != null) {
			dhtId = m_ReceiptRecord.getDHTid();
			hkp.getReceiptModel().addReceipt(m_ReceiptRecord,
														storeRep.getFilename());
		}

		// update the housekeeping pane's list of receipts to process
		hkp.discardReceipt(m_ReceiptRecord);

		// purge the corresponding pending payment record from the SDS
		removeCorrespondingPendingPayment(m_ReceiptRecord);
		m_ReceiptRecord = null;

		if (dhtId != null) {
			// mark this record read on the DHT
			m_ParentHome.setCursor(M_WaitCursor);
			m_TabManager.markReadOnDHT(dhtId, false);
			// NB: this will return us to housekeeping main when reply received
		}
		else {
			Log.error("XMLReceipt record from DHT without DHT id on it");
			showError("could not mark receipt record read on DHT",
						"no DHT id found");
		}

		// clear so they don't click on Store again by mistake
		m_ButtonPane.removeAll();
	}

	/**
	 * method to list the pending payments in our wallet, correlate them with the
	 * receipt we are processing, and remove the appropriate matching one
	 * @param receipt the receipt we are concerned with
	 */
	private void removeCorrespondingPendingPayment(XMLReceipt receipt) {
		if (receipt == null || !receipt.isInit()) {
			Log.error("Null or uninitialized receipt object supplied");
			return;
		}
		final String err
			= "unable to remove pending payment corresponding to receipt";

		// obtain the list of all the pending payments in our folder
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		Hashtable<XMLPayment, String> payments
			= hkp.getPaymentModel().getPaymentList();

		// loop through these looking for a match to our receipt
		boolean matched = false;
		Hashtable<XMLPayment, String> loopHash
			= new Hashtable<XMLPayment, String>(payments);
		for (XMLPayment pmt : loopHash.keySet()) {
			if (pmt.getPayee().equals(receipt.getPayee())
				 && pmt.getSerialNumber().equals(receipt.getVoucherSerial()))
			{
				// perform synchronous deletion of the matching pending payment
				matched = true;
				m_ParentHome.setCursor(M_WaitCursor);
				setButtonState(false);
				SDSMessage delReply = hkp.sendDelPendPmt(pmt, true);
				m_ParentHome.setCursor(null);
				setButtonState(true);
				if (delReply == null) {
					showError(err, "deletion of pending payment failed");
				}
				else if (delReply.getType() == IQ.Type.ERROR) {
					showError(err, delReply.getErrmsg());
				}
				else {
					payments.remove(pmt);
				}
				break;
			}
		}
		if (!matched) {
			Log.error("Could not find pending payment to match this receipt: "
						+ receipt.toXML());
			showError(err, "could not match receipt to a pending payment",
						"this can occur if you tried to reclaim the payment before "
						+ "you processed this receipt");
		}
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// get the receipt we're working with
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_ReceiptRecord = hkp.getReceiptRecord();
		if (m_ReceiptRecord == null) {
			Log.error("Missing receipt record, cannot process");
			showError("cannot do housekeeping item", "new receipt not found");
			return;
		}

		// set our heading pane
		buildValueHeader();

		// set screen label
		buildScreenLabel("Incoming Receipt");

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		if (hkp.automating()) {
			// push our own button, in another thread to avoid recursion
			Runnable doStore = new Runnable() {
				public void run() {
					sendStoreReceipt();
				}
			};
			SwingUtilities.invokeLater(doStore);
			return;
		}
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, the button to skip over this receipt
		JButton skipButton = new JButton(m_SkipAction);
		skipButton.setToolTipText("Click here to skip to the next housekeeping "
											+ "item");
		skipButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer
		m_ButtonPane.add(skipButton);
		m_ButtonPane.add(Box.createHorizontalStrut(105));	 // spacer

		// next, the button to discard this receipt
		JButton discardButton = new JButton(m_DiscardAction);
		discardButton.setToolTipText("Click here to discard this receipt");
		discardButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(discardButton);
		m_ButtonPane.add(Box.createHorizontalStrut(40));	 // spacer

		// last, the button to store this receipt (default)
		JButton storeButton = new JButton(m_StoreAction);
		storeButton.setToolTipText("Click here to store this receipt");
		storeButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(storeButton);

		// place the details about the receipt into a panel
		JPanel rctDets = new JPanel(new GridBagLayout());
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTHWEST;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(Box.createVerticalGlue(), m_GBC);
		rctDets.add(Box.createVerticalStrut(10), m_GBC);

		final Color labColor = UIManager.getColor("standardGray");

		// determine whether signature is valid
		PublicKey payeeKey = m_Plugin.getPKSListener().obtainVSpubkey(
																	m_ReceiptRecord.getPayee());
		if (payeeKey == null) {
			showError("unable to validate receipt signature",
						"could not obtain payee wallet's public key",
						"check validity of payee, or try again later");
			storeButton.setEnabled(false);
			discardButton.setEnabled(false);
		}
		else {
			boolean validSig = m_ReceiptRecord.validateSig(payeeKey);
			if (!validSig) {
				showError("cannot process receipt",
				"signature of payee wallet issuing receipt could not be verified");
				storeButton.setEnabled(false);
			}
		}

		// show the payee
		JLabel sentLab = new JLabel("This receipt is from:  ");
		sentLab.setFont(m_ParentHome.M_LabelFont);
		sentLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		rctDets.add(sentLab, m_GBC);
		JLabel sent
			= new JLabel(m_ReceiptRecord.getPayee(), SwingConstants.TRAILING);
		sent.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(sent, m_GBC);
		rctDets.add(Box.createVerticalStrut(5), m_GBC);

		// pick up date (past)
		JLabel dateLab = new JLabel("Voucher picked up: ");
		dateLab.setFont(m_ParentHome.M_LabelFont);
		dateLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		rctDets.add(dateLab, m_GBC);
		Date pickupDate = new Date(m_ReceiptRecord.getTimestamp() * 1000L);
		SimpleDateFormat sdf = new SimpleDateFormat(XMLVoucher.M_DateFmt);
		JLabel pickup
			= new JLabel(sdf.format(pickupDate), SwingConstants.TRAILING);
		pickup.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(pickup, m_GBC);
		rctDets.add(Box.createVerticalStrut(5), m_GBC);

		// amount in base units
		JLabel amtLab = new JLabel("Amount:");
		amtLab.setFont(m_ParentHome.M_LabelFont);
		amtLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		rctDets.add(amtLab, m_GBC);
		double quan = m_ReceiptRecord.getQuantity();
		StringBuilder amtTxt = new StringBuilder(32);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(0);
		df.setGroupingSize(3);
		String dispUnits = m_ReceiptRecord.getDisplayUnits();
		amtTxt.append(df.format(quan) + " " + dispUnits);
		JLabel amt = new JLabel(amtTxt.toString());
		amt.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(amt, m_GBC);

		// also show amount in preferred fiat currency
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		String lookUnits
			= VPAssetConfig.getPriceLookupUnits(m_ReceiptRecord.getAsset(),
															m_ReceiptRecord.getUnits());
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Double price = apc.getRelativePrice(lookUnits, prefCurrency);
		if (price == null) {
			price = new Double(1.0);
			prefCurrency = dispUnits;
		}
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		StringBuilder fiatAmt = new StringBuilder(32);
		fiatAmt.append("( ");
		fiatAmt.append(df.format(quan * price.doubleValue()));
		fiatAmt.append(" " + prefCurrency + " )");
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		rctDets.add(Box.createHorizontalStrut(50), m_GBC);
		JLabel fiat = new JLabel(fiatAmt.toString(), SwingConstants.TRAILING);
		fiat.setFont(m_ParentHome.M_DataItalFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(fiat, m_GBC);
		rctDets.add(Box.createVerticalStrut(10), m_GBC);

		// see other baggage we have
		TreeMap<String, String> rctBaggage = m_ReceiptRecord.getBaggageList();
		for (String bagFld : rctBaggage.keySet()) {
			String bagVal = rctBaggage.get(bagFld);
			String bagNm = bagFld;
			if (bagFld.equals("Message")) {
				bagNm = "BPayee message";
			}
			else if (bagFld.equals("BMessage")) {
				bagNm = "BYour message";
			}

			// show all pairs, knocking first char ('B') off of names
			JLabel bagLab = new JLabel(bagNm.substring(1) + ":");
			bagLab.setFont(m_ParentHome.M_LabelFont);
			bagLab.setForeground(labColor);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			rctDets.add(bagLab, m_GBC);

			// use a text area so we can get it all to fit
			JTextArea bagTxt = new JTextArea(1, 20);
			bagTxt.setText(bagVal);
			bagTxt.setFont(m_ParentHome.M_DataFont);
			bagTxt.setEditable(false);
			bagTxt.setBorder(null);
			bagTxt.setBackground(m_Plugin.m_LabelBackColor);
			bagTxt.setLineWrap(true);
			bagTxt.setWrapStyleWord(true);
			bagTxt.setFocusable(false);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			rctDets.add(bagTxt, m_GBC);
			rctDets.add(Box.createVerticalStrut(5), m_GBC);
		}
		rctDets.add(Box.createVerticalStrut(15), m_GBC);

		// voucher serial number
		JLabel serialLab = new JLabel("Voucher serial number:");
		serialLab.setFont(m_ParentHome.M_DetailFont);
		serialLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		rctDets.add(serialLab, m_GBC);
		JLabel serial = new JLabel(m_ReceiptRecord.getVoucherSerial(), 
											SwingConstants.TRAILING);
		serial.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		rctDets.add(serial, m_GBC);
		rctDets.add(Box.createVerticalGlue(), m_GBC);

		// install everything into the main WorkPane scroll area
		m_WorkArea.setViewportView(rctDets);
	}

	/**
	 * private utility method to configure the action to store a receipt.
	 */
	private void setStoreAction() {
		m_StoreAction = new AbstractAction("Store") {
			public void actionPerformed(ActionEvent actionEvent) {
				sendStoreReceipt();
			}
		};
	}

	/**
	 * private utility method to request that the SDS store a receipt record.
	 * implied param: m_ReceiptRecord contains the receipt to be stored
	 * NB: this method will block pending the SDS's reply
	 */
	private void sendStoreReceipt() {
		if (m_ReceiptRecord == null || !m_ReceiptRecord.isInit()) {
			Log.error("Missing or uninitialized receipt record cannot be stored");
			return;
		}
		final String genErr = "cannot store receipt";

		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot store receipts while not logged in!");
			showError(genErr, "not logged in", "please log in first");
			return;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot store receipt without SDS connection!");
			showError(genErr, "SDS unavailable", "try again later");
			return;
		}

		// make encrypted blob out of the record
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String encRctRec
			= m_ReceiptRecord.getEncryptedBase64(secrets.getPubKey());
		if (encRctRec.isEmpty()) {
			Log.error("error encrypting XMLReceipt record with public key");
			showError(genErr, "encrypt failed");
			return;
		}

		// build message to send
		SDSMessage storeRct = new SDSMessage();
		storeRct.setType(IQ.Type.SET);
		storeRct.setFrom(sdsConn.getUser());
		storeRct.setTo(sdsConn.getServiceName());
		storeRct.setPacketID("rstore_" + state.getNextSDSid());
		storeRct.setOpcode("REQ_store_receipt");
		String rIdx = new String(secrets.getReceiptIndex());
		storeRct.setFolder_index(rIdx);
		String rwCap = new String(secrets.getReceiptRWCap());
		storeRct.setReadwritecap(rwCap);
		storeRct.setReceipt_blob(encRctRec);

		// add one token for payment
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		/* We'll demand payment tokens from the same Issuer as the first voucher
		 * found which has an asset type matching the receipt's.  Else we'll use
		 * tokens from the default Issuer, which is likely to be the first one
		 * alphabetically, or the last one used when doing a payment, pickup, etc.
		 */
		Hashtable<XMLVoucher, String> allVouchers = vtm.getVoucherList();
		String tokenIssuer = vtm.getSelectedIssuer();
		for (XMLVoucher vouch : allVouchers.keySet()) {
			if (vouch.getAsset().equals(m_ReceiptRecord.getAsset())) {
				tokenIssuer = vouch.getIssuer();
				break;
			}
		}
		vtm.selectIssuer(tokenIssuer);
		m_PaymentTokens = ttm.getAvailableTokens(VoucherRequest.M_ReceiptCost,
															  tokenIssuer);
		if (m_PaymentTokens == null) {
			// Record ourselves with tab manager, so we return here after
			// the user has purchased tokens.  Then go purchase tokens.

			m_TabManager.setPendingTokens(this);
			m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
			m_TabManager.makeTransition();
			return;
		}
		storeRct.addPaymentToken(m_PaymentTokens.get(0).toXML());

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// register synchronous reply collector
		PacketIDFilter pFilter = new PacketIDFilter(storeRct.getPacketID());
		PacketCollector repColl = sdsConn.createPacketCollector(pFilter);

		// send request to the SDS
		sdsConn.sendPacket(storeRct);
		//Log.debug(storeRct.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(storeRct);

		// wait for reply or timeout
		SDSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new SDSMessage();
			rep.setOpcode("REP_receipt_stored");
			rep.setFrom(storeRct.getTo());
			rep.setTo(storeRct.getFrom());
			rep.setPacketID(storeRct.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (SDSMessage) reply;
		}

		// invoke callback with result (will call processRctStoreReply())
		m_Plugin.getSDSListener().processPacket(rep);
	}

	// END methods
}
