/*
 * last modified---
 * 	12-03-16/jet/ improve wording urging copying of wallet Id etc.
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	04-08-14/jet/ sort PKS list so that a path matching the OFS comes first
 * 	10-12-13/jet/ correct for multiple PKS paths
 * 	06-05-13/jet/ SilentVault client version
 * 	12-10-10/kbw/ adjust Q&A formatting in initPhase4() so it lines up better
 * 	11-24-10/kbw/ do not show publisher usernames, just domains
 * 	10-14-10/kbw/ compress layout vertically
 * 	10-12-10/kbw/ make enter in text inputs transfer focus
 * 	10-08-10/kbw/ overhaul for layout changes
 * 	08-03-10/kbw/ add Abort button; shrink horizontally; silently compress
 * 					  recovery input to lowercase; add indicator for successor VS
 * 	07-22-10/kbw/ prevent double-clicks on buttons
 * 	06-03-10/kbw/ use preferences to determine whether to display new keys
 * 	03-19-10/kbw/ PINs need to be String, not Integer
 * 	03-18-10/kbw/ force VS numbers to all lowercase
 * 	02-24-10/kbw/ stand-alone version extending JPanel
 * 	10-06-09/kbw/ display confirmation data in JTextField to permit copy/paste
 * 	09-22-09/kbw/ properly handle errors on REP_safe_init messages
 * 	09-01-09/kbw/ use integer format to display PIN, in case it begins w/ 0
 * 	07-22-09/kbw/ widen hashes for login and recovery indices to SHA512
 * 	07-12-09/kbw/ add entry of VS prefix on phase 1 screen, which we must
 * 					  always do now, even when there is only 1 publisher defined
 * 	07-01-09/kbw/ require 8 chars in passphrase, and suggest up to 24
 * 	05-29-09/kbw/ changes to support creation of a replacement safe
 * 	05-22-09/kbw/ preface recovery hash with VS#
 * 	04-30-09/kbw/ new
 *
 * purpose---
 * 	provide a screen for creating a new voucher safe wallet
 */

package org.silentvault.client.ui.wallet;

import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Font;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Dimension;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.security.*;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.net.InetSocketAddress;
import java.util.List;
import java.util.Vector;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;


/**
 * Extends JPanel to provide a screen for creating a new wallet.
 * This operation has these phases:
 * 1) Initiate login by requesting a list of supported publishers from the OFS.
 *	2) Select a publisher (this step can be obviated if there is only one VP).
 *	3) Tell the OFS the choice of VP, and receive back the keypair parameters.
 *		Generate a private/public keypair according to these parameters.
 *	4) Transmit the public key to the OFS, and receive in reply the PIN length,
 *		walletId, caps blob, and five recovery questions.
 *	5) Prompt the user for their PIN, long-phrase, passphrase, and answers to
 *		the recovery questions.
 *	6) Get the VP's pubkey off the PKS, encrypt the secrets and recovery data
 *		and transmit to OFS.
 *	7) Recapitulate results to user, and transit to login pane.
 *	Thus only step 5, and perhaps step 2, require user input.
 *
 * @author Kevin Wilkerson
 */
public final class NewSafePane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * minimum length of a walletId prefix
	 */
	public final static int	M_Prefix_Min = 3;

	/**
	 * maximum length of a walletId prefix
	 */
	public final static int	M_Prefix_Max = 24;

	/**
	 * minimum length of a long-phrase
	 */
	public final static int	M_Longphrase_Min = 8;

	/**
	 * maximum length of a long-phrase
	 */
	public final static int	M_Longphrase_Max = 15;

	/**
	 * minimum length of a passphrase
	 */
	public final static int	M_Passphrase_Min = 8;

	/**
	 * maximum length of a passphrase
	 */
	public final static int	M_Passphrase_Max = 24;

	/**
	 * the font to use for prompts
	 */
	private final Font		M_PromptFont;

	/**
	 * the plugin/application to which we belong
	 */
	private WalletClient		m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private WTabManager		m_TabManager;

	/**
	 * the PIN entered by the user
	 */
	private String				m_PIN;

	/**
	 * the passphrase entered by the user
	 */
	private byte[]				m_Passphrase;

	/**
	 * the long-phrase entered by the user
	 */
	private char[]				m_LongPhrase;

	/**
	 * the display list of available voucher publishers
	 */
	private ArrayList<String>	m_PubList;

	/**
	 * the actual list of full ids of possible voucher publishers
	 */
	private List<String>		m_AllPublishers;

	/**
	 * the UI element for publisher selection
	 */
	private JComboBox<String>	m_PubSelect;

	/**
	 * the UI element for entry of the walletId prefix
	 */
	private JTextField		m_VSprefix;

	/**
	 * the message giving the key specifications of the selected publisher
	 */
	private OFSMessage		m_KeySpecs;

	/**
	 * the progress bar for key pair generation
	 */
	private JProgressBar		m_KeyProgress;

	/**
	 * the message giving the safe data for the newly created safe
	 */
	private OFSMessage		m_SafeData;

	/**
	 * the capabilities blob, encrypted and base64-encoded
	 */
	private String				m_CapsBlob;

	/**
	 * the current title bar text
	 */
	private String				m_TitlebarText;

	/**
	 * the UI object for the PIN entry
	 */
	private JTextField		m_PINUI;
	/**
	 * the UI object for entry of the PIN confirmation
	 */
	private JTextField		m_PINUIConf;
	/**
	 * the length of the PIN for this publisher
	 */
	private short				m_PINLength;

	/**
	 * the UI object for the long-phrase
	 */
	private JTextField		m_LPhUI;
	/**
	 * the UI object for the long-phrase confirmation
	 */
	private JTextField		m_LPhUIConf;

	/**
	 * the UI object for the passphrase
	 */
	private JTextField		m_PPhUI;
	/**
	 * the UI object for the passphrase confirmation
	 */
	private JTextField		m_PPhUIConf;

	/**
	 * the UI object for the answer to the first recovery question
	 */
	private JTextField		m_Answer1UI;
	/**
	 * the UI object for the answer to the second recovery question
	 */
	private JTextField		m_Answer2UI;
	/**
	 * the UI object for the answer to the third recovery question
	 */
	private JTextField		m_Answer3UI;
	/**
	 * the UI object for the answer to the fourth recovery question
	 */
	private JTextField		m_Answer4UI;
	/**
	 * the UI object for the answer to the fifth recovery question
	 */
	private JTextField		m_Answer5UI;

	/**
	 * the submit button
	 */
	private JButton			m_SubmitButton;

	/**
	 * the action for handling submission of VP selection
	 */
	private Action				m_Phase1Submit;

	/**
	 * the action for handling submission of safe data (PIN, phrases, QnA)
	 */
	private Action				m_Phase2Submit;

	/**
	 * the scroll area used to ensure that this tab always displays everything
	 * regardless of window size
	 */
	private JScrollPane		m_ScrollPane;

	/**
	 * the OFS's reply to safe creation finalization
	 */
	private OFSMessage		m_SafeInitRep;

	/**
	 * whether this safe is a successor/replacement safe for one being closed
	 */
	private boolean			m_IsSuccessor;

	/**
	 * whether we're doing phase1 input or phase2
	 */
	private boolean			m_FirstPage;

	/**
	 * inner class to generate a new key pair, and take down the progress bar
	 * when done
	 */
	class KeygenTask implements Runnable {
		/*
		 * execute main task in background thread
		 */
		public void run() {
			// generate a keypair according to these specifications
			KeyPairGenerator kpg = null;
			try {
				kpg = KeyPairGenerator.getInstance(m_KeySpecs.getAlgorithm());
				kpg.initialize(m_KeySpecs.getMin_length());
			}
			catch (NoSuchAlgorithmException nsae) {
				Log.error("Bad key algorithm supplied", nsae);
				showError("cannot generate key pair", nsae.getMessage());
				return;
			}
			catch (InvalidParameterException ipe) {
				Log.error("Invalid keysize supplied", ipe);
				showError("cannot generate key pair", ipe.getMessage());
				return;
			}
			KeyPair kp = kpg.generateKeyPair();
			if (kp == null) {
				Log.error("Unable to generate specified key pair");
				showError("unable to generate specified key pair");
				m_ScrollPane.remove(m_KeyProgress);
				return;
			}

			// record the keys
			m_Plugin.getLoginSecrets().setKeys(kp);

			// send the public key to the OFS in a REQ_init_key message
			sendInitKey();
		}
	}

	/**
	 * whether we have a wallet Id prefix yet
	 */
	private boolean			m_HavePrefix;

	/**
	 * inner class to provide a DocumentListener for the m_VSprefix field
	 */
	final class SafeDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = "";
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// if this is the first char, it must be a letter
			boolean needUndo = false;
			if (curLen == 1) {
				if (!Character.isLetter(curTxt.charAt(0))) {
					needUndo = true;
				}
			}
			else {
				// if we have a bad character, then undo this edit
				String alteredPref = curTxt.replaceAll("[^a-z0-9]", "");
				if (!curTxt.equals(alteredPref)) {
					needUndo = true;
				}
			}
			// make sure it hasn't gotten too long
			if (curLen > M_Prefix_Max) {
				needUndo = true;
			}
			if (needUndo) {
				// undo this edit
				int changeAt = de.getOffset();
				String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
				queueTextMod(m_VSprefix, modTxt);
				return;
			}

			// min length is 3, max is 24
			if (curLen >= M_Prefix_Min && curLen <= M_Prefix_Max) {
				m_HavePrefix = true;
				updateSubmitButtonState();
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 3, max is 24
			if (curLen < M_Prefix_Min || curLen > M_Prefix_Max) {
				m_HavePrefix = false;
				updateSubmitButtonState();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.error("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * document listener for m_VSprefix text field
	 */
	private SafeDocumentListener	m_PrefixListener;

	/**
	 * whether we're presently undoing a document modification
	 */
	private boolean					m_DoingUndo;

	/**
	 * inner class to provide a DocumentListener for m_PINUI and m_PINUIConf
	 */
	final class PINDocumentListener implements DocumentListener {
		JTextField		m_TextUI;

		/**
		 * constructor
		 * @param textField the text field we're the listener for
		 */
		public PINDocumentListener(JTextField textField) {
			m_TextUI = textField;
		}

		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// see if value parses
			boolean needUndo = false;
			if (!parseValue()) {
				needUndo = true;
			}
			else {
				// see if it's gotten too long
				int curLen = doc.getLength();
				if (curLen > m_PINLength) {
					needUndo = true;
				}
			}
			if (needUndo) {
				// undo this edit
				int changeAt = de.getOffset();
				try {
					curTxt = doc.getText(0, doc.getLength());
					String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
					queueTextMod(m_TextUI, modTxt);
				}
				catch (BadLocationException ble) {
					Log.error("Cannot undo bad PIN edit", ble);
				}
			}
			else {
				updateSubmitButtonState();
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			updateSubmitButtonState();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_TextUI as a valid PIN
		 */
		private boolean parseValue() {
			// examine what's in the PIN entry field
			String pin = m_TextUI.getText();
			if (pin.isEmpty()) {
				return false;
			}

			// format allows l to specify long, but we don't
			if (pin.indexOf('l') != -1 || pin.indexOf('L') != -1) {
				return false;
			}

			// disallow +-
			if (pin.startsWith("-") || pin.startsWith("+")) {
				return false;
			}

			// attempt to parse, very strictly
			int vvv = 0;
			try {
				Integer val = Integer.parseInt(pin);
				vvv = val.intValue();
			}
			catch (NumberFormatException nfe) {
				return false;
			}

			// NB: we check length in updateSubmitButtonState()
			return true;
		}
	}

	/**
	 * inner class to provide a DocumentListener for m_LPhUI and m_LPhUIConf
	 */
	final class LongDocumentListener implements DocumentListener {
		JTextField		m_TextUI;

		/**
		 * constructor
		 * @param txtField the text field for which we're the listener
		 */
		public LongDocumentListener(JTextField txtField) {
			m_TextUI = txtField;
		}

		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";
			int curLen = doc.getLength();
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// make sure we have only a-zA-Z0-9 in the string
			boolean needUndo = false;
			String alteredLong = curTxt.replaceAll("[^a-zA-Z0-9]", "");
			if (!curTxt.equals(alteredLong)) {
				needUndo = true;
			}
			else if (curLen >= M_Longphrase_Min && curLen <= M_Longphrase_Max) {
				// min length is 8 chars, max is 15
				updateSubmitButtonState();
			}
			else if (curLen > M_Longphrase_Max) {
				// see if it's gotten too long
				needUndo = true;
			}
			if (needUndo) {
				// undo this edit
				int changeAt = de.getOffset();
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_TextUI, modTxt);
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 8 chars, max is 15
			if (curLen >= M_Longphrase_Min && curLen <= M_Longphrase_Max) {
				updateSubmitButtonState();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * inner class to provide a DocumentListener for m_PPhUI and m_PPhUIConf
	 */
	final class PhraseDocumentListener implements DocumentListener {
		JTextField		m_TextUI;

		/**
		 * constructor
		 * @param txtField the text field for which we're the listener
		 */
		public PhraseDocumentListener(JTextField txtField) {
			m_TextUI = txtField;
		}

		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 8 chars, max is 24
			if (curLen >= M_Passphrase_Min && curLen <= M_Passphrase_Max) {
				updateSubmitButtonState();
			}
			else if (curLen > M_Passphrase_Max) {
				// undo this edit
				int changeAt = de.getOffset();
				try {
					String curTxt = doc.getText(0, doc.getLength());
					String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
					queueTextMod(m_TextUI, modTxt);
				}
				catch (BadLocationException ble) {
					Log.error("Cannot undo bad passphrase edit", ble);
				}
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			int curLen = doc.getLength();

			// min length is 8 chars, max is 24
			if (curLen >= M_Passphrase_Min && curLen <= M_Passphrase_Max) {
				updateSubmitButtonState();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * inner class to provide a DocumentListener for the answer text entry fields
	 */
	final class AnswerDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen > 0) {
				updateSubmitButtonState();
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 0) {
				updateSubmitButtonState();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * the DocumentListener used for all answer entry fields
	 */
	private AnswerDocumentListener	m_AnswerListener;

	/**
	 * inner class to modify a text field from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		JTextComponent	m_ModComponent;
		String			m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());
			updateSubmitButtonState();
		}
	}

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param replacement whether the safe being created is to receive the value
	 * and tokens in a safe being closed
	 * @param mgr the tab manager which creates us
	 */
	public NewSafePane(WalletClient plugin, boolean replacement, WTabManager mgr)
	{
		super();
		m_Plugin = plugin;
		m_TabManager = mgr;
		m_IsSuccessor = replacement;
		m_TitlebarText = m_PIN = "";
		M_PromptFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		m_FirstPage = true;
		m_PubList = new ArrayList<String>();

		// use grid layout
		setLayout(new GridLayout());
		setBorder(null);

		// phase 1 submission action
		m_Phase1Submit = new AbstractAction("Continue") {
			public void actionPerformed(ActionEvent actionEvent) {
				// Retrieve and validate entered prefix value.  Note that the
				// document listener should eliminate most of these cases.

				final String badPref = "invalid wallet ID prefix";
				String vsPref = m_VSprefix.getText().toLowerCase();
				if (vsPref == null || vsPref.isEmpty()) {
					showError("missing wallet ID prefix", "none entered",
								"please enter an identifier for your wallet");
					return;
				}
				// check length is 3-24
				if (vsPref.length() < M_Prefix_Min
					|| vsPref.length() > M_Prefix_Max)
				{
					showError(badPref, "length must be between " + M_Prefix_Min
								+ " and " + M_Prefix_Max + " characters",
								"please try again");
					return;
				}
				// must also start with a letter
				if (!Character.isLetter(vsPref.charAt(0))) {
					showError(badPref, "wallet identifier must begin with a letter",
								"please try again");
					return;
				}
				// must contain only alphanumeric chars
				String alteredPref = vsPref.replaceAll("[^a-zA-Z0-9]", "");
				if (!vsPref.equals(alteredPref)) {
					showError(badPref,
								"can use only upper and lower case letters, and digits",
								"please try again");
					return;
				}

				// retrieve publisher pick
				String pub = (String) m_PubSelect.getSelectedItem();
				if (pub == null || pub.isEmpty()) {
					showError("missing voucher publisher", "none selected",
								"please pick a voucher publisher");
					return;
				}
				// get the full publisher id which corresponds to the pick
				for (String vPub : m_AllPublishers) {
					if (vPub.endsWith(pub)) {
						m_Plugin.getLoginSecrets().setPublisher(vPub);
						break;
					}
				}

				// disallow double-clicks
				m_SubmitButton.setEnabled(false);

				// send initiate safe creation to OFS by selecting publisher
				sendSelectPublisher();
			}
		};

		// construct some other input objects
		m_PrefixListener = new SafeDocumentListener();
		m_VSprefix = new JTextField(18);
		m_VSprefix.setDropTarget(null);
		m_VSprefix.setHorizontalAlignment(JTextField.LEFT);
		m_VSprefix.setToolTipText("<html>Enter " + M_Prefix_Min + "-"
											+ M_Prefix_Max
											+ " lower case letters and<br/>numbers "
											+ "only, beginning with a letter</html>");
		m_VSprefix.getDocument().addDocumentListener(m_PrefixListener);
		m_VSprefix.addActionListener(this);

		// long-phrase input
		m_LPhUI = new JTextField(12);
		m_LPhUI.setDropTarget(null);
		m_LPhUI.setHorizontalAlignment(JTextField.LEFT);
		m_LPhUI.setToolTipText("Please enter only A-Z, a-z, or 0-9 characters");
		m_LPhUI.getDocument().addDocumentListener(
															new LongDocumentListener(m_LPhUI));
		m_LPhUI.addActionListener(this);
		m_LPhUIConf = new JTextField(12);
		m_LPhUIConf.setDropTarget(null);
		m_LPhUIConf.setHorizontalAlignment(JTextField.LEFT);
		m_LPhUIConf.setToolTipText("Please re-enter your long-phrase to confirm");
		m_LPhUIConf.getDocument().addDocumentListener(
													new LongDocumentListener(m_LPhUIConf));
		m_LPhUIConf.addActionListener(this);

		// passphrase input
		m_PPhUI = new JTextField(16);
		m_PPhUI.setDropTarget(null);
		m_PPhUI.setHorizontalAlignment(JTextField.LEFT);
		m_PPhUI.setToolTipText("Please enter " + M_Passphrase_Min + " to "
										+ M_Passphrase_Max + " characters");
		m_PPhUI.getDocument().addDocumentListener(
														new PhraseDocumentListener(m_PPhUI));
		m_PPhUI.addActionListener(this);
		m_PPhUIConf = new JTextField(16);
		m_PPhUIConf.setDropTarget(null);
		m_PPhUIConf.setHorizontalAlignment(JTextField.LEFT);
		m_PPhUIConf.setToolTipText("Please re-enter your passphrase to confirm");
		m_PPhUIConf.getDocument().addDocumentListener(
													new PhraseDocumentListener(m_PPhUIConf));
		m_PPhUIConf.addActionListener(this);

		// recovery question answer inputs
		final String qTTtxt = "Enter your answer to this question";
		m_AnswerListener = new AnswerDocumentListener();
		m_Answer1UI = new JTextField(15);
		m_Answer1UI.setDropTarget(null);
		m_Answer1UI.setHorizontalAlignment(JTextField.LEFT);
		m_Answer1UI.setToolTipText(qTTtxt);
		m_Answer1UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer1UI.addActionListener(this);

		m_Answer2UI = new JTextField(15);
		m_Answer2UI.setDropTarget(null);
		m_Answer2UI.setHorizontalAlignment(JTextField.LEFT);
		m_Answer2UI.setToolTipText(qTTtxt);
		m_Answer2UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer2UI.addActionListener(this);

		m_Answer3UI = new JTextField(15);
		m_Answer3UI.setDropTarget(null);
		m_Answer3UI.setHorizontalAlignment(JTextField.LEFT);
		m_Answer3UI.setToolTipText(qTTtxt);
		m_Answer3UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer3UI.addActionListener(this);

		m_Answer4UI = new JTextField(15);
		m_Answer4UI.setDropTarget(null);
		m_Answer4UI.setHorizontalAlignment(JTextField.LEFT);
		m_Answer4UI.setToolTipText(qTTtxt);
		m_Answer4UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer4UI.addActionListener(this);

		m_Answer5UI = new JTextField(15);
		m_Answer5UI.setDropTarget(null);
		m_Answer5UI.setHorizontalAlignment(JTextField.LEFT);
		m_Answer5UI.setToolTipText(qTTtxt);
		m_Answer5UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer5UI.addActionListener(this);
	}

	/**
	 * Initialize for phase 1, publisher selection.  The first time we are
	 * called, we'll do nothing but send REQ_create_safe to the OFS.  On the
	 * REP_publisher_list reply, the callback will call us again.  We'll then
	 * proceed to prompt the user for their publisher selection, unless only
	 * one publisher is listed, in which case we'll skip that.  In either case
	 * we'll prompt for a wallet Id prefix from the user, which must be at least
	 * three characters.  We then send a REQ_select_publisher to the OFS.
	 * Once we have the reply we'll transition to stage two and generate keys.
	 */
	public void initPhase1() {
		setCursor(null);
		// see whether we have a publisher list yet
		if (m_PubList == null || m_PubList.isEmpty()) {
			// we do not, request the list
			sendCreateSafe();
			return;
		}
		m_FirstPage = true;

		// we do, check the magnitude of the list
		boolean needPubSelect = true;
		m_PubSelect = new JComboBox<String>(new Vector<String>(m_PubList));
		if (m_PubList.size() == 1) {
			// just assume the only one given is automatically selected
			m_PubSelect.setSelectedItem(0);
			needPubSelect = false;
		}

		// set up to prompt for safe prefix and (possibly) publisher selection
		if (m_IsSuccessor) {
			m_TitlebarText = "Creating a Successor Wallet (Step 1)";
		}
		else {
			m_TitlebarText = "Wallet Login Creation (Step 1)";
		}
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();
	
		// put everything in a grid bag with border, inside the scroll area
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.NONE;
		gbc.anchor = GridBagConstraints.NORTH;
		JPanel selectPane = new JPanel(gbl);
		selectPane.add(Box.createVerticalGlue(), gbc);

		// deal with prefix
		m_HavePrefix = false;
		String prefTxt = "Enter an identifying name for your new wallet";
		JLabel prefPrompt = new JLabel(prefTxt, SwingConstants.CENTER);
		prefPrompt.setFont(M_PromptFont);
		selectPane.add(prefPrompt, gbc);
		selectPane.add(Box.createVerticalStrut(10), gbc);
		selectPane.add(m_VSprefix, gbc);

		// if required, deal with publisher selection
		if (needPubSelect) {
			selectPane.add(Box.createVerticalStrut(45), gbc);
			JLabel pubPrompt = new JLabel("Select the voucher publisher for "
													+ "your wallet login",
													SwingConstants.CENTER);
			pubPrompt.setFont(M_PromptFont);
			m_PubSelect.setToolTipText("Pick desired publisher from list");
			m_PubSelect.addActionListener(this);
			pubPrompt.setLabelFor(m_PubSelect);
			selectPane.add(pubPrompt, gbc);
			selectPane.add(Box.createVerticalStrut(5), gbc);
			selectPane.add(m_PubSelect, gbc);
		}
		selectPane.add(Box.createVerticalStrut(55), gbc);

		// define abort action and button
		AbstractAction abort = new AbstractAction("Abort") {
			public void actionPerformed(ActionEvent actionEvent) {
				removeAll();
				if (!m_IsSuccessor) {
					m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
					m_TabManager.makeTransition();
				}
				else {
					m_TabManager.queueTransition(
													new Integer(WTabManager.M_HOME_RETURN));
					m_TabManager.makeTransition();
					m_TabManager.queueTransition(
													new Integer(WTabManager.M_GO_MAIN_MENU));
					m_TabManager.makeTransition();
				}
			}
		};
		JButton abortButton = new JButton(abort);
		abortButton.setToolTipText("Click here to abort wallet creation");
		abortButton.setFont(m_Plugin.M_ButtonFont);
		abortButton.setFocusable(false);

		// define button for handling submission of publisher and walletId prefix
		m_SubmitButton = new JButton(m_Phase1Submit);
		String buttonTtext = null;
		if (!needPubSelect) {
			buttonTtext = "Enter the identifying name for your wallet";
		}
		else {
			buttonTtext = "<html>Select your publisher, and enter<br/>your wallet "
								+ "identifying name</html>";
		}
		m_SubmitButton.setToolTipText(buttonTtext);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		m_SubmitButton.setEnabled(false);

		// make a borderless panel for the two buttons, one beside the other
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalGlue());
		buttonPane.add(abortButton);
		buttonPane.add(Box.createHorizontalStrut(135));  // spacer
		buttonPane.add(m_SubmitButton);
		buttonPane.add(Box.createHorizontalGlue());

		selectPane.add(buttonPane, gbc);
		selectPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(selectPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		getRootPane().setDefaultButton(m_SubmitButton);
		m_VSprefix.requestFocusInWindow();
	}

	/**
	 * initialize for phase 2, keypair generation
	 */
	public void initPhase2() {
		if (m_KeySpecs == null) {
			Log.error("Null key specs reply, cannot init phase 2 wallet creation");
			return;
		}
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();

		// pop up an indeterminate task progress meter
		m_KeyProgress = new JProgressBar(JProgressBar.HORIZONTAL);
		m_KeyProgress.setIndeterminate(true);
		m_KeyProgress.setString("Generating " + m_KeySpecs.getAlgorithm()
										+ " key pair size " + m_KeySpecs.getMin_length()
										+ ", please wait");
		m_KeyProgress.setStringPainted(true);
		m_KeyProgress.setBackground(Color.LIGHT_GRAY);
		m_KeyProgress.setOpaque(true);
		m_KeyProgress.setFont(new Font(Font.SANS_SERIF, Font.BOLD, 16));
		m_KeyProgress.setPreferredSize(new Dimension(450, 50));

		// create background task to perform keypair generation
		KeygenTask kgt = new KeygenTask();
		new Thread(kgt).start();

		// set wait cursor
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

		// display progress bar
		JPanel prog = new JPanel();
		prog.setLayout(new BoxLayout(prog, BoxLayout.Y_AXIS));
		prog.setBorder(null);
		prog.add(Box.createVerticalStrut(160));
		prog.add(m_KeyProgress);
		add(prog);
	}

	/**
	 * initialize for Phase 3, entry of PIN, long-phrase, passphrase, recovery
	 */
	public void initPhase3() {
		// take down progress bar
		removeAll();
		revalidate();

		setCursor(null);
		if (m_SafeData == null) {
			Log.error("Null safe data reply, cannot init phase 3 wallet creation");
			return;
		}
		m_FirstPage = false;

		// obtain certificate from this base64-encoded string
		String certStr = m_SafeData.getCertificate();
		X509Certificate vsCert = EncodingUtils.getCertFromBase64Str(certStr);
		if (vsCert == null) {
			Log.error("Unable to parse signed certificate");
			return;
		}

		// get and record the key server where the VP's pubkey is stored
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		HashMap<String, ArrayList<InetSocketAddress>> keyServers
			= secrets.getKeyServers();
		String pksDomain = m_SafeData.getKey_server_domain();
		ArrayList<InetSocketAddress> vpKS = m_SafeData.getKey_server_addr();
		if (pksDomain.isEmpty() || vpKS.isEmpty()) {
			Log.error("Missing or invalid VP key server");
			showError("unable to init new wallet", "cannot find key server");
			return;
		}
		ArrayList<InetSocketAddress> keyList = keyServers.get(pksDomain);
		if (keyList == null) {
			keyServers.put(pksDomain, vpKS);
		}
		else {
			keyList.addAll(vpKS);
		}

		/* We now need to obtain the VP's public key, in order to validate
		 * the certificate we were just given.
		 */
		PublicKey vpKey = m_Plugin.getPKSListener().obtainVPpubkey(null);
		if (vpKey == null) {
			return;
		}
		secrets.setVPKey(vpKey);

		// now verify the certificate using the key
		String verfErr = "unable to verify certificate from VP";
		try {
			vsCert.verify(secrets.getVPKey());
		}
		catch (CertificateException ce) {
			Log.error(verfErr, ce);
			showError(verfErr, ce.getMessage());
			return;
		}
		catch (InvalidKeyException ike) {
			Log.error(verfErr, ike);
			showError(verfErr, ike.getMessage());
			return;
		}
		catch (SignatureException se) {
			Log.error(verfErr, se);
			showError(verfErr, se.getMessage());
			return;
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error(verfErr, nsae);
			showError(verfErr, nsae.getMessage());
			return;
		}
		catch (NoSuchProviderException nspe) {
			Log.error(verfErr, nspe);
			showError(verfErr, nspe.getMessage());
			return;
		}
		// if we make it here, it verified okay

		// decrypt capabilities blob
		PrivateKey privKey = secrets.getPrivKey();
		String vsCap
			= EncodingUtils.getStrFromBase64PubkeyEnc(m_SafeData.getCap_blob(),
																	privKey);
		if (vsCap == null) {
			Log.error("Unable to decrypt vs_caps block with private key");
			return;
		}

		/* We now have a string that should be in this form:
		 * <vs_caps>
		 * 	<VS_number/>
		 * 	<voucher_index/>
		 * 	<voucher_rd_cap/>
		 * 	<voucher_rw_cap/>
		 * 	<receipts_index/>
		 * 	<receipts_rd_cap/>
		 * 	<receipts_rw_cap/>
		 * </vs_caps>
		 */
		Element vsCapsElt = EncodingUtils.buildElement(vsCap);
		if (vsCapsElt == null) {
			Log.error("<vs_caps> block did not parse:\n" + vsCap);
			showError("could not set voucher safe caps block", "parse error");
			return;
		}
		String VSnumber = null;
		String vouchIndex = null;
		String vouchRdCap = null;
		String receiptIndex = null;
		String receiptRdCap = null;
		String receiptRwCap = null;
		String vouchRwCap = null;
		try {
			Element vsNumElt = vsCapsElt.element("VS_number");
			if (vsNumElt == null) {
				throw new DocumentException("wallet Id number missing");
			}
			VSnumber = vsNumElt.getTextTrim();
			Element vsIdxElt = vsCapsElt.element("voucher_index");
			if (vsIdxElt == null) {
				throw new DocumentException("voucher index is missing");
			}
			vouchIndex = vsIdxElt.getTextTrim();
			if (vouchIndex.isEmpty()) {
				throw new DocumentException("voucher index is empty");
			}
			Element vouchRdElt = vsCapsElt.element("voucher_rd_cap");
			if (vouchRdElt == null) {
				throw new DocumentException("voucher read cap is missing");
			}
			vouchRdCap = vouchRdElt.getTextTrim();
			Element vouchRwElt = vsCapsElt.element("voucher_rw_cap");
			if (vouchRwElt == null) {
				throw new DocumentException("voucher readwrite cap missing");
			}
			vouchRwCap = vouchRwElt.getTextTrim();
			Element recIdxElt = vsCapsElt.element("receipts_index");
			if (recIdxElt == null) {
				throw new DocumentException("receipt index is missing");
			}
			receiptIndex = recIdxElt.getTextTrim();
			if (receiptIndex.isEmpty()) {
				throw new DocumentException("receipt index is empty");
			}
			Element rectRdElt = vsCapsElt.element("receipts_rd_cap");
			if (rectRdElt == null) {
				throw new DocumentException("receipt read cap is missing");
			}
			receiptRdCap = rectRdElt.getTextTrim();
			Element rectRwElt = vsCapsElt.element("receipts_rw_cap");
			if (rectRwElt == null) {
				throw new DocumentException("receipt readwrite cap missing");
			}
			receiptRwCap = rectRwElt.getTextTrim();
		}
		catch (DocumentException de) {
			Log.error("Error parsing <vs_caps/> blob, XML was: " + vsCap, de);
			showError("could not parse safe capabilities", de.getMessage());
			return;
		}

		// record all of this in the login secrets record
		// NB: we don't set the actual DHT hash at this time due to no passphrase
		if (!secrets.initialize(	VSnumber,
											vsCert,
											privKey,
											vouchIndex.toCharArray(),
											vouchRdCap.toCharArray(),
											receiptIndex.toCharArray(),
											receiptRwCap.toCharArray(),
											// really s/b hash(walletId+PIN+passphrase):
											"**temp**".toCharArray(),
											vouchRwCap	))
		{
			Log.error("Unable to initialize login secrets");
			return;
		}

		// make sure we have valid-looking PIN length and recovery questions
		m_PINLength = m_SafeData.getPin_length();
		final String initErr = "cannot initialize wallet";
		if (m_PINLength <= 0) {
			Log.error("Illegal PIN length specified, " + m_PINLength);
			showError(initErr, "illegal zero PIN length specified by VP");
			return;
		}
		if (m_SafeData.getRecovery_ques_1().isEmpty()) {
			Log.error("First recovery question is missing!");
			showError(initErr, "first recovery question is missing");
			return;
		}
		if (m_SafeData.getRecovery_ques_2().isEmpty()) {
			Log.error("Second recovery question is missing!");
			showError(initErr, "second recovery question is missing");
			return;
		}
		if (m_SafeData.getRecovery_ques_3().isEmpty()) {
			Log.error("Third recovery question is missing!");
			showError(initErr, "third recovery question is missing");
			return;
		}
		if (m_SafeData.getRecovery_ques_4().isEmpty()) {
			Log.error("Fourth recovery question is missing!");
			showError(initErr, "fourth recovery question is missing");
			return;
		}
		if (m_SafeData.getRecovery_ques_5().isEmpty()) {
			Log.error("Fifth recovery question is missing!");
			showError(initErr, "fifth recovery question is missing");
			return;
		}

		// Ask the user to enter a PIN, a long-phrase, and a passphrase.
		// Also request answers to the recovery questions.

		if (m_IsSuccessor) {
			m_TitlebarText = "Creating a Successor Wallet (Step 2)";
		}
		else {
			m_TitlebarText = "Wallet Login Creation (Step 2)";
		}
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));

		// layout setup
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.fill = GridBagConstraints.NONE;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		JPanel createPane = new JPanel(gbl);
		createPane.add(Box.createVerticalGlue(), gbc);

		// display the assigned wallet Id first
		JPanel vsGrp = new JPanel(gbl);
		vsGrp.setBorder(BorderFactory.createTitledBorder("Your new wallet ID"));
		JTextField vsNum = new JTextField(VSnumber);
		vsNum.setDropTarget(null);
		vsNum.setFont(new Font(Font.SANS_SERIF, Font.BOLD, 18));
		vsNum.setBorder(null);
		vsNum.setEditable(false);
		vsNum.setBackground(m_Plugin.m_LabelBackColor);
		vsGrp.add(vsNum, gbc);
		JLabel showVS
			= new JLabel("Important: please record this in a safe place.");
		showVS.setFont(M_PromptFont);
		vsGrp.add(Box.createVerticalStrut(5), gbc);
		vsGrp.add(showVS, gbc);
		createPane.add(vsGrp, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);
		gbc.anchor = GridBagConstraints.NORTHWEST;

		// obtain PIN and its confirmation
		m_PINUI = new JTextField(m_PINLength);
		m_PINUI.setDropTarget(null);
		m_PINUI.setHorizontalAlignment(JTextField.LEFT);
		m_PINUI.setToolTipText("A valid PIN is all digits");
		m_PINUI.getDocument().addDocumentListener(
															new PINDocumentListener(m_PINUI));
		m_PINUI.addActionListener(this);
		m_PINUIConf = new JTextField(m_PINLength);
		m_PINUIConf.setDropTarget(null);
		m_PINUIConf.setHorizontalAlignment(JTextField.LEFT);
		m_PINUIConf.setToolTipText("Please re-enter your PIN to confirm");
		m_PINUIConf.getDocument().addDocumentListener(
														new PINDocumentListener(m_PINUIConf));
		m_PINUIConf.addActionListener(this);
		JLabel pinLab = new JLabel("Please enter a PIN of " + m_PINLength
											+ " digits:", SwingConstants.RIGHT);
		pinLab.setLabelFor(m_PINUI);
		pinLab.setFont(M_PromptFont);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(pinLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_PINUI, gbc);
		JLabel confLab = new JLabel("Confirm PIN:", SwingConstants.RIGHT);
		confLab.setFont(M_PromptFont);
		confLab.setLabelFor(m_PINUIConf);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(confLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_PINUIConf, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);

		// obtain long-phrase and its confirmation
		JTextArea lpDesc = new JTextArea("Please enter a long-phrase of between "
											+ M_Longphrase_Min + " and " + M_Longphrase_Max
											+ " characters, which must be upper or lower "
											+ "case letters, or digits.");
		lpDesc.setEditable(false);
		lpDesc.setLineWrap(true);
		lpDesc.setWrapStyleWord(true);
		lpDesc.setFocusable(false);
		lpDesc.setBackground(m_Plugin.m_LabelBackColor);
		lpDesc.setFont(M_PromptFont);
		lpDesc.setBorder(null);
		gbc.fill = GridBagConstraints.HORIZONTAL;
		createPane.add(lpDesc, gbc);
		createPane.add(Box.createVerticalStrut(5), gbc);
		JLabel lpLab
			= new JLabel("Enter your long-phrase:", SwingConstants.RIGHT);
		lpLab.setLabelFor(m_LPhUI);
		lpLab.setFont(M_PromptFont);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
		createPane.add(lpLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_LPhUI, gbc);
		JLabel lpConfLab
			= new JLabel("Confirm long-phrase:", SwingConstants.RIGHT);
		lpConfLab.setFont(M_PromptFont);
		lpConfLab.setLabelFor(m_LPhUIConf);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(lpConfLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_LPhUIConf, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);

		// obtain passphrase and its confirmation
		JTextArea ppDesc = new JTextArea("Please enter a passphrase of "
													+ M_Passphrase_Min + "-"
													+ M_Passphrase_Max
													+ " characters, of any kind.  This "
													+ "phrase accesses your private key.");
		ppDesc.setEditable(false);
		ppDesc.setLineWrap(true);
		ppDesc.setWrapStyleWord(true);
		ppDesc.setFocusable(false);
		ppDesc.setBackground(m_Plugin.m_LabelBackColor);
		ppDesc.setFont(M_PromptFont);
		ppDesc.setBorder(null);
		gbc.fill = GridBagConstraints.HORIZONTAL;
		createPane.add(ppDesc, gbc);
		createPane.add(Box.createVerticalStrut(5), gbc);
		JLabel ppLab
			= new JLabel("Enter your passphrase:", SwingConstants.RIGHT);
		ppLab.setFont(M_PromptFont);
		ppLab.setLabelFor(m_PPhUI);
		gbc.fill = GridBagConstraints.NONE;
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(ppLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_PPhUI, gbc);
		JLabel ppConfLab
			= new JLabel("Confirm passphrase:", SwingConstants.RIGHT);
		ppConfLab.setFont(M_PromptFont);
		ppConfLab.setLabelFor(m_PPhUIConf);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(ppConfLab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_PPhUIConf, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);

		// panel to get answers to recovery questions
		String recovDesc = "Please enter your answers to the following five "
								+ "standard questions.  Should you ever lose your "
								+ "PIN, long-phrase, or "
								+ "passphrase, you can recover them by entering the "
								+ "same answers to these questions.  Your answers must "
								+ "match exactly (including punctuation and spaces "
								+ "but not case), so "
								+ "please record your answers in a safe place for "
								+ "future reference.";
		JTextArea recovExpl = new JTextArea(recovDesc);
		recovExpl.setEditable(false);
		recovExpl.setLineWrap(true);
		recovExpl.setWrapStyleWord(true);
		recovExpl.setFocusable(false);
		recovExpl.setBackground(m_Plugin.m_LabelBackColor);
		recovExpl.setBorder(null);
		recovExpl.setFont(M_PromptFont);
		gbc.fill = GridBagConstraints.HORIZONTAL;
		createPane.add(recovExpl, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);

		JLabel q1Lab = new JLabel(m_SafeData.getRecovery_ques_1() + ": ",
											SwingConstants.RIGHT);
		q1Lab.setFont(M_PromptFont);
		q1Lab.setLabelFor(m_Answer1UI);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
		createPane.add(q1Lab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_Answer1UI, gbc);

		JLabel q2Lab = new JLabel(m_SafeData.getRecovery_ques_2() + ": ",
											SwingConstants.RIGHT);
		q2Lab.setFont(M_PromptFont);
		q2Lab.setLabelFor(m_Answer2UI);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(q2Lab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_Answer2UI, gbc);

		JLabel q3Lab = new JLabel(m_SafeData.getRecovery_ques_3() + ": ",
											SwingConstants.RIGHT);
		q3Lab.setFont(M_PromptFont);
		q3Lab.setLabelFor(m_Answer3UI);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(q3Lab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_Answer3UI, gbc);

		JLabel q4Lab = new JLabel(m_SafeData.getRecovery_ques_4() + ": ",
											SwingConstants.RIGHT);
		q4Lab.setFont(M_PromptFont);
		q4Lab.setLabelFor(m_Answer4UI);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(q4Lab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_Answer4UI, gbc);

		JLabel q5Lab = new JLabel(m_SafeData.getRecovery_ques_5() + ": ",
											SwingConstants.RIGHT);
		q5Lab.setFont(M_PromptFont);
		q5Lab.setLabelFor(m_Answer5UI);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		createPane.add(q5Lab, gbc);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		createPane.add(m_Answer5UI, gbc);
		createPane.add(Box.createVerticalStrut(10), gbc);

		// define submit button and its action
		setPhase3Action();
		m_SubmitButton = new JButton(m_Phase2Submit);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		updateSubmitButtonState();
		gbc.anchor = GridBagConstraints.NORTH;
		createPane.add(m_SubmitButton, gbc);
		createPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(createPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		getRootPane().setDefaultButton(m_SubmitButton);
		m_PINUI.requestFocusInWindow();
	}

	/**
	 * initialize for Phase 4, confirmation of new safe information
	 */
	public void initPhase4() {
		// display confirmation, vector to login pane on user accept
		m_TitlebarText = "Wallet Login Creation (Confirmation)";
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		removeAll();
		setCursor(null);

		// layout setup
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		lastItemConstraints(gbc);
		JPanel confPane = new JPanel(gbl);
		confPane.setBorder(null);
		confPane.add(Box.createVerticalGlue(), gbc);

		// check for errors
		if (m_SafeInitRep.getType() == IQ.Type.ERROR) {
			String errMsg = "\nUnable to initialize your new wallet.\n"
								+ "Error code " + m_SafeInitRep.getErrcode() + ": "
								+ m_SafeInitRep.getErrmsg() + ".\n\n"
								+ "Please try again later.\n";
			JTextArea errTxt = new JTextArea(errMsg, 8, 40);
			errTxt.setEditable(false);
			errTxt.setLineWrap(true);
			errTxt.setWrapStyleWord(true);
			confPane.add(errTxt, gbc);
			confPane.add(Box.createVerticalStrut(20), gbc);
			AbstractAction confirmed = new AbstractAction("Return to Login") {
				public void actionPerformed(ActionEvent actionEvent) {
					// go to login screen
					m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
					m_TabManager.makeTransition();
				}
			};
			m_SubmitButton = new JButton(confirmed);
			m_SubmitButton.setToolTipText("Click here to return to login screen");
			m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
			confPane.add(m_SubmitButton, gbc);
			m_ScrollPane = new JScrollPane(confPane);
			m_ScrollPane.setBorder(null);
			add(m_ScrollPane);
			return;
		}

		JLabel header
			= new JLabel("Your new wallet's login data", SwingConstants.CENTER);
		header.setFont(M_PromptFont);
		confPane.add(header, gbc);
		confPane.add(Box.createVerticalStrut(10), gbc);

		// recapitulate everything (all login data) inside one text area, so
		// that it can be grabbed using copy and paste

		StringBuilder loginTxt = new StringBuilder(1024);
		JTextArea loginData = new JTextArea(5, 40);
		loginData.setEditable(false);
		loginData.setLineWrap(true);
		loginData.setWrapStyleWord(true);
		loginData.setBackground(m_Plugin.m_LabelBackColor);

		// build data
		final VsSecrets secrets = m_Plugin.getLoginSecrets();
		loginTxt.append("Your wallet ID is:  " + secrets.getVSnumber());
		loginTxt.append("\n\nYour PIN is:\t\t   " + m_PIN + "\n");
		loginTxt.append("Your long-phrase is:\t   " + m_LPhUI.getText() + "\n");
		loginTxt.append("Your passphrase is:\t   " + m_PPhUI.getText());
		loginData.setText(loginTxt.toString());
		confPane.add(loginData, gbc);
		confPane.add(Box.createVerticalStrut(30), gbc);

		// recapitulate recovery data in a separate text area
		JLabel rheader
			= new JLabel("Your wallet's recovery data", SwingConstants.CENTER);
		rheader.setFont(M_PromptFont);
		confPane.add(rheader, gbc);
		confPane.add(Box.createVerticalStrut(10), gbc);
		StringBuilder recovTxt = new StringBuilder(1024);
		JTextArea recovData = new JTextArea(5, 40);
		recovData.setEditable(false);
		recovData.setLineWrap(true);
		recovData.setWrapStyleWord(true);
		recovData.setBackground(m_Plugin.m_LabelBackColor);

		// build data, padding questions to 32 with tabs for each 8 missing chars
		final int desLen = 32;
		final int tabLen = 8;
		String ques = m_SafeData.getRecovery_ques_1();
		recovTxt.append(ques);
		int diff = desLen - ques.length();
		if (diff > 0) {
			for (int iii = 0; iii < diff / tabLen; iii++) {
				recovTxt.append("\t");
			}
		}
		recovTxt.append(":\t");
		recovTxt.append(m_Answer1UI.getText().toLowerCase() + "\n");

		ques = m_SafeData.getRecovery_ques_2();
		diff = desLen - ques.length();
		recovTxt.append(ques);
		if (diff > 0) {
			for (int iii = 0; iii < diff / tabLen; iii++) {
				recovTxt.append("\t");
			}
		}
		recovTxt.append(":\t");
		recovTxt.append(m_Answer2UI.getText().toLowerCase() + "\n");

		ques = m_SafeData.getRecovery_ques_3();
		diff = desLen - ques.length();
		recovTxt.append(ques);
		if (diff > 0) {
			for (int iii = 0; iii < diff / tabLen; iii++) {
				recovTxt.append("\t");
			}
		}
		recovTxt.append(":\t");
		recovTxt.append(m_Answer3UI.getText().toLowerCase() + "\n");

		ques = m_SafeData.getRecovery_ques_4();
		diff = desLen - ques.length();
		recovTxt.append(ques);
		if (diff > 0) {
			for (int iii = 0; iii < diff / tabLen; iii++) {
				recovTxt.append("\t");
			}
		}
		recovTxt.append(":\t");
		recovTxt.append(m_Answer4UI.getText().toLowerCase() + "\n");

		ques = m_SafeData.getRecovery_ques_5();
		diff = desLen - ques.length();
		recovTxt.append(ques);
		if (diff > 0) {
			for (int iii = 0; iii < diff / tabLen; iii++) {
				recovTxt.append("\t");
			}
		}
		recovTxt.append(":\t");
		recovTxt.append(m_Answer5UI.getText().toLowerCase());

		recovData.setText(recovTxt.toString());
		confPane.add(recovData, gbc);
		confPane.add(Box.createVerticalStrut(10), gbc);

		JLabel instructs = new JLabel("Please record this data securely, "
												+ "and keep it strictly private.");
		instructs.setFont(M_PromptFont);
		confPane.add(instructs, gbc);
		confPane.add(Box.createVerticalStrut(20), gbc);

		// if the preferences setting so indicates, show keys for this safe too
		VSPreferences prefs = m_Plugin.getUserPreferences();
		if (prefs.getSafeKeysDisp()) {
			JPanel capGrp = new JPanel(gbl);
			capGrp.setBorder(BorderFactory.createTitledBorder(
																"Your wallet's cap/key data"));
			JLabel hashLab = new JLabel("SDS Voucher/Token folder hash: ",
												 SwingConstants.RIGHT);
			String hashVal = new String(secrets.getVoucherIndex());
			JTextArea hash = new JTextArea(hashVal, 2, 30);
			hash.setEditable(false);
			hash.setLineWrap(true);
			firstItemConstraints(gbc);
			capGrp.add(hashLab, gbc);
			lastItemConstraints(gbc);
			capGrp.add(hash, gbc);
			capGrp.add(Box.createVerticalStrut(20), gbc);

			JLabel capLab = new JLabel("SDS Voucher/Token folder enc cap: ",
												SwingConstants.RIGHT);
			JTextArea cap = new JTextArea(secrets.getVouchRWCap(), 6, 40);
			cap.setEditable(false);
			cap.setLineWrap(true);
			firstItemConstraints(gbc);
			capGrp.add(capLab, gbc);
			lastItemConstraints(gbc);
			capGrp.add(cap, gbc);
			capGrp.add(Box.createVerticalStrut(20), gbc);

			JLabel keyLab = new JLabel("Wallet Id private key: ",
												SwingConstants.RIGHT);
			String keyVal
				= StringUtils.encodeBase64(secrets.getPrivKey().getEncoded(),
													false);
			JTextArea key = new JTextArea(keyVal, 23, 40);
			key.setEditable(false);
			key.setLineWrap(true);
			firstItemConstraints(gbc);
			capGrp.add(keyLab, gbc);
			lastItemConstraints(gbc);
			capGrp.add(key, gbc);
			confPane.add(capGrp, gbc);
			confPane.add(Box.createVerticalStrut(10), gbc);
		}

		// define submit button and its action
		AbstractAction confirmed = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_ScrollPane.removeAll();
				removeAll();

				// see if we're creating this safe as a successor for a closing VS
				if (m_IsSuccessor) {
					// tell close safe pane about new safe's critical data
					m_TabManager.getCloseSafePane().setSuccessorSafeData(
																		secrets.getVSnumber(),
																		secrets.getVoucherIndex(),
																		secrets.getVouchRWCap());

					// go to completion of closure
					m_TabManager.queueTransition(
													new Integer(WTabManager.M_CLOSE_SAFE2));
				}
				else {
					// go to login screen
					m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				}
				m_TabManager.makeTransition();
			}
		};
		m_SubmitButton = new JButton(confirmed);
		if (m_IsSuccessor) {
			m_SubmitButton.setToolTipText("Click here to continue closing your "
													+ "original wallet");
		}
		else {
			m_SubmitButton.setToolTipText("Click here to proceed to login screen");
		}
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		gbc.fill = GridBagConstraints.NONE;
		confPane.add(m_SubmitButton, gbc);
		confPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(confPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		getRootPane().setDefaultButton(m_SubmitButton);
	}

	/**
	 * private utility method to prepare GridBagConstraints for use with the
	 * first item on a line (such as a prompt/label)
	 */
	private void firstItemConstraints(GridBagConstraints gbc) {
		if (gbc != null) {
			gbc.gridwidth = GridBagConstraints.RELATIVE;
			gbc.fill = GridBagConstraints.NONE;
		}
	}

	/**
	 * private utility method to prepare GridBagConstraints for use with the
	 * last item on a line (such as an entry field)
	 */
	private void lastItemConstraints(GridBagConstraints gbc) {
		if (gbc != null) {
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			gbc.fill = GridBagConstraints.HORIZONTAL;
		}
	}

	/**
	 * method to define the action for submit button in Phase3 context
	 */
	private void setPhase3Action() {
		m_Phase2Submit = new AbstractAction("Create") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get and validate the PIN entered, and compare to confirmation
				String pinStr = new String(m_PINUI.getText().trim());
				if (!setPIN(pinStr)) {
					return;
				}
				String pinConfStr = new String(m_PINUIConf.getText().trim());
				if (!pinStr.equals(pinConfStr)) {
					showError("cannot set PIN",
								"PIN and PIN confirmation do not match",
								"please re-enter PIN");
					return;
				}
				m_PINUI.setEditable(false);
				m_PINUIConf.setEditable(false);

				// get and validate the long-phrase entered, and compare to its conf
				String lpStr = m_LPhUI.getText().trim();
				if (!setLongPhrase(lpStr)) {
					return;
				}
				String lpConfStr = m_LPhUIConf.getText().trim();
				if (!lpConfStr.equals(lpStr)) {
					showError("cannot set long-phrase",
								"long-phrase and its confirmation do not match",
								"please re-enter long-phrase");
					return;
				}
				m_LPhUI.setEditable(false);
				m_LPhUIConf.setEditable(false);

				// get and validate the passphrase entered, and compare to its conf
				String ppStr = new String(m_PPhUI.getText().trim());
				if (!setPassphrase(ppStr)) {
					return;
				}
				String ppConfStr = new String(m_PPhUIConf.getText().trim());
				if (!ppConfStr.equals(ppStr)) {
					showError("cannot set passphrase",
								"passphrase and its confirmation do not match",
								"please re-enter passphrase");
					return;
				}
				m_PPhUI.setEditable(false);
				m_PPhUIConf.setEditable(false);

				// get the answer to the first recovery question
				final String notEmp = "answer cannot be empty";
				String ans1 = m_Answer1UI.getText().trim();
				if (ans1 == null || ans1.isEmpty()) {
					showError("unsuitable answer to first recovery question",
								notEmp, "please re-enter your answer");
					return;
				}
				m_Answer1UI.setEditable(false);

				// get the answer to the second recovery question
				String ans2 = m_Answer2UI.getText().trim();
				if (ans2 == null || ans2.isEmpty()) {
					showError("unsuitable answer to second recovery question",
								notEmp, "please re-enter your answer");
					return;
				}
				m_Answer2UI.setEditable(false);

				// get the answer to the third recovery question
				String ans3 = m_Answer3UI.getText().trim();
				if (ans3 == null || ans3.isEmpty()) {
					showError("unsuitable answer to third recovery question",
								notEmp, "please re-enter your answer");
					return;
				}
				m_Answer3UI.setEditable(false);

				// get the answer to the fourth recovery question
				String ans4 = m_Answer4UI.getText().trim();
				if (ans4 == null || ans4.isEmpty()) {
					showError("unsuitable answer to fourth recovery question",
								notEmp, "please re-enter your answer");
					return;
				}
				m_Answer4UI.setEditable(false);

				// get the answer to the fifth recovery question
				String ans5 = m_Answer5UI.getText().trim();
				if (ans5 == null || ans5.isEmpty()) {
					showError("unsuitable answer to fifth recovery question",
								notEmp, "please re-enter your answer");
					return;
				}
				m_Answer5UI.setEditable(false);

				// disallow double-clicks
				m_SubmitButton.setEnabled(false);

				// set wait cursor
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

				// everything looks okay; send init safe to OFS
				sendInitSafe();
			}
		};
	}


	// set methods
	/**
	 * set the PIN number
	 * @param pinStr the string containing the PIN
	 * @return true if PIN was legal (all numbers), false otherwise
	 */
	public boolean setPIN(String pinStr) {
		if (pinStr != null && !pinStr.isEmpty()) {
			if (pinStr.length() != m_SafeData.getPin_length()) {
				showError("cannot set PIN", "incorrect length",
							"please enter indicated number of digits");
				return false;
			}
			try {
				Integer pin = Integer.parseInt(pinStr);
			}
			catch (NumberFormatException nfe) {
				showError("cannot set PIN", "illegal characters",
							"please enter numeric digits only");
				return false;
			}
			m_PIN = pinStr;
			return true;
		}
		showError("cannot set PIN", "not entered", "please select a PIN");
		return false;
	}

	/**
	 * set the passphrase
	 * @param pph the passphrase
	 */
	public boolean setPassphrase(String pph) {
		if (pph != null && !pph.isEmpty()) {
			if (pph.length() < M_Passphrase_Min) {
				showError("cannot set passphrase", "phrase entered is too short",
							"please enter at least " + M_Passphrase_Min
							+ " characters");
				return false;
			}
			if (pph.length() > M_Passphrase_Max) {
				showError("cannot set passphrase", "phrase entered is too long",
							"please enter no more than " + M_Passphrase_Max
							+ " characters");
				return false;
			}
			m_Passphrase = pph.getBytes();
			return true;
		}
		showError("cannot set passphrase", "none entered",
					"please enter your passphase");
		return false;
	}

	/**
	 * set the long-phrase
	 * @param lph the long-phrase
	 */
	public boolean setLongPhrase(String lph) {
		if (lph != null && !lph.isEmpty()) {
			int len = lph.length();
			if (len >= M_Longphrase_Min && len <= M_Longphrase_Max) {
				// make sure only upper and lower case letters and digits are used
				boolean goodChars = true;
				for (int iii = 0; iii < len; iii++) {
					char ch = lph.charAt(iii);
					if (!(Character.isDigit(ch) || Character.isLowerCase(ch)
							|| Character.isUpperCase(ch)))
					{
						goodChars = false;
						break;
					}
				}
				if (goodChars) {
					m_LongPhrase = lph.toCharArray();
					return true;
				}
				showError("cannot set long-phrase", "illegal character entered",
							"please use only upper and lower case letters and digits");
				return false;
			}
			showError("cannot set long-phrase",
						"must be between " + M_Longphrase_Min + " and "
						+ M_Longphrase_Max + " characters long",
						"please enter a long-phrase of suitable length");
			return false;
		}
		showError("cannot set long-phrase", "none entered",
					"please enter your long-phrase");
		return false;
	}

	/**
	 * set the publisher list (invoked by OFS callback)
	 * @param publishers the list of available voucher publishers
	 */
	public boolean setPubList(List<String> publishers) {
		if (publishers == null || publishers.isEmpty()) {
			return false;
		}

		m_AllPublishers = publishers;
		m_PubList.clear();
		for (String vPub : publishers) {
			// for display, strip off userid, i.e. everything before the '@'
			int atIdx = vPub.indexOf('@');
			if (atIdx != -1) {
				vPub = vPub.substring(atIdx);
			}
			m_PubList.add(vPub);
		}
		return true;
	}

	/**
	 * set the key specification message (invoked by OFS callback)
	 * @param keySpec the message giving key generation parameters
	 */
	public void setKeySpecs(OFSMessage keySpec) {
		if (keySpec != null
			 && keySpec.getOpcode().equalsIgnoreCase("REP_key_params"))
		{
			m_KeySpecs = keySpec;
		}
	}

	/**
	 * set the safe data reply message (invoked by OFS callback)
	 * @param safeData the message giving the encoded safe data
	 */
	public void setSafeData(OFSMessage safeData) {
		if (safeData != null
			&& safeData.getOpcode().equalsIgnoreCase("REP_safe_data"))
		{
			m_SafeData = safeData;
		}
		else {
			return;
		}

		// make sure that any domain matching our OFS hostname is listed first
		ArrayList<InetSocketAddress> pksPaths = m_SafeData.getKey_server_addr();
		String ofsDomain = m_Plugin.getOFSConnection().getHost();
		// allow for a "host" that's really a dotted quad IP address
		char d1st = ofsDomain.charAt(0);
		if (!Character.isDigit(d1st)) {
			int dotIdx = ofsDomain.indexOf('.');
			if (dotIdx != -1) {
				ofsDomain = ofsDomain.substring(dotIdx);
			}
		}
		InetSocketAddress prefPath = null;
		for (InetSocketAddress isa : pksPaths) {
			if (isa.getHostName().contains(ofsDomain)) {
				prefPath = isa;
				break;
			}
		}
		if (prefPath != null) {
			// put this one at index zero
			pksPaths.remove(prefPath);
			pksPaths.add(0, prefPath);
		}
	}

	/**
	 * set the safe init reply message (invoked by OFS callback)
	 * @param initSafe the message giving the safe creation reply
	 */
	public void setInitRep(OFSMessage initSafe) {
		if (initSafe != null
			&& initSafe.getOpcode().equalsIgnoreCase("REP_safe_init"))
		{
			m_SafeInitRep = initSafe;
		}
	}


	/**
	 * send a REQ_create_safe to the OFS
	 */
	private void sendCreateSafe() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage creS = new OFSMessage();
		creS.setType(IQ.Type.GET);
		creS.setFrom(conn.getUser());
		creS.setTo(conn.getServiceName());
		creS.setPacketID("create_" + stData.getNextOFSid());
		creS.setOpcode("REQ_create_safe");

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(creS.getPacketID());
		m_Plugin.getOFSListener().addFilter(creS.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(creS);
		//Log.debug("REQ_create_safe sent");

		// record outgoing packet
		stData.setLastActivity(creS);
	}

	/**
	 * send a REQ_select_publisher to the OFS
	 */
	private void sendSelectPublisher() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage specPub = new OFSMessage();
		specPub.setType(IQ.Type.SET);
		specPub.setFrom(conn.getUser());
		specPub.setTo(conn.getServiceName());
		specPub.setPacketID("create_" + stData.getNextOFSid());
		specPub.setOpcode("REQ_select_publisher");
		specPub.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(specPub.getPacketID());
		m_Plugin.getOFSListener().addFilter(specPub.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(specPub);

		// record outgoing packet
		stData.setLastActivity(specPub);
	}
	
	/**
	 * send a REQ_init_key to the OFS
	 */
	private void sendInitKey() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		VsSecrets secrets = m_Plugin.getLoginSecrets();

		// build message to send
		OFSMessage keyInit = new OFSMessage();
		keyInit.setType(IQ.Type.SET);
		keyInit.setFrom(conn.getUser());
		keyInit.setTo(conn.getServiceName());
		keyInit.setPacketID("create_" + stData.getNextOFSid());
		keyInit.setOpcode("REQ_init_key");
		keyInit.setVoucher_publisher(secrets.getPublisher());
		String pubkeyTxt
			= StringUtils.encodeBase64(secrets.getPubKey().getEncoded(), false);
		keyInit.setPublic_key(pubkeyTxt);
		keyInit.setVSprefix(m_VSprefix.getText());

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(keyInit.getPacketID());
		m_Plugin.getOFSListener().addFilter(keyInit.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(keyInit);

		// record outgoing packet
		stData.setLastActivity(keyInit);
	}

	/**
	 * send a REQ_init_safe to the OFS
	 */
	private void sendInitSafe() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		VsSecrets secrets = m_Plugin.getLoginSecrets();

		// prepare outgoing request
		OFSMessage safeInit = new OFSMessage();
		safeInit.setType(IQ.Type.SET);
		safeInit.setFrom(conn.getUser());
		safeInit.setTo(conn.getServiceName());
		safeInit.setPacketID("create_" + stData.getNextOFSid());
		safeInit.setOpcode("REQ_init_safe");
		safeInit.setVoucher_publisher(secrets.getPublisher());

		/* Build the <details_blob/>, which has this format:
		 * <vs_details>
		 * 	<VS_number/>
		 * 	<secret_details>		(encrypted using passphrase)
		 * 		<voucher_publisher/>		(id of VP)
		 * 		<certificate/>		(includes public key, signed by VP)
		 * 		<private_key/>
		 * 		<voucher_index/>
		 * 		<voucher_rd_cap/>
		 * 		<receipts_index/>
		 * 		<receipts_rw_cap/>
		 * 		<voucher_rw_cap/>
		 * 	</secret_details>
		 * 	<hashkey/>				(== hash(walletId+PIN))
		 * 	<long-phrase/>
		 * 	<recovery_details/>	(encrypted using the hash of the walletId+As)
		 * 		<VS_number/>
		 * 		<voucher_publisher/>
		 * 		<pin/>
		 * 		<long-phrase/>
		 * 		<passphrase/>
		 * 	</recovery_details/>
		 * 	<recovery_index/>		(== hash(hash(walletId+recovery answers)))
		 * 	<login_index/>			(== hash(hash(walletId+PIN)+long-phrase))
		 * </vs_details/>
		 */
		// begin details blob
		StringBuilder detBuff = new StringBuilder(4096);
		detBuff.append("<vs_details>\n");
		detBuff.append("\t<VS_number>" + secrets.getVSnumber());
		detBuff.append("</VS_number>\n");
		detBuff.append("\t<secret_details>\n");

		// build <secret_details/>
		StringBuilder secDets = new StringBuilder(4096);
		PublicKey pubKey = secrets.getPubKey();
		byte[] certEnc = null;
		try {
			certEnc = secrets.getCertificate().getEncoded();
		}
		catch (CertificateEncodingException cee) {
			Log.error("Could not get encoded bytes of wallet cert", cee);
			showError("bad certificate encoding", cee.getMessage());
			return;
		}
		secDets.append("<voucher_publisher>");
		secDets.append(secrets.getPublisher());
		secDets.append("</voucher_publisher>\n");
		secDets.append("<certificate>");
		secDets.append(StringUtils.encodeBase64(certEnc));
		secDets.append("</certificate>\n");
		PrivateKey privKey = secrets.getPrivKey();
		secDets.append("<private_key>");
		secDets.append(StringUtils.encodeBase64(privKey.getEncoded(), false));
		secDets.append("</private_key>\n");
		secDets.append("<voucher_index>");
		secDets.append(new String(secrets.getVoucherIndex()));
		secDets.append("</voucher_index>\n");
		secDets.append("<voucher_rd_cap>");
		secDets.append(new String(secrets.getVoucherReadCap()));
		secDets.append("</voucher_rd_cap>\n");
		secDets.append("<receipts_index>");
		secDets.append(new String(secrets.getReceiptIndex()));
		secDets.append("</receipts_index>\n");
		secDets.append("<receipts_rw_cap>");
		secDets.append(new String(secrets.getReceiptRWCap()));
		secDets.append("</receipts_rw_cap>\n");
		secDets.append("<voucher_rw_cap>");
		secDets.append(secrets.getVouchRWCap());
		secDets.append("</voucher_rw_cap>\n");

		// Encrypt it to the passphrase, padded out to 24 bytes
		String encDetails
			= EncodingUtils.makeBase64DESkeyEncStr(secDets.toString(),
																m_Passphrase);
		if (encDetails == null) {
			Log.error("Unable to encrypt secret details blob");
			showError("unable to encrypt secret details blob");
			return;
		}
		detBuff.append(encDetails + "\n");
		detBuff.append("\t</secret_details>\n");

		// compute the hashkey, hash(walletId+PIN)
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			showError("can't construct SHA512 hash", nsae.getMessage());
			return;
		}
		md.update(secrets.getVSnumber().getBytes());
		md.update(m_PIN.getBytes());
		byte[] hashkeyBytes = md.digest();
		detBuff.append("\t<hashkey>");
		detBuff.append(StringUtils.encodeBase64(hashkeyBytes, false));
		detBuff.append("</hashkey>\n");

		// add long-phrase
		String LPstr = new String(m_LongPhrase);
		detBuff.append("\t<long-phrase>" + LPstr + "</long-phrase>\n");

		// now build the <recovery_details/>
		detBuff.append("\t<recovery_details>");
		StringBuilder recovDet = new StringBuilder(2048);
		recovDet.append("<VS_number>" + secrets.getVSnumber());
		recovDet.append("</VS_number>\n");
		recovDet.append("<voucher_publisher>");
		recovDet.append(secrets.getPublisher());
		recovDet.append("</voucher_publisher>\n");
		recovDet.append("<pin>" + m_PIN + "</pin>\n");
		recovDet.append("<long-phrase>" + LPstr + "</long-phrase>\n");
		recovDet.append("<passphrase>" + new String(m_Passphrase));
		recovDet.append("</passphrase>\n");
		byte[] recovBytes = recovDet.toString().getBytes();
		
		// compute the SHA512 hash of the walletId + 5 recovery question answers
		StringBuilder rh = new StringBuilder(1024);
		rh.append(secrets.getVSnumber() + ":");
		rh.append(m_Answer1UI.getText().toLowerCase() + ":");
		rh.append(m_Answer2UI.getText().toLowerCase() + ":");
		rh.append(m_Answer3UI.getText().toLowerCase() + ":");
		rh.append(m_Answer4UI.getText().toLowerCase() + ":");
		rh.append(m_Answer5UI.getText().toLowerCase());
		md.reset();
		byte[] recovHash = md.digest(rh.toString().getBytes());

		// encrypt the recovery details to this hash
		String encRecov
			= EncodingUtils.makeBase64DESkeyEncStr(recovDet.toString(), recovHash);
		if (encRecov == null) {
			Log.error("Could not encrypt recovery details to answer hash");
			showError("unable to encrypt recovery details to answer hash");
			return;
		}
		detBuff.append(encRecov);
		detBuff.append("</recovery_details>\n");

		// compute the hash of the recovery encryption hash
		md.reset();
		byte[] recovIdx = md.digest(recovHash);
		detBuff.append("\t<recovery_index>");
		detBuff.append(StringUtils.encodeBase64(recovIdx, false));
		detBuff.append("</recovery_index>\n");

		// compute the hash(hash(walletId+PIN)+long-phrase) for login blob index
		md.reset();
		md.update(hashkeyBytes);
		md.update(LPstr.getBytes());
		byte[] secDetHash = md.digest();
		detBuff.append("\t<login_index>");
		detBuff.append(StringUtils.encodeBase64(secDetHash, false));
		detBuff.append("</login_index>\n");

		// done!
		detBuff.append("</vs_details>");

		/* Now, set the DHT marking hash in the secrets block.
		 * This is computable as: hash(walletId+PIN+passphrase).
		 */
		md.reset();
		md.update(secrets.getVSnumber().getBytes());
		md.update(m_PIN.getBytes());
		md.update(m_Passphrase);
		byte[] dhtHash = md.digest();
		String hashStr = StringUtils.encodeBase64(dhtHash, false);
		secrets.setDHThash(hashStr);
		//Log.debug("Set DHT read hash to: " + hashStr);

		// encrypt the details blob using the VP's pubkey
		PublicKey vpPubkey = secrets.getVPKey();
		if (vpPubkey != null) {
			String encDetBlob
				= EncodingUtils.makeBase64PubkeyEncStr(detBuff.toString(),
																	vpPubkey);
			if (encDetBlob == null) {
				Log.error("Could not encrypt details blob for VP");
				showError("error encrypting details blob with VP pubkey");
				return;
			}
			safeInit.setDetails_blob(encDetBlob);
		}
		else {
			Log.error("VP pubkey not available");
			showError("missing VP pubkey");
			return;
		}

		// Now send the constructed REQ_init_safe to the OFS.
		// First, register a listener for the OFS's reply.

		PacketIDFilter pFilter = new PacketIDFilter(safeInit.getPacketID());
		m_Plugin.getOFSListener().addFilter(safeInit.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(safeInit);

		// record outgoing packet
		stData.setLastActivity(safeInit);
	}

	/**
	 * method to queue up a modification to a text component from within its
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	/**
	 * utility method to update the status of the submit button
	 */
	private void updateSubmitButtonState() {
		// our tooltips depend on which page we're on
		if (m_FirstPage) {
			if (m_HavePrefix) {
				m_SubmitButton.setEnabled(true);
				m_SubmitButton.setToolTipText("Click here to create your wallet");
			}
			else {
				m_SubmitButton.setEnabled(false);
				if (m_PubList.size() == 1) {
					m_SubmitButton.setToolTipText("Enter your wallet identifying "
															+ "name");
				}
				else {
					m_SubmitButton.setToolTipText("<html>Select your publisher, and "
															+ "enter<br/>your wallet "
															+ "identifying name</html>");
				}
			}
		}
		else {
			// see if we have the PIN and its confirmation yet
			String pin = m_PINUI.getText();
			String pinConf = m_PINUIConf.getText();
			boolean havePin = false;
			if (pin.length() == m_PINLength && pin.equals(pinConf)) {
				havePin = true;
			}

			// see if we have the longphrase and its confirmation yet
			String longPh = m_LPhUI.getText();
			String longPhConf = m_LPhUIConf.getText();
			boolean haveLongphrase = false;
			if (longPh.length() >= M_Longphrase_Min
				 && longPh.length() <= M_Longphrase_Max
				 && longPh.equals(longPhConf))
			{
				haveLongphrase = true;
			}

			// see if we have the passphrase and its confirmation yet
			boolean havePassphrase = false;
			String passPh = m_PPhUI.getText();
			String passPhConf = m_PPhUIConf.getText();
			if (passPh.length() >= M_Passphrase_Min
				 && passPh.length() <= M_Passphrase_Max
				 && passPh.equals(passPhConf))
			{
				havePassphrase = true;
			}

			// see if we have non-empty answers for all five questions
			boolean haveAnswers = false;
			String ans1 = m_Answer1UI.getText();
			String ans2 = m_Answer2UI.getText();
			String ans3 = m_Answer3UI.getText();
			String ans4 = m_Answer4UI.getText();
			String ans5 = m_Answer5UI.getText();
			if (!ans1.isEmpty() && !ans2.isEmpty() && !ans3.isEmpty()
				 && !ans4.isEmpty() && !ans5.isEmpty())
			{
				haveAnswers = true;
			}

			StringBuilder tip = new StringBuilder(256);
			tip.append("Enter and confirm your ");
			if (!havePin) {
				tip.append("PIN, ");
			}
			if (!haveLongphrase) {
				tip.append("long-phrase, ");
			}
			if (!havePassphrase) {
				tip.append("passphrase, ");
			}
			if (!haveAnswers) {
				tip.append("recovery answers");
			}
			if (havePin && haveLongphrase && havePassphrase && haveAnswers) {
				m_SubmitButton.setEnabled(true);
				m_SubmitButton.setToolTipText("Click here to create your wallet");
			}
			else {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText(tip.toString());
			}
		}
	}

	// implement ActionListener
	/**
	 * method to handle ActionEventS from input fields, in order to transfer
	 * focus to the next input
	 * @param ae the event from one of the input objects
	 */
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JTextField) {
			JTextField input = (JTextField) src;
			// move to the next focus item
			input.transferFocus();
		}
		else if (src instanceof JComboBox) {
			JComboBox box = (JComboBox) src;
			// move to the next focus item
			box.transferFocus();
		}
	}


	// error display methods
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	private void showError(String errMsg) {
		m_TabManager.showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	private void showError(String errMsg, String cause) {
		m_TabManager.showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	private void showError(String errMsg, String cause, String action) {
		m_TabManager.showError(errMsg, cause, action);
	}


	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_Passphrase != null) {
				Arrays.fill(m_Passphrase, (byte)0);
			}
			if (m_LongPhrase != null) {
				Arrays.fill(m_LongPhrase, ' ');
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
