/*
 * last modified---
 * 	11-23-16/jet/ use multi-select mode
 * 	07-17-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine value
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	06-05-13/jet/ SilentVault client version
 * 	12-18-12/kbw/ use getNPV() instead of baseUnitValue()
 * 	04-18-11/kbw/ reset TabManager's pending to null on errors
 * 	02-23-11/kbw/ accommodate different voucher types, not just GAU
 * 	02-21-11/kbw/ call getValueOfOneToken() instead of token quantity constant
 * 	10-22-10/kbw/ provide automatic purchase mode
 * 	10-12-10/kbw/ make enter in text input transfer focus
 * 	09-25-10/kbw/ make token quantity processing/updating robust
 * 	09-20-10/kbw/ wording changes
 * 	09-02-10/kbw/ changes for Nimbus L&F
 * 	08-03-10/kbw/ change layout; impose max of 200 tokens; show voucher total
 * 	07-06-10/kbw/ require minimum token lot size of 10 tokens
 * 	05-03-10/kbw/ grey out buttons during messaging operations
 * 	04-26-10/kbw/ show fiat equivalency of tokens
 * 	04-14-10/kbw/ flesh out
 * 	03-29-10/kbw/ new (stub)
 *
 * purpose---
 * 	provide a UI widget to handle token purchases
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.net.HttpURLConnection;
import java.security.*;


/**
 * Extends WorkPane to provide a screen for making token purchases.
 *
 * @author Kevin Wilkerson
 */
public final class TokensPane extends WorkPane {
	// BEGIN data members
	/**
	 * minimum number of tokens which can be purchased
	 */
	public static final int		M_MinTokens = 10;

	/**
	 * maximum number of tokens which can be purchased
	 */
	public static final int		M_MaxTokens = 200;

	/**
	 * the action to perform a token purchase
	 */
	private AbstractAction		m_PurchTokensAction;

	/**
	 * the action to abort token purchase
	 */
	private AbstractAction		m_AbortAction;

	/**
	 * the button to consummate the token purchase
	 */
	private JButton				m_PurchaseButton;

	/**
	 * the button to abort the token purchase (and the triggering action)
	 */
	private JButton				m_AbortButton;

	/**
	 * UI element to hold token everything we add
	 */
	private JPanel					m_TokBuy;

	/**
	 * UI element to hold token buy elements and equivalency display
	 */
	private JPanel					m_QuanPane;

	/**
	 * UI element to hold number of tokens to be purchased
	 */
	private JTextField			m_TokenQuantity;

	/**
	 * UI element to hold token lot size value equivalency statement
	 */
	private JLabel					m_TokenValue;

	/**
	 * UI element to hold token fiat equivalent value
	 */
	private JLabel					m_TokenEquiv;

	/**
	 * the current token lot size which we are buying
	 */
	private int						m_LotSize;

	/**
	 * track whether a token quantity entry undo is in progress
	 */
	private boolean				m_DoingUndo;

	/**
	 * inner class to provide a DocumentListener for the m_TokenQuantity field
	 */
	final class TokDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// see if value parses
			if (parseValue()) {
				buildEquivStrings();
			}
			else {
				// undo this edit
				int changeAt = de.getOffset();
				try {
					curTxt = doc.getText(0, doc.getLength());
					String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
					queueTextMod(m_TokenQuantity, modTxt);
				}
				catch (BadLocationException ble) {
					Log.error("Cannot undo bad token quantity edit", ble);
				}
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// see if value parses
			if (parseValue()) {
				buildEquivStrings();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_TokenQuantity
		 * @return true on success, false on failure
		 * NB: this routine sets m_LotSize as an intentional side-effect
		 */
		private boolean parseValue() {
			// examine what's in the value entry field
			String enteredVal = m_TokenQuantity.getText();
			if (enteredVal.isEmpty() || enteredVal.equals("0")) {
				m_LotSize = 0;
				return true;
			}

			// format allows l to specify long, but we don't
			if (enteredVal.indexOf('l') != -1 || enteredVal.indexOf('L') != -1) {
				return false;
			}

			// disallow +-
			if (enteredVal.startsWith("-") || enteredVal.startsWith("+")) {
				return false;
			}

			// attempt to parse, very strictly
			int vvv = 0;
			try {
				Integer val = Integer.parseInt(enteredVal);
				vvv = val.intValue();
			}
			catch (NumberFormatException nfe) {
				return false;
			}

			// record acceptable value and return success
			m_LotSize = vvv;
			return true;
		}
	}

	/**
	 * document listener for dynamic updates of m_TokenQuantity
	 */
	private TokDocumentListener		m_TokQuantListener;

	/**
	 * inner class to modify a text field from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		/**
		 * the text component being modified
		 */
		JTextComponent    m_ModComponent;

		/**
		 * the new contents for the component
		 */
		String            m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
	   }

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());
			buildEquivStrings();
		}
	}

	/**
	 * whether we're doing a purchase of tokens in automatic mode
	 */
	private boolean						m_AutoMode;

	/**
	 * the progress bar to use in automatic purchase mode
	 */
	private JProgressBar					m_AutoProgress;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public TokensPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_AutoMode = true;

		// provide user instructions
		String buyTxt = "You don't have enough tokens to perform the last "
							+ "operation you requested.<br/><br/>"
							+ "Please select a voucher and enter the number of tokens "
							+ "you wish to buy (minimum " + M_MinTokens + ").  "
							+ "You will receive a voucher back for any change.";
		m_UserInstructs = buyTxt;

		// set up button actions and corresponding buttons
		setPurchTokensAction();
		m_PurchaseButton = new JButton(m_PurchTokensAction);
		m_PurchaseButton.setFont(m_Plugin.M_ButtonFont);
		m_PurchaseButton.setEnabled(false);

		m_AbortAction = new AbstractAction("Abort Purchase") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_LotSize = 0;
				m_ButtonPane.removeAll();

				// return to main menu
				m_TabManager.setPendingTokens(null);
				m_TabManager.queueTransition(
													new Integer(WTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
		m_AbortButton = new JButton(m_AbortAction);
		m_AbortButton.setToolTipText("<html>Click here to abort token purchase"
											+ "<br/>and return to the main menu</html>");
		m_AbortButton.setFont(m_Plugin.M_ButtonFont);

		// other one-time init
		m_HeadingPane.setLayout(new GridLayout(1, 3));
		m_TokQuantListener = new TokDocumentListener();
		m_TokenQuantity = new JTextField(6);
		m_TokenQuantity.setDropTarget(null);
		m_TokenQuantity.setToolTipText("Enter the number of tokens you want");
		m_TokenQuantity.setHorizontalAlignment(JTextField.RIGHT);
		m_TokenQuantity.setPreferredSize(new Dimension(40, 30));
		m_TokenQuantity.setMaximumSize(new Dimension(40, 30));
		m_TokenQuantity.addActionListener(this);

		m_TokenValue = new JLabel();
		m_TokenValue.setHorizontalAlignment(SwingConstants.CENTER);
		m_TokenValue.setFont(m_ParentHome.M_TableFont);

		m_TokenEquiv = new JLabel();
		m_TokenEquiv.setHorizontalAlignment(SwingConstants.CENTER);
		m_TokenEquiv.setFont(new Font(Font.SANS_SERIF, Font.ITALIC, 13));
		m_TokenEquiv.setBorder(m_InsetBorder);

		m_TokBuy = new JPanel(new GridBagLayout());
		m_TokBuy.setBorder(null);
		m_TokBuy.setPreferredSize(UIManager.getDimension("vouchTableDim"));
		m_QuanPane = new JPanel();
		m_QuanPane.setLayout(new BoxLayout(m_QuanPane, BoxLayout.X_AXIS));
		m_QuanPane.setBorder(m_InsetBorder);
		m_QuanPane.setOpaque(false);

		m_AutoProgress = new JProgressBar();
		m_AutoProgress.setStringPainted(true);
		m_AutoProgress.setFont(m_ParentHome.M_DataFont);
		m_AutoProgress.setBackground(Color.LIGHT_GRAY);
		m_AutoProgress.setOpaque(true);
		m_AutoProgress.setIndeterminate(true);

		// no general border
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to establish the Action which can process a request to
	 * purchase a given number of tokens
	 */
	private void setPurchTokensAction() {
		final String stdErr = "could not buy tokens";

		m_PurchTokensAction = new AbstractAction("Buy Now") {
			public void actionPerformed(ActionEvent actionEvent) {
				// examine the quantity text field
				Integer lotSz = null;
				try {
					lotSz = Integer.parseInt(m_TokenQuantity.getText());
				}
				catch (NumberFormatException nfe) {
					// should be impossible given document listener parsing
					showError("cannot purchase tokens",
								"invalid token lot size entered, "
								+ m_TokenQuantity.getText(),
								"please enter a positive integral number");
					return;
				}
				int sz = lotSz.intValue();
				// check versus min and max values
				if (sz < M_MinTokens) {
					showError("cannot purchase tokens",
								"invalid token lot size entered, " + sz,
								"please enter a positive integral number at least "
								+ M_MinTokens);
					return;
				}
				if (sz > M_MaxTokens) {
					showError("cannot purchase tokens",
								"too many tokens requested, " + sz,
								"please enter a number <= " + M_MaxTokens);
					return;
				}
				if (sz != m_LotSize) {
					// override previous value
					m_LotSize = sz;
				}

				// obtain the selected voucher(s)
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> pmtVouchers = vtm.getSelectedVouchers();
				if (pmtVouchers.isEmpty()) {
					if (!m_AutoMode) {
						// should be impossible if button was not greyed out
						showError(stdErr, "no vouchers selected",
								"please select exactly one voucher to use for payment");
					}
					else {
						// we checked for this case at auto-selection
						Log.error("No vouchers for auto-token purchase");
					}
					return;
				}

				// total up and check that voucher selection is enough
				double vVal = 0.0;
				String asset = null;
				String unit = null;
				for (XMLVoucher voucher : pmtVouchers) {
					vVal += voucher.getNPV();
					if (asset == null) {
						asset = voucher.getAsset();
						unit = voucher.getUnits();
					}
				}
				Integer tQuant = new Integer(m_LotSize);
				Double oneToken = XMLToken.getValueOfOneToken(asset, unit);
				double grVal = tQuant.doubleValue() * oneToken.doubleValue();
				if (vVal < grVal) {
					if (!m_AutoMode) {
						// this should be impossible, since button is greyed out until
						// adequate voucher value has been selected

						showError(stdErr, "your voucher is of insufficient size",
									"please select a voucher worth at least " + grVal
									+ " grams");
					}
					else {
						// we checked for this case after auto-selection
						Log.error("Insufficient vouchers for auto-token purchase");
					}
					return;
				}

				// build <voucherRequest/>
				Hashtable<XMLVoucher, String> vouchers = vtm.getVoucherList();
				VsSecrets secData = m_Plugin.getLoginSecrets();
				VoucherRequest vReq = new VoucherRequest();
				if (pmtVouchers.size() == 1) {
					vReq.setAction("split");
				}
				else {
					vReq.setAction("merge");
				}
				vReq.setSigningVS(secData.getVSnumber());
				vReq.setFolderHash(new String(secData.getVoucherIndex()));
				vReq.setFolderCap(secData.getVouchRWCap());
				HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
				for (XMLVoucher vvv : pmtVouchers) {
					// get voucher's storage filename and add to VR
					String sdsFile = vouchers.get(vvv);
					inpVouchers.put(sdsFile, vvv);
				}
				vReq.addOutputValue(grVal, unit);
				vReq.setInitialized();

				// sign the voucher request with our private key
				if (!vReq.signDetails(secData.getPrivKey())) {
					Log.error("Unable to sign VR with VS privkey");
					showError(stdErr, "error signing token purchase request");
					return;
				}

				// encrypt the details to the VP's key so the OFS can't read them
				if (!vReq.encryptDetails(secData.getVPKey(), false)) {
					Log.error("Unable to encrypt VR with VP pubkey");
					showError(stdErr, "error encrypting purchase request for VP");
					return;
				}

				// commit all vouchers to the operation
				vtm.commitVouchers(pmtVouchers);

				// send the request to the OFS/VP for processing
				sendPurchaseTokens(vReq);
			}
		};
	}

	/**
	 * private utility method to send a token purchase request.
	 * @param vReq the <voucherRequest/> object, filled out, signed and encrypted
	 */
	private void sendPurchaseTokens(VoucherRequest vReq) {
		if (vReq == null || !vReq.isEncrypted()) {
			Log.error("Null or unencrypted <voucherRequest/> for token purchase");
			return;
		}

		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot buy tokens while not logged in!");
			showError("cannot buy tokens", "not logged in", "log in first");
			return;
		}

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);
		if (m_AutoMode) {
			// update progress bar
			m_AutoProgress.setString("Purchasing " + m_LotSize
											+ " tokens, please wait");
			JPanel statBar = m_Plugin.getStatusBar();
			statBar.removeAll();
			statBar.add(m_AutoProgress);
			statBar.invalidate();
			statBar.revalidate();
			statBar.repaint();
		}

		// build message to send
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		OFSMessage buyTok = new OFSMessage();
		buyTok.setType(IQ.Type.SET);
		buyTok.setFrom(conn.getUser());
		buyTok.setTo(conn.getServiceName());
		buyTok.setPacketID("purch_" + state.getNextOFSid());
		buyTok.setOpcode("REQ_purchase_tokens");
		buyTok.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());
		buyTok.setVouchReq(vReq);

		// register an asynchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(buyTok.getPacketID());
		m_Plugin.getOFSListener().addFilter(buyTok.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(buyTok);
		//Log.debug(buyTok.getOpcode() + " sent to OFS");

		// record outgoing packet
		state.setLastActivity(buyTok);

		// NB: callback with result goes to processPurchTokenReply() below
	}

	/**
	 * method to process a reply received from OFS to a token purchase
	 * @param purchRep the reply message from the OFS
	 */
	public void processPurchTokenReply(OFSMessage purchRep) {
		if (purchRep == null
			|| !purchRep.getOpcode().equalsIgnoreCase("REP_tokens_purchased"))
		{
			Log.error("Bad token purchase reply received");
			return;
		}

		// clean up UI states
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (m_AutoMode) {
			// clear status bar
			JPanel statBar = m_Plugin.getStatusBar();
			statBar.remove(m_AutoProgress);
			m_Plugin.setDefaultStatusBar();
		}

		// see whether this reply is an error
		final String err = "could not process token purchase";
		if (purchRep.getType() == IQ.Type.ERROR) {
			Log.error("Error purchasing tokens, code " + purchRep.getErrcode()
						+ ": " + purchRep.getErrmsg());
			showError(err, "code " + purchRep.getErrcode() + ": "
						+ purchRep.getErrmsg());
			m_TabManager.setPendingTokens(null);
			return;
		}

		// decode/decrypt the <purchased/> - show wait cursor
		m_ParentHome.setCursor(M_WaitCursor);
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String decPurch
			= EncodingUtils.getStrFromBase64PubkeyEnc(purchRep.getPurchased(),
																	secrets.getPrivKey());
		if (decPurch == null || decPurch.isEmpty()) {
			Log.error("Unable to decode <purchased/> of " + purchRep.getOpcode());
			m_ParentHome.setCursor(null);
			showError(err, "could not decrypt reply confirmation",
						"log in again to recheck your voucher and token lists");
			m_TabManager.setPendingTokens(null);
			return;
		}

		// parse it
		Element purchElt = EncodingUtils.buildElement(decPurch);
		if (purchElt == null) {
			Log.error("Unable to parse <purchased/> of " + purchRep.getOpcode());
			m_ParentHome.setCursor(null);
			showError(err, "could not parse confirmation details");
			m_TabManager.setPendingTokens(null);
			return;
		}
		int tokenCount = 0;
		XMLVoucher changeVoucher = null;
		String chngVouchFile = null;
		try {
			// get list of new tokens
			Element newToksElt = purchElt.element("tokenCount");
			if (newToksElt == null) {
				throw new DocumentException("Missing tokenCount list");
			}
			tokenCount = Integer.parseInt(newToksElt.getText());

			// get change voucher -- note this could conceivably be an empty field
			Element vouchElt = purchElt.element("voucher");
			if (vouchElt == null) {
				throw new DocumentException("Missing change voucher");
			}
			if (!vouchElt.asXML().equals("<voucher/>")) {
				changeVoucher = XMLVoucher.getVoucherRep(vouchElt);

				// get the voucher's storage filename
				Element fileElt = purchElt.element("voucherFile");
				if (fileElt == null) {
					throw new DocumentException("Missing change voucher filename");
				}
				chngVouchFile = fileElt.getTextTrim();
			}

			// get VP signature and validate it
			Element sigElt = purchElt.element("signature");
			if (sigElt == null) {
				throw new DocumentException("Missing VP signature");
			}
			String b64Sig = sigElt.getTextTrim();
			int sigIdx = decPurch.lastIndexOf("<signature");
			String sigData = decPurch.substring(0, sigIdx);
			Signature vpSig = null;
			try {
				vpSig = Signature.getInstance("SHA1withRSA");
				vpSig.initVerify(secrets.getVPKey());
				vpSig.update(sigData.getBytes());
				if (!vpSig.verify(StringUtils.decodeBase64(b64Sig))) {
					throw new DocumentException("Invalid VP signature on reply");
				}
			}
			catch (NoSuchAlgorithmException nsae) {
				// impossible
				throw new DocumentException("No such sig algorithm", nsae);
			}
			catch (InvalidKeyException ike) {
				throw new DocumentException("Invalid VP pubkey", ike);
			}
			catch (SignatureException se) {
				throw new DocumentException("Unable to validate token purchase "
													+ "reply signature", se);
			}
		}
		catch (DocumentException de) {
			Log.error("Parse error on token purchase reply", de);
			m_ParentHome.setCursor(null);
			showError(err, "parse error: " + de.getMessage());
			m_TabManager.setPendingTokens(null);
			return;
		}

		// if we have one, add the change voucher to the voucher table model
		if (changeVoucher != null) {
			m_ParentHome.addVoucher(changeVoucher, chngVouchFile);
		}
		m_ParentHome.setCursor(null);

		// and delete the vouchers we previously committed to the purchase
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> commitList = vtm.getCommittedVouchers();
		m_ParentHome.spendVouchers(commitList);

		if (!m_AutoMode) {
			// now display message indicating success and no. tokens purchased
			m_TabManager.showInfo("Your purchase of " + tokenCount
										+ " tokens was successful.");
		}

		// next, display warning if we have sync pending
		if (purchRep.syncPending()) {
			showError("warning: folder updates were queued", "SDS timeout",
						"wait a few minutes then use Resync for tokens to appear");
		}
		else {
			// synchronously query token contents to reflect purchase
			XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
			m_ParentHome.setCursor(M_WaitCursor);
			setButtonState(false);
			m_TabManager.queryTokens(sdsConn, true);
			m_ParentHome.setCursor(null);
		}

		// reset
		m_LotSize = 0;
		m_ButtonPane.removeAll();

		// Send user back to whatever screen they came from (main menu is default)
		// which continues whatever operation was about to occur.

		m_TabManager.queueTransition(new Integer(WTabManager.M_PENDING_TOKENS));
		m_TabManager.makeTransition();
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// set heading pane
		buildValueHeader();

		// set screen label
		buildScreenLabel("Buy Tokens");

		// calculate how many tokens were wanted from last operation
		JPanel pendPane = m_TabManager.getPendingTokensPane();
		int tokReq = 0;
		String what = "";
		if (pendPane instanceof ValidatePane) {
			tokReq = VoucherRequest.M_ValidateCost;
			what = "to validate a voucher";
		}
		else if (pendPane instanceof MergePane) {
			tokReq = VoucherRequest.M_MergeCost;
			what = "to merge vouchers";
		}
		else if (pendPane instanceof SplitPane) {
			tokReq = VoucherRequest.M_SplitCost;
			what = "to split a voucher";
		}
		else if (pendPane instanceof PaymentPane) {
			tokReq = VoucherRequest.M_PaymentCost;
			what = "to make a payment";
		}
		else if (pendPane instanceof NewReceiptPane) {
			tokReq = VoucherRequest.M_ReceiptCost;
			what = "to store a receipt";
		}
		else if (pendPane instanceof IncomingPane) {
			tokReq = VoucherRequest.M_ReceiveCost;
			what = "to receive a payment";
		}
		else if (pendPane instanceof ConfirmPane) {
			tokReq = VoucherRequest.M_ReceiptCost;
			what = "to confirm a payment";
		}
		else if (pendPane instanceof CloseSafePane) {
			tokReq = VoucherRequest.M_CloseCost;
			what = "to close a wallet";
		}
		m_TokenCost.setText(tokReq + " tokens " + what);

		// obtain default quantity from user preferences
		Integer tokLotSize = null;
		if (m_LotSize == 0) {
			tokLotSize
				= new Integer(m_Plugin.getUserPreferences().getTokenLotSize());
			m_LotSize = tokLotSize.intValue();
		}
		else {
			tokLotSize = new Integer(m_LotSize);
		}
		m_TokenQuantity.setText(tokLotSize.toString());

		/* If we have a pending token pane (i.e. a place we came from when we
		 * decided we lacked sufficient tokens), then we want to perform this
		 * operation invisibly, updating only the progress bar.  Otherwise, if
		 * we got here by selection of Buy Tokens off the Advanced menu, then
		 * we want to show the screen as laid out below.
		 */
		JPanel fromPane = m_TabManager.getPendingTokensPane();
		if (fromPane != null) {
			m_AutoMode = true;

			// say what we're doing in main screen
			JLabel whatLab = new JLabel("Purchasing tokens " + what);
			whatLab.setFont(m_ParentHome.M_HeaderFont);
			m_TokBuy.removeAll();
			m_GBC.gridheight = 3;
			m_GBC.fill = GridBagConstraints.HORIZONTAL;
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			m_GBC.anchor = GridBagConstraints.NORTH;
			m_TokBuy.add(Box.createVerticalStrut(150), m_GBC);
			m_TokBuy.add(whatLab, m_GBC);
			m_TokBuy.add(Box.createVerticalStrut(150), m_GBC);
			m_WorkArea.setViewportView(m_TokBuy);
			revalidate();
			repaint();

			// auto-select some vouchers
			Integer tQuant = new Integer(m_LotSize);
			VoucherTableModel vtm = m_ParentHome.getVoucherModel();
			Hashtable<XMLVoucher, String> issuerVouchers = vtm.getIssuerVouchers();
			if (issuerVouchers.isEmpty()) {
				showError("unable to purchase tokens for requested operation",
							"your wallet contains no vouchers");
				return;
			}
			// compute value of tokens to be purchased
			XMLVoucher firstVouch = vtm.getFirstVoucher();
			Double oneToken = XMLToken.getValueOfOneToken(firstVouch.getAsset(),
																		 firstVouch.getUnits());
			double grVal = tQuant.doubleValue() * oneToken.doubleValue();
			vtm.deselectAll();
			vtm.setSelectMode(VoucherTableModel.M_SelectMulti);
			ArrayList<XMLVoucher> selVouchers = vtm.autoSelectVouchers(grVal);
			if (selVouchers == null) {
				showError("unable to purchase tokens for requested operation",
							"your wallet contains insufficient voucher value",
							"reduce the token lot size in your user preferences");
			}
			else {
				// push the button programmatically
				vtm.selectVouchers(selVouchers);
				ActionEvent trigger
					= new ActionEvent(m_PurchaseButton, m_LotSize, "AUTOBUY");
				m_PurchTokensAction.actionPerformed(trigger);
			}
			return;
		}
		m_AutoMode = false;

		// add abort button as sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_AbortButton);

		// add Purchase button in button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(350));	// spacer
		m_PurchaseButton.setEnabled(false);
		m_PurchaseButton.setToolTipText("Select voucher(s) to pay for your token "
												+ "purchase");
		m_ButtonPane.add(m_PurchaseButton);

		// place the voucher table into a scroll pane, set for multi select mode
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.deselectAll();
		vtm.setSelectMode(VoucherTableModel.M_SelectMulti);
		JTable vt = m_ParentHome.getVoucherTable();
		JScrollPane vouchSelect = new JScrollPane(vt);
		vouchSelect.setOpaque(true);
		vouchSelect.setBorder(null);
		vouchSelect.setPreferredSize(new Dimension(450, 150));

		// put this scroll pane into another pane
		m_TokBuy.removeAll();
		m_GBC.gridheight = 4;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.WEST;
		m_TokBuy.add(vouchSelect, m_GBC);
		m_TokBuy.add(Box.createVerticalStrut(20), m_GBC);

		// prompt and quantity/equivalency go in a separate panel
		m_QuanPane.removeAll();

		// get number of tokens desired
		JLabel tokCntLab = new JLabel("Buy: ");
		tokCntLab.setLabelFor(m_TokenQuantity);
		tokCntLab.setFont(m_ParentHome.M_LabelFont);
		m_QuanPane.add(tokCntLab);

		// define document listener
		m_DoingUndo = false;
		m_TokenQuantity.getDocument().addDocumentListener(m_TokQuantListener);
		m_QuanPane.add(m_TokenQuantity);

		// show value equivalency
		buildEquivStrings();
		m_QuanPane.add(m_TokenValue);
		m_TokBuy.add(m_QuanPane, m_GBC);
		m_TokBuy.add(m_TokenEquiv, m_GBC);

		// install everything into scroll area
		m_WorkArea.setViewportView(m_TokBuy);
	}

	/**
	 * private utility method to build the token value equivalency strings
	 * (m_TokenValue and m_TokenEquiv)
	 */
	private void buildEquivStrings() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(2);
		df.setMinimumIntegerDigits(1);
		df.setGroupingUsed(false);

		// get value of one token in the selected issuer type
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		Hashtable<XMLVoucher, String> issuerVouchers = vtm.getIssuerVouchers();
		if (issuerVouchers.isEmpty()) {
			showError("unable to purchase tokens for requested operation",
						"your wallet contains no vouchers");
			return;
		}
		XMLVoucher firstv = vtm.getFirstVoucher();
		String asset = firstv.getAsset();
		Double oneToken = XMLToken.getValueOfOneToken(asset, firstv.getUnits());
		String dispUnits = XMLVoucher.getDisplayUnits(asset);
		
		// build string describing total token value
		StringBuilder eqv = new StringBuilder(32);
		eqv.append(" Tokens @ " + df.format(oneToken.doubleValue()) + " ");
		eqv.append(dispUnits + " = ");
		double tokTot = (double) m_LotSize * oneToken.doubleValue();
		eqv.append(df.format(tokTot) + " " + dispUnits);
		m_TokenValue.setText(eqv.toString());

		if (m_LotSize < M_MinTokens || m_LotSize > M_MaxTokens) {
			m_TokenQuantity.setForeground(Color.RED);
			m_TokenQuantity.setToolTipText("Enter at least " + M_MinTokens
													+ " and at most " + M_MaxTokens
													+ " tokens");
			m_PurchaseButton.setEnabled(false);
			m_PurchaseButton.setToolTipText("Enter a number of tokens between "
														+ M_MinTokens + " and "
														+ M_MaxTokens);
		}
		else {
			m_TokenQuantity.setForeground(Color.BLACK);
			m_TokenQuantity.setToolTipText("Enter the number of tokens you want");

			// see if Buy Now button can be turned on
			ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
			if (!selVouchers.isEmpty()) {
				XMLVoucher voucher = selVouchers.get(0);
				if (voucher.getNPV() >= tokTot) {
					m_PurchaseButton.setToolTipText("Click here to perform your "
															+ "token purchase");
					m_PurchaseButton.setEnabled(true);
				}
				else {
					m_PurchaseButton.setToolTipText("Select voucher(s) at least "
																+ tokTot + " " + dispUnits
																+ " in value");
					m_PurchaseButton.setEnabled(false);
				}
			}
			else {
				m_PurchaseButton.setToolTipText("Select voucher(s) to pay for your "
														+ "token purchase");
				m_PurchaseButton.setEnabled(false);
			}
		}

		// build fiat equiv in user's preferred currency
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		String lookUnits
			= VPAssetConfig.getPriceLookupUnits(asset, firstv.getUnits());
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Double price = apc.getRelativePrice(lookUnits, prefCurrency);
		if (price == null) {
			price = new Double(1.0);
			prefCurrency = dispUnits;
		}
		df.setMaximumFractionDigits(2);
		String fiat = "( Equivalent to " + df.format(tokTot * price.doubleValue())
							+ " " + prefCurrency + " )";
		m_TokenEquiv.setText(fiat);
	}

	/**
	 * method to receive notification of a change in voucher selection
	 * (called from VoucherSelector)
	 */
	public void voucherSelectChanged() {
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (selVouchers.isEmpty()
			 || vtm.getSelectMode() != VoucherTableModel.M_SelectMulti)
		{
			// not meant for us; do nothing
			m_PurchaseButton.setEnabled(false);
			return;
		}

		// see if the vouchers selected are big enough to buy the tokens
		Integer tQuant = new Integer(m_LotSize);
		XMLVoucher voucher = selVouchers.get(0);
		String dispUnits = XMLVoucher.getDisplayUnits(voucher.getAsset());
		Double oneToken
			= XMLToken.getValueOfOneToken(voucher.getAsset(), voucher.getUnits());
		double grVal = tQuant.doubleValue() * oneToken.doubleValue();
		double vVal = 0.0;
		for (XMLVoucher vouch : selVouchers) {
			vVal += vouch.getNPV();
		}
		if (vVal < grVal) {
			if (m_LotSize < M_MinTokens || m_LotSize > M_MaxTokens) {
				m_PurchaseButton.setToolTipText("Enter a number of tokens between "
															+ M_MinTokens + " and "
															+ M_MaxTokens);
			}
			else {
				m_PurchaseButton.setToolTipText("Select voucher(s) at least "
															+ grVal + " " + dispUnits
															+ " in value");
			}
			m_PurchaseButton.setEnabled(false);
		}
		else {
			if (m_LotSize >= M_MinTokens && m_LotSize <= M_MaxTokens) {
				m_PurchaseButton.setToolTipText("Click here to perform your token "
														+ "purchase");
				m_PurchaseButton.setEnabled(true);
			}
			else {
				m_PurchaseButton.setToolTipText("Enter a number of tokens between "
															+ M_MinTokens + " and "
															+ M_MaxTokens);
				m_PurchaseButton.setEnabled(false);
			}
		}
	}

	/**
	 * method to queue up a modification to a text component from within its
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	// implement ActionListener
	/**
	 * method to handle ActionEventS from the m_TokenQuantity field
	 * @param ae the event from m_TokenQuantity
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src == m_IssuerSel) {
			super.actionPerformed(ae);
			return;
		}
		JTextField input = (JTextField) src;
		// move to the next focus item
		input.transferFocus();
	}

	// END methods
}
