/*
 * last modified---
 * 	06-05-13/jet/ SilentVault client version
 * 	09-20-10/kbw/ wording changes, grey button until selections made
 * 	09-02-10/kbw/ changes for Nimbus L&F
 * 	08-16-10/kbw/ remove spurious Done button
 * 	05-03-10/kbw/ grey out buttons during messaging operations
 * 	04-12-10/kbw/ flesh out
 * 	03-29-10/kbw/ new (stub)
 *
 * purpose---
 * 	provide a UI widget to handle voucher validations
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.Color;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for doing voucher (re)validations.
 *
 * @author Kevin Wilkerson
 */
public final class ValidatePane extends WorkPane {
	// BEGIN data members
	/**
	 * action to validate a voucher
	 */
	private AbstractAction		m_ValidateAction;

	/**
	 * action button to trigger revalidation
	 */
	private JButton				m_ValidateButton;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public ValidatePane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_HeadingPane.setLayout(new GridLayout(1, 3));

		// provide user textual instructions
		String validTxt = "Select a voucher to validate.  (Expired vouchers "
								+ "can't be validated.)<br/><br/>"
								+ "Note: validating is not usually necessary.  Vouchers"
								+ " are validated automatically when used during "
								+ "payments.  When a voucher is nearing its expiry "
								+ "date, validating will renew it.";
		m_UserInstructs = validTxt;

		// set token cost
		m_TokenCost.setText("Validation uses " + VoucherRequest.M_ValidateCost
								+ " tokens");

		// establish button actions
		setValidateAction();
		m_ValidateButton = new JButton(m_ValidateAction);
		m_ValidateButton.setFont(m_Plugin.M_ButtonFont);
	}

	/**
	 * utility method to establish the Action which can process a click on the
	 * Validate voucher ops button.
	 * NB: this method will block waiting on the OFS's reply
	 */
	private void setValidateAction() {
		m_ValidateAction = new AbstractAction("Validate Now") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need exactly one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() < 1) {
					showError("no vouchers to validate", "none selected",
								"please select exactly one voucher to validate");
					return;
				}
				// this case should never occur, but...
				if (selVouchers.size() > 1) {
					showError("too many vouchers to validate",
								"more than one selected",
								"please select exactly one voucher to validate");
					return;
				}

				// ensure we have enough tokens
				TokenTableModel ttm = m_ParentHome.getTokenModel();
				ArrayList<XMLToken> safeTokens = ttm.getAvailableTokens(
																VoucherRequest.M_ValidateCost,
																		vtm.getSelectedIssuer());
				if (safeTokens == null) {
					// Record ourselves with tab manager, so we return here after
					// the user has purchased tokens.  Then go purchase tokens.

					m_TabManager.setPendingTokens(m_TabManager.getValidatePane());
					m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_TabManager.makeTransition();
					return;
				}

				// get the voucher and its filename
				XMLVoucher voucher = selVouchers.get(0);
				Hashtable<XMLVoucher, String> vouchers = vtm.getVoucherList();
				String sdsFile = vouchers.get(voucher);

				// send the message to the OFS
				final XMPPConnection conn = m_Plugin.getOFSConnection();
				VscState stData = m_Plugin.getVscState();
				VsSecrets secData = m_Plugin.getLoginSecrets();

				// build message to send
				OFSMessage val = new OFSMessage();
				val.setType(IQ.Type.SET);
				val.setFrom(conn.getUser());
				val.setTo(conn.getServiceName());
				val.setPacketID("valid_" + stData.getNextOFSid());
				val.setOpcode("REQ_revalidate");
				val.setVoucher_publisher(secData.getPublisher());

				// build voucher request and add it to message
				VoucherRequest vReq = new VoucherRequest();
				vReq.setAction("validate");
				vReq.setSigningVS(secData.getVSnumber());
				vReq.setFolderHash(new String(secData.getVoucherIndex()));
				vReq.setFolderCap(secData.getVouchRWCap());
				HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
				inpVouchers.put(sdsFile, voucher);
				vReq.setInitialized();
				val.setVouchReq(vReq);
				// NB: the <output/> value is implicitly the same as the input

				// supply payment tokens (validate costs 4)
				for (int iii = 0; iii < VoucherRequest.M_ValidateCost; iii++) {
					vReq.addPaymentToken(safeTokens.get(iii));
				}

				// sign the voucher request with our private key
				if (!vReq.signDetails(secData.getPrivKey())) {
					Log.error("Unable to sign VR with VS privkey");
					showError("could not sign voucherRequest!");
					return;
				}

				// encrypt the details to the VP's key so the OFS can't read them
				if (!vReq.encryptDetails(secData.getVPKey(), true)) {
					Log.error("Unable to encrypt VR with VP pubkey");
					showError("could not encrypt voucherRequest for VP!");
					return;
				}

				// flag this voucher and the tokens as committed to an operation
				vtm.commitVouchers(selVouchers);
				ttm.commitTokens(safeTokens);
				vtm.deselectAll();

				// register a collector for the reply
				PacketIDFilter pFilter = new PacketIDFilter(val.getPacketID());
				PacketCollector repColl = conn.createPacketCollector(pFilter);

				// send it to the server
				m_ParentHome.setCursor(M_WaitCursor);
				setButtonState(false);
				conn.sendPacket(val);

				// record outgoing packet
				stData.setLastActivity(val);

				// wait for reply or timeout
				OFSMessage rep = null;
				IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
				m_ParentHome.setCursor(null);
				setButtonState(true);
				if (reply == null) {
					// got timeout; build error
					rep = new OFSMessage();
					rep.setOpcode("REP_validation");
					rep.setFrom(val.getTo());
					rep.setTo(val.getFrom());
					rep.setPacketID(val.getPacketID());
					rep.setType(IQ.Type.ERROR);
					rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
					rep.setErrmsg("timed out");
				}
				else {
					rep = (OFSMessage) reply;
				}

				// invoke callback with result
				m_Plugin.getOFSListener().processPacket(rep);
			}
		};
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// set heading pane
		buildValueHeader();

		// set screen label
		buildScreenLabel("Validate a Voucher");

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, the button to validate the selected voucher
		m_ValidateButton.setToolTipText("Select a voucher to validate");
		m_ValidateButton.setEnabled(false);
		m_ButtonPane.add(Box.createHorizontalStrut(325));	// spacer
		m_ButtonPane.add(m_ValidateButton);

		// place the voucher table into work area, set for radio select mode
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.deselectAll();
		vtm.setSelectMode(VoucherTableModel.M_SelectSingle);
		JTable vt = m_ParentHome.getVoucherTable();
		m_WorkArea.setViewportView(vt);
	}

	/**
	 * method to receive notification of a change in voucher selection
	 */
	public void voucherSelectChanged() {
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (vtm.getSelectMode() != VoucherTableModel.M_SelectSingle) {
			// not meant for us; do nothing
			return;
		}

		if (!selVouchers.isEmpty()) {
			// make the button active and change tooltip
			m_ValidateButton.setToolTipText("Click here to validate the selected "
														+ "voucher");
			m_ValidateButton.setEnabled(true);
		}
		else {
			m_ValidateButton.setToolTipText("Select a voucher to validate");
			m_ValidateButton.setEnabled(false);
		}
	}

	// END methods
}
