/*
 * last modified---
 * 	07-17-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine value
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	01-31-14/jet/ do not allow a valid issuer selection to be overridden in
 * 					  buildValueHeader()
 * 	06-05-13/jet/ SilentVault client version
 * 	02-23-11/kbw/ remove assumption that all vouchers are in GAU
 * 	09-03-10/kbw/ changes for Nimbus L&F
 * 	08-16-10/kbw/ dispense with rollover buttons and icons
 * 	08-02-10/kbw/ move from horizontal layout to vertical one
 * 	07-30-10/kbw/ add generic m_MainButton for use in subclasses
 * 	05-03-10/kbw/ do not include expired vouchers in safe value total; add
 * 					  setButtonState()
 * 	04-22-10/kbw/ use BoxLayout for button pane
 * 	03-26-10/kbw/ rewrite as base class for new UI spec
 *
 * purpose---
 * 	provide a UI widget to act as the work area for doing user input/output
 */

package org.silentvault.client.ui.wallet;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Component;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.silentvault.client.*;
import org.silentvault.client.ui.*;

import java.util.TreeSet;
import java.util.Vector;
import java.util.Hashtable;
import java.util.ArrayList;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends JPanel to provide a base screen for doing wallet-related I/O.
 * This class will never be used directly, but subclasses of it will be
 * swapped into the HomePane by the WTabManager, which can be invoked by
 * messaging callbacks.
 *
 * @author Kevin Wilkerson
 */
public abstract class WorkPane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * standard wait (hourglass) cursor
	 */
	protected final Cursor	M_WaitCursor = 
											Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);

	/**
	 * width of items in this pane
	 */
	private final int			M_WpWidth = 485;

	/**
	 * the plugin/application to which we belong
	 */
	protected WalletClient	m_Plugin;

	/**
	 * the tab manager that manages us
	 */
	protected WTabManager	m_TabManager;

	/**
	 * the home pane in which we are installed
	 */
	protected HomePane		m_ParentHome;

	/**
	 * the panel for top buttons with which we interact
	 */
	protected JPanel			m_WorkBar;

	/**
	 * a pane to contain all of our UI objects
	 */
	private JPanel				m_WorkPane;

	/**
	 * our standard GridBagConstraints definition
	 */
	protected GridBagConstraints	m_GBC;

	/**
	 * the layout to use for m_WorkArea, m_ButtonPane, and m_HeadingPane
	 */
	private GridBagLayout	m_GBLayout;

	/**
	 * the heading, typically containing Issuer select, text and a displayed
	 * token count (set by subclasses)
	 */
	protected JPanel			m_HeadingPane;

	/**
	 * the work area, whose contents are set by our subclasses
	 */
	protected JScrollPane	m_WorkArea;

	/**
	 * a label for the screen title
	 */
	private JLabel				m_ScreenTitle;

	/**
	 * a pane to contain the screen title, icon and token cost/count
	 */
	private JPanel				m_ScreenLabel;

	/**
	 * the instructional text (used as a icon tooltip in m_ScreenLabel)
	 */
	protected String			m_UserInstructs;

	/**
	 * the token cost indicator (for middle element in m_ScreenLabel)
	 */
	protected JLabel			m_TokenCost;

	/**
	 * the available token count indicator (for last element in m_ScreenLabel)
	 */
	protected JLabel			m_TokenCount;

	/**
	 * a panel in which to install action buttons (used by subclasses)
	 */
	protected JPanel			m_ButtonPane;

	/**
	 * selector tool for selecting Issuers (typically found in m_HeadingPane)
	 */
	protected JComboBox<String>	m_IssuerSel;

	/**
	 * pre-defined button to return to the MenuPane, for use in subclasses
	 */
	protected JButton			m_MainButton;

	/**
	 * pre-defined border object to supply insets
	 */
	protected EmptyBorder	m_InsetBorder;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which created us
	 */
	public WorkPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_ParentHome = home;
		m_TabManager = mgr;
		m_WorkBar = m_ParentHome.getMenuBar();
		m_IssuerSel = new JComboBox<String>();
		setLayout(new GridLayout());
		setOpaque(true);

		// define a return to menu action and button, for installation into the
		// m_WorkBar in subclasses

		AbstractAction goMain = new AbstractAction("Back to Overview") {
			public void actionPerformed(ActionEvent actionEvent) {
				// switch to main overview screen
				WTabManager tm = m_Plugin.getWalletTabManager();
				tm.queueTransition(new Integer(WTabManager.M_GO_MAIN_MENU));
				tm.makeTransition();
			}
		};
		m_MainButton = new JButton(goMain);
		m_MainButton.setToolTipText("Return to the main menu");
		m_MainButton.setFont(m_Plugin.M_ButtonFont);

		// establish panes and layout controls
		m_WorkPane = new JPanel();
		m_InsetBorder = new EmptyBorder(4, 8, 4, 8);
		m_WorkPane.setPreferredSize(new Dimension(M_WpWidth, 350));
		m_GBLayout = new GridBagLayout();
		m_WorkPane.setLayout(m_GBLayout);
		m_GBC = new GridBagConstraints();
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.CENTER;
		m_GBC.fill = GridBagConstraints.NONE;

		// label containing screen title
		m_ScreenTitle = new JLabel("Default WorkPane");
		m_ScreenTitle.setFont(m_ParentHome.M_HeaderFont);

		// label containing token cost
		m_TokenCost = new JLabel("No tokens required");
		m_TokenCost.setHorizontalAlignment(SwingConstants.LEADING);
		m_TokenCost.setFont(m_ParentHome.M_HeaderFont);

		// label containing available token count
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		String tokCnt
			= " (" + ttm.getAvailableCount(vtm.getSelectedIssuer()) + ")";
		m_TokenCount = new JLabel(tokCnt);
		m_TokenCount.setFont(m_ParentHome.M_HeaderFont);
		m_TokenCount.setHorizontalAlignment(SwingConstants.TRAILING);

		// create empty screen label panel, using blue background
		m_ScreenLabel = new JPanel();
		m_ScreenLabel.setLayout(new BoxLayout(m_ScreenLabel, BoxLayout.X_AXIS));
		m_ScreenLabel.setPreferredSize(new Dimension(M_WpWidth, 23));
		m_ScreenLabel.setBackground(m_ParentHome.M_TitleBackgroundColor);
		EmptyBorder margin = new EmptyBorder(0, 4, 0, 4);
		m_ScreenLabel.setBorder(margin);
		m_ScreenLabel.setVisible(true);

		// heading pane, for titles, Issuer selection and voucher values
		m_HeadingPane = new JPanel();
		m_HeadingPane.setPreferredSize(new Dimension(M_WpWidth, 20));
		m_HeadingPane.setBackground(m_ParentHome.M_TitleBackgroundColor);
		m_HeadingPane.setBorder(margin);

		// the work area will be populated by subclasses via setViewPort()
		m_WorkArea = new JScrollPane();
		m_WorkArea.setPreferredSize(new Dimension(M_WpWidth, 260));
		m_WorkArea.setOpaque(true);

		// create panel in which to install buttons
		m_ButtonPane = new JPanel();
		m_ButtonPane.setLayout(new BoxLayout(m_ButtonPane, BoxLayout.X_AXIS));
		m_ButtonPane.setPreferredSize(new Dimension(M_WpWidth, 25));

		// put it all top to bottom in the work pane
		m_GBC.gridheight = 7;
		m_WorkPane.add(m_ScreenLabel, m_GBC);
		m_WorkPane.add(Box.createVerticalStrut(3), m_GBC);
		m_WorkPane.add(m_HeadingPane, m_GBC);
		m_WorkPane.add(Box.createVerticalStrut(5), m_GBC);
		m_WorkPane.add(m_WorkArea, m_GBC);
		m_WorkPane.add(Box.createVerticalStrut(5), m_GBC);
		m_WorkPane.add(m_ButtonPane, m_GBC);

		// everything is now in the work pane
		add(m_WorkPane);
	}

	/**
	 * method to generate a heading bar containing: an Issuer selection drop-
	 * down list; a total of the voucher value in the safe; and the token count
	 */
	protected void buildValueHeader() {
		// from the voucher list, build a set of possible Issuers
		TreeSet<String> issuers = new TreeSet<String>();
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		Hashtable<XMLVoucher, String> vouchList = vtm.getVoucherList();
		String currIssuer = vtm.getSelectedIssuer();
		for (XMLVoucher vouch : vouchList.keySet()) {
			issuers.add(vouch.getIssuer());
		}
		if (issuers.isEmpty()) {
			// must not have any vouchers yet
			issuers.add("-none-");
		}
		if (m_IssuerSel != null) {
			m_IssuerSel.removeActionListener(this);
			m_IssuerSel.removeAllItems();
		}
		m_IssuerSel = new JComboBox<String>(new Vector<String>(issuers));
		// the selected one will match selected Issuer, if found
		int idx = 0;
		int sel = 0;
		for (String iss : issuers) {
			if (iss.equals(vtm.getSelectedIssuer())) {
				sel = idx;
				break;
			}
			idx++;
		}
		m_IssuerSel.setSelectedIndex(sel);

		// provide default
		String iss = (String) m_IssuerSel.getSelectedItem();
		if (!iss.equals("-none-")) {
			/* Do not override a previously selected Issuer, because each subclass
			 * has its own m_HeadingPane but there is only one VoucherTableModel.
			 * So we don't want to set the wrong value, in case the Issuer list
			 * we just built does not contain some particular Issuer at the moment.
			 */
			if (currIssuer.isEmpty() || currIssuer.equals("-none-")) {
				vtm.selectIssuer(iss);
			}
			vtm.setEquivHeader(m_Plugin.getUserPreferences().getValueCurrency());
			JTable vouchTab = m_ParentHome.getVoucherTable();
			m_ParentHome.configTableColumns(vouchTab);
		}
		m_IssuerSel.addActionListener(this);
		m_IssuerSel.setToolTipText("Select voucher/token Issuer");
		m_IssuerSel.setFont(m_ParentHome.M_HeaderFont);

		// add selector to the heading
		m_HeadingPane.removeAll();
		m_HeadingPane.setLayout(new BoxLayout(m_HeadingPane, BoxLayout.X_AXIS));
		m_HeadingPane.add(m_IssuerSel);

		// add the value in the safe (total of all vouchers from selected Issuer)
		double vTot = vtm.getVoucherSum();
		XMLVoucher vouch = vtm.getFirstVoucher();
		String dispUnits = "GAU";
		String asset = "GOLD";
		String lookUnits = dispUnits;
		if (vouch != null) {
			// should be present, or this Issuer wouldn't be listed...
			asset = vouch.getAsset();
			dispUnits = XMLVoucher.getDisplayUnits(asset);
			lookUnits = VPAssetConfig.getPriceLookupUnits(asset, vouch.getUnits());
		}

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);
		String val = "Balance: " + df.format(vTot) + " " + dispUnits;
		JLabel valFld = new JLabel(val);
		valFld.setFont(m_ParentHome.M_HeaderFont);
		valFld.setBorder(m_InsetBorder);
		m_HeadingPane.add(valFld);

		// add the equivalent in the user's currency preference
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Double assetPrice = apc.getRelativePrice(lookUnits, prefCurrency);
		if (assetPrice == null) {
			// provide a default in case prices not available
			assetPrice = new Double(1.0);
			prefCurrency = dispUnits;
		}
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		val = "( " + df.format(vTot * assetPrice.doubleValue()) + " "
				+ prefCurrency + " )";
		JLabel equivFld = new JLabel(val);
		equivFld.setHorizontalAlignment(SwingConstants.TRAILING);
		equivFld.setFont(new Font(Font.SANS_SERIF, Font.ITALIC, 14));
		equivFld.setBorder(m_InsetBorder);
		m_HeadingPane.add(equivFld);
		m_HeadingPane.add(Box.createHorizontalStrut(150));

		// update count of tokens in label
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		m_TokenCount.setText(" (" + ttm.getAvailableCount(vtm.getSelectedIssuer())										+ ")");
		m_HeadingPane.revalidate();
		m_HeadingPane.repaint();
	}

	/**
	 * method to build and display the screen label pane
	 * @param screenLabel the label text for the screen
	 * NB: implied parameters: m_UserInstructs, m_TokenCost
	 */
	protected void buildScreenLabel(	String screenLabel) {
		if (screenLabel == null) {
			Log.error("Null screen label");
			return;
		}

		m_ScreenTitle.setText(screenLabel);
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		m_TokenCount.setText(" (" + ttm.getAvailableCount(vtm.getSelectedIssuer())										+ ")");
		StringBuilder instTip = new StringBuilder(512);
		instTip.append("<html><table border=\"0\" width=\"300\"><tr><td>");
		instTip.append(m_UserInstructs);
		instTip.append("</td></tr></table></html>");
		m_TabManager.M_InfoIcon.setToolTipText(instTip.toString());

		// in order to preserve alignments, we need to add two separate panels
		// with opposite flow alignments

		JPanel part1 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		part1.setOpaque(false);
		EmptyBorder eb = new EmptyBorder(0, 0, 4, 0);
		part1.setBorder(eb);

		m_ScreenLabel.removeAll();
		part1.add(m_ScreenTitle);
		part1.add(Box.createHorizontalStrut(5));
		part1.add(m_TabManager.M_InfoIcon);
		m_ScreenLabel.add(part1);

		JPanel part2 = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		part2.setOpaque(false);
		part2.setBorder(eb);

		part2.add(m_TokenCost);
		part2.add(Box.createHorizontalStrut(5));
		part2.add(m_TokenCount);
		m_ScreenLabel.add(part2);
	}

	/**
	 * method to enable/disable all buttons in the m_ButtonPane
	 * @param enabled status to set
	 */
	protected void setButtonState(boolean enabled) {
		Component[] buttons = m_ButtonPane.getComponents();
		for (int iii = 0; iii < buttons.length; iii++) {
			buttons[iii].setEnabled(enabled);
		}
	}

	/**
	 * method required to implement ActionListener (when an Issuer is chosen)
	 * @param ae the event from the issuer select list
	 */
	public void actionPerformed(ActionEvent ae) {
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		String iss = (String) m_IssuerSel.getSelectedItem();
		if (!iss.equals("-none-")) {
			if (vtm != null) {
				vtm.selectIssuer(iss);
			}
		}
		Runnable rebuildHdr = new Runnable() {
			public void run() {
				buildValueHeader();
			}
		};
		SwingUtilities.invokeLater(rebuildHdr);
	}


	// error display methods
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	protected void showError(String errMsg) {
		m_ParentHome.setCursor(null);
		m_Plugin.getWalletTabManager().showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	protected void showError(String errMsg, String cause) {
		m_ParentHome.setCursor(null);
		m_Plugin.getWalletTabManager().showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	protected void showError(String errMsg, String cause, String action) {
		m_ParentHome.setCursor(null);
		m_Plugin.getWalletTabManager().showError(errMsg, cause, action);
	}

	/**
	 * method to be overridden by subclasses, to prepare display in whatever
	 * ways might be required (for example, by making the menu bar visible or
	 * setting the heading)
	 */
	abstract void prepDisplay();

	// END methods
}
